// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: arraystatistics.inc
// Last updated: 28 May 2021

// Description: This file contains a basic set of  array statistic macros.

//  Mean               - Long in math.inc and earlier includes.
//  Variance           -
//  Minimum            -
//  Maximum            -
//  StandardDeviation  - 'Std_Dev' long in math.inc and earlier includes.
//  Statistics         - 'GetStats' long in math.inc and earlier includes.
//  PrintStatistics    -
//  Histogram          - Long in math.inc and earlier includes.

// Variance and StandardDeviation full population based - not sample based.

// Mean, Variance and StandardDeviation macros also work for arrays of vectors
// so long as the vector types are of constant dimension. If passing v3.8 mixed
// type arrays the internal types cannot change from the first encountered.

// The versions herein primarily coded by Tor Olav Kristensen and posted to
// povray.general and the thread
//      "Why only a +1 clamp with VAngle, VAngleD macros in math.inc?"
// on May 08, 2021.

// May 24, 2021  - WFP Updates to check array types do not change if incoming
//                 arrays are of v3.8 mixed type. This the reason internal fixed
//                 array variables are used in loops over simple floats.

//------------------------------------------------------------------------------

#ifndef(ARRAYSTATISTICS_INC_TEMP)
#declare ARRAYSTATISTICS_INC_TEMP = version;
#version 3.8;   // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including arraystatistics.inc\n"
#end


//  // ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7

#macro Mean(Values)

    #local N = dimension_size(Values, 1);
    #local Sum = array[1] // Fixed array types cannot change.
    #local Sum[0] = Values[0];
    #for (I, 1, N - 1)
        #local Sum[0] = Sum[0] + Values[I];
    #end // for

    (Sum[0]/N)

#end // macro Mean


#macro Variance(Values)

    #local N = dimension_size(Values, 1);
    #local M = Mean(Values);
    #local D = Values[0] - M;
    #local SquareSum = D*D;
    #for (I, 1, N - 1)
        #local D = Values[I] - M;
        #local SquareSum = SquareSum + D*D;
    #end

    (SquareSum/N)

#end // macro Variance


#macro Minimum(Values)

    #local N = dimension_size(Values, 1);
    #local Min = array[2] {1,1} // Fixed array types cannot change.
    #local Min[0] = Values[0];  // Better err msg - min() floats only.
    #for (I, 1, N - 1)
        #local Min[1] = Values[I]; // Better err msg.
        #local Min[0] = min(Min[0], Min[1]);
    #end // for

    Min[0]

#end // macro Minimum


#macro Maximum(Values)

    #local N = dimension_size(Values, 1);
    #local Max = array[2] {1,1} // Fixed array types cannot change.
    #local Max[0] = Values[0];  // Better err msg - min() floats only.
    #for (I, 1, N - 1)
        #local Max[1] = Values[I]; // Better err msg.
        #local Max[0] = max(Max[0], Max[1]);
    #end // for

    Max[0]

#end // macro Maximum


#macro StandardDeviation(Values)

    sqrt(Variance(Values))

#end // macro StandardDeviation


#macro Statistics(Values)

    #local N = dimension_size(Values, 1);
    #local Value = array[2] {1,1} // Fixed array types cannot change.
    #local Value[0] = Values[0];  // Better err msg - floats only.
    #local Min = Value[0];
    #local Max = Value[0];
    #local Sum = Value[0];
    #local SquareSum = pow(Value[0], 2);
    #for (I, 1,  N - 1)
        #local Value[1] = Values[I]; // Better err msg
        #local Min = min(Min, Value[1]);
        #local Max = max(Max, Value[1]);
        #local Sum = Sum + Value[1];
        #local SquareSum = SquareSum + pow(Value[1], 2);
    #end // for
    #local M = Sum/N;
    #local Var = SquareSum/N - pow(M, 2);

    dictionary {
        .Count: N,
        .Min: Min,
        .Max: Max,
        .Mean: M,
        .Var: Var,
        .StdDev: sqrt(Var)
    }

#end // macro Statistics


#macro PrintStatistics(Stats)

    #local S = array[6] { "Count", "Min", "Max", "Mean", "Var", "StdDev" };
    #debug "\n"
    #for (I, 0, 5)
        #debug concat(S[I], " = ", str(Stats[S[I]], 0, -1), "\n")
    #end // for
    #debug "\n"

#end // macro PrintStatistics


#macro Histogram(Values, NoOfBins)

    #local Stats = Statistics(Values);
    #local N = Stats.Count;
    #local Min = Stats.Min;
    #local Max = Stats.Max;
    #local BinWidth = (Max - Min)/NoOfBins;
    #local Bins = array[NoOfBins][3];
    #local BinLimitLow = Min;
    #for (I, 0, NoOfBins - 1)
        #local BinLimitHigh = BinLimitLow + BinWidth;
        #local Bins[I][0] = BinLimitLow;
        #local Bins[I][1] = BinLimitHigh;
        #local Bins[I][2] = 0; // Set count for bin to zero
        #local BinLimitLow = BinLimitHigh;
    #end // for
    #for (I, 0, N - 1)
        // Find the right bin for the value
        #local BinNo = int((Values[I] - Min)/BinWidth);
        // Make sure rounding errors do not lead to a non existent bin
        #local BinNo = min(max(0, BinNo), NoOfBins - 1);
        // Increase the count for the bin
        #local Bins[BinNo][2] = Bins[BinNo][2] + 1;
    #end // for

    Bins

#end // macro Histogram

// ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="arraystatistics.inc")

    #fopen FID "SelfTestArrayStatisticsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n\n")
    #write (FID,"\n")

    // Statistics, PrintStatistics
    #declare A = array[5]
    #declare A[0] = 1; #declare A[1] = 2; #declare A[2] = 3; #declare A[3] = 4; #declare A[4] = 5;
    #declare Stats = Statistics(A)
    PrintStatistics(Stats)
    #ifdef (A) #undef A #end
    #ifdef (Stats) #undef Stats #end
    #write (FID,"\n")

    #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))
    #fclose FID

    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestArrayStatisticsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n\n")
    // If all macro call tests above work create output for what was tested.
    #write (FID,concat("Statistics 1.000000 ",str(1.0,9,6)," Statistics OK\n"))
    #write (FID,concat("PrintStatistics 1.000000 ",str(1.0,9,6)," PrintStatistics OK\n"))
    #write (FID,"\n")

    // Mean macro
    #declare A = array[5]
    #declare A[0] = 1; #declare A[1] = 2; #declare A[2] = 3; #declare A[3] = 4; #declare A[4] = 5;
    #declare Val = Mean(A);
    #write (FID,concat("Mean 3.000000 ",str(Val,9,6)," \n"))
    #write (FID,"-\n")
    #declare AV = array mixed[5]
    #declare AV[0] = <1,1,1,1,5>;
    #declare AV[1] = <2,2,2,2,4>;
    #declare AV[2] = <3,3,3,3,3>;
    #declare AV[3] = <4,4,4,4,2>;
    #declare AV[4] = <5,5,5,5,1>;
    #declare Val = Mean(AV);
    #write (FID,concat("Mean 3.000000 ",str(Val.x,9,6)," x\n"))
    #write (FID,concat("Mean 3.000000 ",str(Val.y,9,6)," y\n"))
    #write (FID,concat("Mean 3.000000 ",str(Val.z,9,6)," z\n"))
    #write (FID,concat("Mean 3.000000 ",str(Val.t,9,6)," t\n"))
    #write (FID,concat("Mean 3.000000 ",str(Val.transmit,9,6)," transmit\n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // Variance
    #declare Val = Variance(A);
    #write (FID,concat("Variance 2.000000 ",str(Val,9,6)," \n"))
    #write (FID,"-\n")
    #declare Val = Variance(AV);
    #write (FID,concat("Variance 2.000000 ",str(Val.x,9,6)," x\n"))
    #write (FID,concat("Variance 2.000000 ",str(Val.y,9,6)," y\n"))
    #write (FID,concat("Variance 2.000000 ",str(Val.z,9,6)," z\n"))
    #write (FID,concat("Variance 2.000000 ",str(Val.t,9,6)," t\n"))
    #write (FID,concat("Variance 2.000000 ",str(Val.transmit,9,6)," transmit\n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // Minimum
    #declare Val = Minimum(A);
    #write (FID,concat("Minimum 1.000000 ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // Maximum
    #declare Val = Maximum(A);
    #write (FID,concat("Maximum 5.000000 ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // StandardDeviation
    #declare Val = StandardDeviation(A);
    #write (FID,concat("StandardDeviation 1.414214 ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // Histogram
    #declare HistogramArray = Histogram(A,5)
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[0][0],9,6)," 0 lo \n"))
    #write (FID,concat("Histogram 1.800000 ",str(HistogramArray[1][0],9,6)," 1 lo \n"))
    #write (FID,concat("Histogram 2.600000 ",str(HistogramArray[2][0],9,6)," 2 lo \n"))
    #write (FID,concat("Histogram 3.400000 ",str(HistogramArray[3][0],9,6)," 3 lo \n"))
    #write (FID,concat("Histogram 4.200000 ",str(HistogramArray[4][0],9,6)," 4 lo \n"))
    #write (FID,"---\n")
    #write (FID,concat("Histogram 1.800000 ",str(HistogramArray[0][1],9,6)," 0 hi \n"))
    #write (FID,concat("Histogram 2.600000 ",str(HistogramArray[1][1],9,6)," 1 hi \n"))
    #write (FID,concat("Histogram 3.400000 ",str(HistogramArray[2][1],9,6)," 2 hi \n"))
    #write (FID,concat("Histogram 4.200000 ",str(HistogramArray[3][1],9,6)," 3 hi \n"))
    #write (FID,concat("Histogram 5.000000 ",str(HistogramArray[4][1],9,6)," 4 hi \n"))
    #write (FID,"---\n")
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[0][2],9,6)," 0 cnt \n"))
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[1][2],9,6)," 1 cnt \n"))
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[2][2],9,6)," 2 cnt \n"))
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[3][2],9,6)," 3 cnt \n"))
    #write (FID,concat("Histogram 1.000000 ",str(HistogramArray[4][2],9,6)," 4 cnt \n"))
    #ifdef (HistogramArray) #undef HistogramArray #end
    #ifdef (A) #undef A #end
    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"


#end // End self testing

#version ARRAYSTATISTICS_TEMP;
#end //arraystatistics.inc

