// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: math.inc
// Last updated: 28 May 2021

// Description: This file contains various math macros.

//------------------------------------------------------------------------------

#ifndef(MATH_INC_TEMP)
#declare MATH_INC_TEMP = version;
#version 3.8;   // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including math.inc\n"
#end

#include "functions.inc"
#include "vectors.inc"


// Returns the r,g,b pigment values at a point in 3D space.
// Note. In POV-Ray proper the macro called eval_pigment pretending
// to be a scene description language (SDL) inbuilt.
#macro Eval_Pigment(_pigm, _vec)
    #local Fn = function { pigment { _pigm } }
    #local Result = (Fn(_vec.x,_vec.y,_vec.z));
    Result
#end

// Interpolate author: Margus Ramst
// Interpolation
// GC - global current
// GS - global start
// GE - global end
// TS - target start
// TE - target end
// Method - interpolation method:
//          Method = 0 - cosine interpolation
//          Method > 0 - exponential (1 - linear, etc)
#macro Interpolate(GC, GS, GE, TS, TE, Method)
    (
    #if(Method!=0)
        (TS+(TE-TS)*pow((GC-GS)/(GE-GS),Method))
    #else
        #local X=(GC-GS)/(GE-GS);
        #local F=(1-cos(X*pi))*.5;
        (TS*(1-F)+TE*F)
    #end
    )
#end

// --------------------------------------------------------
// Vector macros:
// --------------------------------------------------------

// Squares the components of a vector
#macro VSqr(V) (V*V) #end

// Raises the components of a vector to a given power
#macro VPow(V, P) <pow(V.x, P), pow(V.y, P), pow(V.z, P)> #end
#macro VPow5D(V, P) <pow(V.x, P), pow(V.y, P), pow(V.z, P), pow(V.filter, P), pow(V.transmit, P)> #end

// Returns true if vectors are equal, otherwise false
#macro VEq(V1, V2) (V1.x = V2.x & V1.y = V2.y & V1.z = V2.z) #end
#macro VEq5D(V1, V2)
    ( V1.x = V2.x
    & V1.y = V2.y
    & V1.z = V2.z
    & V1.filter = V2.filter
    & V1.transmit = V2.transmit)
#end

// Returns true if vector is <0,0,0>, otherwise false
#macro VZero(V1) (V1.x = 0 & V1.y = 0 & V1.z = 0) #end
// Returns true if vector is <0,0,0,0,0>, otherwise false
#macro VZero5D(V1) (V1.x = 0 & V1.y = 0 & V1.z = 0 & V1.filter = 0 & V1.transmit = 0) #end

#macro VLength5D(V) sqrt(V.x*V.x + V.y*V.y + V.z*V.z + V.filter*V.filter + V.transmit*V.transmit) #end

#macro VNormalize5D(V) (V/sqrt(V.x*V.x + V.y*V.y + V.z*V.z + V.filter*V.filter + V.transmit*V.transmit)) #end

#macro VDot5D(V1, V2) (V1.x*V2.x + V1.y*V2.y + V1.z*V2.z + V1.filter*V2.filter + V1.transmit*V2.transmit) #end

//  Cosine of angle between V1 and V2
//  #macro VCos_Angle(V1,V2) max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2)))) #end
//  Update as able to VectorCos() in vectors.inc
#macro VCos_Angle(V1,V2)
  //#error "VCos_Angle in math.inc called. Use VectorCos()"
    VectorCos(V1,V2)
#end

//  Angle in radians between V1 and V2
//  #macro VAngle(V1, V2) acos(max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2))))) #end
//  See also: AngleBetweenVectors, AccurateAngleBetweenVectors in vectors.inc
#macro VAngle(V1,V2) f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1) #end

//  Angle in degrees between V1 and V2
//  #macro VAngleD(V1, V2) degrees(acos(max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2)))))) #end
//  See also: AngleBetweenVectors, AccurateAngleBetweenVectors in vectors.inc
#macro VAngleD(V1,V2) degrees(f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1)) #end

// VRotation() will find the rotation angle from V1 to V2
// around Axis. Axis should be perpendicular to both V1
// and V2. The output will be in the range between -pi and
// pi radians or between -180 degrees and 180 degrees if
// you are using the degree version. However, if Axis is
// set to <0,0,0> the output will always be positive or
// zero, the same result you will get with the VAngle() macros.
// Author: Rune S. Johansen
//  #macro VRotation(V1, V2, Axis)
//     (acos(max(min(vdot(vnormalize(V1),vnormalize(V2)),1),-1))
//     *(vdot(Axis,vcross(V1,V2))<0?-1:1))
//  #end
#macro VRotation(V1,V2,Axis)
    (f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1)
    *(vdot(Axis,vcross(V1,V2))<0?-1:1))
#end
//  #macro VRotationD(V1, V2, Axis)
//     (degrees(acos(max(min(vdot(vnormalize(V1),vnormalize(V2)),1),-1)))
//     *(vdot(Axis,vcross(V1,V2))<0?-1:1))
//  #end
#macro VRotationD(V1,V2,Axis)
    (degrees(f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1))
    *(vdot(Axis,vcross(V1,V2))<0?-1:1))
#end

// Distance between V1 and V2. With povr have f_length too.
#macro VDist(V1, V2)
    #error "VDist from math.inc called. Use inbuilt vlength()"
    vlength(V1 - V2)
#end

// Returns a vector perpendicular to V
// Author: Tor Olav Kristensen
// See: OrthogonalVector() and AltOrthogonalVector() in vectors.inc
#macro VPerp_To_Vector(v0)
  //#error "VPerp_To_Vector in math.inc called. Use OrthogonalVector()"
    OrthogonalVector(v0)
#end

// Returns a vector perpendicular to V1 and V2
#macro VPerp_To_Plane(V1, V2) (vnormalize(vcross(V1, V2))) #end

// Find a vector perpendicular to Axis and in the plane of
// V and Axis. In other words, the new vector is a version
// of V adjusted to be perpendicular to Axis.
#macro VPerp_Adjust(V, Axis)
    vnormalize(vcross(vcross(Axis, V), Axis))
#end

// Projects a vector onto the plane defined by Axis.
// Based on code by Ron Parker
// Use instead VectorReject from vectors.inc
// Kept for backward compatibility.
#macro VProject_Plane(V, Axis)
    #local A = vnormalize(Axis);
    (V - vdot(V, A)*A)
#end

// Projects a vector onto an axis.
// Based on code by Ron Parker
// Use instead VectorProject from vectors.inc
// Kept for backward compatibility.
#macro VProject_Axis(V, Axis)
    (Axis*vdot(V, Axis)/vdot(Axis, Axis))
#end

// Smallest component of V
#macro VMin(V) (F_min3(V.x, V.y, V.z)) #end

// Largest component of V
#macro VMax(V) (F_max3(V.x, V.y, V.z)) #end

// Creates a vector going in the direction of the
// given vector with the specified length
#macro VWith_Len(V, Len) (Len*vnormalize(V)) #end


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="math.inc")

 // --- Currently not 'just' calling macros herein.
 // #fopen FID "SelfTestMathInc.txt" write
 // #write (FID,"\nExpect value followed by value generated.\n")
 // // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
 // #write (FID,"\n---\n")
 // #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))
 // #fclose FID

    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestMathInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
 // // If all macro call tests above work create output for what was tested.
 // #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Center_Object OK\n"))
    #write (FID,"\n")

    // Eval_Pigment
    #declare Pigm00 = pigment { rgb 1/3 }
    #declare Vec00 = <0,0,0>;
    #declare Vec00 = Eval_Pigment(Pigm00,Vec00);
    #write (FID,concat("Eval_Pigment 0.333333 ",str(Vec00.x,9,6)," R\n"))
    #write (FID,concat("Eval_Pigment 0.333333 ",str(Vec00.y,9,6)," G\n"))
    #write (FID,concat("Eval_Pigment 0.333333 ",str(Vec00.z,9,6)," B\n"))
    #ifdef (Pigm00) #undef Pigm00 #end
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // Interpolate
    #declare TS = <0,0,0>;
    #declare TE = <1,2,3>;
    //                           now gs  ge
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,1.0);
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.x,9,6)," x lin\n"))
    #write (FID,concat("Interpolate 1.000000 ",str(Vec00.y,9,6)," y lin\n"))
    #write (FID,concat("Interpolate 1.500000 ",str(Vec00.z,9,6)," z lin\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"---\n")
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,2.0);
    #write (FID,concat("Interpolate 0.250000 ",str(Vec00.x,9,6)," x ^2\n"))
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.y,9,6)," y ^2\n"))
    #write (FID,concat("Interpolate 0.750000 ",str(Vec00.z,9,6)," z ^2\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"---\n")
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,0.0);
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.x,9,6)," x cos\n"))
    #write (FID,concat("Interpolate 1.000000 ",str(Vec00.y,9,6)," y cos\n"))
    #write (FID,concat("Interpolate 1.500000 ",str(Vec00.z,9,6)," z cos\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (TS) #undef TS #end
    #ifdef (TE) #undef TE #end
    #write (FID,"\n")
    // VSqr
    #declare Vec00 = <0,0.5,1,2,-3>;
    #declare Vec00 = VSqr(Vec00);
    #write (FID,concat("VSqr 0.000000 ",str(Vec00.red,9,6)," 0.0\n"))
    #write (FID,concat("VSqr 0.250000 ",str(Vec00.green,9,6)," 0.5\n"))
    #write (FID,concat("VSqr 1.000000 ",str(Vec00.blue,9,6)," 1.0\n"))
    #write (FID,concat("VSqr 4.000000 ",str(Vec00.filter,9,6)," 2.0\n"))
    #write (FID,concat("VSqr 9.000000 ",str(Vec00.transmit,9,6)," -3.0\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // VPow
    #declare Vec00 = <0,-0.5,1>;
    #declare Vec00 = VPow(Vec00,2);
    #write (FID,concat("VPow 0.000000 ",str(Vec00.red,9,6)," 0.0\n"))
    #write (FID,concat("VPow 0.250000 ",str(Vec00.green,9,6)," -0.5\n"))
    #write (FID,concat("VPow 1.000000 ",str(Vec00.blue,9,6)," 1.0\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // VPow5D
    #declare Vec00 = <0,0.5,1,2,-3>;
    #declare Vec00 = VPow5D(Vec00,2);
    #write (FID,concat("VPow5D 0.000000 ",str(Vec00.red,9,6)," 0.0\n"))
    #write (FID,concat("VPow5D 0.250000 ",str(Vec00.green,9,6)," 0.5\n"))
    #write (FID,concat("VPow5D 1.000000 ",str(Vec00.blue,9,6)," 1.0\n"))
    #write (FID,concat("VPow5D 4.000000 ",str(Vec00.filter,9,6)," 2.0\n"))
    #write (FID,concat("VPow5D 9.000000 ",str(Vec00.transmit,9,6)," -3.0\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // VEq
    #declare Vec00 = <0,0.5,1>;
    #declare Vec01 = <0,0.5,1>;
    #declare Vec02 = <9,0.5,1>;
    #if (VEq(Vec00,Vec01))
        #write (FID,concat("VEq 1.000000 ",str(1.0,9,6)," pass\n"))
    #else
        #write (FID,concat("VEq 1.000000 ",str(0.0,9,6)," fail\n"))
    #end
    #if (VEq(Vec00,Vec02))
        #write (FID,concat("VEq 1.000000 ",str(0.0,9,6)," fail\n"))
    #else
        #write (FID,concat("VEq 1.000000 ",str(1.0,9,6)," pass\n"))
    #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // VEq5D
    #declare Vec00 = <0,0.5,1,2,-3>;
    #declare Vec01 = <0,0.5,1,2,-3>;
    #declare Vec02 = <9,0.5,1,2,-3>;
    #if (VEq5D(Vec00,Vec01))
        #write (FID,concat("VEq5D 1.000000 ",str(1.0,9,6)," pass\n"))
    #else
        #write (FID,concat("VEq5D 1.000000 ",str(0.0,9,6)," fail\n"))
    #end
    #if (VEq5D(Vec00,Vec02))
        #write (FID,concat("VEq5D 1.000000 ",str(0.0,9,6)," fail\n"))
    #else
        #write (FID,concat("VEq5D 1.000000 ",str(1.0,9,6)," pass\n"))
    #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // VZero
    #declare Vec00 = <0.0,+0.0,-0.0>;
    #declare Vec01 = <9.9,+0.0,-0.0>;
    #if (VZero(Vec00))
        #write (FID,concat("VZero 1.000000 ",str(1.0,9,6)," pass\n"))
    #else
        #write (FID,concat("VZero 1.000000 ",str(0.0,9,6)," fail\n"))
    #end
    #if (VZero(Vec01))
        #write (FID,concat("VZero 1.000000 ",str(0.0,9,6)," fail\n"))
    #else
        #write (FID,concat("VZero 1.000000 ",str(1.0,9,6)," pass\n"))
    #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #write (FID,"\n")
    // VZero5D
    #declare Vec00 = <0.0,+0.0,-0.0,0,0>;
    #declare Vec01 = <9.9,+0.0,-0.0,0,0>;
    #if (VZero5D(Vec00))
        #write (FID,concat("VZero5D 1.000000 ",str(1.0,9,6)," pass\n"))
    #else
        #write (FID,concat("VZero5D 1.000000 ",str(0.0,9,6)," fail\n"))
    #end
    #if (VZero5D(Vec01))
        #write (FID,concat("VZero5D 1.000000 ",str(0.0,9,6)," fail\n"))
    #else
        #write (FID,concat("VZero5D 1.000000 ",str(1.0,9,6)," pass\n"))
    #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #write (FID,"\n")
    // VLength5D
    #declare Vec00 = <0,0.0,1,2,-3>;
    #declare Vec01 = <1,2,-3>;
    #declare VLen = VLength5D(Vec00);
    #declare vLen = vlength(Vec01);
    #write (FID,concat("VLength5D 3.741657 ",str(VLen,9,6)," <0,0.0,1,2,-3>\n"))
    #write (FID,concat("vlength 3.741657 ",str(vLen,9,6)," <1,2,-3>\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VLen) #undef VLen #end
    #ifdef (vLen) #undef vLen #end
    #write (FID,"\n")
    // VNormalize5D
    #declare Vec00 = <0,0.0,1,2,-3>;
    #declare Vec01 = <1,2,-3>;
    #declare VNrm = VNormalize5D(Vec00);
    #declare vNrm = vnormalize(Vec01);
    #write (FID,concat("VNormalize5D 0.000000 ",str(VNrm.red,9,6)," <0,0.0,1,2,-3>\n"))
    #write (FID,concat("VNormalize5D 0.000000 ",str(VNrm.green,9,6)," \n"))
    #write (FID,concat("VNormalize5D 0.267261 ",str(VNrm.blue,9,6)," \n"))
    #write (FID,concat("VNormalize5D 0.534522 ",str(VNrm.filter,9,6)," \n"))
    #write (FID,concat("VNormalize5D -0.801784 ",str(VNrm.transmit,9,6)," \n"))
    #write (FID,"---\n")
    #write (FID,concat("vnormalize 0.267261 ",str(vNrm.x,9,6)," <1,2,-3>\n"))
    #write (FID,concat("vnormalize 0.534522 ",str(vNrm.y,9,6)," \n"))
    #write (FID,concat("vnormalize -0.801784 ",str(vNrm.z,9,6)," \n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VNrm) #undef VNrm #end
    #ifdef (vNrm) #undef vNrm #end
    #write (FID,"\n")
    // VDot5D
    #declare Vec00 = <0,0.0,1,2,-3>;
    #declare Vec00 = VNormalize5D(Vec00);
    #declare Vec01 = <1,2,-3>;
    #declare Vec01 = vnormalize(Vec01);
    #declare V00 = VDot5D(Vec00,Vec00);
    #declare vV00 = vdot(Vec01,Vec01);
    #write (FID,concat("VDot5D 1.000000 ",str(V00,9,6)," <0,0.0,1,2,-3>\n"))
    #write (FID,concat("vdot 1.000000 ",str(vV00,9,6)," <1,2,-3>\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (V00) #undef V00 #end
    #ifdef (v00) #undef v00 #end
    #write (FID,"\n")
    // VCos_Angle
    // Plus minor testing of new v3.8 tuple #declare style. (Parser test)
    #declare (Vec00,Vec01,Vec02,Vec03,Vec04) =
        (<+1,0,+0>,<-1,0,+0>,<+0,1,+0>,<+0,0,-1>,<+1,0,+1>);
    #declare (A00,A01,A02,A03,A04) =
        (
        VCos_Angle(Vec00,Vec00),VCos_Angle(Vec00,Vec01),
        VCos_Angle(Vec00,Vec02),VCos_Angle(Vec00,Vec03),VCos_Angle(Vec00,Vec04)
        );
    #write (FID,concat("VCos_Angle +1.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VCos_Angle -1.000000 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VCos_Angle +0.000000 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VCos_Angle +0.000000 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VCos_Angle +0.707107 ",str(A04,9,6)," A04\n"))
    #write (FID,"\n")
    // VAngle
    // Plus minor testing of new v3.8 vector components to variables style. (Parser test)
    #declare VecTmp = <VAngle(Vec00,Vec00),VAngle(Vec00,Vec01),VAngle(Vec00,Vec02),
        VAngle(Vec00,Vec03),VAngle(Vec00,Vec04)>;
    #declare <A00,A01,A02,A03,A04> = VecTmp;
    #undef VecTmp
    #write (FID,concat("VAngle +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngle +3.141593 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VAngle +1.570796 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VAngle +1.570796 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VAngle +0.785398 ",str(A04,9,6)," A04\n"))
    #write (FID,"---\n")
    #declare VecD = <+1.73,+1.73,+1.73>;
    #declare VecTmp = <VAngle(VecD,VecD),VAngle(VecD,-VecD)>;
    #declare <A00,A01> = VecTmp;
    #write (FID,concat("VAngle +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngle +3.141593 ",str(A01,9,6)," A01\n"))
    #ifdef (VecD) #undef VecD #end
    #write (FID,"\n")
    // VAngleD
    #declare A00 = VAngleD(Vec00,Vec00);
    #declare A01 = VAngleD(Vec00,Vec01);
    #declare A02 = VAngleD(Vec00,Vec02);
    #declare A03 = VAngleD(Vec00,Vec03);
    #declare A04 = VAngleD(Vec00,Vec04);
    #write (FID,concat("VAngleD +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngleD +180.000000 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VAngleD +90.000000  ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VAngleD +90.000000  ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VAngleD +45.000000  ",str(A04,9,6)," A04\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #write (FID,"\n")
    // VRotation
    // #declare (Vec00,Vec01,Vec02,Vec03,Vec04) =
    //   (<+1,0,+0>,<-1,0,+0>,<+0,1,+0>,<+0,0,-1>,<+1,0,+1>);
    #declare A00 = VRotation(Vec00,Vec00,Vec04);
    #declare A01 = VRotation(Vec00,Vec02,Vec04);
    #declare A02 = VRotation(Vec00,Vec01,Vec04);
    #declare A03 = VRotation(Vec00,Vec00,Vec03);
    #declare A04 = VRotation(Vec00,Vec02,Vec03);
    #declare A05 = VRotation(Vec00,Vec01,Vec03);
    #declare A06 = VRotation(Vec00,<-1,+1,+0>,Vec03);
    #write (FID,concat("VRotation +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VRotation +1.570796   ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VRotation +3.141593   ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VRotation +0.000000   ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VRotation -1.570796   ",str(A04,9,6)," A04\n"))
    #write (FID,concat("VRotation +3.141593   ",str(A05,9,6)," A05\n"))
    #write (FID,concat("VRotation -2.356194   ",str(A06,9,6)," A05\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #ifdef (A05) #undef A05 #end
    #ifdef (A06) #undef A06 #end
    #write (FID,"\n")
    // VRotationD
    #declare A00 = VRotationD(Vec00,Vec00,Vec04);
    #declare A01 = VRotationD(Vec00,Vec02,Vec04);
    #declare A02 = VRotationD(Vec00,Vec01,Vec04);
    #declare A03 = VRotationD(Vec00,Vec00,Vec03);
    #declare A04 = VRotationD(Vec00,Vec02,Vec03);
    #declare A05 = VRotationD(Vec00,Vec01,Vec03);
    #declare A06 = VRotationD(Vec00,<-1,+1,+0>,Vec03);
    #write (FID,concat("VRotationD +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VRotationD +90.000000  ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VRotationD +180.000000 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VRotationD +0.000000   ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VRotationD -90.000000  ",str(A04,9,6)," A04\n"))
    #write (FID,concat("VRotationD +180.000000 ",str(A05,9,6)," A05\n"))
    #write (FID,concat("VRotationD -135.000000 ",str(A06,9,6)," A05\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #ifdef (A05) #undef A05 #end
    #ifdef (A06) #undef A06 #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #ifdef (Vec03) #undef Vec03 #end
    #ifdef (Vec04) #undef Vec04 #end
    #write (FID,"\n")
    // VDist (Changed to #error recommending vlength()).
    // A few vlength tests below.
    #declare Vec00 = <+0.707,+0.707,+0.707>;
    #declare Vec01 = <-0.707,-0.707,-0.707>;
    #declare D00 = vlength(Vec00 - Vec01);
    #declare D01 = vlength(Vec01 - Vec00);
    #declare D02 = vlength(Vec00 - Vec00);
    #write (FID,concat("vlength +2.449120   ",str(D00,9,6)," D00\n"))
    #write (FID,concat("vlength +2.449120   ",str(D01,9,6)," D01\n"))
    #write (FID,concat("vlength +0.000000   ",str(D02,9,6)," D02\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (D00) #undef D00 #end
    #ifdef (D01) #undef D01 #end
    #ifdef (D02) #undef D02 #end
    #write (FID,"\n")
    // VPerp_To_Vector
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_To_Vector(Vec00);
    #declare VP01 = VPerp_To_Vector(Vec01);
    #write (FID,concat("VPerp_To_Vector +1.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #write (FID,concat("VPerp_To_Vector +0.707107   ",str(VP01.x,9,6)," VP01.x\n"))
    #write (FID,concat("VPerp_To_Vector -0.707107   ",str(VP01.y,9,6)," VP01.y\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP01.z,9,6)," VP01.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #ifdef (VP01) #undef VP01 #end
    #write (FID,"\n")
    // VPerp_To_Plane
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_To_Plane(Vec00,Vec01);
    #write (FID,concat("VPerp_To_Plane +0.707107   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_To_Plane +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_To_Plane -0.707107   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VPerp_Adjust
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_Adjust(Vec00,Vec01);
    #write (FID,concat("VPerp_Adjust -0.408248   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_Adjust +0.816497   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_Adjust -0.408248   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VProject_Plane
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VProject_Plane(Vec01,Vec00);
    #write (FID,concat("VProject_Plane +0.707000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VProject_Plane +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VProject_Plane +0.707000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VProject_Axis
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VProject_Axis(Vec01,Vec00);
    #write (FID,concat("VProject_Axis +0.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VProject_Axis +0.707000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VProject_Axis +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VMin, Vmax
    #declare Vec00 = <+1.0,+2.0,+3.0>;
    #declare VarMin = VMin(Vec00);
    #declare VarMax = VMax(Vec00);
    #write (FID,concat("Vmin +1.000000   ",str(VarMin,9,6)," Vmin\n"))
    #write (FID,concat("Vmax +3.000000   ",str(VarMax,9,6)," Vmax\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (VarMin) #undef VarMin #end
    #ifdef (VarMax) #undef VarMax #end
    #write (FID,"\n")
    // VWith_Len
    #declare Vec00 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VWith_Len(Vec00,1.0);
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"


#end // End self testing

#version MATH_INC_TEMP;
#end //math.inc

