//******************************************************************************
///
/// @file base/pov_err.h
///
/// Declaration of identifiers for internal errors.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2020 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

#ifndef POVRAY_BASE_POV_ERR_H
#define POVRAY_BASE_POV_ERR_H

// Module config header file must be the first file included within POV-Ray unit header files
#include "base/configbase.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <stdexcept>
#include <string>

// POV-Ray header files (base module)
//  (none at the moment)

namespace pov_base
{

//##############################################################################
///
/// @defgroup PovBasePovErr Error Handling
/// @ingroup PovBase
///
/// @{

// fatal errors
enum
{
    kNoError                  = 0,
    kNoErr                    = kNoError,
    kParamErr                 = -1,
    kMemFullErr               = -2,
    kOutOfMemoryErr           = kMemFullErr,
    kInvalidDataSizeErr       = -3,              // Few uses.    (have too: povms.cpp -> povms.c)
    kCannotHandleDataErr      = -4,              // povmscpp.cpp (have too: povms.cpp -> povms.c)
    kNullPointerErr           = -5,              // povmscpp.cpp (have too: povms.cpp -> povms.c)
    kChecksumErr              = -6,              // Only.        (have: povms.cpp -> povms.c)
    kParseErr                 = -7,              // Much used. Interesting: mMessageFactory.Error(POV_EXCEPTION(kParseErr, localvsbuffer), "%s", localvsbuffer);
    kCannotOpenFileErr        = -8,              // Various.
    kInvalidDestAddrErr       = -9,              // Not used.
    kCannotConnectErr         = -10,             // Not used.
    kDisconnectedErr          = -11,             // Not used.
    kHostDisconnectedErr      = -12,             // Not used.
    kNetworkDataErr           = -13,             // Not used.
    kNetworkConnectionErr     = -14,             // backendscenedata.cpp   Some code to grab fie from URL... 1/2 uses commented.
    kObjectAccessErr          = -15,             // processrenderoptions.cpp  (have too: povms.cpp -> povms.c)
    kVersionErr               = -16,             // povmscpp.cpp  (have too: povms.cpp -> povms.c)  Object header checks?
    kFileDataErr              = -17,             // Great use with image in/out  bsptree.cpp  truetype.cpp
    kAuthorisationErr         = -18,             // povray.cpp renderbackend.cpp
    kDataTypeErr              = -19,             // parser.cpp throw POV_EXCEPTION(kDataTypeErr, "Unknown file type."); (have too: povms.cpp -> povms.c)
    kTimeoutErr               = -20,             // Only.        (have: povms.cpp -> povms.c)
    kInvalidContextErr        = -21,             // Only.        (have: povms.cpp -> povms.c)
    kIncompleteDataErr        = -22,             // Only.        (have: povms.cpp -> povms.c)
    kInvalidIdentifierErr     = -23,             // renderbackend.cpp
    kCannotHandleRequestErr   = -24,             // Render backend + misc
    kImageAlreadyRenderedErr  = -25,             // Tangled in continue render code.
    kAccessViolationErr       = -26,             // Not used
    kDivideByZeroErr          = -27,             // Not used
    kStackOverflowErr         = -28,             // Not used
    kNativeExceptionErr       = -29,             // Not used
    kInternalLimitErr         = -30,             // In ./source/core/support/simplevector.h  (Barely still used?)
    kUncategorizedError       = -31,             // #define POV_ASSERT_SOFT(expr) SINGLE_STATEMENT( if(!(expr)) throw POV_EXCEPTION_CODE(kUncategorizedError); )
    kNumericalLimitErr        = -32,             // One uncommented use in ./source/base/safemath.h template and there are uses of template.
    kFlagFormErr              = -33              // I added this trying out returned better flag parsing.
                                                 // Getting sigkill and defined string on console, but not this message.
                                                 // Is not getting trapped? Is this the POV_EXCEPTION macro the wrong access?
};

// non fatal errors
enum
{
    kFalseErr                 = 1,               // Used. Especially in options stuff. Really meaning?
    kOutOfSyncErr             = 2,               // Not used directly. Only as below?
    kNotNowErr                = kOutOfSyncErr,
    kQueueFullErr             = 3,               // Not used
    kUserAbortErr             = 4
};

// Macro which builds an exception with file name, line and function.
// Requires an error message text string to be specified.
//
#define POV_EXCEPTION_STRING(str) pov_base::Exception(__func__, __FILE__, __LINE__, (str))

// Macro which builds an exception with file name, line and function.
// Requires an error code to be specified.
// The string will be derived from the error code.
//
#define POV_EXCEPTION_CODE(err) pov_base::Exception(__func__, __FILE__, __LINE__, (err))

// Macro which builds an exception with file name, line and function.
// Requires an error code and string to be specified.
//
#define POV_EXCEPTION(err,str) pov_base::Exception(__func__, __FILE__, __LINE__, (err), (str))

// POV-Ray exception class.
//
class Exception : public std::runtime_error
{
    public:
        // Create a new exception without location information.
        // @param  str  Error message string.
        //
        Exception(const char *str) noexcept :
            std::runtime_error(str),
            xfunction(nullptr), xfile(nullptr), xline(0),
            xcode(0), xcodevalid(false), xfrontendnotified(false)
        {}

        // Create a new exception with location information; looks up message from code.
        // @param  fn   `__FUNCTION__` or `nullptr`
        // @param  fi   `__FILE__`
        // @param  li   `__LINE__`
        // @param  err  Error number.
        //
        Exception(const char *fn, const char *fi, unsigned int li, int err) noexcept :
            std::runtime_error(Exception::lookup_code(err, fi, li)),
            xfunction(fn), xfile(fi), xline(li),
            xcode(err), xcodevalid(true), xfrontendnotified(false)
        {}

        // Create a new exception with code, location information, and an explicit message.
        // @param  fn   `__FUNCTION__` or `nullptr`
        // @param  fi   `__FILE__`
        // @param  li   `__LINE__`
        // @param  str  Error message string.
        //
        Exception(const char *fn, const char *fi, unsigned int li, const char *str) noexcept :
            std::runtime_error(str),
            xfunction(fn), xfile(fi), xline(li),
            xcode(0), xcodevalid(false), xfrontendnotified(false)
        {}

        // Create a new exception with code, location information, and an explicit message.
        // @param  fn   `__FUNCTION__` or `nullptr`
        // @param  fi   `__FILE__`
        // @param  li   `__LINE__`
        // @param  str  Error message string.
        //
        Exception(const char *fn, const char *fi, unsigned int li, const std::string& str) noexcept :
            std::runtime_error(str),
            xfunction(fn), xfile(fi), xline(li),
            xcode(0), xcodevalid(false), xfrontendnotified(false)
        {}

        // Create a new exception with code, location information, and an explicit message.
        // @param  fn   `__FUNCTION__` or `nullptr`
        // @param  fi   `__FILE__`
        // @param  li   `__LINE__`
        // @param  err  Error number.
        // @param  str  Error message string.
        //
        Exception(const char *fn, const char *fi, unsigned int li, int err, const char *str) noexcept :
            std::runtime_error(str),
            xfunction(fn), xfile(fi), xline(li),
            xcode(err), xcodevalid(true), xfrontendnotified(false)
        {}

        // Create a new exception with code, location information, and an explicit message.
        // @param  fn   `__FUNCTION__` or `nullptr`
        // @param  fi   `__FILE__`
        // @param  li   `__LINE__`
        // @param  err  Error number.
        // @param  str  Error message string.
        //
        Exception(const char *fn, const char *fi, unsigned int li, int err, const std::string& str) noexcept :
            std::runtime_error(str),
            xfunction(fn), xfile(fi), xline(li),
            xcode(err), xcodevalid(true), xfrontendnotified(false)
        {}

        // Destructor.
        virtual ~Exception() noexcept override { }

        // Determine the function name where the exception occurred.
        // @return  Function name or `nullptr`.
        //
        const char *function() const { return xfunction; }

        // Determine the name of the source file where the exception occurred.
        // @return  File name or `nullptr`.
        //
        const char *file() const { return xfile; }

        // Determine the line number in the source file where the exception occurred.
        // @return  File line or 0.
        //
        unsigned int line() const { return xline; }

        //  Determine the error code.
        //  @return  Error code (n.b. 0 is a legal errorcode)
        //
        int code() const { return xcode; }

        // Determine the error code or default value if code not supplied at construction.
        // @return  Error code (n.b. 0 is a legal errorcode)
        //
        int code(int defval) const { return xcodevalid ? xcode : defval; }

        // Find out if an errorcode was supplied
        // @return  true if this exception was created with an int param
        //
        bool codevalid() const { return xcodevalid; }

        // Find out if the front-end has been notified about this exception
        // @return  true if the front-end has been told about the exception
        //
        bool frontendnotified(void) const { return xfrontendnotified; }

        // Set the front-end notification flag
        // @param  notified  true to indicate notification has been passed on
        // @return           previous value of notification flag
        //
        bool frontendnotified(bool notified) { bool oldval = xfrontendnotified; xfrontendnotified = notified; return oldval; }

        // Return the error description for the given code.
        // @param  err  Error code.
        // @return      Error string for error code.
        //
        static std::string lookup_code(int err, const char *file = nullptr, unsigned int line = 0);

    private:
        /// Function where the exception occurred
        const char *xfunction;
        /// File name of the source file where the exception occurred
        const char *xfile;
        // Line number in the source file where the exception occurred
        unsigned int xline;
        // Error code
        int xcode;
        // flag so we can tell if this exception was created with an errorcode
        bool xcodevalid;
        // flag that tells us if this exception has been notified to the front-end
        bool xfrontendnotified;
};

/// @}
///
//##############################################################################

}
// end of namespace pov_base

#endif // POVRAY_BASE_POV_ERR_H
