//******************************************************************************
///
/// @file core/material/normal.cpp
///
/// Implementations related to surface normal perturbation.
///
/// The code in this file implements solid texturing functions that perturb the
/// surface normal to create a faux bumpy effect.
///
/// @copyright
/// @parblock
///
/// Some texture ideas garnered from SIGGRAPH '85 Volume 19 Number 3, "An Image
/// Synthesizer" By Ken Perlin. Further Ideas Garnered from "The RenderMan
/// Companion" (Addison Wesley).
///
/// ----------------------------------------------------------------------------
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2021 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/material/normal.h"

// C++ variants of C standard header files
// C++ standard header files
//  (none at the moment)

// POV-Ray header files (base module)
#include "base/pov_err.h"
#include "base/povassert.h"

// POV-Ray header files (core module)
#include "core/material/blendmap.h"
#include "core/material/noise.h"
#include "core/material/pattern.h"
#include "core/material/pigment.h"
#include "core/material/warp.h"
#include "core/math/randomsequence.h"
#include "core/scene/object.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/imageutil.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

// Local preprocessor defines

enum BevyType
{
    kBevy_WrinkleOrig=0,
    kBevy_Wrinkle0a=1,
    kBevy_Wrinkle0b=2,
    kBevy_Wrinkle0c=3,
    kBevy_Wrinkle1=4,
    kBevy_Wrinkle1a=5,
    kBevy_Wrinkle1b=6,
    kBevy_RawMultBump=7,    // This and next could chain too?
    kBevy_RawMultBumpDir=8,
    kBevy_Direction0=98,
    kBevy_Direction1=99
};

// Local typedefs


// Local constants

static const Vector3d Pyramid_Vect [4]=
    { Vector3d( 0.942809041,-0.333333333, 0.0),
      Vector3d(-0.471404521,-0.333333333, 0.816496581),
      Vector3d(-0.471404521,-0.333333333,-0.816496581),
      Vector3d( 0.0        , 1.0        , 0.0) };

static RandomDoubleSequence gPatternRands(0.0, 1.0, 32768);

// Static functions

static void ripples(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& Vector,
        const TraceThreadData *Thread);
static void waves(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& Vector,
        const TraceThreadData *Thread);
static void bumps(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal);
static void dents(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        const TraceThreadData *Thread);
static void bevy(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        Intersection *pInter, const Ray *ray, TraceThreadData *Thread);
static double quilt_cubic(double t, double p1, double p2);
static void grooved(const Vector3d& EPoint, bool First, const TNORMAL *Tnormal, Vector3d& normal);
static void quilted(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal);
static void tiled(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal);
static double Hermite_Cubic(double T1, const Vector2d& UV1, const Vector2d& UV2);
static double Do_Slope_Map(double value, const SlopeBlendMap *Blend_Map);
static void Do_Average_Normals(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        Intersection *Inter, const Ray *ray, TraceThreadData *Thread);
static int PickInCube(const Vector3d& tv, Vector3d& p1);
static void facets(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        TraceThreadData *Thread);
static void RotationAboutVec(const double& angle, const Vector3d& axisVec, Vector3d& normal);


static void ripples(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        const TraceThreadData *Thread)
{
    double length, scalar, index;
    Vector3d point;

    POV_PATTERN_ASSERT(dynamic_cast<RipplesPattern*>(Tnormal->pattern.get()) != nullptr);
    RipplesPattern* pPat = static_cast<RipplesPattern*>(Tnormal->pattern.get());

    for (unsigned int i = 0; i < Thread->numberOfWaves; i++)
    {
        point = EPoint - Thread->waveSources[i];
        length = point.length();

        // Code v1.0->v3.8 set to 1.0 if length zero.
        if (fabs(length) <= gkDBL_epsilon)
            length = gkDBL_epsilon;

        index = length * pPat->frequency + pPat->phase;

        scalar = cycloidal(index) * Tnormal->Amount;

        normal += (scalar / (length * (double)Thread->numberOfWaves)) * point;
    }
}


static void waves(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        const TraceThreadData *Thread)
{
    double length, scalar, index;
    Vector3d point;

    POV_PATTERN_ASSERT(dynamic_cast<WavesPattern*>(Tnormal->pattern.get()) != nullptr);
    WavesPattern* pPat = static_cast<WavesPattern*>(Tnormal->pattern.get());

    for (unsigned int i = 0; i < Thread->numberOfWaves; i++)
    {
        point = EPoint - Thread->waveSources[i];
        length = point.length();

        // Code v1.0->v3.8 set to 1.0 if length zero.
        if (fabs(length) <= gkDBL_epsilon)
            length = gkDBL_epsilon;

        index = length * pPat->frequency * Thread->waveFrequencies[i] + pPat->phase;

        scalar = cycloidal(index) * Tnormal->Amount / Thread->waveFrequencies[i];

        normal += (scalar / (length * (double)Thread->numberOfWaves)) * point;
    }
}


static void bumps(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal)
{
    Vector3d bump_turb;

    DNoise (bump_turb, EPoint);

    normal += Tnormal->Amount * bump_turb;
}


//  Dents is similar to bumps, but uses noise() to control the amount of
//  dnoise() perturbation of the object normal...
//
static void dents(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        const TraceThreadData *Thread)
{
    Vector3d stucco_turb;

    double noise = Noise(EPoint,GetNoiseGen(Tnormal,Thread));

    noise = noise * noise * noise * Tnormal->Amount;

    DNoise(stucco_turb, EPoint);

    normal += noise * stucco_turb;
}


//--- bevy: a collection of normal perturbation methods.

// Easier when trying things than creating many keywords.

// Type 0 the original biased perturbation wrinkles. Avoid it!
// Type 3 Using direction vector to indicate lean.

// -- Code comments for original wrinkles follow.

// Wrinkles - This is my(who?) implementation of the dented() routine, using a
// surface iterative fractal derived from DTurbulence.

// This is a 3-D version (thanks to DNoise()...) of the usual version using the
// singular Noise()...

// Idea garnered from the April 89 Byte Graphics Supplement on RenderMan,
// refined from "The RenderMan Companion, by Steve Upstill of Pixar, (C) 1990
// Addison-Wesley.

static void bevy(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        Intersection *pInter, const Ray *ray, TraceThreadData *Thread)
{
    POV_PATTERN_ASSERT(dynamic_cast<BevyPattern*>(Tnormal->pattern.get()) != nullptr);
    const BevyPattern *pattern = static_cast<BevyPattern*>(Tnormal->pattern.get());

    Vector3d value, value2;
    Vector3d result = Vector3d(0.0, 0.0, 0.0);
    double amount = Tnormal->Amount;
    double scale = 1.0;

    switch (pattern->bevyType)
    {
        // Original biased perturbation wrinkles. Avoid it!
        // Normal inversions possible.
        case kBevy_WrinkleOrig:  // 0
        case kBevy_Wrinkle0a:    // 1
        case kBevy_Wrinkle0b:    // 2
        case kBevy_Wrinkle0c:    // 3
            for (auto i = 0; i < 10; scale *= 2.0, i++)
            {
                value2 = EPoint * scale;
                DNoise(value, value2);

                result[X] += fabs(value[X] / scale);  // fabs causes bias in pnormals all + x,y,z axis
                result[Y] += fabs(value[Y] / scale);
                result[Z] += fabs(value[Z] / scale);
            }
            if (pattern->bevyType == kBevy_WrinkleOrig)
            {
                normal += amount * result; // 0 kBevy_WrinkleOrig
                break;
            }
            else
            {
                result[X] = (value[Z]>=0) ? result[X] : -1.0*result[X]; // Something like this balances bias at cost
                result[Y] = (value[X]>=0) ? result[Y] : -1.0*result[Y]; // a "spreading bias" and at small scales grainy
                result[Z] = (value[Y]>=0) ? result[Z] : -1.0*result[Z]; // while at very big ones kinda camouflage
                if ((pattern->bevyType >= kBevy_Wrinkle0b) && (pattern->bevyType <= kBevy_Wrinkle0c))
                {
                    Vector3d rotV      = (normal + amount * result).normalized();
                    double   Angle     = acos(std::max(-1.0,std::min(1.0,dot(normal.normalized(),rotV))));
                    double   rotLimMag = std::min(M_PI_2*0.997,M_PI_2*amount);
                             Angle     = std::min(+rotLimMag,std::max(-rotLimMag,Angle));
                    if (pattern->bevyType == kBevy_Wrinkle0b)
                    {
                        rotV = cross(rotV,normal.normalized());
                        RotationAboutVec(Angle,rotV,normal); // 2 kBevy_Wrinkle0b
                    }
                    else
                    {
                        RotationAboutVec(Angle,pattern->bevyDirection,normal); // 3 kBevy_Wrinkle0c
                    }
                }
                else
                {
                    normal += amount * result; // 1 kBevy_Wrinkle0a
                }
                break;
            }
        // Like original biased perturbation wrinkles, without directional bias.
        // Normal inversions possible.
        case kBevy_Wrinkle1:  // 4
        case kBevy_Wrinkle1a: // 5
        case kBevy_Wrinkle1b: // 6
            for (auto i = 0; i < 10; scale *= 2.0, i++)
            {
                value2 = EPoint * scale;
                DNoise(value, value2);

                result[X] += value[X] / scale;  // fabs causes bias in pnormals all + x,y,z axis
                result[Y] += value[Y] / scale;
                result[Z] += value[Z] / scale;
            }
            if (pattern->bevyType == kBevy_Wrinkle1)
            {
                normal += amount * result; // 1 kBevy_Wrinkle1
                break;
            }
            else
            {
                if ((pattern->bevyType >= kBevy_Wrinkle1a) && (pattern->bevyType <= kBevy_Wrinkle1b))
                {
                    Vector3d rotV      = (normal + amount * result).normalized();
                    double   Angle     = acos(std::max(-1.0,std::min(1.0,dot(normal.normalized(),rotV))));
                    double   rotLimMag = std::min(M_PI_2*0.997,M_PI_2*amount);
                             Angle     = std::min(+rotLimMag,std::max(-rotLimMag,Angle));
                    if (pattern->bevyType == kBevy_Wrinkle1a)
                    {
                        rotV = cross(rotV,normal.normalized());
                        RotationAboutVec(Angle,rotV,normal); // 5 kBevy_Wrinkle1a
                    }
                    else
                    {
                        RotationAboutVec(Angle,pattern->bevyDirection,normal); // 6 kBevy_Wrinkle1b
                    }
                }
                else
                {
                   throw POV_EXCEPTION_STRING("Internal Error. Type of bevy not valid.");
                }
            break;
            }

        // multiply the incoming raw normal by bump_size amount as perturbed normal
        case kBevy_RawMultBump:
            normal *= amount;
            break;
        // multiply the incoming raw normal by (amount * direction vector) as perturbed normal
        case kBevy_RawMultBumpDir:
            normal *= (amount * pattern->bevyDirection);
            break;
        // Lean toward point specified via 'direction' vector. Use this version
        // to just lean or to use/average with other patterns in the
        // normal_map.
        // ...
        // average
        // normal_map {
        //      [1 bevy type  4 bump_size 0.5]
        //      [1 bevy type 98 bump_size 0.5]
        // }
        // ...
        case kBevy_Direction0:
            normal -= amount * pattern->bevyDirection;
            break;
        // Lean toward point specified via 'direction' vector. Use this version
        // after another pattern with weight 0 in an average normal_map to lean
        // the previously perturbed normal instead of the raw normal vector.
        // ...
        // average
        // normal_map {
        //      [0 bevy type  4 bump_size 0.5]
        //      [1 bevy type 99 bump_size 0.5]
        // }
        // ...
        case kBevy_Direction1:
            if (pInter->PNormal.IsNull())
                normal -= amount * pattern->bevyDirection;
            else
                normal = pInter->PNormal - amount * pattern->bevyDirection;
            break;
        default:
            if (pInter)
            {
                result = (EPoint + pInter->IPoint).normalized();
#if 0
                normal += amount * result;  // bolder (especially neg bump_size), unclamped
#else
                Vector3d rotV      = (normal + amount * result).normalized();
                double   Angle     = acos(std::max(-1.0,std::min(1.0,dot(normal.normalized(),rotV))));
                double   rotLimMag = std::min(M_PI_2*0.997,M_PI_2*amount);
                         Angle     = std::min(+rotLimMag,std::max(-rotLimMag,Angle));
                rotV = cross(rotV,normal.normalized());
                RotationAboutVec(Angle,rotV,normal);
              //RotationAboutVec(Angle,pattern->bevyDirection,normal); // Single rotation vector tends to wash out result.
#endif
            }
            break;
    }
}


static void grooved(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal)
{

    // This pattern similar to tiled except it works in only one direction
    // across the primary surface as determined by the bool First.

    POV_PATTERN_ASSERT(dynamic_cast<GroovedPattern*>(Tnormal->pattern.get()) != nullptr);
    const GroovedPattern *pattern = static_cast<GroovedPattern*>(Tnormal->pattern.get());

    bool flip = pattern->groovedFlip;
    double lo = pattern->groovedLo, hi = pattern->groovedHi;
    if (hi < lo)
        hi = lo + gkMinIsectDepthReturned;

    // rlo and rhi the range of returned values for the cube face. In other
    // words, this the range of values fed to the quilt_cubic curve function.
    const double rlo = 0.0, rhi = 0.5;
    auto lambdaGroovedAdjustValRange = [&] (double val) -> double
    {
        val = ((val-rlo)/(rhi-rlo))*(hi-lo) + lo;
        return (val);
    };

    bool adjRange = ((lo != rlo) || (hi != rhi)) ? true : false;

    double nx = normal.x(), ny = normal.y(), nz = normal.z();

    double x = EPoint.x(), y = EPoint.y(), z = EPoint.z();
    double flx = floor(x), fly = floor(y), flz = floor(z);
    double sx = (x-flx < 0.5) ? -1 : 1, sy = (y-fly < 0.5) ? -1 : 1,
           sz = (z-flz < 0.5) ? -1 : 1; // track sign
    double xm = flx+(ceil(flx+gkMinIsectDepthReturned)-flx)/2.0;
    double ym = fly+(ceil(fly+gkMinIsectDepthReturned)-fly)/2.0;
    double zm = flz+(ceil(flz+gkMinIsectDepthReturned)-flz)/2.0;

    double c0 = pattern->Control0, c1 = pattern->Control1;
    double na = Tnormal->Amount; // matches the normal bump_size (scale value) specified.

    double ax,ay,az;
    if (1)  // Not understood, but on i3 code just below very slightly faster.
    {
        Vector3d nn = normal.normalized();
        ax = fabs(dot(Vector3d(1,0,0),nn)); // alignedWith_
        ay = fabs(dot(Vector3d(0,1,0),nn));
        az = fabs(dot(Vector3d(0,0,1),nn));
    }
    else
    {
        ax = fabs(nx); // alignedWith_
        ay = fabs(ny);
        az = fabs(nz);
    }

    if ((ax >= ay) && (ax >= az))
    {
        if (!flip)
        {
            double vy = fabs(y-ym);
            if (adjRange)
                vy = lambdaGroovedAdjustValRange(vy);
            vy = quilt_cubic(vy,c0,c1)*sy;
          //nx   Let this pass through so no_bump_scale works.
            ny = ny + (na * vy);
          //nz   Only y perturbed
        }
        else
        {
            double vz = fabs(z-zm);
            if (adjRange)
                vz = lambdaGroovedAdjustValRange(vz);
            vz = quilt_cubic(vz,c0,c1)*sz;
          //nx   Let this pass through so no_bump_scale works.
          //ny   Only z perturbed
            nz = nz + (na * vz);
        }
    }
    else if ((ay >= ax) && (ay >= az))
    {
        if (!flip)
        {
            double vx = fabs(x-xm);
            if (adjRange)
                vx = lambdaGroovedAdjustValRange(vx);
            vx = quilt_cubic(vx,c0,c1)*sx;
            nx = nx + (na * vx);
          //ny   Let this pass through so no_bump_scale works.
          //nz   Only x is perturbed
        }
        else
        {
            double vz = fabs(z-zm);
            if (adjRange)
                vz = lambdaGroovedAdjustValRange(vz);
            vz = quilt_cubic(vz,c0,c1)*sz;
          //nx   Only z is perturbed
          //ny   Let this pass through so no_bump_scale works.
            nz = nz + (na * vz);
        }
    }
    else
    {
        if (!flip)
        {
            double vx = fabs(x-xm);
            if (adjRange)
                vx = lambdaGroovedAdjustValRange(vx);
            vx = quilt_cubic(vx,c0,c1)*sx;
            nx = nx + (na * vx);
          //ny   Only x is perturbed
          //nz   Let this pass through so no_bump_scale works.
        }
        else
        {
            double vy = fabs(y-ym);
            if (adjRange)
                vy = lambdaGroovedAdjustValRange(vy);
            vy = quilt_cubic(vy,c0,c1)*sy;
          //nx   Only y is perturbed
            ny = ny + (na * vy);
          //nz   Let this pass through so no_bump_scale works.
        }
    }

    normal = Vector3d(nx,ny,nz);
}


//  Dan Farmer 1994
//
static void quilted(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal)
{

    // If (1) will use a cleaner but slower option than the original.
    // On biscuit.pov it was around +6.8% slower. Faster floor and ceil
    // functions would likely improve speed.

#if (1)
    POV_PATTERN_ASSERT(dynamic_cast<QuiltedPattern*>(Tnormal->pattern.get()) != nullptr);
    const QuiltedPattern *pattern = static_cast<QuiltedPattern*>(Tnormal->pattern.get());

    double lo = pattern->quiltedLo, hi = pattern->quiltedHi;
    if (hi < lo)
        hi = lo + gkMinIsectDepthReturned;

    // rlo and rhi the range of returned values to length at full cube face. In other
    // words this the range of values fed to the quilt_cubic curve function.
    const double rlo = 0.5, rhi = 0.866025403;
    auto lambdaQuiltedAdjustValRange = [&] (double val) -> double
    {
        val = ((val-rlo)/(rhi-rlo))*(hi-lo) + lo;
        return (val);
    };

    bool adjRange = ((lo != rlo) || (hi != rhi)) ? true : false;

    double nx = normal.x(), ny = normal.y(), nz = normal.z();

    double x = EPoint.x(), y = EPoint.y(), z = EPoint.z();
    double flx = floor(x), fly = floor(y), flz = floor(z);
    double sx = (x-flx < 0.5) ? -1 : 1, sy = (y-fly < 0.5) ? -1 : 1,
           sz = (z-flz < 0.5) ? -1 : 1;
    double xm = flx+(ceil(flx+gkMinIsectDepthReturned)-flx)/2.0;
    double ym = fly+(ceil(fly+gkMinIsectDepthReturned)-fly)/2.0;
    double zm = flz+(ceil(flz+gkMinIsectDepthReturned)-flz)/2.0;

    double c0 = pattern->Control0, c1 = pattern->Control1;
    double na = Tnormal->Amount; // This value matches the normal bump_size (scale value) specified.

    double ax,ay,az;
    if (1)  // Not understood, but on i3 just below very slightly faster.
    {
        Vector3d nn = normal.normalized();
        ax = fabs(dot(Vector3d(1,0,0),nn)); // alignedWith_
        ay = fabs(dot(Vector3d(0,1,0),nn));
        az = fabs(dot(Vector3d(0,0,1),nn));
    }
    else
    {
        ax = fabs(nx); // alignedWith_
        ay = fabs(ny);
        az = fabs(nz);
    }

    double len, v;
    if ((ax >= ay) && (ax >= az))
    {
        y = fabs(y-ym);
        z = fabs(z-zm);
        len = Vector3d(0.5,y,z).length();
        if (adjRange)
            len = lambdaQuiltedAdjustValRange(len);
        v   = quilt_cubic(len,c0,c1);
        double vy = v * y * sy;
        double vz = v * z * sz;
        ny = ny + (na * vy);
        nz = nz + (na * vz);
    }
    else if ((ay >= ax) && (ay >= az))
    {
        x = fabs(x-xm);
        z = fabs(z-zm);
        len = Vector3d(x,0.5,z).length();
        if (adjRange)
            len = lambdaQuiltedAdjustValRange(len);
        v   = quilt_cubic(len,c0,c1);
        double vx = v * x * sx;
        double vz = v * z * sz;
        nx = nx + (na * vx);
        nz = nz + (na * vz);
    }
    else
    {
        x = fabs(x-xm);
        y = fabs(y-ym);
        len = Vector3d(x,y,0.5).length();
        if (adjRange)
            len = lambdaQuiltedAdjustValRange(len);
        v   = quilt_cubic(len,c0,c1);
        double vx = v * x * sx;
        double vy = v * y * sy;
        nx = nx + (na * vx);
        ny = ny + (na * vy);
    }
    normal = Vector3d(nx,ny,nz);
#else
   // This the old coding. At best it looks more quilt / pillow like,
   // but it has many stability issues and is less flexible with respect
   // to end result using more extreme control values(1.0 best for this).
   Vector3d vec;
   vec.x() = EPoint.x()-floor(EPoint.x())-0.5;
   vec.y() = EPoint.y()-floor(EPoint.y())-0.5;
   vec.z() = EPoint.z()-floor(EPoint.z())-0.5;

   double t = vec.length();

   POV_PATTERN_ASSERT(dynamic_cast<QuiltedPattern*>(Tnormal->pattern.get()) != nullptr);
   const QuiltedPattern *pattern = static_cast<QuiltedPattern*>(Tnormal->pattern.get());

   t = quilt_cubic(t, pattern->Control0, pattern->Control1);

   vec *= t;

   normal += Tnormal->Amount * vec;
#endif
}


static void tiled(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal)
{

    // This pattern similar in some respects to the updated quilted pattern. It
    // differs in using the two perpendicular half distances along the fave as
    // inputs to the quilt_cubic curve function instead of the length from the
    // cube center (a spherical pattern).

    POV_PATTERN_ASSERT(dynamic_cast<TiledPattern*>(Tnormal->pattern.get()) != nullptr);
    const TiledPattern *pattern = static_cast<TiledPattern*>(Tnormal->pattern.get());

    double lo = pattern->tiledLo, hi = pattern->tiledHi;
    if (hi < lo)
        hi = lo + gkMinIsectDepthReturned;

    // rlo and rhi the range of returned values for the cube face. In other
    // words, this the range of values fed to the quilt_cubic curve function.
    const double rlo = 0.0, rhi = 0.5;
    auto lambdaTiledAdjustValRange = [&] (double val) -> double
    {
        val = ((val-rlo)/(rhi-rlo))*(hi-lo) + lo;
        return (val);
    };

    bool adjRange = ((lo != rlo) || (hi != rhi)) ? true : false;

    double nx = normal.x(), ny = normal.y(), nz = normal.z();

    double x = EPoint.x(), y = EPoint.y(), z = EPoint.z();
    double flx = floor(x), fly = floor(y), flz = floor(z);
    double sx = (x-flx < 0.5) ? -1 : 1, sy = (y-fly < 0.5) ? -1 : 1,
           sz = (z-flz < 0.5) ? -1 : 1; // track sign
    double xm = flx+(ceil(flx+gkMinIsectDepthReturned)-flx)/2.0;
    double ym = fly+(ceil(fly+gkMinIsectDepthReturned)-fly)/2.0;
    double zm = flz+(ceil(flz+gkMinIsectDepthReturned)-flz)/2.0;

    double c0 = pattern->Control0, c1 = pattern->Control1;
    double na = Tnormal->Amount; // matches the normal bump_size (scale value) specified.

    double ax,ay,az;
    if (1)  // Not understood, but on i3 code just below very slightly faster.
    {
        Vector3d nn = normal.normalized();
        ax = fabs(dot(Vector3d(1,0,0),nn)); // alignedWith_
        ay = fabs(dot(Vector3d(0,1,0),nn));
        az = fabs(dot(Vector3d(0,0,1),nn));
    }
    else
    {
        ax = fabs(nx); // alignedWith_
        ay = fabs(ny);
        az = fabs(nz);
    }

    if ((ax >= ay) && (ax >= az))
    {
        double vy = fabs(y-ym), vz = fabs(z-zm);
        if (adjRange)
        {
            vy = lambdaTiledAdjustValRange(vy);
            vz = lambdaTiledAdjustValRange(vz);
        }
        vy = quilt_cubic(vy,c0,c1)*sy;
        vz = quilt_cubic(vz,c0,c1)*sz;
      //nx   Let this pass through so no_bump_scale works.
        ny = ny + (na * vy);
        nz = nz + (na * vz);
    }
    else if ((ay >= ax) && (ay >= az))
    {
        double vx = fabs(x-xm), vz = fabs(z-zm);
        if (adjRange)
        {
            vx = lambdaTiledAdjustValRange(vx);
            vz = lambdaTiledAdjustValRange(vz);
        }
        vx = quilt_cubic(vx,c0,c1)*sx;
        vz = quilt_cubic(vz,c0,c1)*sz;
        nx = nx + (na * vx);
      //ny   Let this pass through so no_bump_scale works.
        nz = nz + (na * vz);
    }
    else
    {
        double vx = fabs(x-xm), vy = fabs(y-ym);
        if (adjRange)
        {
            vx = lambdaTiledAdjustValRange(vx);
            vy = lambdaTiledAdjustValRange(vy);
        }
        vx = quilt_cubic(vx,c0,c1)*sx;
        vy = quilt_cubic(vy,c0,c1)*sy;
        nx = nx + (na * vx);
        ny = ny + (na * vy);
      //nz   Let this pass through so no_bump_scale works.
    }

    normal = Vector3d(nx,ny,nz);
}


//  This pattern is based on the "Crackle" pattern and creates a faceted
//  look on a curved surface.
//
//  Author: Ronald Parker 1998
//
static void facets(const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal, TraceThreadData *Thread)
{
    int      i;
    int      thisseed;
    double   sum, minsum;
    Vector3d sv, tv, dv, t1, add, newnormal, pert;
    double   scale;
    int      UseSquare;
    int      UseUnity;
    double   Metric;

    POV_PATTERN_ASSERT(dynamic_cast<FacetsPattern*>(Tnormal->pattern.get()) != nullptr);
    const FacetsPattern *pattern = static_cast<FacetsPattern*>(Tnormal->pattern.get());

    Vector3d *cv = Thread->Facets_Cube;
    Metric = pattern->facetsMetric;

    UseSquare = (Metric == 2 );
    UseUnity  = (Metric == 1 );

    normal.normalize();

    if (pattern->facetsCoords)
    {
        tv = EPoint;
    }
    else
    {
        tv = normal;
    }

    if (pattern->facetsSize < 1e-6)
    {
        scale = 1e6;
    }
    else
    {
        scale = 1. / pattern->facetsSize;
    }

    tv *= scale;

    // Check to see if the input point is in the same unit cube as the last
    // call to this function, to use cache of cubelets for speed.

    thisseed = PickInCube(tv, t1);

    if (thisseed != Thread->Facets_Last_Seed)
    {

        // No, not same unit cube.  Calculate the random points for this new
        // cube and its 80 neighbours which differ in any axis by 1 or 2. Why
        // distance of 2? If there is 1 point in each cube, located randomly,
        // it is possible for the closest random point to be in the cube 2
        // over, or the one two over and one up.  It is NOT possible for it to
        // be two over and two up. Picture a 3x3x3 cube with 9 more cubes
        // glued onto each face.

        // Now store a points for this cube and each of the 80 neighbour cubes.

        int cvc = 0;

        for (add[X] = -2.0; add[X] < 2.5; add[X] +=1.0)
        {
            for (add[Y] = -2.0; add[Y] < 2.5; add[Y] += 1.0)
            {
                for (add[Z] = -2.0; add[Z] < 2.5; add[Z] += 1.0)
                {
                    /* For each cubelet in a 5x5 cube. */

                    if ((fabs(add[X])>1.5)+(fabs(add[Y])>1.5)+(fabs(add[Z])>1.5) <= 1.0)
                    {
                        /* Yes, it's within a 3d knight move away. */

                        sv = tv + add;

                        PickInCube(sv, t1);

                        cv[cvc] = t1;
                        cvc++;
                    }
                }
            }
        }

        Thread->Facets_Last_Seed = thisseed;
        Thread->Facets_CVC = cvc;
    }

    // Find the point with the shortest distance from the input point.

    dv = cv[0] - tv;
    if ( UseSquare )
    {
        minsum  = dv.lengthSqr();
    }
    else if ( UseUnity )
    {
        minsum = dv[X]+dv[Y]+dv[Z];
    }
    else
    {
        minsum = pow(fabs(dv[X]), Metric)+
                 pow(fabs(dv[Y]), Metric)+
                 pow(fabs(dv[Z]), Metric);
    }

    newnormal = cv[0];

    /* Loop for the 81 cubelets to find closest. */

    for (i = 1; i < Thread->Facets_CVC; i++)
    {
        dv = cv[i] - tv;

        if ( UseSquare )
        {
            sum  = dv.lengthSqr();
        }
        else
        {
            if ( UseUnity )
            {
                sum = dv[X]+dv[Y]+dv[Z];
            }
            else
            {
                sum = pow(fabs(dv[X]), Metric)+
                      pow(fabs(dv[Y]), Metric)+
                      pow(fabs(dv[Z]), Metric);
            }
        }

        if (sum < minsum)
        {
            minsum = sum;
            newnormal = cv[i];
        }
    }

    if (pattern->facetsCoords)
    {
        DNoise( pert, newnormal );
        sum = dot(pert, normal);
        newnormal = normal * sum;
        pert -= newnormal;
        normal += pattern->facetsCoords * pert;
    }
    else
    {
        normal = newnormal;
    }
}


// Allocate memory for new Tnormal and initialize it to system default values.
// Returns pointer to the created Tnormal.
//
TNORMAL *Create_Tnormal ()
{
    TNORMAL *New;

    New = new TNORMAL;

    Init_TPat_Fields(New);

    New->Amount = 0.5;

    /* NK delta */
    New->Delta = 0.02; /* this is a good starting point for delta */

    return (New);
}


TNORMAL *Copy_Tnormal (TNORMAL *Old)
{
    TNORMAL *New;

    if (Old != nullptr)
    {
        New = Create_Tnormal();

        Copy_TPat_Fields (New, Old);
        New->Blend_Map = Copy_Blend_Map(Old->Blend_Map);

        New->Amount = Old->Amount;
        New->Delta = Old->Delta;
    }
    else
    {
        New = nullptr;
    }

    return (New);
}


void Destroy_Tnormal(TNORMAL *Tnormal)
{
    if (Tnormal != nullptr)
        delete Tnormal;
}


void Post_Tnormal (TNORMAL *Tnormal)
{
    GenericNormalBlendMapPtr Map;

    if (Tnormal != nullptr)
    {
        if (Tnormal->Flags & POST_DONE)
        {
            return;
        }

        if (Tnormal->Type == NO_PATTERN)
        {
            throw POV_EXCEPTION_STRING("No normal type given.");
        }

        Tnormal->Flags |= POST_DONE;

        if ((Map = Tnormal->Blend_Map) != nullptr)
        {
            Map->Post((Tnormal->Flags & DONT_SCALE_BUMPS_FLAG) != 0);
        }
    }
}

void SlopeBlendMap::Post(bool dontScaleBumps)
{
    POV_BLEND_MAP_ASSERT(Type == kBlendMapType_Slope);
}

void NormalBlendMap::Post(bool dontScaleBumps)
{
    POV_BLEND_MAP_ASSERT(Type == kBlendMapType_Normal);
    for(auto i = Blend_Map_Entries.begin(); i != Blend_Map_Entries.end(); i++)
    {
        if (dontScaleBumps)
            i->Vals->Flags |= DONT_SCALE_BUMPS_FLAG;
        Post_Tnormal(i->Vals);
    }
}


//   Added intersectin parameter for UV mapping - NK 1998
//
void Perturb_Normal(Vector3d& Layer_Normal, const TNORMAL *Tnormal, const Vector3d& EPoint,
        Intersection *Intersection, const Ray *ray, TraceThreadData *Thread)
{
    Vector3d TPoint,P1;
    double value1,Amount;
    std::shared_ptr<NormalBlendMap> Blend_Map;

    if (Tnormal == nullptr)
    {
        return;
    }

    /* If normal_map present, use it and return */

    Blend_Map = std::dynamic_pointer_cast<NormalBlendMap>(Tnormal->Blend_Map); // TODO Avoid dynamic...

    if (Blend_Map != nullptr)
    {
        if (Tnormal->Type == UV_MAP_PATTERN)
        {
            Vector2d UV_Coords;

            /* Don't bother warping, simply get the UV vect of the intersection */
            Intersection->Object->UVCoord(UV_Coords, Intersection);
            TPoint[X] = UV_Coords[U];
            TPoint[Y] = UV_Coords[V];
            TPoint[Z] = 0;

            Perturb_Normal(Layer_Normal,Blend_Map->Blend_Map_Entries[0].Vals,TPoint,Intersection,ray,Thread);
            Layer_Normal.normalize();
            Intersection->PNormal = Layer_Normal; /* -hdf- June 98 */

            return;
        }
        else if (Tnormal->Type != AVERAGE_PATTERN)
        {
            const NormalBlendMapEntry *Prev, *Cur;
            double prevWeight, curWeight;

            /* NK 19 Nov 1999 added Warp_EPoint */
            Warp_EPoint(TPoint, EPoint, Tnormal);
            value1 = Evaluate_TPat(Tnormal, TPoint, Intersection, ray, Thread);

            Blend_Map->Search(value1,Prev,Cur,prevWeight,curWeight);

            Warp_Normal(Layer_Normal,Layer_Normal, Tnormal, Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));
            P1 = Layer_Normal;

            Warp_EPoint(TPoint, EPoint, Tnormal);

            Perturb_Normal(Layer_Normal,Cur->Vals,TPoint,Intersection,ray,Thread);

            if (Prev != Cur)
            {
                Perturb_Normal(P1,Prev->Vals,TPoint,Intersection,ray,Thread);

                Layer_Normal = prevWeight * P1 + curWeight * Layer_Normal;
            }

            UnWarp_Normal(Layer_Normal,Layer_Normal, Tnormal, Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));

            Layer_Normal.normalize();

            Intersection->PNormal = Layer_Normal; /* -hdf- June 98 */

            return;
        }
        // TODO - what if Tnormal->Type == AVERAGE_PATTERN?
    }

    /* No normal_map. */

    if (Tnormal->Type <= LAST_SPECIAL_NORM_PATTERN)
    {
        Warp_Normal(Layer_Normal,Layer_Normal, Tnormal,
                    Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));

        Warp_EPoint (TPoint, EPoint, Tnormal);

        switch (Tnormal->Type)
        {
            case BEVY_PATTERN:      bevy        (TPoint, Tnormal, Layer_Normal, Intersection, ray, Thread); break;
            case BITMAP_PATTERN:    bump_map    (TPoint, Tnormal, Layer_Normal);            break;
            case BUMPS_PATTERN:     bumps       (TPoint, Tnormal, Layer_Normal);            break;
            case DENTS_PATTERN:     dents       (TPoint, Tnormal, Layer_Normal, Thread);    break;
            case RIPPLES_PATTERN:   ripples     (TPoint, Tnormal, Layer_Normal, Thread);    break;
            case WAVES_PATTERN:     waves       (TPoint, Tnormal, Layer_Normal, Thread);    break;
            case GROOVED_PATTERN:   grooved     (TPoint, Tnormal, Layer_Normal);            break;
            case QUILTED_PATTERN:   quilted     (TPoint, Tnormal, Layer_Normal);            break;
            case TILED_PATTERN:     tiled       (TPoint, Tnormal, Layer_Normal);            break;
            case FACETS_PATTERN:    facets      (TPoint, Tnormal, Layer_Normal, Thread);    break;
            case AVERAGE_PATTERN:   Do_Average_Normals (TPoint, Tnormal, Layer_Normal, Intersection, ray, Thread); break;
            default:
                throw POV_EXCEPTION_STRING("Normal pattern not yet implemented.");
        }

        UnWarp_Normal(Layer_Normal,Layer_Normal, Tnormal,
                      Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));
    }
    else
    {

        // Value for map patterns. The pattern is sampled in a pyramid of
        // locations and these are turned into a perturbation vector which is
        // added to the incoming surface normal.

        // TODO  Look to change from dynamic_pointer_cast for better performance.
        std::shared_ptr<SlopeBlendMap> slopeMap = std::dynamic_pointer_cast<SlopeBlendMap>(Tnormal->Blend_Map);

        Warp_Normal(Layer_Normal,Layer_Normal, Tnormal,
                    Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));

        // TODO FIXME - two magic fudge factors
        Amount  = Tnormal->Amount * -5.0; // fudge factor
        Amount *= 0.02 / Tnormal->Delta;  // NK delta

        // warp the center point first - this is the last warp. TODO. Assuming this applies all warps in list?
        Warp_EPoint(TPoint,EPoint,Tnormal);

        for(int i=0; i<=3; i++)
        {
            P1 = TPoint + Tnormal->Delta * Pyramid_Vect[i]; /* NK delta */
            value1 = Do_Slope_Map(Evaluate_TPat(Tnormal, P1, Intersection, ray, Thread), slopeMap.get());
            Layer_Normal += (value1*Amount) * Pyramid_Vect[i];
        }

        UnWarp_Normal(Layer_Normal,Layer_Normal,Tnormal,
                      Test_Flag(Tnormal,DONT_SCALE_BUMPS_FLAG));
    }

    // If the incoming intersection pointer valid, copy the updated normal to
    // the perturbed normal in the intersection structure.

    if ( Intersection )
        Intersection->PNormal = Layer_Normal; /* -hdf- June 98 */
}

//----------- Some static functions.

//---- quilt_cubic  (moved here from pattern.h/.cpp)
// Originally coded as:
//
//    val= (DBL)(itcubed*n1+(tcubed)*n2+3*t*(itsqrd)*p1+3*(tsqrd)*(it)*p2);
//
// re-written by CEY to optimise because n1=0 n2=1 always. DBL
// itcubed=it*itsqrd;
//
// const DBL INV_SQRT_3_4 = 1.154700538;
// TODO g++ offers constant versions of basic math functions, but not standard.  Jan09_2021 fixable?
//
// #if (__cpp_constexpr >= 201304)
// constexpr double INV_SQRT_3_4 = 1.0/std::sqrt(3.0/4.0);
// ...
const double INV_SQRT_3_4 = 1.0/std::sqrt(3.0/4.0);

static double quilt_cubic(double t, double p1, double p2)
{
    double it=(1-t);
    double itsqrd=it*it;
    double tsqrd=t*t;
    double tcubed=t*tsqrd;
    double val = (tcubed + 3.0*t*itsqrd*p1 + 3.0*tsqrd*it*p2) * INV_SQRT_3_4;

    return(val);
}


static double Do_Slope_Map (double value, const SlopeBlendMap *Blend_Map)
{
    double prevWeight, curWeight;
    const SlopeBlendMapEntry *Prev, *Cur;

    if (Blend_Map == nullptr)
    {
        return(value);
    }

    Blend_Map->Search(value, Prev, Cur, prevWeight, curWeight);

    if (Prev == Cur)
    {
        return(Cur->Vals[0]);
    }

    return (Hermite_Cubic(curWeight, Prev->Vals, Cur->Vals));
}


static double Hermite_Cubic(double T1, const Vector2d& UV1, const Vector2d& UV2)
{
    double TT=T1*T1;
    double TTT=TT*T1;
    double rv;       // Simplified equation for Symantec.
                     // TODO. Is there a better/faster representation?
                     // Symantec stopped compiler work on Apple move to OSX in the early 2000s.

    rv  = TTT*(UV1[1]+UV2[1]+2.0*(UV1[0]-UV2[0]));
    rv += -TT*(2.0*UV1[1]+UV2[1]+3.0*(UV1[0]-UV2[0]));
    rv += T1*UV1[1] +UV1[0];

    return (rv);
}


//   Added intersection parameter for UV mapping - NK 1998
//
static void Do_Average_Normals (const Vector3d& EPoint, const TNORMAL *Tnormal, Vector3d& normal,
        Intersection *Inter, const Ray *ray, TraceThreadData *Thread)
{
    Tnormal->Blend_Map->ComputeAverage(EPoint, normal, Inter, ray, Thread);
}


//---  PickInCube(tv, p1)
//
// Input tv is input test vector(point). Returns the base generated seed based
// upon the containing cube and updates the p1 vector with three random values
// generated fro that base and two increments of the base seed.
//
// This code originally written for crackle which has since adopted a faster
// method. The only remaining use is the the nomral perturbation facets pattern.
//
// Author: Jim McElhiney
//
// Original descriptsion:
//
// A subroutine to go with crackle.
//
// Pick a random point in the same unit-sized cube as tv, in a
// predictable way, so that when called again with another point in
// the same unit cube, p1 is picked to be the same.
//
static int PickInCube(const Vector3d& tv, Vector3d& p1)
{
    size_t seed;
    Vector3d flo;

    flo[X] = floor(tv.x() - gkMinIsectDepthReturned);
    flo[Y] = floor(tv.y() - gkMinIsectDepthReturned);
    flo[Z] = floor(tv.z() - gkMinIsectDepthReturned);

    seed = size_t(Hash3d((int)flo.x(), (int)flo.y(), (int)flo.z()));

    p1[X] = flo.x() + gPatternRands(seed);
    p1[Y] = flo.y() + gPatternRands(seed + 1);
    p1[Z] = flo.z() + gPatternRands(seed + 2);

    return (int)seed;
}


SlopeBlendMap::SlopeBlendMap() : BlendMap<Vector2d>(kBlendMapType_Slope) {}

SlopeBlendMap::~SlopeBlendMap()
{
    POV_BLEND_MAP_ASSERT(Type == kBlendMapType_Slope);
}


NormalBlendMap::NormalBlendMap() : BlendMap<TNORMAL*>(kBlendMapType_Normal) {}

NormalBlendMap::~NormalBlendMap()
{
    POV_BLEND_MAP_ASSERT(Type == kBlendMapType_Normal);
    for(auto i = Blend_Map_Entries.begin(); i != Blend_Map_Entries.end(); i++)
        Destroy_Tnormal(i->Vals);
}

void SlopeBlendMap::ComputeAverage (const Vector3d& EPoint, Vector3d& normal, Intersection *Inter,
        const Ray *ray, TraceThreadData *Thread)
{
    POV_BLEND_MAP_ASSERT(false);
}

void NormalBlendMap::ComputeAverage (const Vector3d& EPoint, Vector3d& normal, Intersection *Inter,
        const Ray *ray, TraceThreadData *Thread)
{
    double Value;
    double Total = 0.0;
    Vector3d V1,V2;

    POV_BLEND_MAP_ASSERT(Type == kBlendMapType_Normal);

    V1 = Vector3d(0.0, 0.0, 0.0);

    for(auto i = Blend_Map_Entries.cbegin(); i != Blend_Map_Entries.cend(); i++)
    {
        Value = i->value;

        V2 = normal;

        Perturb_Normal(V2,i->Vals,EPoint,Inter,ray,Thread);

        V1 += Value * V2;

        Total += Value;
    }

    normal = V1 / Total;
}


static void RotationAboutVec(const double& angle, const Vector3d& axisVec, Vector3d& normal)
{
    TRANSFORM Trans;
#if 0
    Vector3d rotVec = axisVec.normalized();
    double   cosA   = cos(angle);
    double   sinA   = sin(angle);
    MATRIX Matrix;  // Thought local would be faster, but in testing not. See comment below.
    double rx = rotVec.x(), ry = rotVec.y(), rz = rotVec.z();
    Matrix[0][0]=rx*rx*(1.0-cosA)+cosA;    Matrix[0][1]=rx*ry*(1.0-cosA)+rz*sinA; Matrix[0][2]=rx*rz*(1.0-cosA)-ry*sinA; Matrix[0][3]=0.0;
    Matrix[1][0]=rx*ry*(1.0-cosA)-rz*sinA; Matrix[1][1]=ry*ry*(1.0-cosA)+cosA;    Matrix[1][2]=ry*rz*(1.0-cosA)+rx*sinA; Matrix[1][3]=0.0;
    Matrix[2][0]=rx*rz*(1.0-cosA)+ry*sinA; Matrix[2][1]=ry*rz*(1.0-cosA)-rx*sinA; Matrix[2][2]=rz*rz*(1.0-cosA)+cosA;    Matrix[2][3]=0.0;
    Matrix[3][0]=0.0;                      Matrix[3][1]=0.0;                      Matrix[3][2]=0.0;                      Matrix[3][3]=1.0;
    Compute_Matrix_Transform(&Trans, Matrix);
#else
    Compute_Axis_Rotation_Transform(&Trans,axisVec,angle);  // In testing forced +a0.0, simple scene, -2.8% faster.
#endif
    MTransDirection(normal,normal,&Trans);
}


}
// end of namespace pov
