//******************************************************************************
///
/// @file core/shape/isosurface.cpp
///
/// Implementation of the isosurface geometric primitive.
///
/// @author D.Skarda, T.Bily (original code)
/// @author R.Suzuki (modifications)
/// @author Thorsten Froehlich (porting to POV-Ray v3.5)
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2020 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/shape/isosurface.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <algorithm>

// POV-Ray header files (base module)
#include "base/messenger.h"

// POV-Ray header files (core module)
#include "core/math/matrix.h"
#include "core/render/ray.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

using std::min;
using std::max;

struct IsosurfaceCache final
{
    const IsoSurface *current;
    Vector3d Pglobal;
    Vector3d Dglobal;
    double fmax;
    IsosurfaceCache();
};

IsosurfaceCache::IsosurfaceCache() :
    current(nullptr)
{}

struct ISO_ThreadData final
{
    IsosurfaceCache cache;
    GenericScalarFunctionInstance* pFn;
    double Vlength;
    double tl;
    int Inv3;
};

// Local preprocessor defines

struct ISO_Max_Gradient final
{
    double max_gradient, gradient;

    ISO_Max_Gradient() :
        max_gradient(0.0),
        gradient(0.0),
        mRefCounter(0)
    {}

    bool IsShared() const { return mRefCounter > 1; }

private:
    mutable size_t mRefCounter;
    friend void intrusive_ptr_add_ref(ISO_Max_Gradient* f);
    friend void intrusive_ptr_release(ISO_Max_Gradient* f);
};

inline void intrusive_ptr_add_ref(ISO_Max_Gradient* f) { ++f->mRefCounter; }
inline void intrusive_ptr_release(ISO_Max_Gradient* f) { if (!(--f->mRefCounter)) delete f; }


bool IsoSurface::All_Intersections(const Ray& ray, IStack& Depth_Stack, TraceThreadData *Thread)
{
    int Side1 = 0, Side2 = 0, itrace = 0;
    double Depth1 = 0.0, Depth2 = 0.0;
    BasicRay New_Ray;
    Vector3d IPoint;
    Vector3d Plocal, Dlocal;
    double tmax = 0.0, tmin = 0.0, tmp = 0.0;
    double maxg = max_gradient;
    int i = 0 ; /* count of intervals in stack - 1      */
    int IFound = false;
    int begin = 0, end = 0;
    bool in_shadow_test = false;
    Vector3d VTmp;
    thread_local ISO_ThreadData isoData;

    Thread->Stats()[Ray_IsoSurface_Bound_Tests]++;

    if (container->Intersect(ray, Trans, Depth1, Depth2, Side1, Side2)) /* IsoSurface_Bound_Tests */
    {
        Thread->Stats()[Ray_IsoSurface_Bound_Tests_Succeeded]++;

        GenericScalarFunctionInstance fn(Function, Thread);

        in_shadow_test = ray.IsShadowTestRay();

        if (Depth1 < 0.0)
        {
            Depth1 = 0.0;
        }

        if (Trans != nullptr)
        {
            MInvTransPoint(Plocal, ray.Origin, Trans);
            MInvTransDirection(Dlocal, ray.Direction, Trans);
        }
        else
        {
            Plocal = ray.Origin;
            Dlocal = ray.Direction;
        }

        isoData.Inv3 = 1;

        if (closed != false)
        {
            VTmp = Plocal + Depth1 * Dlocal;
            tmp = EvaluateFunction(fn, VTmp);
            if (Depth1 > accuracy)
            {
                if (tmp < 0.0) // The ray hits the bounding shape
                {
                    IPoint = ray.Evaluate(Depth1);
                    if (Clip.empty() || Point_In_Clip(IPoint, Clip, Thread))
                    {
                        Depth_Stack->push(Intersection(Depth1, IPoint, this, 1, Side1));
                        IFound = true;
                        itrace++;
                        isoData.Inv3 *= -1;
                    }
                }
            }
            else
            {
                if (tmp < (maxg * accuracy * 4.0))
                {
                    Depth1 = accuracy * 5.0;
                    VTmp = Plocal + Depth1 * Dlocal;
                    if (IsInside(fn, VTmp))
                    {
                        // Change sign of the function (IPoint is in the bounding shape.)
                        isoData.Inv3 = -1;
                    }
                }
                VTmp = Plocal + Depth2 * Dlocal;
                if (IsInside (fn, VTmp))
                {
                    IPoint = ray.Evaluate(Depth2);
                    if (Clip.empty() || Point_In_Clip(IPoint, Clip, Thread))
                    {
                        Depth_Stack->push(Intersection(Depth2, IPoint, this, 1, Side2));
                        IFound = true;
                    }
                }
            }
        }

        /*  METHOD 2   by R. Suzuki */
        tmax = Depth2 = min(Depth2, BOUND_HUGE);
        tmin = Depth1 = min(Depth2, Depth1);
        if ((tmax - tmin) < accuracy)
        {
            if (IFound)
            {
                Depth_Stack->pop(); // we added an intersection already, so we need to undo that
            }
            return (false);
        }
        Thread->Stats()[Ray_IsoSurface_Tests]++;
        if ((Depth1 < accuracy) && (isoData.Inv3 == 1))
        {
            // IPoint is on the isosurface
            VTmp = Plocal + tmin * Dlocal;
            if (EvaluateAbs (fn, VTmp) < (maxg * accuracy * 4.0))
            {
                tmin = accuracy * 5.0;
                VTmp = Plocal + tmin * Dlocal;
                if (IsInside (fn, VTmp))
                {
                    // Change the sign and go into the isosurface
                    isoData.Inv3 = -1;
                }
            }
        }

        isoData.pFn = &fn;

        for (; itrace < max_trace; itrace++)
        {
            if (Function_Find_Root(isoData, Plocal, Dlocal, &tmin, &tmax, maxg, in_shadow_test, Thread) == false)
            {
                break;
            }
            else
            {
                IPoint = ray.Evaluate(tmin);
                if (Clip.empty() || Point_In_Clip(IPoint, Clip, Thread))
                {
                    Depth_Stack->push(Intersection(tmin, IPoint, this, 0, 0 /*Side1*/));
                    IFound = true;
                }
            }
            tmin += accuracy * 5.0;
            if ((tmax - tmin) < accuracy)
            {
                break;
            }
            isoData.Inv3 *= -1;
        }

        if (IFound)
        {
            Thread->Stats()[Ray_IsoSurface_Tests_Succeeded]++;
        }

        isoData.pFn = nullptr;
    }

    return (IFound);
}


bool IsoSurface::Inside(const Vector3d& IPoint, TraceThreadData *Thread) const
{
    Vector3d New_Point;

    // Transform the point into containing shape space.
    if (Trans != nullptr)
    {
        MInvTransPoint(New_Point, IPoint, Trans);
    }
    else
    {
        New_Point = IPoint;
    }

    if (!container->Inside(New_Point))
    {
        return (Test_Flag(this, INVERTED_FLAG));
    }

    GenericScalarFunctionInstance fn(Function, Thread);
    if (!IsInside (fn, New_Point))
    {
        return (Test_Flag(this, INVERTED_FLAG));
    }

    // Inside the containing shape.
    return (!Test_Flag(this, INVERTED_FLAG));
}


double IsoSurface::GetPotential (const Vector3d& globalPoint, bool subtractThreshold, TraceThreadData *pThread) const
{
    Vector3d localPoint;

    if (Trans != nullptr)
    {
        MInvTransPoint (localPoint, globalPoint, Trans);
    }
    else
    {
        localPoint = globalPoint;
    }

    double potential = GenericScalarFunctionInstance(Function, pThread).Evaluate (localPoint);
    if (subtractThreshold)
    {
        potential -= threshold;
    }

    if (Test_Flag (this, INVERTED_FLAG))
    {
        return -potential;
    }
    else
    {
        return  potential;
    }
}


void IsoSurface::Normal(Vector3d& Result, Intersection *Inter, TraceThreadData *Thread) const
{
    Vector3d New_Point, TPoint;
    double funct;
    bool containerHit = Inter->i1;

    if (containerHit)
    {
        container->Normal(Inter->IPoint, Trans, Inter->i2, Result);
    }
    else
    {
        GenericScalarFunctionInstance fn(Function, Thread);

        /* Transform the point into the isosurface space */
        if (Trans != nullptr)
        {
            MInvTransPoint(New_Point, Inter->IPoint, Trans);
        }
        else
        {
            New_Point = Inter->IPoint;
        }

        TPoint = New_Point;
        funct = fn.Evaluate(TPoint);

        TPoint     = New_Point;
        TPoint[X] += accuracy;
        Result[X]  = fn.Evaluate(TPoint) - funct;

        TPoint     = New_Point;
        TPoint[Y] += accuracy;
        Result[Y]  = fn.Evaluate(TPoint) - funct;

        TPoint     = New_Point;
        TPoint[Z] += accuracy;
        Result[Z]  = fn.Evaluate(TPoint) - funct;

        if ((Result[X] == 0) && (Result[Y] == 0) && (Result[Z] == 0))
        {
            Result[X] = 1.0;
        }
        Result.normalize();

        // Transform the point into the boxes space.
        if (Trans != nullptr)
        {
            MTransNormal(Result, Result, Trans);
            Result.normalize();
        }
    }
}


void IsoSurface::Translate(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void IsoSurface::Rotate(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void IsoSurface::Scale(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void IsoSurface::Transform(const TRANSFORM* tr)
{
    if (Trans == nullptr)
    {
        Trans = Create_Transform();
    }

    Compose_Transforms(Trans, tr);

    Compute_BBox();
}


IsoSurface::IsoSurface() :
    ObjectBase(ISOSURFACE_OBJECT),
    jitter(false), report(true)
{
    container = std::shared_ptr<ContainedByShape>(new ContainedByBox());

    Make_BBox(BBox, -1.0, -1.0, -1.0, 2.0, 2.0, 2.0);

    Trans = Create_Transform();

    Function  = nullptr;

    accuracy     = 0.001;
    closed       = true;
    gradient     = 0.0;
    max_gradient = 1.1;
    max_trace    = 1;
    threshold    = 0.0;

    mginfo = boost::intrusive_ptr<ISO_Max_Gradient>(new ISO_Max_Gradient());
}


ObjectPtr IsoSurface::Copy()
{
    IsoSurface *New = new IsoSurface();
    Destroy_Transform(New->Trans);
    *New = *this;

    New->Function = Function->Clone();
    New->Trans = Copy_Transform(Trans);

    New->mginfo = mginfo;

    New->jitter = jitter;
    New->report = report;

    New->container = std::shared_ptr<ContainedByShape>(container->Copy());

    return (New);
}


IsoSurface::~IsoSurface()
{
    delete Function;
}


// Input : messenger: destination of messages
//
// Author Chris Cason
//
// If any max_gradient messages need to be sent to the user, they are
// dispatched via the supplied GenericMessenger.
//
void IsoSurface::DispatchShutdownMessages(GenericMessenger& messenger) {

    // Suppress max gradient warnings?  With specialized functions often have
    // jumps /large gradients well away from the surface. The inbuilt function
    // f_elliptical_sphrswp() a good example - with faster two step version
    // turning off the report is always necessary.
    //
    if (report == false)
    {
        return;
    }

    // TODO Of minor import, but, why the max operations below? Threading
    // concern?  Perhaps that max_gradient variable with evaluate functionality
    // was marked volatile?
    //
    mginfo->gradient = max(gradient, mginfo->gradient);
    mginfo->max_gradient = max(max_gradient, mginfo->max_gradient);

    if (mginfo->IsShared())
    {
        // Other instances shall take care of reporting the max gradient.
        mginfo.reset();
        return;
    }

    const CustomFunctionSourceInfo* fnInfo = Function->GetSourceInfo();

    if (fnInfo != nullptr)
    {
        // Only show the warning if necessary!
        // BTW, not being too picky here is a feature and not a bug ;-)  [trf]
        if ((mginfo->gradient > EPSILON) && (mginfo->max_gradient > EPSILON))
        {
            double diff = mginfo->max_gradient - mginfo->gradient;
            double prop = fabs(mginfo->max_gradient / mginfo->gradient);

            if (((prop <= 0.9) && (diff <= -0.5)) ||
                (((prop <= 0.95) || (diff <= -0.1)) && (mginfo->max_gradient < 10.0)))
            {
                messenger.WarningAt(kWarningGeneral, *fnInfo,
                                    "The maximum gradient found was %0.3f, but max_gradient of the\n"
                                    "isosurface was set to %0.3f. The isosurface may contain holes!\n"
                                    "Adjust max_gradient to get a proper rendering of the isosurface.",
                                    (float)(mginfo->gradient),
                                    (float)(mginfo->max_gradient));
            }
            else if ((diff >= 10.0) || ((prop >= 1.1) && (diff >= 0.5)))
            {
                messenger.WarningAt(kWarningGeneral, *fnInfo,
                                    "The maximum gradient found was %0.3f, but max_gradient of\n"
                                    "the isosurface was set to %0.3f. Adjust max_gradient to\n"
                                    "get a faster rendering of the isosurface.",
                                    (float)(mginfo->gradient),
                                    (float)(mginfo->max_gradient));
            }
        }
    }
}


void IsoSurface::Compute_BBox()
{
    container->ComputeBBox(BBox);
    if (Trans != nullptr)
    {
        Recompute_BBox(&BBox, Trans);
    }
}


bool IsoSurface::Function_Find_Root(ISO_ThreadData& itd, const Vector3d& PP, const Vector3d& DD,
        double* Depth1, double* Depth2, double& maxg, bool in_shadow_test, TraceThreadData* pThreadData)
{
    double dt, t21, l_b, l_e, oldmg;
    ISO_Pair EP1, EP2;
    Vector3d VTmp;

    pThreadData->Stats()[Ray_IsoSurface_Find_Root]++;

    itd.Vlength = DD.length();

    if (itd.cache.current == this)
    {
        pThreadData->Stats()[Ray_IsoSurface_Cache]++;
        VTmp = PP + *Depth1 * DD;
        VTmp -= itd.cache.Pglobal;
        l_b = VTmp.length();
        VTmp = PP + *Depth2 * DD;
        VTmp -= itd.cache.Dglobal;
        l_e = VTmp.length();
        if ((itd.cache.fmax - maxg * max(l_b, l_e)) > 0.0)
        {
            pThreadData->Stats()[Ray_IsoSurface_Cache_Succeeded]++;
            return false;
        }
    }

    itd.cache.Pglobal = PP;
    itd.cache.Dglobal = DD;

    itd.cache.current = nullptr;
    EP1.t = *Depth1;
    EP1.f = Float_Function(itd, *Depth1);
    itd.cache.fmax = EP1.f;
    if ((closed == false) && (EP1.f < 0.0))
    {
        itd.Inv3 *= -1;
        EP1.f *= -1;
    }

    EP2.t = *Depth2;
    EP2.f = Float_Function(itd, *Depth2);
    itd.cache.fmax = min(EP2.f, itd.cache.fmax);

    oldmg = maxg;
    t21 = (*Depth2 - *Depth1);
    dt = maxg * itd.Vlength * t21;

    //  It can be nearly impossible to eliminate the accuracy artifacts no
    //  matter how small accuracy(1) in some shadow and camera situations.
    //  True with something as simple as an isosurface f_sphere with an
    //  orthographic camera. The idea below is to always jitter the accuracy on
    //  each solver decent by +-0.5 (accuracy*0.967) whatever the accuracy is.
    //  In testing thus far results are quite good enabling clean result where
    //  often nearly impossible to get previously.
    //
    //  The most visible aspects of this are tangled in the fixed
    //  shadow_tolerance testing. Plans are to make changes there too.
    //
    //  In adaptive AA modes, the jitter below tends to hurt performance some
    //  where the accuracy is set quite small (<= 0.0001). Otherwise, it tends
    //  to help performance with adaptive AA.
    //

    double jitterAccuracy;
    if (jitter)
    {
        jitterAccuracy = accuracy + ((((double)std::rand()/(double)RAND_MAX)-0.5)*(accuracy*(0.9+(2/3)*0.1 )));
    }
    else
    {
        jitterAccuracy = accuracy;
    }

    if (Function_Find_Root_R(itd, &EP1, &EP2, dt, t21, 1.0 / (itd.Vlength * t21), maxg,
                jitterAccuracy, pThreadData))
    {
        *Depth1 = itd.tl;

        return true;
    }
    else if (!in_shadow_test)  // Bumping forward to not find same root again?
    {
        itd.cache.Pglobal = PP + EP1.t * DD;
        itd.cache.Dglobal = PP + EP2.t * DD;
        itd.cache.current = this;

        return false;
    }

    return false;
}


bool IsoSurface::Function_Find_Root_R(ISO_ThreadData& itd, const ISO_Pair* EP1, const ISO_Pair* EP2,
        double dt, double t21, double len, double& maxg, double jitterAccuracy, TraceThreadData* pThreadData)
{
    ISO_Pair EPa;
    double temp;

    temp = fabs((EP2->f - EP1->f) * len);
    if (gradient < temp)
    {
        gradient = temp;
    }

    if (t21 < jitterAccuracy)
    {
        if (EP2->f < 0)
        {
            itd.tl = EP2->t;
            return true;
        }
        else
        {
            return false;
        }
    }

    if ((EP1->f + EP2->f - dt) < 0)
    {
        t21 *= 0.5;
        dt *= 0.5;
        EPa.t = EP1->t + t21;
        EPa.f = Float_Function(itd, EPa.t);

        itd.cache.fmax = min(EPa.f, itd.cache.fmax);
        if (!Function_Find_Root_R(itd, EP1, &EPa, dt, t21, len * 2.0, maxg, jitterAccuracy, pThreadData))
        {
            return (Function_Find_Root_R(itd, &EPa, EP2, dt, t21, len * 2.0,maxg, jitterAccuracy, pThreadData));
        }
        else
        {
            return true;
        }
    }
    else
    {
        return false;
    }
}


double IsoSurface::Float_Function(ISO_ThreadData& itd, double t) const
{
    Vector3d VTmp;

    VTmp = itd.cache.Pglobal + t * itd.cache.Dglobal;

    return ((double)itd.Inv3 * EvaluateFunction(*itd.pFn, VTmp));
}


double IsoSurface::EvaluateAbs (GenericScalarFunctionInstance& fn, Vector3d& p) const
{
    return fabs(threshold - fn.Evaluate(p));
}


double IsoSurface::EvaluateFunction (GenericScalarFunctionInstance& fn, Vector3d& p) const
{
    return fn.Evaluate(p) - threshold;
}


bool IsoSurface::IsInside(GenericScalarFunctionInstance& fn, Vector3d& p) const
{
    return fn.Evaluate(p) < threshold;
}


}
// end of namespace pov
