//******************************************************************************
///
/// @file core/material/pattern.cpp
///
/// Implementations related to patterns.
///
/// @copyright
/// @parblock
///
/// Some texture ideas garnered from SIGGRAPH '85 Volume 19 Number 3, "An Image
/// Synthesizer" By Ken Perlin. Further Ideas Garnered from "The RenderMan
/// Companion" (Addison Wesley).
///
/// ----------------------------------------------------------------------------
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2021 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/material/pattern.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <algorithm>
#include <limits>
#include <vector>

// POV-Ray header files (base module)
#include "base/fileinputoutput.h"
#include "base/povassert.h"

// POV-Ray header files (core module)
#include "core/material/blendmap.h"
#include "core/material/noise.h"
#include "core/material/pigment.h"
#include "core/material/warp.h"
#include "core/math/matrix.h"
#include "core/math/randomsequence.h"
#include "core/render/ray.h"
#include "core/scene/object.h"
#include "core/scene/scenedata.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/cracklecache.h"
#include "core/support/imageutil.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

using std::min;
using std::max;

// Local variables

static RandomDoubleSequence gPatternRands(0.0, 1.0, 32768);
static int gaBinomialCoefficients[((kFractalMaxExponent+1)*(kFractalMaxExponent+2))/2]; // GLOBAL VARIABLE
static int gaCrackleCubeTable[81*3];


static const ColourBlendMapEntry gaDefaultBlendMapEntries_BlackWhite[2] =
    {{0.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(1.0, 1.0, 1.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Gray (new ColourBlendMap(2, gaDefaultBlendMapEntries_BlackWhite));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Brick[2] =
    {{0.0, TransColour(ToMathColour(RGBColour(0.2140, 0.2140, 0.2140)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.3185, 0.0196, 0.0196)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Brick (new ColourBlendMap(2, gaDefaultBlendMapEntries_Brick));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Checker[2] =
    {{0.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 1.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 0.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Checker (new ColourBlendMap(2, gaDefaultBlendMapEntries_Checker));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Cubic[6] =
    {{0.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 1.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(1.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 1.0)))},
     {2.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 1.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Cubic (new ColourBlendMap(6, gaDefaultBlendMapEntries_Cubic));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Hexagon[3] =
    {{0.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 1.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 0.0)))},
     {2.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 0.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Hexagon (new ColourBlendMap(3, gaDefaultBlendMapEntries_Hexagon));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Mandel[5] =
    {{0.001, TransColour(ToMathColour(RGBColour(0.0, 0.0, 0.0)))},
     {0.001, TransColour(ToMathColour(RGBColour(0.0, 1.0, 1.0)))},
     {0.012, TransColour(ToMathColour(RGBColour(1.0, 1.0, 0.0)))},
     {0.015, TransColour(ToMathColour(RGBColour(1.0, 0.0, 1.0)))},
     {0.1,   TransColour(ToMathColour(RGBColour(0.0, 1.0, 1.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Mandel (new ColourBlendMap(5, gaDefaultBlendMapEntries_Mandel));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Square[6] =
    {{0.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 1.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(1.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 1.0)))},
     {2.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 1.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Square (new ColourBlendMap(4, gaDefaultBlendMapEntries_Square));


static const ColourBlendMapEntry gaDefaultBlendMapEntries_Triangular[6] =
    {{0.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 0.0, 1.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(1.0, 1.0, 0.0)))},
     {1.0, TransColour(ToMathColour(RGBColour(0.0, 1.0, 1.0)))},
     {2.0, TransColour(ToMathColour(RGBColour(1.0, 0.0, 1.0)))}};

static ColourBlendMapConstPtr gpDefaultBlendMap_Triangular (new ColourBlendMap(6, gaDefaultBlendMapEntries_Triangular));


// Static functions

static inline const ClassicTurbulence* GetTurb(const WarpList warps);
static unsigned short ReadUShort(IStream *pInFile);
static unsigned int ReadUInt(IStream *pInFile);

#define SQRT3_2     0.86602540378443864676372317075294  ///< sqrt(3)/2
#define SQRT3       1.7320508075688772935274463415059   ///< sqrt(3)
#define SQRT2       1.4142135623730950488016887242097   ///< sqrt(2)
#define SQRT2_2     0.70710678118654752440084436210485  ///< sqrt(2)/2

#define SIN18       0.30901699437494742410229341718282  ///< sin(18 deg)
#define SIN36       0.58778525229247312916870595463907  ///< sin(36 deg)
#define SIN54       0.80901699437494742410229341718282  ///< sin(54 deg)
#define SIN72       0.95105651629515357211643933337938  ///< sin(72 deg)
#define SIN108      SIN72                               ///< sin(108 deg)
#define SIN144      SIN36                               ///< sin(144 deg)
#define SIN162      SIN18                               ///< sin(162 deg)
#define COS18       SIN72                               ///< cos(18 deg)
#define COS36       SIN54                               ///< cos(36 deg)
#define COS54       SIN36                               ///< cos(54 deg)
#define COS72       SIN18                               ///< cos(72 deg)
#define COS108      (-COS72)                            ///< cos(108 deg)
#define COS126      (-COS54)                            ///< cos(126 deg)
#define COS144      (-COS36)                            ///< cos(144 deg)
#define COS162      (-COS18)                            ///< cos(162 deg)
#define TAN18       0.32491969623290632615587141221513  ///< tan(18 deg)
#define TAN36       0.72654252800536088589546675748062  ///< tan(36 deg)
#define TAN54       1.3763819204711735382072095819109   ///< tan(54 deg)
#define TAN72       3.0776835371752534025702905760369   ///< tan(72 deg)
#define TAN108      (-TAN72)                            ///< tan(108 deg)
#define TAN126      (-TAN54)                            ///< tan(126 deg)
#define TAN144      (-TAN36)                            ///< tan(144 deg)
#define TAN162      (-TAN18)                            ///< tan(162 deg)

#define PHI         1.6180339887498948482045868343656   ///< golden ratio = (1+sqrt(5))/2
#define INVPHI      0.61803398874989484820458683436564  ///< inverse of golden ratio (= golden ratio -1)
#define SQRPHI      2.6180339887498948482045868343656   ///< square of golden ratio (= golden ratio +1)
#define INVSQRPHI   0.38196601125010515179541316563436  ///< inverse square of golden ratio (= 2 - golden ratio)

/*****************************************************************************/

int GetNoiseGen (const TPATTERN *TPat, const TraceThreadData *pThread)
{
    int noise_gen = TPat->pattern->noiseGenerator;
    if (noise_gen == kNoiseGen_Default)
        noise_gen = pThread->GetSceneData()->noiseGenerator;
    return noise_gen;
}


BasicPattern::BasicPattern() :
    noiseGenerator(kNoiseGen_Default)
{}

BasicPattern::BasicPattern(const BasicPattern& obj) :
    noiseGenerator(obj.noiseGenerator)
{
    Copy_Warps(warps, obj.warps);
}

BasicPattern::~BasicPattern()
{
    Destroy_Warps(warps);
}

bool BasicPattern::Precompute()
{
    return true;
}

int BasicPattern::GetNoiseGen(const TraceThreadData *pThread) const
{
    int noise_gen = noiseGenerator;
    if (noise_gen == kNoiseGen_Default)
        noise_gen = pThread->GetSceneData()->noiseGenerator;
    return noise_gen;
}

ColourBlendMapConstPtr BasicPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Gray; }

bool BasicPattern::HasSpecialTurbulenceHandling() const { return false; }  // Effective pattern default


ImagePatternImpl::ImagePatternImpl() :
    pImage(nullptr)
{}

ImagePatternImpl::ImagePatternImpl(const ImagePatternImpl& obj) :
    pImage(obj.pImage ? Copy_Image(obj.pImage) : nullptr)
{}

ImagePatternImpl::~ImagePatternImpl()
{
    if (pImage)
        Destroy_Image(pImage);
}


ColourPattern::ColourPattern() :
    BasicPattern()
{}

ColourPattern::ColourPattern(const ColourPattern& obj) :
    BasicPattern(obj)
{}

DBL ColourPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    TransColour colour;
    if (Evaluate(colour, EPoint, pIsection, pRay, pThread))
        return colour.Greyscale();
    else
        return 0.0;
}

unsigned int ColourPattern::NumDiscreteBlendMapEntries() const { return 999999; }
bool ColourPattern::CanMap() const { return false; }


ContinuousPattern::ContinuousPattern() :
    BasicPattern(),
    waveType(kWaveType_Ramp),
    waveFrequency(1.0),
    wavePhase(0.0),
    waveExponent(0.0),
    functionInterval(false),
    waveModifiers(false),
    waveInverse(false)
{}

ContinuousPattern::ContinuousPattern(const ContinuousPattern& obj) :
    BasicPattern(obj),
    waveType(obj.waveType),
    waveFrequency(obj.waveFrequency),
    wavePhase(obj.wavePhase),
    waveExponent(obj.waveExponent),
    functionInterval(obj.functionInterval),
    waveModifiers(obj.waveModifiers),
    waveInverse(obj.waveInverse)
{}


// The Evaluate method below implements what the documentation and users
// typically discuss as continuous pattern "wave shape modifications." The
// documentation - and to an extent some code like ripples and waves - further
// discuss and handle a subset of these modifiers calling them "wave type
// modifiers" such as sine_wave and triangle_wave.

// In povr the intent is to discuss only various "wave shape modifications."
// Where there might have been some need for distinction to support (in my
// experience nonworking) dual use of keywords, new keywords will be created of
// the ip_ prefix variety to indicate internal pattern forms. To fix ripples
// and waves these keywords are ip_frequency and ip_phase. With these we can
// support frequency and phase internal to the pattern and as "wave shape
// modifications."

// Further, in povr the wave type cubic_wave can act as both a wave type and a
// value selection mechanism. The duality making the wave type vs wave shape
// modification distinct foggy.

// More: http://news.povray.org/povray.binaries.images/thread/%3C5f5a5a21%40news.povray.org%3E/

DBL ContinuousPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value = EvaluateRaw(EPoint, pIsection, pRay, pThread);

    if (waveType == kWaveType_Raw)
        return value;

    DBL tmpValue = value * waveFrequency + wavePhase;

    // Careful not to snap 1.0 multiples to 0.0 as that causes rare artifacts
    // in pattern and function usage where ray/function evaluation lands
    // exactly on the multiple. In other words, simple fmod snapping works only
    // with a ramp wave and even there it snaps 1.0 map values to the zero
    // entry - a behavior not optimal.

    if(fabs(tmpValue) > 1.0)
    {
        DBL tmpCount = (unsigned long long)std::abs(tmpValue/1.0);
        value = fmod(tmpValue, 1.0);
        if(fabs(value) < gkDBL_epsilon)
        {
            value = tmpValue / tmpCount;
        }
    }
    else
    {
        value = tmpValue;
    }

    if (!functionInterval)
    {   // Work in a range of [0.0 to 1.0]

        // Flip any negative values so continuous at 0.0. Allows too, for
        // example, negative phase values and frequencies not 1.0.

        if ((value < 0.0) && (waveType != kWaveType_Cubic))
            value -= floor(value);

        switch(waveType)
        {
            case kWaveType_Ramp:
                break;
            case kWaveType_Sine:
                value = (1.0 + cycloidal(value)) * 0.5;
                break;
            case kWaveType_Triangle:
                value = Triangle_Wave(value);
                break;
            case kWaveType_Scallop:
                value = fabs(cycloidal(value * 0.5));
                break;
            case kWaveType_Cubic:

                // Has been messed up since neg value inversion added eons ago.
                // Now returns a single continuous about zero cubic bump.  A
                // little like a traditional bump function.

                if (fabs(tmpValue) <= 1.0)
                    value = (Sqr(value) * ((-2.0 * fabs(value)) + 3.0));
                else
                    value = 1.0;
                break;
            default:
                throw POV_EXCEPTION_STRING("Unknown Wave Type for 0 to 1 range.");
        }

        if (waveModifiers)
        {
            if (waveInverse)
            {
                value = 1.0 - value;
            }
            if (waveExponent != 0.0) // Value other than 0.0 back door path to waveModiers use.
            {
                value = pow(value,waveExponent);
            }
        }
    }
    else
    { // Work in a range of [-1.0 to 1.0]
        switch(waveType)
        {
            case kWaveType_Ramp:
                break;
            case kWaveType_Sine:
                value = cycloidal(value);
                break;
            case kWaveType_Triangle:
                value = 2.0 * Triangle_Wave(value) - 1.0;
                break;
            case kWaveType_Scallop:
                value = 2.0 * fabs(cycloidal(value * 0.5)) - 1.0;
                break;
            case kWaveType_Cubic:
                // See description above.
                if (fabs(tmpValue) <= 1.0)
                    value = (Sqr(value) * ((-2.0 * fabs(value)) + 3.0));
                else
                    value = 1.0;
                value = (tmpValue < 0.0) ? -value : value;
                break;
            default:
                throw POV_EXCEPTION_STRING("Unknown Wave Type for -1 to 1 range.");
        }

        if (waveModifiers)
        {
            if (waveInverse)
            {
                value *= -1.0;
            }
            if (waveExponent != 0.0) // Value other than 0.0 back door path to waveModiers use.
            {

                // Note. Unlike using poly_wave with a 0 to 1 interval which distorts or shapes
                // the scalar values returned. With the function_interval poly_wave offers a
                // way to twiddle with the gradient of the incoming function for isosurfaces in
                // addition to some shaping if scalar values being used in a map. This
                // twiddling not generally useful if the gradients are already near 1.0 and it
                // WON'T change the surface position which is at value 0.0

                if (value < 0.0)
                {
                    value = -pow(fabs(value),waveExponent);
                }
                else
                {
                    value = pow(value,waveExponent);
                }
                value = max(-1.0,min(1.0,value)); // Clamping due floating point noise.
            }
        }
    }

    return value;
}

unsigned int ContinuousPattern::NumDiscreteBlendMapEntries() const { return 999999; }
bool ContinuousPattern::CanMap() const { return true; }

bool DiscretePattern::CanMap() const { return false; }

unsigned int NormalOnlyPattern::NumDiscreteBlendMapEntries() const { return 999999; }
bool NormalOnlyPattern::CanMap() const { return false; }
DBL NormalOnlyPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const { return 0.0; }

unsigned int PlainPattern::NumDiscreteBlendMapEntries() const { return 1; }

unsigned int AveragePattern::NumDiscreteBlendMapEntries() const { return 999999; }
bool AveragePattern::CanMap() const { return true; }
DBL AveragePattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const { return 0.0; }


AgatePattern::AgatePattern() :
    agateFrequency(1.0),
    agateOctaves(6),
    agateOmega(0.5),
    agateLambda(2)
{}


BevyPattern::BevyPattern() : bevyDirection(0,1,0), bevyType(0) {}


BrickPattern::BrickPattern() : brickSize(8.0,3.0,4.5), mortar(0.5-gkDBL_epsilon*2.0) {}

ColourBlendMapConstPtr BrickPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Brick; }
unsigned int BrickPattern::NumDiscreteBlendMapEntries() const { return 2; }

ColourBlendMapConstPtr CheckerPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Checker; }
unsigned int CheckerPattern::NumDiscreteBlendMapEntries() const { return 2; }


CracklePattern::CracklePattern() :
    crackleForm(-1,1,0),
    crackleMetric(2),
    crackleOffset(0),
    repeat(0,0,0),
    crackleIsSolid(false)
{}


ColourBlendMapConstPtr CubicPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Cubic; }
unsigned int CubicPattern::NumDiscreteBlendMapEntries() const { return 6; }


ColourFunctionPattern::ColourFunctionPattern()
{
    for (int iChannel = 0; iChannel < 5; ++iChannel)
    {
        pFn[iChannel] = nullptr;
    }
}

ColourFunctionPattern::ColourFunctionPattern(const ColourFunctionPattern& obj) :
    ColourPattern(obj)
{
    for (int iChannel = 0; iChannel < 5; ++iChannel)
    {
        if (obj.pFn[iChannel])
            pFn[iChannel] = obj.pFn[iChannel]->Clone();
        else
            pFn[iChannel] = nullptr;
    }
}

ColourFunctionPattern::~ColourFunctionPattern()
{
    for (int iChannel = 0; iChannel < 5; ++iChannel)
    {
        if (pFn[iChannel])
            delete pFn[iChannel];
    }
}

bool ColourFunctionPattern::Evaluate(TransColour& result, const Vector3d& EPoint,
        const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    for (int iChannel = 0; iChannel < 5; ++iChannel)
    {
        ColourChannel channelValue = 0.0;
        if (pFn[iChannel])
            channelValue = GenericScalarFunctionInstance(pFn[iChannel], pThread).Evaluate(EPoint);
        switch (iChannel)
        {
        case 3:  result.filter()           = channelValue; break;
        case 4:  result.transm()           = channelValue; break;
        default: result.colour()[iChannel] = channelValue; break;
        }
    }
    return true;
}

bool ColourFunctionPattern::HasTransparency() const
{
    return (pFn[3] || pFn[4]);
}


bool ColourImagePattern::Evaluate(TransColour& result, const Vector3d& EPoint,
        const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    // TODO ALPHA - the caller does expect non-premultiplied data, but maybe he could profit from premultiplied data?

    int reg_number;
    DBL xcoor = 0.0, ycoor = 0.0;

    // If outside map coverage area, return clear

    if (map_pos(EPoint, pImage, &xcoor, &ycoor))
    {
        result = ToTransColour(RGBFTColour(1.0, 1.0, 1.0, 0.0, 1.0));
        return false;
    }
    else
    {
        RGBFTColour rgbft;
        image_colour_at(pImage, xcoor, ycoor, rgbft, &reg_number, false);
        result = ToTransColour(rgbft);
        return true;
    }
}

bool ColourImagePattern::HasTransparency() const
{
    return (!pImage || pImage->Once_Flag || !is_image_opaque(pImage));
}


DensityFilePattern::DensityFilePattern() :
    densityFile(nullptr)
{}

DensityFilePattern::DensityFilePattern(const DensityFilePattern& obj) :
    ContinuousPattern(obj),
    densityFile(nullptr)
{
    if (obj.densityFile)
        densityFile = Copy_Density_File(obj.densityFile);
}

DensityFilePattern::~DensityFilePattern()
{
    if (densityFile)
        Destroy_Density_File(densityFile);
}


FacetsPattern::FacetsPattern() : facetsSize(0.1), facetsCoords(0), facetsMetric(2) {}


FractalPattern::FractalPattern() :
    interiorType(0),
    exteriorType(1),
    maxIterations(),        // no explicit default; must be set by caller
    exteriorFactor(1),
    interiorFactor(1)
{}


FunctionPattern::FunctionPattern() :
    pFn(nullptr)
{}

FunctionPattern::FunctionPattern(const FunctionPattern& obj) :
    ContinuousPattern(obj),
    pFn(obj.pFn->Clone())
{}

FunctionPattern::~FunctionPattern()
{
    if (pFn)
        delete pFn;
}


// See normal perturbation code (normal.cpp) as to reasons for values.
GroovedPattern::GroovedPattern() :
    Control0(1.0), Control1(1.0),
    groovedLo(0.0), groovedHi(0.5),
    groovedFlip(false)
{}


HardObjectPattern::HardObjectPattern() :
    pObject(NULL),
    radius(1.0),
    recursion_limit(4),
    samples(22)
{
    waveType = kWaveType_Ramp;
}

HardObjectPattern::HardObjectPattern(const HardObjectPattern& obj) :
    ContinuousPattern(obj),
    pObject(NULL),
    radius(obj.radius),
    recursion_limit(obj.recursion_limit),
    samples(obj.samples)
{
    if (obj.pObject)
        pObject = Copy_Object(obj.pObject);
}

HardObjectPattern::~HardObjectPattern()
{
    if (pObject)
        Destroy_Object(pObject);
}


ColourBlendMapConstPtr HexagonPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Hexagon; }
unsigned int HexagonPattern::NumDiscreteBlendMapEntries() const { return 3; }


DBL ImagePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    return image_pattern(EPoint, this);
}


MarblePattern::MarblePattern() :
    marbleFrequency((1.0/3.0)*0.1),  // 0.0333...
    marbleOctaves(6),
    marbleOmega(0.5),
    marbleLambda(2)
{
    waveType = kWaveType_Triangle;
}


NoiseCubedPattern::NoiseCubedPattern() : noiseCubedScale(1.0), noiseCubedType(1) {}


ObjectPattern::ObjectPattern() :
    pObject(nullptr)
{}

ObjectPattern::ObjectPattern(const ObjectPattern& obj) :
    DiscretePattern(obj),
    pObject(nullptr)
{
    if (obj.pObject)
        pObject = Copy_Object(obj.pObject);
}

ObjectPattern::~ObjectPattern()
{
    if (pObject)
        Destroy_Object(pObject);
}

ColourBlendMapConstPtr ObjectPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Checker; } // sic!
unsigned int ObjectPattern::NumDiscreteBlendMapEntries() const { return 2; }


PavementPattern::PavementPattern() : Side(3), Tile(1), Number(1), Exterior(0), Interior(0), Form(0) {}


PigmentPattern::PigmentPattern() :
    pPigment(nullptr)
{}

PigmentPattern::PigmentPattern(const PigmentPattern& obj) :
    ContinuousPattern(obj),
    pPigment(nullptr)
{
    if (obj.pPigment)
        pPigment = Copy_Pigment(obj.pPigment);
}

PigmentPattern::~PigmentPattern()
{
    if (pPigment)
        Destroy_Pigment(pPigment);
}


PotentialPattern::PotentialPattern() :
    pObject(nullptr),
    subtractThreshold(false)
{
    waveType = kWaveType_Ramp;
}

PotentialPattern::PotentialPattern(const PotentialPattern& obj) :
    ContinuousPattern(obj),
    pObject(nullptr),
    subtractThreshold(obj.subtractThreshold)
{
    if (obj.pObject)
        pObject = Copy_Object(obj.pObject);
}

PotentialPattern::~PotentialPattern()
{
    if (pObject)
        Destroy_Object(pObject);
}


// Ripples code uses frequency and phase values internally. Set by
// ip_frequency <float> and ip_phase <float> keywords.
RipplesPattern::RipplesPattern() : frequency(1.0), phase(0.0) {}


// See normal perturbation code (normal.cpp) as to reasons for values.
QuiltedPattern::QuiltedPattern() :
    Control0(1.0), Control1(1.0),
    quiltedLo(0.5), quiltedHi(0.866025403)
{}


SeedfiPattern::SeedfiPattern() : seedfiType(0), seedfiOffset(0.0) {}


SlopePattern::SlopePattern() :
    altitudeDirection(0.0, 0.0, 0.0),
    slopeDirection(),       // no explicit default; must be set by caller
    altitudeLen(),          // no explicit default; must be set by caller
    altitudeModLow(0.0),
    altitudeModWidth(0.0),
    slopeLen(),             // no explicit default; must be set by caller
    slopeModLow(0.0),
    slopeModWidth(0.0),
    altitudeAxis(),         // no explicit default; must be set by caller
    slopeAxis(),            // no explicit default; must be set by caller
    pointAt(false)
{}


SoftObjectPattern::SoftObjectPattern() :
    pObject(NULL),
    spacing(0.25+gkDBL_epsilon), // Prevents virtual grid ray alignment and FltPt noise.
    strength(1.0)
{
    waveType = kWaveType_Ramp;
}

SoftObjectPattern::SoftObjectPattern(const SoftObjectPattern& obj) :
    ContinuousPattern(obj),
    pObject(NULL),
    spacing(obj.spacing),
    strength(obj.strength)
{
    if (obj.pObject)
        pObject = Copy_Object(obj.pObject);
}

SoftObjectPattern::~SoftObjectPattern()
{
    if (pObject)
        Destroy_Object(pObject);
}


// Setting defaults for spiral1 and spiral2 in common parent class.
SpiralPattern::SpiralPattern() :
    arms(),                 // no explicit default; must be set by caller
    spiralFrequency(0.0),
    spiralOctaves(6),
    spiralOmega(0.5),
    spiralLambda(2)
{
    waveType = kWaveType_Triangle;
}


// See normal perturbation code (normal.cpp) as to reasons for values.
TiledPattern::TiledPattern() :
    Control0(1.0), Control1(1.0),
    tiledLo(0.0), tiledHi(0.5)
{}


ColourBlendMapConstPtr SquarePattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Square; }
unsigned int SquarePattern::NumDiscreteBlendMapEntries() const { return 4; }

ColourBlendMapConstPtr TriangularPattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Triangular; }
unsigned int TriangularPattern::NumDiscreteBlendMapEntries() const { return 6; }

// Waves code uses frequency and phase values internally. Set by
// ip_frequency <float> and ip_phase <float> keywords.
WavesPattern::WavesPattern() : frequency(1.0), phase(0.0) {}


WoodPattern::WoodPattern() :
    woodFrequency((1.0/3.0)*0.1),  // 0.0333...
    woodOctaves(6),
    woodOmega(0.5),
    woodLambda(2)
{
    waveType = kWaveType_Triangle;
}


JuliaPattern::JuliaPattern() :
    FractalPattern(),
    juliaCoord()
{}

JuliaPattern::JuliaPattern(const JuliaPattern& obj) :
    FractalPattern(obj),
    juliaCoord(obj.juliaCoord)
{}


Julia3Pattern::Julia3Pattern() :
    JuliaPattern()
{}

Julia3Pattern::Julia3Pattern(const JuliaPattern& obj) :
    JuliaPattern(obj)
{}


Julia4Pattern::Julia4Pattern() :
    JuliaPattern()
{}

Julia4Pattern::Julia4Pattern(const JuliaPattern& obj) :
    JuliaPattern(obj)
{}


JuliaXPattern::JuliaXPattern() :
    JuliaPattern(),
    fractalExponent(0.0)
{}

JuliaXPattern::JuliaXPattern(const JuliaPattern& obj) :
    JuliaPattern(obj),
    fractalExponent(0.0)
{}

JuliaXPattern::JuliaXPattern(const JuliaXPattern& obj) :
    JuliaPattern(obj),
    fractalExponent(obj.fractalExponent)
{}


Mandel2Pattern::Mandel2Pattern() :
    MandelPattern()
{}

Mandel2Pattern::Mandel2Pattern(const MandelPattern& obj) :
    MandelPattern(obj)
{}

ColourBlendMapConstPtr Mandel2Pattern::GetDefaultBlendMap() const { return gpDefaultBlendMap_Mandel; }


Mandel3Pattern::Mandel3Pattern() :
    MandelPattern()
{}

Mandel3Pattern::Mandel3Pattern(const MandelPattern& obj) :
    MandelPattern(obj)
{}


Mandel4Pattern::Mandel4Pattern() :
    MandelPattern()
{}

Mandel4Pattern::Mandel4Pattern(const MandelPattern& obj) :
    MandelPattern(obj)
{}


MandelXPattern::MandelXPattern() :
    MandelPattern(),
    fractalExponent(0.0)
{}

MandelXPattern::MandelXPattern(const MandelPattern& obj) :
    MandelPattern(obj),
    fractalExponent(0.0)
{}

MandelXPattern::MandelXPattern(const MandelXPattern& obj) :
    MandelPattern(obj),
    fractalExponent(obj.fractalExponent)
{}

//---------- 61 pattern names
std::string AOIPattern::Name() const { return "aoi"; }
std::string AgatePattern::Name() const { return "agate"; }
std::string AveragePattern::Name() const { return "average"; }
std::string BevyPattern::Name() const { return "bevy"; }
std::string BoxedPattern::Name() const { return "boxed"; }
std::string BozoPattern::Name() const { return "bozo"; }
std::string BrickPattern::Name() const { return "brick"; }
std::string BumpsPattern::Name() const { return "bumps"; }
std::string CellsPattern::Name() const { return "cells"; }
std::string CheckerPattern::Name() const { return "checker"; }
std::string ColourFunctionPattern::Name() const { return "colourfunction"; }
std::string ColourImagePattern::Name() const { return "colourimage"; }
std::string CracklePattern::Name() const { return "crackle"; }
std::string CubicPattern::Name() const { return "cubic"; }
std::string CylindricalPattern::Name() const { return "cylindrical"; }
std::string DensityFilePattern::Name() const { return "density_file"; }
std::string DentsPattern::Name() const { return "dents"; }
std::string FacetsPattern::Name() const { return "facets"; }
std::string FunctionPattern::Name() const { return "function"; }
std::string GradientPattern::Name() const { return "gradient"; }
std::string GranitePattern::Name() const { return "granite"; }
std::string GroovedPattern::Name() const { return "grooved"; }
std::string HardObjectPattern::Name() const { return "hard_object"; }
std::string HexagonPattern::Name() const { return "hexagon"; }
std::string ImagePattern::Name() const { return "image"; }
std::string Julia3Pattern::Name() const { return "julia3"; }
std::string Julia4Pattern::Name() const { return "julia4"; }
std::string JuliaPattern::Name() const { return "julia"; }
std::string JuliaXPattern::Name() const { return "juliax"; }
std::string LeopardPattern::Name() const { return "leopard"; }
std::string Magnet1JPattern::Name() const { return "magnet1j"; }
std::string Magnet1MPattern::Name() const { return "magnet1m"; }
std::string Magnet2JPattern::Name() const { return "magnet2j"; }
std::string Magnet2MPattern::Name() const { return "magnet2m"; }
std::string Mandel2Pattern::Name() const { return "mandel2"; }
std::string Mandel3Pattern::Name() const { return "mandel3"; }
std::string Mandel4Pattern::Name() const { return "mandel4"; }
std::string MandelXPattern::Name() const { return "mandelx"; }
std::string MarblePattern::Name() const { return "marble"; }
std::string NoiseCubedPattern::Name() const { return "noise_cubed"; }
std::string ObjectPattern::Name() const { return "object"; }
std::string OnionPattern::Name() const { return "onion"; }
std::string PavementPattern::Name() const { return "pavement"; }
std::string PigmentPattern::Name() const { return "pigment"; }
std::string PlainPattern::Name() const { return "plain"; }
std::string PlanarPattern::Name() const { return "planar"; }
std::string PotentialPattern::Name() const { return "potential"; }
std::string QuiltedPattern::Name() const { return "quilted"; }
std::string RadialPattern::Name() const { return "radial"; }
std::string RipplesPattern::Name() const { return "ripples"; }
std::string SeedfiPattern::Name() const { return "seedfi"; }
std::string SlopePattern::Name() const { return "slope"; }
std::string SoftObjectPattern::Name() const { return "soft_object"; }
std::string SphericalPattern::Name() const { return "spherical"; }
std::string Spiral1Pattern::Name() const { return "spiral1"; }
std::string Spiral2Pattern::Name() const { return "spiral2"; }
std::string SquarePattern::Name() const { return "square"; }
std::string TiledPattern::Name() const { return "tiled"; }
std::string TilingPattern::Name() const { return "tiling"; }
std::string TriangularPattern::Name() const { return "triangular"; }
std::string WavesPattern::Name() const { return "waves"; }
std::string WoodPattern::Name() const { return "wood"; }
std::string WrinklesPattern::Name() const { return "wrinkles"; }



// EPoint       - The point in 3d space at which the pattern is evaluated.
// TPat         - Texture pattern struct
// Intersection - intersection structure
//
// Returns: DBL result usual 0.0 to 1.0 but may be 2.0 in hexagon
//
// Author: Adapted from Add_Pigment by Chris Young
//
DBL Evaluate_TPat (const TPATTERN *TPat, const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread)
{
    return TPat->pattern->Evaluate(EPoint, pIsection, pRay, pThread);
}


void Init_TPat_Fields (TPATTERN *Tpat)
{
    Tpat->Type       = NO_PATTERN;
    Tpat->Flags      = NO_FLAGS;
    Tpat->pattern    = PatternPtr(new PlainPattern());
}


void Copy_TPat_Fields (TPATTERN *New, const TPATTERN *Old)
{
    *New = *Old;

    if (Old->pattern)
        New->pattern = Old->pattern->Clone();
    else
        New->pattern.reset();
}


void Translate_Tpattern(TPATTERN *Tpattern, const Vector3d& Vector)
{
    TRANSFORM Trans;

    if (Tpattern != nullptr)
    {
        Compute_Translation_Transform (&Trans, Vector);

        Transform_Tpattern (Tpattern, &Trans);
    }
}


void Rotate_Tpattern(TPATTERN *Tpattern, const Vector3d& Vector)
{
    TRANSFORM Trans;

    if (Tpattern != nullptr)
    {
        Compute_Rotation_Transform (&Trans, Vector);

        Transform_Tpattern (Tpattern, &Trans);
    }
}


void Scale_Tpattern(TPATTERN *Tpattern, const Vector3d& Vector)
{
    TRANSFORM Trans;

    if (Tpattern != nullptr)
    {
        Compute_Scaling_Transform (&Trans, Vector);

        Transform_Tpattern (Tpattern, &Trans);
    }
}


void Transform_Tpattern(TPATTERN *Tpattern, const TRANSFORM *Trans)
{
    if ((Tpattern != nullptr) && (Tpattern->pattern != nullptr))
    {
        TransformWarp* temp = nullptr;
        if (!Tpattern->pattern->warps.empty())
            temp = dynamic_cast<TransformWarp*>(Tpattern->pattern->warps.back());
        if (!temp)
        {
            temp = new TransformWarp();
            Tpattern->pattern->warps.push_back(temp);
        }
        Compose_Transforms (&temp->Trans, Trans);
    }
}


template<typename DATA_T>
void BlendMap<DATA_T>::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const
{
    size_t Max_Ent = Blend_Map_Entries.size()-1;

    if (value >= Blend_Map_Entries[Max_Ent].value)
    {
        /* if greater than last, use last. */

        rpPrev = rpNext = &(Blend_Map_Entries[Max_Ent]);
        rPrevWeight = 0.0;
        rNextWeight = 1.0;
    }
    else
    {
        // TODO - we might use a binary search instead

        typename std::vector<Entry>::const_iterator iP;
        typename std::vector<Entry>::const_iterator iN;

        iP = iN = Blend_Map_Entries.begin();

        while (value > iN->value)
        {
            iP = iN;
            iN++;
        }

        if ((value == iN->value) || (iP == iN))
        {
            rpPrev = rpNext = &(*iN);
            rPrevWeight = 0.0;
            rNextWeight = 1.0;
        }
        else
        {
            rpPrev = &(*iP);
            rpNext = &(*iN);
            rPrevWeight = (iN->value - value) / (iN->value - iP->value);
            rNextWeight = 1.0 - rPrevWeight;
        }
    }
}

template void BlendMap<TransColour> ::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const;
template void BlendMap<PIGMENT*>    ::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const;
template void BlendMap<Vector2d>    ::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const;
template void BlendMap<TNORMAL*>    ::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const;
template void BlendMap<TexturePtr>  ::Search (DBL value, EntryConstPtr& rpPrev, EntryConstPtr& rpNext, DBL& rPrevWeight, DBL& rNextWeight) const;


static inline const ClassicTurbulence *GetTurb(const WarpList warps)
{
    POV_PATTERN_ASSERT(!warps.empty());
    POV_PATTERN_ASSERT(dynamic_cast<const ClassicTurbulence*>(*warps.begin()));
    return static_cast<const ClassicTurbulence*>(*warps.begin());
}


// Assumed pattern until a pattern specified. Should never be the pattern used.
DBL PlainPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    return 0.0;
}

// Think harmless, but why set true?
bool PlainPattern::HasSpecialTurbulenceHandling() const { return true; }


/* Tiling & Pavement */


/*****************************************************************************
*
* FUNCTIONS
*
*   related to Tiling & Pavement
*
* DESCRIPTION
*
*   New pattern function by J. Grimbert
*
*   Fill the X-Z plane with repeating pattern
*   When there is only one kind of form, 1 is the outside, 0 is most inside
*   When there is two kind of form, 0 & 1 are the most inside, 1/2 is
*   the outside
*
******************************************************************************/

static DBL tiling_square (const Vector3d& EPoint)
{
    /*
     ** Classical square tiling
     */
    DBL x,z;
    x = fabs(EPoint.x());
    x -= floor(x);
    x = 2*fabs( x-0.5 );

    z = fabs(EPoint.z());
    z -= floor(z);
    z = 2*fabs( z-0.5 );

    return max(x,z);
}

static DBL tiling_hexagon (const Vector3d& EPoint)
{
    /*
     ** Classical Hexagon tiling
     */
    DBL x,z;
    DBL dist1,dist2;
    DBL answer;

    x=EPoint.x();
    z=EPoint.z();
    x += 0.5;
    x -= 3.0*floor(x/3.0);
    z -= SQRT3*floor(z/(SQRT3));
    /* x,z is in { [0.0, 3.0 [, [0.0, SQRT3 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > SQRT3_2 )
    {
        z = SQRT3 - z;
    }
    /*
     ** Now only [0,3[,[0,SQRT3/2[
     */
    if (x > 1.5)
    {
        x -= 1.5; /* translate */
        z = SQRT3_2 -z; /* mirror */
    }
    /*
     ** And now, it is even simpler :  [0,1.5],[0,SQRT3/2]
     ** on the bottom left corner, part of some other hexagon
     ** on the top right corner, center of the hexagon
     */
    if ((SQRT3*x+z)<SQRT3_2)
    {
        x = 0.5 - x;
        z = SQRT3_2 -z; /* mirror */
    }
    if (x > 1.0)
    {
        x = 2.0 -x; /* mirror */
    }
    /* Hexagon */
    dist1 = 1.0 - ( z / SQRT3_2 );
    dist2 = 1.0 - (((SQRT3 * x + z - SQRT3_2) ) / SQRT3 );
    answer = max(dist1,dist2);
    answer = min(1.0,answer);
    answer = max(0.0,answer);
    return answer;
}

static DBL tiling_triangle (const Vector3d& EPoint)
{
    DBL x,z;
    DBL slop1;
    DBL dist1,dist2;
    int delta;
    x=EPoint.x();
    z=EPoint.z();
    delta = 0;
    x -= floor(x);
    z -= SQRT3*floor(z/SQRT3);
    /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > SQRT3_2 )
    {
        z = SQRT3 - z; /* mirror */
        delta = 1 - delta;
    }
    if (x > 0.5)
    {
        x = 1.0 - x; /* mirror */
    }
    if (x != 0.0)
    {
        slop1 = z/x;
        if (slop1>SQRT3)
        {
            z = SQRT3_2 - z;
            x = 0.5 -x;
            delta = 1 - delta;
        }
    }
    else
    {
        z = SQRT3_2 -z;
        x = 0.5;
    }
    dist1 = 1.0 - (z * 2 * SQRT3 );
    dist2 = 1.0 - ((SQRT3 * x - z) * SQRT3 );
    return delta/2.0+(0.5)*max(dist1,dist2);
}

static DBL tiling_lozenge (const Vector3d& EPoint)
{
    DBL x,z;
    DBL slop1;
    DBL dist1,dist2;

    x=EPoint.x();
    z=EPoint.z();

    x -= floor(x);
    z -= SQRT3*floor(z/SQRT3);
    /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
     ** There is some mirror to reduce the problem
     */
    if ( z > SQRT3_2 )
    {
        z -= SQRT3_2;
        x += 0.5;
    }
    if ( (2.0*z) > SQRT3_2 )
    {
        z = SQRT3_2 - z;
        x = 1.5 - x;
    }
    if (x > 0.75)
    {
        x -= 1.0;
    }
    if (x != 0.0)
    {
        slop1 = z/x;
        if (slop1>SQRT3)
        {
            z = SQRT3_2 - z;
            x = 0.5 -x;
        }
    }
    dist1 = 1.0 - (z * 4.0 * SQRT3 / 3.0 );
    dist2 = 1.0 - (fabs(SQRT3 * x - z) * SQRT3 *2.0 / 3.0);
    return max(dist1,dist2);
}

static DBL tiling_rhombus (const Vector3d& EPoint)
{
    DBL x,z;
    DBL answer;
    DBL slop1;
    DBL dist1,dist2;
    int delta;
    x=EPoint.x();
    z=EPoint.z();
    delta = 0;
    x += 0.5;
    x -= 3.0*floor(x/3.0);
    z -= SQRT3*floor(z/SQRT3);
    /* x,z is in { [0.0, 3.0 [, [0.0, SQRT3 [ }
     ** There is some mirror to reduce the problem
     */
    if ( z > SQRT3_2 )
    {
        z = SQRT3 -z; /* mirror */
        delta = 2 - delta;
    }
    if (x > 1.5)
    {
        x -= 1.5; /* translate */
        z = SQRT3_2 -z; /* mirror */
        delta = 2 - delta;
    }
    /* Now in [0,1.5],[0,SQRT3/2]
     ** from left to right
     ** part of a horizontal (z=0)
     ** half a vertical
     ** part of a horizontal
     */
    if (x < 0.5)
    {
        /* mirrror */
        x = 1.0 - x;
        delta = 2 - delta;
    }
    /*
     ** Let shift the [0.5,1.5],[0,SQRT3/2] to [0,1]....
     */
    x -= 0.5;
    if (x != 0.0)
    {
        slop1 = z/x;
        if (slop1>SQRT3)
        { /* rotate the vertical to match the horizontal on the right */
            dist1 = ( x / 2.0 ) + ( z * SQRT3_2 );
            dist2 = ( z / 2.0 ) - ( x * SQRT3_2 );
            z = dist2;
            x = dist1;
            delta = 1;
        }
    }
    else
    {
        /* rotate the vertical to match the horizontal on the right */
        dist1 = ( x / 2.0 ) + ( z * SQRT3_2 );
        dist2 = ( z / 2.0 ) - ( x * SQRT3_2 );
        z = dist2;
        x = dist1;
        delta = 1;
    }
    /* It may be similar to lozenge (in fact, IT IS !), now */

    if ( (2.0*z) > SQRT3_2 )
    {
        z = SQRT3_2 - z;
        x = 1.5 - x;
    }
    if (x > 0.75)
    {
        x -= 1.0;
    }
    if (x != 0.0)
    {
        slop1 = z/x;
        if (slop1>SQRT3)
        {
            z = SQRT3_2 - z;
            x = 0.5 -x;
        }
    }
    dist1 = 1.0 - (z * 4.0 * SQRT3 / 3.0 );
    dist2 = 1.0 - (fabs(SQRT3 * x - z) * SQRT3 *2.0 / 3.0);
    answer = max(dist1,dist2);
    answer /= 3.0;
    answer += delta/3.0;
    answer = min(1.0,answer);
    answer = max(0.0,answer);
    return answer;
}

static DBL tiling_rectangle (const Vector3d& EPoint)
{
    /*
     ** Tiling with rectangles
     ** resolve to square [0,4][0,4]
     ** then 16 cases
     **
     **  +-----+--+  +
     **  |     |  |  |
     **  +--+--+  +--+
     **     |  |  |
     **  +--+  +--+--+
     **  |  |  |     |
     **  +  +--+--+--+
     **  |  |     |  |
     **  +--+-----+  +
     */
    DBL x,z;
    DBL delta;
    DBL answer;
    int valueX,valueZ;
    x=EPoint.x();
    z=EPoint.z();
    x -= 4.0*floor(x/4.0);
    z -= 4.0*floor(z/4.0);
    valueX = (int)x;
    valueZ = (int)z;
    delta = 1.0;
    switch((valueX+valueZ*4))
    {
        case 0:
        case 4:
            z -= 1.0;
            break;
        case 1:
        case 2:
            x -= 2.0;
            delta = 0.0;
            break;
        case 3:
            x -= 3.0;
            break;
        case 5:
        case 9:
            x -= 1.0;
            z -= 2.0;
            break;
        case 6:
        case 7:
            x -= 3.0;
            z -= 1.0;
            delta = 0.0;
            break;
        case 8:
            z -= 2.0;
            delta = 0.0;
            break;
        case 10:
        case 14:
            x -= 2.0;
            z -= 3.0;
            break;
        case 11:
            x -= 4.0;
            z -= 2.0;
            delta = 0.0;
            break;
        case 12:
        case 13:
            x -= 1.0;
            z -= 3.0;
            delta = 0.0;
            break;
        case 15:
            x -= 3.0;
            z -= 4.0;
            break;
    }
    if (delta == 1.0)
    {
        x = 2*fabs(x-0.5);
        z = 2*(max(fabs(z),0.5)-0.5);
    }
    else
    {
        x = 2*(max(fabs(x),0.5)-0.5);
        z = 2*fabs(z-0.5);
    }
    answer = fabs(max(x,z)/2.0 + delta/2);
    return answer;
}

static DBL tiling_octa_square (const Vector3d& EPoint)
{
    /*
     ** Tiling with a square and an octagon
     */
    DBL answer;
    DBL x,z;
    DBL dist1,dist2;
    z=EPoint.z();
    z -= (SQRT2+1.0)*floor(z/(SQRT2+1.0));
    z -= SQRT2_2+0.5;
    z = fabs(z);
    x = EPoint.x();
    x -= (SQRT2+1.0)*floor(x/(SQRT2+1.0));
    x -= SQRT2_2+0.5;
    x = fabs(x);
    if (z > x)
    {
        answer = x;
        x = z;
        z = answer;
    }
    if ((x+z) < SQRT2_2)
    {
        /* Square tile */
        return ((x+z)/SQRT2);
    }
    dist1 = 1.0-z;
    dist2 = (SQRT2+SQRT2_2-(x+z))/SQRT2;

    return max(0.5+gkDBL_epsilon,max(dist1,dist2));
}

static DBL tiling_square_triangle (const Vector3d& EPoint)
{
    DBL x,z;
    DBL slop1;
    DBL dist1,dist2;
    int delta;
    x=EPoint.x();
    z=EPoint.z();
    delta = 0;
    x -= floor(x);
    z -= (2.0+SQRT3)*floor(z/(SQRT3+2.0));
    /* x,z is in { [0.0, 1.0 [, [0.0, 2+SQRT3 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > SQRT3_2+1.0 )
    {
        z -= SQRT3_2+1.0;
        x += (x>0.5)?-0.5:0.5;
    }
    if (x > 0.5)
    {
        x = 1.0 - x; /* mirror */
    }
    z -= 1.0;
    if (z > 0.0)
    { /* triangle */
        if (x != 0.0)
        {
            slop1 = z/x;
            if (slop1>SQRT3)
            {
                z = SQRT3_2 - z;
                x = 0.5 -x;
                delta = 1 - delta;
            }
        }
        else
        {
            z = SQRT3_2 -z;
            x = 0.5;
            delta = 1 - delta;
        }
        dist1 = 1.0 - (2 * z * SQRT3);
        dist2 = 1.0 - ((SQRT3 * x - z) * SQRT3);
        return (delta+max(dist1,dist2))/3.0;
    }
    else
    { /* square */
        if (z < -0.5)
        {
            z = -1.0 -z;
        }
        if (x > 0.5)
        {
            x = 1.0 - x;
        }
        dist1 = 2 + 2 * SQRT3 * fabs(x);
        dist2 = 2 + 2 * SQRT3 * fabs(z);
        dist1 = min(dist1,3.0);
        dist2 = min(dist2,3.0);
        // +1e-7 add & magnitude due numerical fuzziness.
        return min(1.0,((5.0+1e-7)-min(dist1,dist2))/3.0);
    }
}

static DBL tiling_hexa_triangle (const Vector3d& EPoint)
{
    /*
     ** Tiling with a hexagon and 2 triangles
     */
    DBL x,z;
    DBL dist1,dist2;
    DBL answer=0;
    int delta;
    x=EPoint.x();
    z=EPoint.z();
    delta = 0;
    x += 0.5;
    x -= 2.0*floor(x/2.0);
    z -= 2.0*SQRT3*floor(z/(SQRT3*2.0));
    /* x,z is in { [0.0, 2.0 [, [0.0, 2*SQRT3 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > SQRT3)
    {
        z -= SQRT3;
        x += (x<1.0)?1.0:-1.0;
    }
    /*
     ** Now only [0,2[,[0,SQRT3[
     */
    if (z > SQRT3_2)
    {
        z = SQRT3 - z; /* mirror */
        delta = 1 - delta;
    }

    if (x > 1.0)
    {
        x = 2.0 - x; /* mirror */
    }
    /*
     ** And now, it is even simpler :  [0,1],[0,SQRT3/2]
     ** on the bottom left corner, part of the triangle
     ** on the top right corner, center of the hexagon
     */
    if ((SQRT3*x+z)<SQRT3_2)
    {
        /* Triangle */
        dist1 = 1.0 - (z * 2 * SQRT3);
        dist2 = 1.0 + ((SQRT3 * x + z) - SQRT3_2) * SQRT3; /* really substracting */
        // -1e-7 sub & magnitude due numerical fuzziness.
        answer = (delta + max(dist1,dist2) - 1e-7)/3.0;
    }
    else
    {
        /* Hexagon */
        dist1 = 2 + 2 * (z * SQRT3);
        dist2 = 2 + 2 * ((SQRT3 * x + z - SQRT3_2)) * SQRT3_2;
        answer = 5.0-min(dist1,dist2);
        // +1e-7 add & magnitude due numerical fuzziness.
        answer = max(answer,2.0 + 1e-7)/3.0;
    }
    return min(1.0,answer);
}

static DBL tiling_square_offset (const Vector3d& EPoint)
{
    /*
     ** Tiling with a square, offset of half size
     ** Reduce to rectangle [0,1][0,2]
     ** move x,[1,2] to [0,1][0,1] with new x = x+1/2
     */
    DBL x,z;
    z = EPoint.z();
    z -= 2.0*floor(z/2.0);
    x = EPoint.x();
    if (z > 1.0)
    {
        x += 0.5;
        z -= 1.0;
    }
    x -= floor(x);
    x = 2*fabs(x-0.5);
    z = 2*fabs(z-0.5);

    return max(x,z);
}

static DBL tiling_square_rectangle (const Vector3d& EPoint)
{
    /*
     ** tiling with a central square and 4 rectangle (2x1)
     ** orbiting around the square
     ** Reduce to [0,3][0,3]
     ** then 9 cases
     **
     **  +-----+--+
     **  |     |  |
     **  +--+--+  |
     **  |  |  |  |
     **  |  +--+--+
     **  |  |     |
     **  +--+-----+
     */
    DBL x,z;
    DBL delta;
    int valueX,valueZ;
    x = EPoint.x();
    x -= 3.0*floor(x/3.0);
    z = EPoint.z();
    z -= 3.0*floor(z/3.0);
    valueX = (int)x;
    valueZ = (int)z;
    delta = 2.0;
    switch((valueX+valueZ*3))
    {
        case 0:
        case 3:
            z -= 1.0;
            break;
        case 1:
        case 2:
            x -= 2.0;
            delta = 1.0;
            break;
        case 4:
            x -= 1.0;
            z -= 1.0;
            delta = 0.0;
            break;
        case 5:
        case 8:
            x -= 2.0;
            z -= 2.0;
            break;
        case 6:
        case 7:
            x -= 1.0;
            z -= 2.0;
            delta = 1.0;
            break;
    }
    if (delta == 1.0)
    {
        x = fabs(x);
        x = 2.0*(max(x,0.5)-0.5);
        z = 2.0*fabs(z-0.5);
    }
    if (delta == 2.0)
    {
        x = 2.0*fabs(x-0.5);
        z = fabs(z);
        z = 2.0*(max(z,0.5)-0.5);
    }
    if (delta == 0.0)
    {
        x = 2.0*fabs(x-0.5);
        z = 2.0*fabs(z-0.5);
    }
    return ((max(x,z))+delta)/3.0;
}

static DBL tiling_rectangle_square (const Vector3d& EPoint)
{
    /*
     ** Tiling with a central square and 4 rectangles (2x1)
     ** which turns around the square in both directions
     ** Reduce to [0,6][0,6], fold in four and back
     ** to 9 cases.
     **
     **  +-----+--+
     **  |     |  |
     **  +--+--+  |
     **  |  |  |  |
     **  |  +--+--+
     **  |  |     |
     **  +--+--+--+
     */
    DBL x,z;
    DBL delta;
    int valueX,valueZ;
    x = EPoint.x();
    x -= 6.0*floor(x/6.0);
    x -= 3.0;
    x = fabs(x);
    z = EPoint.z();
    z -= 6.0*floor(z/6.0);
    z -= 3.0;
    z = fabs(z);
    valueX = (int)x;
    valueZ = (int)z;
    delta = 2.0;
    switch((valueX+valueZ*3))
    {
        case 0:
        case 3:
            z -= 1.0;
            break;
        case 1:
        case 2:
            x -= 2.0;
            delta = 1.0;
            break;
        case 4:
            x -= 1.0;
            z -= 1.0;
            delta = 0.0;
            break;
        case 5:
        case 8:
            x -= 2.0;
            z -= 2.0;
            break;
        case 6:
        case 7:
            x -= 1.0;
            z -= 2.0;
            delta = 1.0;
            break;
    }
    if (delta == 1.0)
    {
        x = fabs(x);
        x = 2.0*(max(x,0.5) -0.5);
        z = 2.0*fabs(z -0.5);
    }
    if (delta == 2.0)
    {
        x = 2.0*fabs(x-0.5);
        z = fabs(z);
        z = 2.0*(max(z,0.5) - 0.5);
    }
    if (delta == 0.0)
    {
        x = 2.0*fabs(x-0.5);
        z = 2.0*fabs(z-0.5);
    }
    return ((max(x,z))+delta)/3.0;
}

static DBL tiling_square_internal (const Vector3d& EPoint)
{
    DBL answer=0;
    DBL x,z;
    DBL dist1,dist2;
    int valueX,valueZ;
    x = EPoint.x();
    x *= SQRT2;
    x -= 4.0 * floor(x/4.0);
    x -= 2.0;
    x = fabs(x);
    z = EPoint.z();
    z *= SQRT2;
    z -= 4.0 * floor(z/4.0);
    z -= 2.0;
    z = fabs(z);
    valueX=(int)x;
    valueZ=(int)z;
    switch((valueX+valueZ*2))
    {
        case 0:
            x -= 0.5;
            x = max(x,0.0);
            x *= 2.0;
            z -= 0.5;
            z = max(z,0.0);
            z *= 2.0;
            answer=max(x,z)/3.0;
            break;
        case 1:
            answer=(2.0+fabs(1.5-x)*2.0)/3.0;
            if (z>0.5)
            {
                dist2=(3.0-SQRT2*fabs(x-z))/3.0;
                answer = max(answer,dist2);
            }
            break;
        case 2:
            answer=(1.0+fabs(1.5-z)*2.0)/3.0;
            if (x>0.5)
            {
                dist2=(2.0-SQRT2*fabs(x-z))/3.0;
                answer = max(answer,dist2);
            }
            break;
        case 3:
            if (x > z)
            {
                dist1=(2.0+fabs(1.5-x)*2.0)/3.0;
                dist2=(3.0-SQRT2*fabs(z-x))/3.0;
                answer=max(dist1,dist2);
            }
            else
            {
                dist1=(1.0+fabs(1.5-z)*2.0)/3.0;
                dist2=(2.0-SQRT2*fabs(x-z))/3.0;
                answer=max(dist1,dist2);
            }
            break;
    }
    return answer;
}

static DBL tiling_square_internal_5 (const Vector3d& EPoint)
{
    DBL answer=0;
    DBL x,z;
    DBL dist1,dist2;
    int mirX,mirZ;
    int valueX,valueZ;
    mirX = mirZ = 0;
    x = EPoint.x();
    x *= SQRT2;
    x -= 4.0 * floor(x/4.0);
    x -= 2.0;
    mirX = (x < 0)?1:0;
    x = fabs(x);
    z = EPoint.z();
    z *= SQRT2;
    z -= 4.0 * floor(z/4.0);
    z -= 2.0;
    mirZ = (z < 0)?2:3;
    z = fabs(z);
    valueX = (int)x;
    valueZ = (int)z;
    switch((valueX+valueZ*2))
    {
        case 0:
            x -= 0.5;
            x = max(x,0.0);
            x *= 2.0;
            z -= 0.5;
            z = max(z,0.0);
            z *= 2.0;
            // +1e-6 add & magnitude due numerical fuzziness.
            answer = ((4.0 + 1e-6) + max(x,z))/5.0;
            break;
        case 1:
            answer=fabs(1.5-x)*2.0;
            if (z>0.5)
            {
                dist2 = 1.0-SQRT2*fabs(x-z);
                answer = max(answer,dist2);
            }
            answer += mirX;
            answer /= 5.0;
            break;
        case 2:
            answer=fabs(1.5-z)*2.0;
            if (x>0.5)
            {
                dist2 = 1.0-SQRT2*fabs(x-z);
                answer = max(answer,dist2);
            }
            answer += mirZ;
            answer /= 5.0;
            break;
        case 3:
            if (x > z)
            {
                dist1 = fabs(1.5-x)*2.0;
                dist2 = 1.0-SQRT2*fabs(z-x);
                answer = max(dist1,dist2) + mirX;
            }
            else
            {
                dist1 = fabs(1.5-z)*2.0;
                dist2 = 1.0-SQRT2*fabs(x-z);
                answer = max(dist1,dist2) + mirZ;
            }
            answer /= 5.0;
            break;
    }
    return answer;
}

static DBL tiling_square_double (const Vector3d& EPoint)
{
    /*
     ** Tiling with a square (1x1) and a square (2x2)
     ** Reduce to [0,5][0,5] then 25 cases
     **
     **  +--+     +--+--+
     **     |     |  |
     **     +--+--+--+
     **     |  |     |
     **  +--+--+     +--+
     **  |     |     |  |
     **  +     +--+--+--+
     **  |     |  |     |
     **  +--+--+--+     +
     **  |  |     |     |
     **  +--+     +--+--+
     */
    DBL x,z;
    DBL delta;
    int valueX,valueZ;
    x = EPoint.x();
    x -= 5.0*floor(x/5.0);
    z = EPoint.z();
    z -= 5.0*floor(z/5.0);
    valueX = (int)x;
    valueZ = (int)z;
    // +1e-7 add & magnitude due numerical fuzziness.
    delta = 0.5 + 1e-7;
    switch((valueX+valueZ*5))
    {
        case 0:
            delta = 0.0;
            break;
        case 1:
        case 2:
            x -= 2.0;
            break;
        case 3:
        case 4:
        case 8:
        case 9:
            x -= 4.0;
            z -= 1.0;
            break;
        case 5 :
        case 6 :
        case 10:
        case 11:
            z -= 2.0;
            x -= 1.0;
            break;
        case 7:
            delta = 0.0;
            x -= 2.0;
            z -= 1.0;
            break;
        case 12:
        case 13:
        case 17:
        case 18:
            x -= 3.0;
            z -= 3.0;
            break;
        case 14:
            x -= 4.0;
            z -= 2.0;
            delta = 0.0;
            break;
        case 15:
        case 20:
            z -= 4.0;
            break;
        case 16:
            x -= 1.0;
            z -= 3.0;
            delta = 0.0;
            break;
        case 21:
        case 22:
            x -= 2.0;
            z -= 5.0;
            break;
        case 23:
            x -= 3.0;
            z -= 4.0;
            delta = 0.0;
            break;
        case 24:
        case 19:
            x -= 5.0;
            z -= 4.0;
            break;
    }
    if (delta)
    {
        x = fabs(x);
        x = 2*( max(x,0.5) -0.5);
        z = fabs(z);
        z = 2*( max(z,0.5) -0.5);
    }
    else
    {
        x = 2*fabs(x-0.5);
        z = 2*fabs(z-0.5);
    }
    return min(1.0,fabs((max(x,z))/2.0+delta));
}

static DBL tiling_hexa_square_triangle (const Vector3d& EPoint)
{
    /*
     ** tiling with 1 hexagon, squares and triangles
     */
    DBL x,z;
    DBL dist1,dist2;
    DBL answer=0;

    x = EPoint.x();
    z = EPoint.z();
    x += 0.5;
    x -= (3.0+SQRT3)*floor(x/(3.0+SQRT3));
    z -= 2.0*(1.0+SQRT3)*floor(z/((SQRT3+1.0)*2.0));
    /* x,z is in { [0.0, 3.0+SQRT3 [, [0.0, 2*(1+SQRT3) [ }
     ** but there is some symmetry to simplify the testing
     */
    if (x > (SQRT3_2+1.5))
    {
        x -= (SQRT3_2+1.5);
        z += (z < SQRT3+1.0)?(SQRT3_2+0.5):(-1*(SQRT3_2+0.5));
    }
    if (z > (SQRT3+1.0))
    {
        z -= SQRT3+1.0;
    }
    /*
     ** Now only [0, SQRT3/2+1.5 ], [0,SQRT3+1.0]
     */
    if (z > (0.5+SQRT3_2))
    {
        z = SQRT3+1.0 - z; /* mirror */
    }

    if (x > (1.0+SQRT3_2))
    {
        x = SQRT3+2.0 - x; /* mirror */
    }
    /*
     ** And now, it is even simpler :  [0,1+(SQRT3/2)],[0,1+(SQRT3/2)]
     ** on the top right corner, center of the hexagon
     ** on the bottom right, middle of square, a triangle on the left
     ** on the top left corner, half a triangle
     ** bottom
     */
    if (((SQRT3*x+z)<(SQRT3_2+2.0))&&(x < 0.5+SQRT3_2))
    {
        /* Triangle or square */
        /* rotate in the lower part */
        z -= 0.5 + SQRT3_2;
        x -= 1.0 + SQRT3_2;
        dist1 = (x / 2.0) - (z * SQRT3_2);
        dist2 = (z / 2.0) + (x * SQRT3_2);
        z = dist2;
        x = dist1;
        z += 0.5 + SQRT3_2;
        x += 1.0 + SQRT3_2;
        if (z < 0)
        {
            z *= -1;
        }
        if (x > (1.0+SQRT3_2))
        {
            x = SQRT3+2.0 - x; /* mirror */
        }
    }
    if ((!((SQRT3*x+z)<(SQRT3_2+2.0)))&&(!(z < 0.5)))
    {
        /* Hexagon */
        z -= 0.5;
        x -= SQRT3_2;
        z= fabs(z);
        x= fabs(x);
        dist1 = 2* z * SQRT3;
        dist2 = ((SQRT3 * x + z ) ) * SQRT3 - 1.5;
        answer = 3.0-min(dist1,dist2);
        // +1e-7 add & magnitude due numerical fuzziness.
        answer = max(answer,2.0 + 1e-7);
        answer /= 3.0;
        return answer;
    }
    if (z < 0.5 )
    {
        if ( x > SQRT3_2+0.5)
        {
            /* square */
            x -= SQRT3_2+0.5;
            z -= 0.5;
            dist1 = 1.0 + z*2 * SQRT3;
            dist2 = 1.0 - x*2 * SQRT3;
            dist1 = max(dist1,0.0);
            dist2 = max(dist2,0.0);
            answer = max(dist1,dist2)/3.0;
        }
        else
        {
            /* triangle */
            x -= SQRT3_2 + 0.5;
            x = fabs(x);
            dist2 = ((SQRT3 * z + x)) * SQRT3 +0.5;
            dist1 = 2.0 - (x * 2.0 * SQRT3);
            answer = (max(dist1,dist2))/3.0;
        }
    }
    return answer;
}

static DBL tiling_hexa_square_triangle_6 (const Vector3d& EPoint)
{
    /*
     ** tiling with 1 hexagon, squares and triangles
     ** all tiles get its own colour (according to its orientation)
     ** 1 hexagon, 3 squares, 2 triangles
     */
    DBL x,z;
    DBL dist1,dist2;
    DBL answer=0;
    int mirX=0;
    int mirZ=0;
    int rota=0;
    x = EPoint[X];
    z = EPoint[Z];
    x += 0.5;
    x -= (3.0+SQRT3)*floor(x/(3.0+SQRT3));
    z -= 2.0*(1.0+SQRT3)*floor(z/((SQRT3+1.0)*2.0));
    /* x,z is in { [0.0, 3.0+SQRT3 [, [0.0, 2*(1+SQRT3) [ }
     ** but there is some symmetry to simplify the testing
     */
    if (x > (SQRT3_2+1.5))
    {
        x -= (SQRT3_2+1.5);
        z += (z < SQRT3+1.0)?(SQRT3_2+0.5):(-1*(SQRT3_2+0.5));
    }
    if (z > (SQRT3+1.0))
    {
        z -= SQRT3+1.0;
    }
    /*
     ** Now only [0, SQRT3/2+1.5 ], [0,SQRT3+1.0]
     */
    if (z > (0.5+SQRT3_2))
    {
        z = SQRT3+1.0 - z; /* mirror */
        mirZ = 1 - mirZ;
    }

    if (x > (1.0+SQRT3_2))
    {
        x = SQRT3+2.0 - x; /* mirror */
        mirX = 1 - mirX;
    }
    /*
     ** And now, it is even simpler :  [0,1+(SQRT3/2)],[0,1+(SQRT3/2)]
     ** on the top right corner, center of the hexagon
     ** on the bottom right, middle of square, a triangle on the left
     ** on the top left corner, half a triangle
     ** bottom
     */
    if (((SQRT3*x+z)<(SQRT3_2+2.0))&&(x < 0.5+SQRT3_2))
    {
        /* Triangle or square */
        /* rotate in the lower part */
        z -= 0.5 + SQRT3_2;
        x -= 1.0 + SQRT3_2;
        dist1 = (x / 2.0) - (z * SQRT3_2);
        dist2 = (z / 2.0) + (x * SQRT3_2);
        z = dist2;
        x = dist1;
        z += 0.5 + SQRT3_2;
        x += 1.0 + SQRT3_2;
        rota = 1 - rota;
        if (z < 0)
        {
            z *= -1;
        }
        if (x > (1.0+SQRT3_2))
        {
            x = SQRT3+2.0 - x; /* mirror */
            mirX = 1 - mirX;
            mirZ = 1 - mirZ;
        }
    }
    if ((!((SQRT3*x+z)<(SQRT3_2+2.0)))&&(!(z < 0.5)))
    {
        /* Hexagon */
        z -= 0.5;
        x -= SQRT3_2;
        z= fabs(z);
        x= fabs(x);
        dist1 = 2* z * SQRT3;
        dist2 = ((SQRT3 * x + z ) ) * SQRT3 - 1.5;
        answer = 6.0 - min(dist1,dist2);
        // +1e-6 add & magnitude due numerical fuzziness.
        answer = max(answer,5.0 + 1e-6);
        answer /= 6.0;
        return answer;
    }
    if (z < 0.5 )
    {
        if ( x > SQRT3_2+0.5)
        {
            /* square */
            x -= SQRT3_2+0.5;
            z -= 0.5;
            dist1 = 1.0 + z*2 * SQRT3;
            dist2 = 1.0 - x*2 * SQRT3;
            dist1 = max(dist1,0.0);
            dist2 = max(dist2,0.0);
            // +1e-6 add & magnitude due numerical fuzziness.
            answer = (max(dist1,dist2) + rota * ((1.0 + 1e-6) + ((mirX + mirZ) % 2)))/6.0;
        }
        else
        {
            /* triangle */
            x -= SQRT3_2 + 0.5;
            x = fabs(x);
            dist2 = ((SQRT3 * z + x)) * SQRT3 +0.5;
            dist1 = 2.0 - (x * 2.0 * SQRT3);
            answer = max(dist1,dist2);
            if ((rota + mirX)%2)
            {
                answer = 2.0 + answer;
            }
            else
            {
                answer = 3.0 + answer;
            }
            answer /= 6.0;
        }
    }
    return answer;
}

static DBL tiling_rectangle_pair (const Vector3d& EPoint)
{

    /*
     ** Tiling with 2 rectangles (2x1)
     ** Reduce to [0,4][0,4] then 16 cases
     **
     **  +-----+--+--+
     **  |     |  |  |
     **  +-----+  +  +
     **  |     |  |  |
     **  +-----+--+--+
     **  |  |  |     |
     **  |  |  +-----+
     **  |  |  |     |
     **  +--+--+-----+
     */
    DBL x,z;
    DBL delta;
    DBL answer;
    int valueX,valueZ;
    x = EPoint.x();
    z = EPoint.z();
    x -= 4.0*floor(x/4.0);
    z -= 4.0*floor(z/4.0);
    valueX = (int)x;
    valueZ = (int)z;
    delta = 1.0;
    switch((valueX+valueZ*4))
    {
        case 0:
        case 4:
            z -= 1.0;
            break;
        case 1:
        case 5:
            x -= 1.0;
            z -= 1.0;
            break;
        case 2:
        case 3:
            x -= 3.0;
            delta = 0.0;
            break;
        case 6:
        case 7:
            x -= 3.0;
            z -= 1.0;
            delta = 0.0;
            break;
        case 8:
        case 9:
            x -= 1.0;
            z -= 2.0;
            delta = 0.0;
            break;
        case 12:
        case 13:
            x -= 1.0;
            z -= 3.0;
            delta = 0.0;
            break;
        case 10:
        case 14:
            x -= 2.0;
            z -= 3.0;
            break;
        case 11:
        case 15:
            x -= 3.0;
            z -= 3.0;
            break;
    }
    if (delta == 1.0)
    {
        x = 2*fabs(x-0.5);
        z = 2*(max(fabs(z),0.5) -0.5);
    }
    else
    {
        x = 2*(max(fabs(x),0.5) -0.5);
        z = 2*fabs(z -0.5);
    }
    answer = fabs((max(x,z)+delta)/2.0);
    return answer;
}

static DBL tiling_hexa_tri_right (const Vector3d& EPoint)
{
    DBL x,z;
    DBL answer;
    DBL slop1;
    DBL dist1,dist2;
    int zzof;
    int delta;
    x = EPoint.x();
    z = EPoint.z();
    /* First, resume to a simple pattern */
    zzof = z / SQRT3_2 - gkDBL_epsilon;  // epsilon helps at +z edge.
    zzof /= 3;
    if (z < 0.0)
    {
        zzof -= 1;
    }
    x += zzof / 2.0; /* right handed */
    // +1e-7 add & magnitude due numerical fuzziness.
    z -= (3*SQRT3_2*floor(z/(3*SQRT3_2)) + 1e-7); // helps with triangles at +z
    x += 7.0;
    x -= 7.0 * floor(x/7.0);
    if ((x > 4.5) && (z > SQRT3))
    {
        x -= 4.5;
        z -= SQRT3_2;
    }
    if ((x > 5.0) && (z < SQRT3_2))
    {
        x -= 5;
        z += SQRT3;
    }
    if ((x > 2.5) && (z < SQRT3))
    {
        x -= 2.5;
        z += SQRT3_2;
    }
    delta = 0;
    zzof = z / SQRT3_2;
    if (zzof == 2)
    {
        zzof = 1;
        z = 2 * SQRT3 - z;
        delta = 1 - delta;
    }
    if ( (!zzof) || (x > 2.0) ||
         ((z + SQRT3*x < SQRT3) || (SQRT3*x - z > SQRT3)) )
    {
        /* triangle */
        x -= 1.0 *floor(x/1.0);
        z -= SQRT3*floor(z/SQRT3);
        /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
         ** but there is some symmetry to simplify the testing
         */
        if (z > SQRT3_2)
        {
            z = SQRT3 - z; /* mirror */
            delta = 1 - delta;
        }
        if (x > 0.5)
        {
            x = 1.0 - x; /* mirror */
        }
        if (x != 0.0)
        {
            slop1 = z/x;
            if (slop1>SQRT3)
            {
                z = SQRT3_2 - z;
                x = 0.5 -x;
                delta = 1 - delta;
            }
        }
        else
        {
            z = SQRT3_2 - z;
            x = 0.5;
            delta = 1 - delta;
        }
        dist1 = 1.0 - (z * 2 * SQRT3);
        dist2 = 1.0 - ((SQRT3 * x - z) * SQRT3);
        // -1e-7 sub & magnitude due numerical fuzziness.
        answer = (delta + max(dist1,dist2) - 1e-7)/3.0;
    }
    else
    { /* hexagon */
        z -= SQRT3_2;
        if (x > 1.0)
        {
            x = 2.0 - x; /* mirror */
        }
        /* Hexagon */
        dist1 = 2 + 2 * (z * SQRT3);
        dist2 = 2 + 2 * ((SQRT3 * x + z - SQRT3_2)) * SQRT3_2;
        answer = 5.0 - min(dist1,dist2);
        // +1e-7 add & magnitude due numerical fuzziness.
        answer = max(answer, 2.0 + 1e-7)/3.0;
    }
    return min(1.0,answer);
}

static DBL tiling_hexa_tri_left (const Vector3d& EPoint)
{
    DBL x,z;
    DBL slop1;
    DBL dist1,dist2;
    DBL answer;
    int zzof;
    int delta;
    x = EPoint.x();
    z = EPoint.z();
    /* First, resume to a simple pattern */
    zzof = z / SQRT3_2 - gkDBL_epsilon;  // epsilon helps at +z edge.
    zzof /= 3;
    if (z < 0)
    {
        zzof -= 1;
    }
    x -= zzof / 2.0; /* left handed */
    // +1e-7 add & magnitude due numerical fuzziness.
    z -= (3*SQRT3_2*floor(z/(3*SQRT3_2)) + 1e-7); // helps with triangles at +z
    x += 7.0;
    x -= 7.0 * floor(x/7.0);
    if ((x > 2.0) && (z < SQRT3_2))
    {
        x -= 2.0;
        z += SQRT3;
    }
    if ((x > 4.5) && (z < SQRT3))
    {
        x -= 4.5;
        z += SQRT3_2;
    }
    if ((x > 2.5) && (z > SQRT3))
    {
        x -= 2.5;
        z -= SQRT3_2;
    }
    delta = 0;
    zzof = z/SQRT3_2;
    if (zzof == 2)
    {
        zzof = 1;
        z = 2 * SQRT3 - z;
        delta = 1 - delta;
    }
    if ( (!zzof) || (x > 2.0) ||
         ((z + SQRT3*x < SQRT3) || (SQRT3*x -z > SQRT3)) )
    {
        /* triangle */
        x -= 1.0 * floor(x/1.0);
        z -= SQRT3 * floor(z/SQRT3);
        /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
         ** but there is some symmetry to simplify the testing
         */
        if (z > SQRT3_2)
        {
            z = SQRT3 - z; /* mirror */
            delta = 1 - delta;
        }
        if (x > 0.5)
        {
            x = 1.0 - x; /* mirror */
        }
        if (x != 0.0)
        {
            slop1 = z/x;
            if (slop1>SQRT3)
            {
                z = SQRT3_2 - z;
                x = 0.5 -x;
                delta = 1 - delta;
            }
        }
        else
        {
            z = SQRT3_2 -z;
            x = 0.5;
            delta = 1 - delta;
        }
        dist1 = 1.0 - (z * 2 * SQRT3);
        dist2 = 1.0 - ((SQRT3 * x - z) * SQRT3);
        // -1e-7 sub & magnitude due numerical fuzziness.
        answer = (delta + max(dist1,dist2) - 1e-7)/3.0;
    }
    else
    { /* hexagon */
        z -= SQRT3_2;
        if (x > 1.0)
        {
            x = 2.0 - x; /* mirror */
        }
        /* Hexagon */
        dist1 = 2 + 2 * (z * SQRT3);
        dist2 = 2 + 2 * ((SQRT3 * x + z - SQRT3_2)) * SQRT3_2;
        answer = 5.0 - min(dist1,dist2);
        // +1e-7 add & magnitude due numerical fuzziness.
        answer = max(answer,2.0 + 1e-7)/3.0;
    }
    return min(1.0,answer);
}

static DBL tiling_square_tri (const Vector3d& EPoint)
{
    DBL x,z;
    DBL slop1;
    DBL dist1,dist2;
    int delta;
    int gamma,beta;
    int xflop,zflop;
    x = EPoint.x();
    z = EPoint.z();
    delta = 0;
    gamma = 0;
    beta = 0;
    x -= (1.0+SQRT3)*floor(x/(1.0+SQRT3));
    z -= (1.0+SQRT3)*floor(z/(1.0+SQRT3));
    /* x,z is in { [0.0, SQRT3+1.0 [, [0.0, SQRT3+1.0 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > 0.5 + SQRT3_2)
    {
        z = 1.0 + SQRT3 - z; /* mirror */
        delta = 1 - delta;
    }
    if (x > 0.5 + SQRT3_2)
    {
        x = 1.0 + SQRT3 - x; /* mirror */
        delta = 1 - delta;
        beta = 1;
    }
    /* x,z is in { [0.0, (SQRT3+1)/2 ], [0.0, (SQRT3+1)/2 ] }
     ** but there is still a symmetry and a rotation
     */
    xflop = (2.0 * x + (SQRT3 -1.0)* 2.0 * z / (SQRT3 + 1.0) > (SQRT3)) ? 1: 0;
    zflop = (2.0 * z * SQRT3 + (SQRT3 -3.0)* 2.0 * x / (SQRT3 + 1.0) > (SQRT3)) ? 1: 0;
    switch (xflop + 2*zflop)
    {
        case 0: /* do nothing */
            gamma = 2 * beta;
            break;
        case 1: /* rotate clockwise */
            gamma = beta ? 1 + 2 * delta : 3 - 2 *delta;
            slop1 = x;
            x = z;
            z = SQRT3_2 + 0.5 - slop1;
            break;
        case 2: /* rotate normal */
            gamma = beta ? 3 - 2 * delta : 1 + 2 *delta;
            slop1 = z;
            z = x;
            x = SQRT3_2 +0.5 - slop1;
            break;
        case 3: /* symmetry */
            gamma = beta ? 0: 2;
            x = SQRT3_2+0.5 -x;
            z = SQRT3_2+0.5 -z;
            break;
    }

    if (x == 0.0)
    {
        z = 0.0;
        x = 0.0;
    }
    slop1 = (z * 2 * SQRT3) - SQRT3 + x * 2;
    if (slop1 < 0)
    {
        /* triangle */

        dist1 = -1.5 / SQRT3 * slop1;
        dist2 = 2.0 * x * SQRT3;
        slop1 = min(dist1,dist2);
        return (gamma+1.0-slop1)/6.0;
    }
    else
    {
        /* square */
        slop1 *= 1.5 / SQRT3;
        slop1 = min(slop1,1.0);
        if (delta)
        {
            slop1 *= -1;
            slop1 += 6.0;
            slop1 /= 6.0;
            // +1e-6 add & magnitude due numerical fuzziness.
            slop1 = max(slop1,(5.0 + 1e-6)/6.0);
            slop1 = min(slop1,1.0);
        }
        else
        {
            slop1 *= -1;
            slop1 += 5.0;
            slop1 /= 6.0;
            slop1 = min(slop1,5.0/6.0);
            // +1e-6 add & magnitude due numerical fuzziness.
            slop1 = max(slop1,(4.0 + 1e-6)/6.0);
        }
        return slop1;
    }
}

static DBL tiling_dodeca_tri (const Vector3d& EPoint)
{
    DBL x,z;
    DBL dist1,dist2,dist3,dist4,ret_value;
    DBL tmpx,tmpz;
    int toggle = 0; /* switched each time a triangle get toggled */
    x = EPoint.x();
    z = EPoint.z();
    x -= (2.0+SQRT3)*floor(x/(2.0+SQRT3));
    z -= (3.0+2.0*SQRT3)*floor(z/(3.0+2.0*SQRT3));
    /* x,z is in { [0.0, SQRT3+2.0 [, [0.0, 2*SQRT3+3.0 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > SQRT3+1.5)
    {
        /* translate */
        z -= SQRT3+1.5;
        x += (x<(1.0+SQRT3_2) ? 1.0: -1.0)*(1.0+SQRT3_2);
    }
    if (x > 1.0+SQRT3_2)
    {
        x = 2.0 + SQRT3 -x;
    }
    if (z > 1.0+SQRT3_2)
    {
        z = 2.0 + SQRT3 -z;
        toggle = 1 - toggle;
    }
    dist2 = x - SQRT3_2 - 0.5 + z * SQRT3;
    if (dist2 < 0.0)
    {
        tmpx = x;
        tmpz = z;
        x = (1.0+SQRT3)/4.0 + 0.5 * tmpx - SQRT3_2 * tmpz;
        z = (3.0+SQRT3)/4.0 - SQRT3_2 * tmpx - 0.5 * tmpz;
        dist2 *= -1.0;
    }
    dist1 = (z * 3.0); /* from the bottom line */
    dist3 =  z - SQRT3_2 - 0.5 + x * SQRT3;
    dist3 *= SQRT3;
    dist2 *= SQRT3;
    dist4 = (x * 3.0); /* from the vertical line */
    if (dist3 < 0.0)
    {
        ret_value = max(dist3,-1.0);
        ret_value += 1+toggle;
        ret_value /= 3.0;
    }
    else
    {
        /* dodecagon */
        ret_value = min(dist1,dist2);
        ret_value = min(ret_value,dist3);
        ret_value = min(ret_value,dist4);
        ret_value = min(ret_value,1.0);
        // +1e-7 add & magnitude due numerical fuzziness.
        ret_value = (3.0 + 1e-7) - ret_value;
        ret_value /= 3.0;
    }
    return min(1.0,ret_value);
}

static DBL tiling_dodeca_hex (const Vector3d& EPoint)
{
    DBL x,z;
    DBL dist1,dist2,dist3,dist4,ret_value;
    DBL dist5,dist6,dist7;
    DBL tmpx,tmpz;
    x = EPoint.x();
    z = EPoint.z();
    x -= (3.0+SQRT3)*floor(x/(3.0+SQRT3));
    z -= (3.0+3.0*SQRT3)*floor(z/(3.0+3.0*SQRT3));
    /* x,z is in { [0.0, SQRT3+3.0 [, [0.0, 3*SQRT3+3.0 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > 1.5*SQRT3+1.5)
    {
        /* translate */
        z -= 1.5*SQRT3+1.5;
        x += (x<(1.5+SQRT3_2) ? 1.0: -1.0)*(1.5+SQRT3_2);
    }
    if (x > 1.5+SQRT3_2)
    {
        x = 3.0 + SQRT3 -x;
    }
    if (z > 1.0+SQRT3)
    {
        z = 2.0 + 2.0*SQRT3 -z;
    }
    dist2 = x - SQRT3_2 - 1.5 + z * SQRT3;
    if (dist2 < 0.0)
    {
        tmpx = x;
        tmpz = z;
        x = (3.0+SQRT3)/4.0 + 0.5 * tmpx - SQRT3_2 * tmpz;
        z = (3.0+3.0*SQRT3)/4.0 - SQRT3_2 * tmpx - 0.5 * tmpz;
    }
    dist2 = x - SQRT3_2 - 2.5 + z * SQRT3;
    dist1 = (z * 2.0) - SQRT3; /* from the bottom line */
    dist3 = z - 1.5*SQRT3 - 0.5 + x * SQRT3;
    dist4 = ((x-0.5) * 2.0); /* from the vertical line */
    if ( (dist2 >= 0.0) &&
         (dist1 >= 0.0) &&
         (dist3 >= 0.0) &&
         (dist4 >= 0.0) )
    {
        /* dodecagon */
        ret_value = min(dist1,dist2);
        ret_value = min(ret_value,dist3);
        ret_value = min(ret_value,dist4);
        ret_value = min(ret_value,1.0);
        // +1e-6 add & magnitude due numerical fuzziness.
        ret_value = (3.0 + 1e-6) - ret_value;
    }
    else
    {
        dist5 = 2*z - 2*SQRT3 - 1.0;
        if (dist5 >= 0)
        {
            dist4 *= -1.0;
            ret_value = min(dist5,dist4);
            ret_value = 2.0 - ret_value;
        }
        else
        {
            dist6 = SQRT3 * x - z -SQRT3_2 +0.5;
            dist7 = dist6 - 2.0;
            switch((dist6 >= 0?0:1)+(dist7 >= 0.0 ?2:0))
            {
                case 1: /* left hexagon */
                    dist5 *= -1.0;
                    dist6 *= -1.0;
                    dist3 *= -1.0;
                    ret_value = min(dist6,dist3);
                    ret_value = min(ret_value,dist5);
                    ret_value = min(ret_value,1.0);
                    ret_value = 1.0 - ret_value;
                    break;
                case 2: /* bottom hexagon */
                    dist1 *= -1.0;
                    ret_value = min(dist7,dist1);
                    ret_value = min(ret_value,1.0);
                    ret_value = 1.0 - ret_value;
                    break;
                case 0:
                default: /* slanted square */
                    dist2 *= -1.0;
                    dist7 *= -1.0;
                    ret_value = min(dist6,dist2);
                    ret_value = min(ret_value,dist7);
                    ret_value = 2.0 - ret_value;
                    break;
            }
        }
    }
    ret_value /= 3.0;
    return ret_value;
}

static DBL tiling_dodeca_hex_5 (const Vector3d& EPoint)
{
    DBL x,z;
    DBL dist1,dist2,dist3,dist4,ret_value;
    DBL dist5,dist6,dist7;
    DBL tmpx,tmpz;
    int mirX,mirZ,rota;
    mirX = mirZ = rota = 0;
    x = EPoint.x();
    z = EPoint.z();
    x -= (3.0+SQRT3)*floor(x/(3.0+SQRT3));
    z -= (3.0+3.0*SQRT3)*floor(z/(3.0+3.0*SQRT3));
    /* x,z is in { [0.0, SQRT3+3.0 [, [0.0, 3*SQRT3+3.0 [ }
     ** but there is some symmetry to simplify the testing
     */
    if (z > 1.5*SQRT3+1.5)
    {
        /* translate */
        z -= 1.5*SQRT3+1.5;
        x += (x<(1.5+SQRT3_2) ? 1.0: -1.0)*(1.5+SQRT3_2);
    }
    if (x > 1.5+SQRT3_2)
    {
        x = 3.0 + SQRT3 -x;
        mirX = 1 - mirX;
    }
    if (z > 1.0+SQRT3)
    {
        z = 2.0 + 2.0*SQRT3 -z;
        mirZ = 1 - mirZ;
    }
    dist2 = x - SQRT3_2 - 1.5 + z * SQRT3;
    if (dist2 < 0.0)
    {
        tmpx = x;
        tmpz = z;
        x = (3.0+SQRT3)/4.0 + 0.5 * tmpx - SQRT3_2 * tmpz;
        z = (3.0+3.0*SQRT3)/4.0 - SQRT3_2 * tmpx - 0.5 * tmpz;
    }
    dist2 = x - SQRT3_2 - 2.5 + z * SQRT3;
    dist1 = (z * 2.0) - SQRT3; /* from the bottom line */
    dist3 = z - 1.5*SQRT3 - 0.5 + x * SQRT3;
    dist4 = ((x-0.5) * 2.0); /* from the vertical line */
    if ( (dist2 >= 0.0) &&
         (dist1 >= 0.0) &&
         (dist3 >= 0.0) &&
         (dist4 >= 0.0) )
    {
        /* dodecagon */
        ret_value = min(dist1,dist2);
        ret_value = min(ret_value,dist3);
        ret_value = min(ret_value,dist4);
        ret_value = min(ret_value,1.0);
        // +1e-6 add & magnitude due numerical fuzziness.
        ret_value = (5.0 + 1e-6) - ret_value;
    }
    else
    {
        dist5 = 2*z - 2*SQRT3 - 1.0;
        if (dist5 >= 0)
        {
            dist4 *= -1.0;
            ret_value = min(dist5,dist4);
            ret_value = 2.0 - ret_value;
        }
        else
        {
            dist6 = SQRT3 * x - z -SQRT3_2 +0.5;
            dist7 = dist6 - 2.0;
            switch((dist6 >= 0 ? 0:1)+(dist7 >= 0.0 ? 2:0))
            {
                case 1: /* left hexagon */
                    dist5 *= -1.0;
                    dist6 *= -1.0;
                    dist3 *= -1.0;
                    ret_value = min(dist6,dist3);
                    ret_value = min(ret_value,dist5);
                    ret_value = min(ret_value,1.0);
                    ret_value = 1.0 - ret_value;
                    break;
                case 2: /* bottom hexagon */
                    dist1 *= -1.0;
                    ret_value = min(dist7,dist1);
                    ret_value = min(ret_value,1.0);
                    ret_value = 1.0 - ret_value;
                    break;
                case 0:
                default: /* slanted square */
                    dist2 *= -1.0;
                    dist7 *= -1.0;
                    ret_value = min(dist6,dist2);
                    ret_value = min(ret_value,dist7);
                    ret_value = (mirZ + mirX) %2 + 3.0 - ret_value;
                    break;
            }
        }
    }
    ret_value /= 5.0;
    return ret_value;
}

static DBL tiling_penrose_halfkite (DBL pX, DBL pZ, int depth, bool rhombs);
static DBL tiling_penrose_halfdart (DBL pX, DBL pZ, int depth, bool rhombs);

static DBL tiling_penrose_halfkite (DBL pX, DBL pZ, int depth, bool rhombs)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (z < (x - INVPHI) * TAN144)
        {
            return tiling_penrose_halfdart(x*PHI, z*PHI, depth-1, rhombs);
        }
        else
        {
            x -= COS36;
            z -= SIN36;
            DBL rotX = x*COS108 - z*SIN108;
            DBL rotZ = z*COS108 + x*SIN108;
            return tiling_penrose_halfkite(rotX*PHI, rotZ*PHI, depth-1, rhombs);
        }
    }
    else if (rhombs)
    {
        if (z < (x - INVPHI) * TAN72)
        {
            DBL dist1 = abs(SIN72  * (x-INVPHI) - COS72  * z) * 5.55; // TODO FIXME - the factor is just an empiric value
            DBL dist2 = abs(SIN108 * (x-1)      - COS108 * z) * 5.55;
            return max3(1.0-dist1/2,1.0-dist2/2,0.5+1e-6);
        }
        else
        {
            DBL dist1 = abs(z)                              * 5.55; // TODO FIXME - the factor is just an empiric value
            DBL dist2 = abs(SIN72 * (x-INVPHI) - COS72 * z) * 5.55;
            return min(max3(0.5-dist1/2,0.5-dist2/2,0.0),0.5-1e-6);
        }
    }
    else
    {
        DBL dist1 = abs(SIN36  *  x    - COS36  * z) * 4.46; // TODO FIXME - the factor is just an empiric value
        DBL dist2 = abs(SIN108 * (x-1) - COS108 * z) * 4.46;
        return min(max3(0.5-dist1/2,0.5-dist2/2,0.0),0.5-1e-6);
    }
}

static DBL tiling_penrose_halfdart (DBL pX, DBL pZ, int depth, bool rhombs)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (z < (x - INVPHI) * TAN108)
        {
            DBL rotX = x*COS36 + z*SIN36;
            DBL rotZ = z*COS36 - x*SIN36;
            return tiling_penrose_halfkite(rotX*PHI, rotZ*PHI, depth-1, rhombs);
        }
        else
        {
            x -= 1;
            DBL rotX = x*COS144 + z*SIN144;
            DBL rotZ = z*COS144 - x*SIN144;
            return tiling_penrose_halfdart(rotX*PHI, rotZ*PHI, depth-1, rhombs);
        }
    }
    else if (rhombs)
    {
        DBL dist1 = abs(SIN36  *  x    - COS36  * z) * 5.55; // TODO FIXME - the factor is just an empiric value
        DBL dist2 = abs(SIN144 * (x-1) - COS144 * z) * 5.55;
        return min(max3(0.5-dist1/2,0.5-dist2/2,0.0),0.5-1e-6);
    }
    else
    {
        DBL dist1 = abs(z)                           * 4.46; // TODO FIXME - the factor is just an empiric value
        DBL dist2 = abs(SIN144 * (x-1) - COS144 * z) * 4.46;
        return max3(1.0-dist1/2,1.0-dist2/2,0.5+1e-6);
    }
}

static DBL tiling_penrose (const Vector3d& EPoint, bool rhombs, bool centerFlag)
{
    // Penrose tiling
    // rhombs=false: P2 ("kite and dart") Penrose tiling
    //   centerFlag=false: 5 darts at center forming a star
    //   centerFlag=true:  5 kites at center forming a regular decagon ("sun")
    // rhombs=true:  P3 ("rhombus") Penrose tiling
    //   centerFlag=false: 5 wide rhombs at center, surrounded by 5 narrow rhombs to form a regular decagon
    //   centerFlag=true:  5 wide rhombs at center, surrounded by 10 narrow rhombs to form a pointed star

    DBL x,z;
    x = EPoint[X];
    z = EPoint[Z];

    DBL r = sqrt(x*x+z*z);
    if (r <= gkDBL_epsilon)
        return 1.0;

    // exploit trivial mirror symmetry
    z = abs(z);

    // exploit rotational & mirror symmetry
    if (x < r * COS36)
    {
        DBL rotSin;
        DBL rotCos;
        if (x < r *COS108)
        {
            rotSin = SIN144;
            rotCos = COS144;
        }
        else
        {
            rotSin = SIN72;
            rotCos = COS72;
        }
        DBL rotX = x*rotCos + z*rotSin;
        DBL rotZ = z*rotCos - x*rotSin;

        x =     rotX;
        z = abs(rotZ);
    }

    if (rhombs)
    {
        x *= INVPHI;
        z *= INVPHI;
    }

    DBL dist = abs(SIN108 * x - COS108 * z) / COS18;

    int depth = max(0, (int)ceil(log(dist)/log(SQRPHI)));

    x *= pow(INVSQRPHI,depth);
    z *= pow(INVSQRPHI,depth);

    if (depth % 2)
    {
        DBL rotX = x*COS36 + z*SIN36;
        DBL rotZ = z*COS36 - x*SIN36;
        x = rotX;
        z = abs(rotZ);
    }

    depth *= 2;

    if (centerFlag)
    {
        depth += 1;
        x *= INVPHI;
        z *= INVPHI;
    }

    return tiling_penrose_halfkite(x, z, depth, rhombs);
}

static DBL tiling_penrose1_pentagon1 (DBL pX, DBL pZ, int depth);
static DBL tiling_penrose1_pentagon2 (DBL pX, DBL pZ, int depth, bool insideQuad);
static DBL tiling_penrose1_pentagon3 (DBL pX, DBL pZ, int depth, bool insideWedge);
static DBL tiling_penrose1_star (DBL pX, DBL pZ, int depth);
static DBL tiling_penrose1_boat (DBL pX, DBL pZ, int depth, bool insideWedge);
static DBL tiling_penrose1_diamond (DBL pX, DBL pZ, int depth, bool sideA);

static void tiling_penrose1_pentagon_symmetry(DBL& x, DBL& z, DBL r)
{
    z = abs(z);

    if (x < r * COS36)
    {
        DBL rotSin;
        DBL rotCos;
        if (x < r *COS108)
        {
            rotSin = SIN144;
            rotCos = COS144;
        }
        else
        {
            rotSin = SIN72;
            rotCos = COS72;
        }
        DBL rotX = x*rotCos + z*rotSin;
        DBL rotZ = z*rotCos - x*rotSin;

        x =     rotX;
        z = abs(rotZ);
    }
}

static void tiling_penrose1_pentagon_symmetry(DBL& x, DBL& z)
{
    tiling_penrose1_pentagon_symmetry (x, z, sqrt(x*x+z*z));
}

static DBL tiling_penrose1_pentagon_dist (DBL pX, DBL pZ)
{
    return abs(pX - 0.5/TAN36) * 5.55 * INVPHI; // TODO FIXME - the factor is just an empiric value
}

static DBL tiling_penrose1_pentagon1 (DBL pX, DBL pZ, int depth)
{
    DBL x = pX;
    DBL z = abs(pZ);

    tiling_penrose1_pentagon_symmetry (x, z);

    if (depth > 0)
    {
        if (z < (x - 0.5/TAN36) * TAN54 + 0.5)
        {
            DBL rotX = x - 0.5/TAN36 - INVPHI*0.5*COS72/SIN36;
            DBL rotZ = z;
            return tiling_penrose1_pentagon2 (rotX*PHI, rotZ*PHI, depth-1, false);
        }
        else
        {
            DBL rotX = x*COS36 + z*SIN36;
            DBL rotZ = z*COS36 - x*SIN36;
            return tiling_penrose1_star (rotX*PHI, rotZ*PHI, depth-1);
        }
    }
    else
    {
        DBL dist = tiling_penrose1_pentagon_dist (x, z);
        return min(max(1.0/6-dist/6,0.0),1.0/6-1e-6);
    }
}

static DBL tiling_penrose1_pentagon2 (DBL pX, DBL pZ, int depth, bool insideQuad)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (insideQuad)
        {
            if (z < (x - INVSQRPHI*0.5/SIN36) * TAN54)
            {
                DBL rotX = x - 0.5/SIN36;
                DBL rotZ = z;
                return tiling_penrose1_pentagon1 (rotX*PHI, rotZ*PHI, depth-1);
            }
            else if (z < (x - INVSQRPHI*0.5/SIN36) * TAN162)
            {
                DBL rotX = -x;
                DBL rotZ =  z;
                return tiling_penrose1_diamond (rotX*PHI, rotZ*PHI, depth-1, true);
            }
            else
            {
                DBL rotX = x*COS108 - z*SIN108 + INVPHI*0.5/SIN36;
                DBL rotZ = z*COS108 + x*SIN108;
                return tiling_penrose1_pentagon2 (rotX*PHI, rotZ*PHI, depth-1, true);
            }
        }
        else if (z < (x + 0.5/SIN36) * TAN18)
        {
            DBL rotX = x + 0.5/SIN36 - INVSQRPHI*0.5/SIN36;
            DBL rotZ = z;
            return tiling_penrose1_diamond (rotX*PHI, rotZ*PHI, depth-1, false);
        }
        else
        {
            DBL rotX = - (x*COS36 + z*SIN36) - COS72*0.5/SIN36;
            DBL rotZ =   (z*COS36 - x*SIN36) - SIN72*0.5/SIN36;
            return tiling_penrose1_pentagon3 (rotX*PHI, rotZ*PHI, depth-1, false);
        }
    }
    else
    {
        tiling_penrose1_pentagon_symmetry (x, z);
        DBL dist = tiling_penrose1_pentagon_dist (x, z);
        return min(max(2.0/6-dist/6,1.0/6+1e-6),2.0/6-1e-6);
    }
}

static DBL tiling_penrose1_pentagon3 (DBL pX, DBL pZ, int depth, bool insideWedge)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (insideWedge && (x > INVSQRPHI*0.5*COS72/SIN36))
        {
            DBL rotX = -(x - INVSQRPHI*0.5*COS72/SIN36 - INVPHI*0.5/TAN36);
            DBL rotZ =   z;
            return tiling_penrose1_pentagon2 (rotX*PHI, rotZ*PHI, depth-1, true);
        }
        else if (!insideWedge && (x < 0.5*COS108/SIN36))
        {
            DBL rotX = x*COS144 + z*SIN144 - 0.5/SIN36;
            DBL rotZ = z*COS144 - x*SIN144;
            return tiling_penrose1_pentagon2 (rotX*PHI, rotZ*PHI, depth-1, false);
        }
        else if (!insideWedge && (z > (x - INVSQRPHI*0.5/SIN36)*TAN126))
        {
            DBL rotX = - (x*COS36 - z*SIN36) - COS72*0.5/SIN36;
            DBL rotZ =   (z*COS36 + x*SIN36) - SIN72*0.5/SIN36;
            return tiling_penrose1_pentagon3 (rotX*PHI, rotZ*PHI, depth-1, false);
        }
        else
        {
            return tiling_penrose1_boat (x*PHI, z*PHI, depth-1, insideWedge);
        }
    }
    else
    {
        tiling_penrose1_pentagon_symmetry (x, z);
        DBL dist = tiling_penrose1_pentagon_dist (x, z);
        return min(max(3.0/6-dist/6,2.0/6+1e-6),3.0/6-1e-6);
    }
}

static DBL tiling_penrose1_star (DBL pX, DBL pZ, int depth)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (x < INVPHI * 0.5/TAN36)
        {
            return tiling_penrose1_pentagon1(x*PHI, z*PHI, depth-1);
        }
        else
        {
            DBL rotX = -(x - INVPHI/TAN36);
            DBL rotZ =   z;
            return tiling_penrose1_pentagon3(rotX*PHI, rotZ*PHI, depth-1, true);
        }
    }
    else
    {
        DBL dist = abs( SIN162 * (x - PHI*0.5/SIN36) - COS162 * z) * 5.55 * INVPHI; // TODO FIXME - the factor is just an empiric value
        return min(max(4.0/6-dist/6,3.0/6+1e-6),4.0/6-1e-6);
    }
}

static DBL tiling_penrose1_boat (DBL pX, DBL pZ, int depth, bool insideWedge)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (insideWedge && (x > PHI*0.5*COS108/SIN36))
        {
            DBL rotX = -x;
            DBL rotZ =  z;
            return tiling_penrose1_pentagon1 (rotX*PHI, rotZ*PHI, depth-1);
        }
        else
        {
            DBL rotX, rotZ;
            if (insideWedge)
            {
                rotX = x;
                rotZ = z;
            }
            else
            {
                rotX = x*COS72 - z*SIN72;
                rotZ = z*COS72 + x*SIN72;
            }
            rotX += 0.5/SIN36;
            return tiling_penrose1_pentagon3 (rotX*PHI, rotZ*PHI, depth-1, true);
        }
    }
    else
    {
        DBL dist1 = abs(x - INVPHI*0.5*COS72/SIN36) * 5.55 * INVPHI; // TODO FIXME - the factor is just an empiric value
        x = -x;
        tiling_penrose1_pentagon_symmetry (x, z);
        DBL dist2 = abs(SIN162 * (x - PHI*0.5/SIN36) - COS162 * z) * 5.55 * INVPHI; // TODO FIXME - the factor is just an empiric value
        return min(max3(5.0/6-dist1/6,5.0/6-dist2/6,4.0/6+1e-6),5.0/6-1e-6);
    }
}

static DBL tiling_penrose1_diamond (DBL pX, DBL pZ, int depth, bool sideA)
{
    DBL x = pX;
    DBL z = abs(pZ);

    if (depth > 0)
    {
        if (sideA)
        {
            return tiling_penrose1_pentagon1(x*PHI, z*PHI, depth-1);
        }
        else
        {
            return tiling_penrose1_pentagon3(x*PHI, z*PHI, depth-1, true);
        }
    }
    else
    {
        DBL dist = abs(SIN18 * (x + INVPHI*0.5/SIN36) - COS18 * z) * 5.55 * INVPHI; // TODO FIXME - the factor is just an empiric value
        return min(max(6.0/6-dist/6,5.0/6+1e-6),6.0/6-1e-6);
    }
}

static DBL tiling_penrose1 (const Vector3d& EPoint, bool centerFlag)
{
    // Penrose P1 ("pentagon, star, boat and diamond") tiling
    //   centerFlag=false: pentagon at center
    //   centerFlag=true:  pentagram (star) at center

    DBL x,z;
    x = EPoint[X];
    z = EPoint[Z];

    DBL r = sqrt(x*x+z*z);
    if (r <= gkDBL_epsilon)
        return 1.0;

    tiling_penrose1_pentagon_symmetry (x, z, r);

    DBL dist = x * 2 * TAN36;

    int depth = max(0, (int)ceil(log(dist)/log(SQRPHI*SQRPHI)));

    x *= pow(INVSQRPHI*INVSQRPHI,depth);
    z *= pow(INVSQRPHI*INVSQRPHI,depth);

    depth *= 4;

    /*
    if (centerFlag)
    {
        depth += 1;
        x *= INVPHI;
        z *= INVPHI;
    }
    */

    return tiling_penrose1_pentagon1(x, z, depth);
}

static unsigned char digon[]={ 0x0E, 0x0B };
static unsigned char trigon[][6]=
{
    { 0x0D, 0x05, 0x07 , 0x0D, 5,7 },
    { 0x0E, 0x0D, 0x16, 0x49, 7, 0x0B }
};

static unsigned char tetragon[][16]=
{
    { 0x0D, 5, 5, 7, 0x0D, 5, 5, 7, 0x0D, 5, 5, 7, 0x0D, 5, 5, 7 },
    { 0x0E, 0x0D, 5, 0x16, 0x49, 5, 7, 0x0B, 0x0E, 0x0D, 5, 0x16, 0x49, 5, 7, 0x0B },
    { 0x0D, 0x34, 0x7, 0x0B, 0x07, 0x0B, 0x0D, 0x34, 0x0D, 0x34, 0x7, 0x0B, 0x07, 0x0B, 0x0D, 0x34},
    { 0x0C, 0x06, 0x0C, 0x06, 0x09, 0x03, 0x09, 0x03, 0x0C, 0x06, 0x0C, 0x06, 0x09, 0x03, 0x09, 0x03 },
    { 7, 0x2C, 7, 0x2C, 0x0D, 0x83, 0x0D, 0x83, 0x2C, 7, 0x2C,7, 0x83, 0x0D, 0x83, 0x0D  }
};

static unsigned char pentagon[][10]=
{
    { 0x0D, 5,5,5,7 , 0x0D, 5,5,5,7},
    { 0x0E, 0x0D,5,5,0x16, 0x49, 5,5,7,0x0B},
    { 7, 0x0C, 6, 0x0C, 0x24, 0x0D, 0x81, 3, 9, 3 },
    { 0x0D, 5, 0x34, 7, 0x0B, 15,15,15,15,15 },
    { 0x0D, 0xF0, 7, 0x0B, 0x0E , 15,15,15,15,15 },
    { 0x0E, 0x0A, 0x49, 5, 7, 0x0D, 5, 0x16, 0x0A, 0x0B },
    { 0x0B, 0x0E, 0x49, 0x16, 0x49, 7, 0x0D, 0x16, 0x49, 0x16 },
    { 0x2C, 5, 0x16, 0x0B, 0x0E, 0x0B, 0x0E, 0x49, 5, 0x83 },
    { 0x0E, 0x49, 5, 0x16, 0x0B, 0x0E, 0x49, 5, 0x16, 0x0B },
    { 0x0E, 0x0D, 0xC1, 0x16, 0x0B, 0x49, 0x34, 7, 0x0B, 0x0E },
    { 7, 0x2C, 5, 7, 0x2C, 0x0D, 0x83, 0x0D, 5, 0x83 },
    { 0x0D, 0x34, 7, 0x0D, 0xC1, 7, 0x0A, 0x0B, 0x0E, 0x0A}
};

static unsigned char hexagon[][12]=
{
    { 0x0D, 5, 5, 5, 5, 7, 0x0D, 5, 5, 5, 5, 7},
    { 0x2C, 5, 5, 5, 7, 0x0E, 0x0B, 0x0D, 5, 5, 5, 0x83 },
    { 0x0D, 0x34, 5, 5, 7, 0x0E, 7, 0x0B, 0x0D, 5, 5, 0xC1 },
    { 0x0D, 5, 0x34, 5, 7, 0x0E, 5, 7, 0x0B, 0x0D, 5, 0xC1 },
    { 0x0D, 0x83, 0x2C, 5, 5, 7 , 15,15,15,15,15,15},

    { 0x0C, 0x24, 5, 7, 0x0C, 6, 9, 3, 0x0D, 5, 0x81, 3 },
    { 7, 0x0C, 6, 0x0D, 0x14, 0x24, 0x0D, 0x81, 0x41, 7, 9, 3},
    { 0x0D, 5, 0x83, 0x2C, 5, 7, 15,15,15,15,15,15},
    { 7, 0x0C, 0x24, 7, 0x0C, 0x24, 0x0D, 0x81, 3, 0x0D, 0x81, 3},
    { 0x0C, 4, 6, 0x0C, 4, 6, 9, 1, 3, 9, 1, 3 },

    { 0x0E, 0x0D, 0x14, 6, 0x48, 6, 9, 0x12, 9, 0x41, 7, 0x0B},
    { 0x0D, 5, 0x16, 0x49, 0x16, 0x0B, 15,15,15,15,15,15},
    { 0x0D, 0x34, 5, 0x16, 0x0E, 0x0B, 0x0E, 0x0B, 0x49, 5, 0xC1, 7},
    { 0x0D, 0x34, 7, 0x16, 0x0A, 0x0D, 0x0A, 0x49, 7, 0xC1, 7, 0x0D},
    { 0x0D, 0x14, 0x82, 0x0E, 9, 3, 15,15,15,15,15,15},

    { 0x0E, 0x0D, 0xD0, 6, 9, 3, 15,15,15,15,15,15},
    { 0x0D, 0xF0, 5, 7, 0x0E, 0x0B, 0x0E, 0x0B, 0x0D, 5, 0xF0, 7},
    { 0x0E, 0x0B, 0x0D, 5, 5, 0x92, 0x68, 5, 5, 7, 0x0E, 0x0B},
    { 0x0E, 0x49, 5, 0x16, 0x0A, 0x0B, 15,15,15,15,15,15},
    { 0x0B, 0x2C, 5, 5, 0x83, 0x0E, 15,15,15,15,15,15},

    { 0x0B, 0x0A, 0x2C, 5, 5, 7, 15,15,15,15,15,15},
    { 0x2C, 5, 5, 0x16, 0x0B, 0x0E, 0x0E, 0x0B, 5, 0x83, 0x49, 5},
    { 0x0A, 0x0D, 0xC1, 0x16, 0x49, 0x34, 7, 0x0A, 0x0E, 0x0B, 0x0E, 0x0B},
    { 0x0D, 0x16, 0x0D, 5, 0x92, 0x0B, 0x0E, 0x68, 5, 7, 0x49, 7},
    { 0x0D, 0x16, 0x0D, 0x16, 0x16, 0x49, 0x16, 0x49, 0x49, 7, 0x49, 7},

    { 0x0C, 6, 0x0C, 6, 9, 0x12, 9, 0x12, 7, 0x49, 7, 0x49 },
    { 0x2C, 7, 0x2C, 7, 0x92, 0x0D, 0x92, 0x0D, 0x49, 7, 0x49, 7},
    { 0x0E, 0x49, 5, 0x16, 0x49, 7, 15,15,15,15,15,15},
    { 0x0D, 0x16, 0x0B, 0x0E, 0x49, 7, 0x16, 0x49, 5, 0x83, 0x2C, 5},
    { 7, 0x0E, 0x49, 0x34, 7, 0x49, 0x0D, 0xC1, 0x16, 0x0B, 0x0D, 0x16},

    { 0x2C, 5, 0xC1, 7, 0x0E, 0x0B, 0x0B, 0x0D, 0x34, 5, 0x83, 0x0E },
    { 0x0E, 0x0B, 0x0D, 0x34, 0xC1, 7, 15,15,15,15,15,15},
    { 0x0D, 5, 0x16, 0x2C, 7, 0x0A, 0x0A, 0x0D, 0x83, 0x49, 5, 7 },
    { 0x2C, 0xC1, 5, 7, 0x0E, 0x0B, 0x0E, 0x0B, 0x0D, 5, 0x34, 0x83 },
    { 0x0A, 0x0D, 5, 0x92, 0x68, 5, 7, 0x0A, 0x0B, 0x0E, 0x0E, 0x0B}
};

DBL PavementPattern::tetragonal (const Vector3d& EPoint) const
{
    unsigned char how;
    long xv,zv;
    DBL return_value=0;
    DBL value;
    DBL value1;
    DBL value2;
    DBL x,y;
    int lng=0;
    zv = floor(y=EPoint[Z]);
    xv = floor(x=EPoint[X]);
    switch(Tile)
    {
        case 6:
            switch(Number-1)
            {
                case 0:
                case 1:
                case 2:
                case 3:
                case 5:
                case 6:
                case 8:
                case 9:
                    setWrapInt (xv, 6);
                    setWrapInt (zv, 2);
                    lng = 6;
                    break;
                case 4:
                case 7:
                case 19:
                case 20:
                    lng = 0;
                    setWrapInt (zv, 6);
                    xv += 5*zv;
                    setWrapInt (xv, 6);
                    break;
                case 11:
                case 18:
                case 27:
                    lng = 0;
                    setWrapInt (zv, 6);
                    xv += zv;
                    setWrapInt (xv, 6);
                    break;
                case 10:
                case 12:
                case 21:
                case 22:
                case 24:
                case 25:
                case 26:
                    lng = 4;
                    setWrapInt (xv, 4);
                    setWrapInt (zv, 3);
                    break;
                case 13:
                case 32:
                    lng = 3;
                    setWrapInt (zv, 4);
                    setWrapInt (xv, 3);
                    break;
                case 14:
                    lng = 3;
                    setWrapInt (zv, 6);
                    xv += 2 * (zv/2);
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 3);
                    break;
                case 15:
                    lng = 2;
                    setWrapInt (xv, 6);
                    zv += (xv/2);
                    setWrapInt (xv, 2);
                    setWrapInt (zv, 3);
                    break;
                case 16:
                case 17:
                    lng = 6;
                    setWrapInt (zv, 12);
                    xv += zv/2;
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 6);
                    break;
                case 23:
                case 28:
                    lng = 6;
                    setWrapInt (zv, 12);
                    xv += 4* (zv/2);
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 6);
                    break;
                case 29:
                case 30:
                    lng = 6;
                    setWrapInt (zv, 4);
                    xv += 3* (zv/2);
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 6);
                    break;
                case 31:
                    lng = 0;
                    setWrapInt (zv, 3);
                    xv += 4* zv;
                    setWrapInt (xv, 6);
                    break;
                case 33:
                    lng = 0;
                    setWrapInt (zv, 12);
                    xv += 7*zv;
                    setWrapInt (xv, 12);
                    break;
                case 34:
                    lng = 4;
                    setWrapInt (zv, 6);
                    xv += 2 * (zv/3);
                    setWrapInt (xv, 4);
                    setWrapInt (zv, 3);
                    break;
            }
            how = hexagon[Number-1][xv+zv*lng];
            break;
        case 5:
            switch(Number-1)
            {
                case 0:
                case 1:
                    setWrapInt (xv, 5);
                    setWrapInt (zv, 2);
                    break;
                case 2:
                case 9:
                    setWrapInt (zv, 10);
                    xv += 3 * (zv/2);
                    setWrapInt (xv, 5);
                    setWrapInt (zv, 2);
                    break;
                case 10:
                    setWrapInt (zv, 10);
                    xv += 4*(zv/2);
                    setWrapInt (xv, 5);
                    setWrapInt (zv, 2);
                    break;
                case 3:
                    setWrapInt (zv, 5);
                    xv += 2*zv;
                    setWrapInt (xv, 5);
                    zv = 0;
                    break;
                case 4:
                    setWrapInt (zv, 5);
                    xv += 2 * zv;
                    setWrapInt (xv, 5);
                    zv = 0;
                    break;
                case 5:
                case 6:
                case 8:
                    setWrapInt (zv, 10);
                    xv += zv;
                    setWrapInt (xv, 10);
                    zv = 0;
                    break;
                case 11:
                    setWrapInt (zv, 10);
                    xv += 8* zv;
                    setWrapInt (xv, 10);
                    zv = 0;
                    break;
                case 7:
                    setWrapInt (zv, 10);
                    xv += 3*zv;
                    setWrapInt (xv, 10);
                    zv = 0;
                    break;
            }
            how = pentagon[Number-1][xv+zv*5];
            break;
        case 4:
            setWrapInt (xv, 4);
            setWrapInt (zv, 4);
            how = tetragon[Number-1][xv+zv*4];
            break;
        case 3:
            setWrapInt (xv, 3);
            setWrapInt (zv, 2);
            how = trigon[Number-1][xv+zv*3];
            break;
        case 2:
            setWrapInt (zv, 2);
            how = digon[zv];
            break;
        case 1:
        default:
            how = 0x0F;
            break;
    }
    /*
     **   5---1---6
     **   |       |
     **   4       2
     **   |       |
     **   8---3---7
     */
    x -= floor(x);
    y -= floor(y);
    switch(Form)
    {
        case 2:
            if ((how & 0x16) == 0x16)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x2C) == 0x2C)
            {
                value1 = 2 - 2*x;
                value2 = 2 - 2*y;
                value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x49) == 0x49)
            {
                value1 = 2 - 2*x;
                value2 = 2*y;
                value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x83) == 0x83)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = fabs(sqrt(value1*value1+value2*value2) - 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        case 1:
            if ((how & 0x16) == 0x16)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = fabs(value1 - value2 + 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x2C) == 0x2C)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = fabs(value2 - value1 + 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x49) == 0x49)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = fabs(value2 - value1 + 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if ((how & 0x83) == 0x83)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = fabs(value1 - value2 + 1.0);
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        case 0:
            if ((how & 0x16) == 0x16)
            {
                value1 = 2*x -1;
                value2 = 1 - 2*y;
                return_value = max(value1,value2);
                value1 = 1 - 2*x;
                value2 = 2*y - 1;
                value = min(value1,value2);
                return_value = max(return_value,value);
                return return_value;
            }
            if ((how & 0x2C) == 0x2C)
            {
                value1 = 1 - 2*x;
                value2 = 1 - 2*y;
                return_value = max(value1,value2);
                value1 = 2*x - 1;
                value2 = 2*y - 1;
                value = min(value1,value2);
                return_value = max(return_value,value);
                return return_value;
            }
            if ((how & 0x49) == 0x49)
            {
                value1 = 1 - 2*x;
                value2 = 2*y - 1;
                return_value = max(value1,value2);
                value1 = 2*x - 1;
                value2 = 1 - 2*y;
                value = min(value1,value2);
                return_value = max(return_value,value);
                return return_value;
            }
            if ((how & 0x83) == 0x83)
            {
                value1 = 2*x -1;
                value2 = 2*y -1;
                return_value = max(value1,value2);
                value1 = 1 - 2*x;
                value2 = 1 - 2*y;
                value = min(value1,value2);
                return_value = max(return_value,value);
                return return_value;
            }
            break;
        default:
        case 3:
            break;
    }
    if (how & 0x01)
    {
        value = 2*y - 1;
        return_value = max(return_value,value);
    }
    if (how & 0x02)
    {
        value = 2*x -1;
        return_value = max(return_value,value);
    }
    if (how & 0x04)
    {
        value = 1 - 2*y;
        return_value = max(return_value,value);
    }
    if (how & 0x08)
    {
        value = 1 - 2*x;
        return_value = max(return_value,value);
    }
    switch(Interior)
    {
        case 2:
            if (how & 0x40)
            {
                value1 = 2 - 2*x;
                value2 = 2*y;
                value = 1.0- sqrt(value1*value1+value2*value2);
                return_value = max(return_value,value);
            }
            if (how & 0x80)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = 1.0- sqrt(value1*value1+value2*value2);
                return_value = max(return_value,value);
            }
            if (how & 0x10)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = 1.0- sqrt(value1*value1+value2*value2);
                return_value = max(return_value,value);
            }
            if (how & 0x20)
            {
                value1 = 2 - 2*x;
                value2 = 2 - 2*y;
                value = 1.0- sqrt(value1*value1+value2*value2);
                return_value = max(return_value,value);
            }
            break;
        case 1:
            if (how & 0x40)
            {
                value1 = 2 - 2*x;
                value2 = 2*y;
                value = 1.0- (value1+value2);
                return_value = max(return_value,value);
            }
            if (how & 0x80)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = 1.0- (value1+value2);
                return_value = max(return_value,value);
            }
            if (how & 0x10)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = 1.0- (value1+value2);
                return_value = max(return_value,value);
            }
            if (how & 0x20)
            {
                value1 = 2 - 2*x;
                value2 = 2 - 2*y;
                value = 1.0- (value1+value2);
                return_value = max(return_value,value);
            }
            break;
        default:
        case 0:
            if (how & 0x10)
            {
                value1 = 1 - 2*x;
                value2 = 2*y - 1;
                value = min(value1,value2);
                return_value = max(return_value,value);
            }
            if (how & 0x20)
            {
                value1 = 2*x - 1;
                value2 = 2*y - 1;
                value = min(value1,value2);
                return_value = max(return_value,value);
            }
            if (how & 0x40)
            {
                value1 = 2*x - 1;
                value2 = 1 - 2*y;
                value = min(value1,value2);
                return_value = max(return_value,value);
            }
            if (how & 0x80)
            {
                value1 = 1 - 2*x;
                value2 = 1 - 2*y;
                value = min(value1,value2);
                return_value = max(return_value,value);
            }
            break;
    }
    switch(Exterior)
    {
        case 2:
            value1 = 2*x - 1;
            value2 = 2*y - 1;
            if ( (((how & 0x06) == 0x06)&&(value1>=0.0)&&(value2<=0.0)) ||
                 (((how & 0x0C) == 0x0C)&&(value1<=0.0)&&(value2<=0.0)) ||
                 (((how & 0x09) == 0x09)&&(value1<=0.0)&&(value2>=0.0)) ||
                 (((how & 0x03) == 0x03)&&(value1>=0.0)&&(value2>=0.0)) )
            {
                value = sqrt(value1*value1+value2*value2);
                value = min(value,1.0);
                return_value = max(return_value,value);
            }
            break;
        case 1:
            if ((how & 0x06) == 0x06)
            {
                value1 = 2 - 2*x;
                value2 = 2*y;
                value = 2.0- (value1+value2);
                value = min(value,1.0);
                return_value = max(return_value,value);
            }
            if ((how & 0x0C) == 0x0C)
            {
                value1 = 2*x;
                value2 = 2*y;
                value = 2.0- (value1+value2);
                value = min(value,1.0);
                return_value = max(return_value,value);
            }
            if ((how & 0x09) == 0x09)
            {
                value1 = 2*x;
                value2 = 2 - 2*y;
                value = 2.0- (value1+value2);
                value = min(value,1.0);
                return_value = max(return_value,value);
            }
            if ((how & 0x03) == 0x03)
            {
                value1 = 2 - 2*x;
                value2 = 2 - 2*y;
                value = 2.0- (value1+value2);
                value = min(value,1.0);
                return_value = max(return_value,value);
            }
            break;
        default:
        case 0:
            break;
    }
    return return_value;
}

static unsigned short tritrigon[][6]=
{
    {0x215,0x344,0x126,   0x126,0x344,0x215}
};

static unsigned short tritetragon[][8]=
{
    {0x126,0x126,0x144,0x144, 0x126,0x126,0x144,0x144},
    {0x611,0x344,0x126,0x423, 0x611,0x344,0x126,0x423},
    {5,0,6,3, 0,5,3,6}
};

static unsigned short tripentagon[][10]=
{
    {0x215,0x244,4,0x144,0x126, 0x126,0x144,4,0x244,0x215},
    {0x215,0x244,0x244,0x611,0x611, 0x244,0x244,0x215,0x423,0x423},
    {0x146,0x126,0x344,0x611,0x611, 0x344, 0x126,0x146,0x522,0x522},
    {5,0,0x244,0x215,3, 3,0x215,0x244,0,5}
};

static unsigned short trihexagon[][12]=
{
    {0x215,0x244,4,4,0x244,0x215, 0x215,0x244,4,4,0x244,0x215},
    {0x413,0x522,0x144,4,0x244,0x215, 0x413,0x522,0x144,4,0x244,0x215},
    {3,0x126,0x144,0,0x244,0x215, 3,0x126,0x144,0,0x244,0x215},
    {0x215,0x244,4,0,6,3, 4,0x244,0x215,3,6,0},
    {5,0,0x244,0x215,0x245,0x211, 5,0x211,0x245,0x215,0x244,0},
    {0x245,0x211,0x215,0x244,0x244,0x211, 0x245,0x211,0x244,0x244,0x215,0x211},
    {0x215,0x244,0x244,0x611, 0x522,0x146,0x146,0x522, 0x611,0x244,0x244,0x215},
    {5,0,0x244,0x215, 0x146,0x122,0x122,0x146, 0x215,0x244,0,5},
    {5,0,6,0x126,0x344,0x211, 5,0x211,0x344,0x126,6,0},
    {0x215,0x344,0x122,0x122,0x344,0x215, 0x215,0x344,0x122,0x122,0x344,0x215},
    {5,0,6,6,0,5, 6,0,5,5,0,6},
    {0x691,0x3C4,0x5A2,0x5A2,0x3C4,0x691, 0x5A2,0x3C4,0x691,0x691,0x3C4,0x5A2},
};

DBL PavementPattern::trigonal (const Vector3d& EPoint) const
{
    unsigned short how;
    long xv,zv;
    DBL return_value=0;
    DBL value;
    DBL value1;
    DBL value2;
    DBL dist1;
    DBL dist2;
    DBL dist3;
    DBL x,z;
    int lng=0;

    x=EPoint[X];
    z=EPoint[Z];

    xv = floor(x);
    zv = floor(z/SQRT3);
    x -= xv;
    z -= SQRT3*zv;
    /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
     ** There is some mirror to reduce the problem
     */
    zv *= 2;
    xv *= 2;
    if ( z > SQRT3_2 )
    {
        z -= SQRT3_2;
        if (x>0.5)
        {
            x -= 0.5;
            xv++;
        }
        else
        {
            x += 0.5;
            xv--;
        }
        zv++;
    }
    if ((x == 0.0)||(z/x>SQRT3))
    {
        z = SQRT3_2 - z;
        x = 0.5 -x;
        xv--;
    }
    if ((x == 1.0)||(z/(1.0-x)>SQRT3))
    {
        z = SQRT3_2 - z;
        x = 1.5 -x;
        xv++;
    }
    switch(Tile)
    {
        case 6:
            switch(Number-1)
            {
                case 0:
                case 1:
                case 9:
                    xv += 5*zv;
                    zv = 0;
                    setWrapInt (xv, 6);
                    lng = 0;
                    break;
                case 2:
                case 10:
                case 11:
                    setWrapInt (zv, 2);
                    xv += 3*zv;
                    setWrapInt (xv, 6);
                    lng = 0;
                    break;
                case 3:
                    xv += 14*(wrapInt (zv, 6)/2);
                    setWrapInt (xv, 6);
                    lng = 6;
                    setWrapInt (zv, 2);
                    break;
                case 4:
                case 8:
                    xv += 8*(wrapInt (zv, 6)/2);
                    setWrapInt (xv, 6);
                    lng = 6;
                    setWrapInt (zv, 2);
                    break;
                case 5:
                    setWrapInt (xv, 6);
                    lng = 6;
                    setWrapInt (zv, 2);
                    break;
                case 6:
                case 7:
                    xv -= wrapInt (zv, 12)/3;
                    setWrapInt (xv, 4);
                    lng = 4;
                    setWrapInt (zv, 3);
                    break;
            }
            how = trihexagon[Number-1][xv+zv*lng];
            break;
        case 5:
            switch(Number-1)
            {
                case 0:
                case 1:
                case 2:
                    setWrapInt (zv, 2);
                    xv += 5*zv;
                    setWrapInt (xv, 10);
                    zv = 0;
                    break;
                case 3:
                    setWrapInt (zv, 10);
                    xv += 3*zv;
                    setWrapInt (xv, 10);
                    zv = 0;
                    break;
            }
            how = tripentagon[Number-1][xv];
            break;
        case 4:
            setWrapInt (zv, 4);
            xv += zv;
            setWrapInt (xv, 4);
            setWrapInt (zv, 2);
            how = tritetragon[Number-1][xv+zv*4];
            break;
        case 3:
            setWrapInt (zv, 2);
            xv += 3*zv;
            setWrapInt (xv, 6);
            zv = 0;
            how = tritrigon[Number-1][xv];
            break;
        case 2:
            how = 0x166;
            break;
        case 1:
        default:
            how = 0x07;
            break;
    }

    /*
     **      / \
     **     1   \
     **    /     2
     **   /       \
     **  -----3-----
     ** *3/2
     */
    if (how & 0x01)
    {
        dist1 = 1.0 - (fabs(SQRT3 * x - z) * SQRT3 );
        return_value = max(return_value,dist1);
    }
    if (how & 0x02)
    {
        dist2 = 1.0 - (fabs(SQRT3 * (1.0-x) - z) * SQRT3  );
        return_value = max(return_value,dist2);
    }
    if (how & 0x04)
    {
        dist3 = 1.0 - (z * 2.0 * SQRT3 );
        return_value = max(return_value,dist3);
    }
    switch(Interior)
    {
        case 1:
            dist1 = (1.0 - (fabs(SQRT3 * z + x) ));
            dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) ));
            dist3 = (1.0 - (x * 2.0 ));
            if (((how & 0x83) == 0x00)&&(dist1<0)&&(dist2<0))
            {
                value1 = (3.0 / 2.0 *(fabs(SQRT3 * z + x) ) - 2.0);
                value2 = (3.0 / 2.0 *(fabs(SQRT3 * z - x + 1.0) ) - 2.0);
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            if (((how & 0x85) == 0x00)&&(dist1>0)&&(dist3>0))
            {
                value1 = (1.0 - 3.0 / 2.0 * (fabs(SQRT3 * z + x) ));
                value2 = (1.0 - (x * 3.0 ));
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            if (((how & 0x86) == 0x00)&&(dist3<0)&&(dist2>0))
            {
                value1 = (1.0 - 3.0 / 2.0 *(fabs(SQRT3 * z - x + 1.0) ));
                value2 = ((x * 3.0 ) - 2.0);
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            break;
        case 2:
            if ((how & 0x83) == 0x00)
            {
                dist1 = x - 0.5;
                dist2 = z - SQRT3_2;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2))*3.0 );
                return_value = max(return_value,dist3);
            }
            if ((how & 0x85) == 0x00)
            {
                dist1 = x;
                dist2 = z;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
                return_value = max(return_value,dist3);
            }
            if ((how & 0x86) == 0x00)
            {
                dist1 = x - 1.0;
                dist2 = z;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
                return_value = max(return_value,dist3);
            }
            break;
        case 0:
            if ((how & 0x83) == 0x00)
            {
                dist3 = 1.0 - ((SQRT3_2 - z) * 2.0 * SQRT3  );
                return_value = max(return_value,dist3);
            }
            if ((how & 0x85) == 0x00)
            {
                dist2 = 1.0 - (fabs(SQRT3 * x + z) * SQRT3 );
                return_value = max(return_value,dist2);
            }
            if ((how & 0x86) == 0x00)
            {
                dist1 = 1.0 - (fabs(SQRT3 * (x -1.0) - z) * SQRT3 );
                return_value = max(return_value,dist1);
            }
            break;
    }
    switch(Exterior)
    {
        case 2:
            dist1 = (1.0 - (fabs(SQRT3 * z + x) ));
            dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) ));
            dist3 = (1.0 - (x * 2.0 ));
            if ( (((how & 0x03) == 0x03)&&(dist1<=0.0)&&(dist2<=0.0)) ||
                 (((how & 0x06) == 0x06)&&(dist2>=0.0)&&(dist3<=0.0)) ||
                 (((how & 0x05) == 0x05)&&(dist1>=0.0)&&(dist3>=0.0)) )
            {
                value1 = x - 0.5;
                value2 = z - SQRT3_2/3.0;
                value = 2 * SQRT3 *sqrt(value1*value1+value2*value2);
                return_value = min(1.0,value);
            }
            break;
        case 1:
            /* dist1 = (1.0 - (fabs(SQRT3 * z + x) )); */
            dist1 = (1.0 - (fabs(SQRT3 * x - z)*SQRT3 ));
            /* dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) )); */
            dist2 = (1.0 - (fabs(SQRT3 * (1.0 -x ) -z ) *SQRT3));
            /* dist3 = (1.0 - (x * 2.0 )); */
            dist3 = (1.0 - (z * 2.0 * SQRT3));
            value1 = (x - 0.5);
            value2 = (z - SQRT3_2/3.0);
            if  (((how & 0x03) == 0x03)&&(dist1>=0.0)&&(dist2>=0.0))
            {
                value = fabs(value2 * 2.0 * SQRT3 );
                return_value = min(1.0,value);
            }
            if  (((how & 0x06) == 0x06)&&(dist2>=0.0)&&(dist3>=0.0))
            {
                value = fabs(SQRT3 * value1 - value2) * SQRT3;
                return_value = min(1.0,value);
            }
            if  (((how & 0x05) == 0x05)&&(dist1>=0.0)&&(dist3>=0.0))
            {
                value = fabs(SQRT3 * value1 + value2) * SQRT3;
                return_value = min(1.0,value);
            }
            break;
        case 0:
        default:
            break;
    }
    dist1 = (1.0 - (fabs(SQRT3 * z + x) ));
    dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) ));
    dist3 = (1.0 - (x * 2.0 ));
    switch(Form)
    {
        case 2:
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value1 = x;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
            {
                value1 = x - 0.5;
                value2 = z- SQRT3_2;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
            {
                value1 = x - 1.0;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
            {
                value1 = x - 0.5;
                value2 = z - SQRT3_2;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
            {
                value1 = x - 1.0;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x420) == 0x420)&&(dist3<0)&&(dist2>0))
            {
                value1 = x;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        case 1:
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value = -dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
            {
                value = dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
            {
                value = -dist2 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
            {
                value = dist2 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
            {
                value = dist3 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x420) == 0x420)&&(dist2>0)&&(dist3<0))
            {
                value = -dist3 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value = -dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        default:
        case 0:
            break;
    }
    return return_value;
}

/*
** open face (3 bits)
** special (1 bit)
** nexus vertex/open face (3 bits)
** unused (1 bits)
** close face (3 bits)
*/
static unsigned short hexmonogon[][6]=
{
    {0x691,0x3C4,0x5A2,0x5A2,0x3C4,0x691}
};

static unsigned short hexdigon[][12]=
{
    {0x691,0x3C4,0x1A2,0x1A2,0x3C4,0x691,0x5A2,0x1C4,0x6080,0x6080,0x1C4,0x5A2 }
};

static unsigned short hextrigon[][36]=
{
    {
        0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
        0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2
    },

    {
        0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
        0x5A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
        0x491,0x1080,0x1080, 0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x2080, 0x4080,0x1C4,0x5A2
    },

    {
        0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
        0x5A2,0x1C4,0x6080,0x6080,0x0C4,0x5080,0x5080,0x2C4,0x691,
        0x5A2,0x1C4,0x6080,0x6080,0x0C4,0x5080,0x5080,0x2C4,0x691,
        0x691,0x3C4,0x1A2, 0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2
    }
};

static unsigned short hextetragon[][48]=
{
    {
        0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
        0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2
    },

    {
        0x691,0x2C4,0x4080, 0x1080,0x80,0x2080,
        0x4080,0x2C4,0x691, 0x291,0x3C4,0x1A2,
        0x4080,0x1C4,0x5A2, 0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x4080, 0x2080,0x80,0x1080,
        0x691,0x2C4,0x4080, 0x1080,0x80,0x2080,
        0x4080,0x2C4,0x691, 0x291,0x3C4,0x1A2,
        0x4080,0x1C4,0x5A2, 0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x4080, 0x2080,0x80,0x1080
    },

    {
        0x691,0x2C4,0x5080,  0x5080,0x0C4,0x6080,
        0x6080,0x0C4,0x5080, 0x5080,0x2C4,0x691,
        0x5A2,0x3C4,0x291, 0x291,0x3C4,0x1A2,
        0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
        0x691,0x2C4,0x5080,  0x5080,0x0C4,0x6080,
        0x6080,0x0C4,0x5080, 0x5080,0x2C4,0x691,
        0x5A2,0x3C4,0x291, 0x291,0x3C4,0x1A2,
        0x1A2,0x3C4,0x291, 0x291,0x3C4,0x5A2,
    },

    {
        0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
        0x1A2,0x3C4,0x691,  0x691,0x3C4,0x5A2,
        0x6080,0x1C4,0x5A2, 0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2,
        0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2
    },

    {
        0x691,0x3C4,0x1A2,  0x5A2,0x1C4,0x6080,
        0x5080,0x2C4,0x691, 0x491,0x3080,0x0A2,
        0x291,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
        0x5A2,0x3C4,0x291,  0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x1A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
        0x6080,0x1C4,0x5A2, 0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2
    },

    {
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
        0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
        0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
        0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2
    },

    {
        0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691, 0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x291,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,
        0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x291, 0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
    }
};

static unsigned short hexpentagon[][60]=
{ /* 0 */
    {
        0x691,0x3C4,0x5A2,  0x5A2,0x3C4, 0x691,
        0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2
    },
    /* 1 */
    {
        0x6080,0x1C4,0x5A2, 0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2 ,
        0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
        0x1A2,0x3C4,0x691,  0x691,0x3C4,0x5A2
    },
    /* 2 */
    {
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4, 0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
        0x5A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x091, 0x691,0x3C4,0x5A2,
        0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2
    },
    /* 3 */
    {
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x691,0x2C4,0x4080, 0x1080,0x1080,0x491,
        0x5A2,0x3C4,0x691,  0x091,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x091, 0x691,0x3C4,0x5A2,
        0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
        0x491,0x3080,0x0A2, 0x5A2,0x3C4, 0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2
    },
    /* 4 */
    {
        0x6080,0x1C4,0x5A2, 0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2,
        0x491,0x3080,0x0A2, 0x5A2,0x1C4,0x6080,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x0A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x491,
        0x5A2,0x3C4,0x691,  0x691,0x3C4,0x5A2,
        0x491,0x3080,0x4A2, 0x5A2,0x3C4,0x691,
        0x0A2,0x3080,0x491, 0x491,0x3080,0x4A2,
    },
    /* 5 */
    {
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4, 0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x491,0x1080,0x080, 0x080,0x1080,0x491,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
        0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x491,0x1080,0x080, 0x080,0x1080,0x491,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2
    },
    /* 6 */
    {
        0x691,0x2C4,0x4080, 0x1080,0x1080, 0x491,
        0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
        0x491,0x1080,0x1080,0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x080, 0x080,0x2080,0x4A2,
        0x491,0x1080,0x080, 0x080,0x1080,0x491,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2,
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x491,0x1080,0x080, 0x080,0x1080,0x491,
        0x4A2,0x2080,0x080, 0x080,0x2080,0x4A2
    },
    /* 7 */
    {
        0x6080,0x1C4,0x5A2,
        0x5A2,0x1C4,0x4080,
        0x691,0x2C4,0x4080,
        0x4080,0x2C4,0x691,
        0x4080,0x1C4,0x5A2,
        0x5A2,0x1C4,0x6080,
        0x691,0x3C4,0x1A2,

        0x5A2,0x1C4,0x6080,
        0x491,0x3080,0x0A2,

        0x4A2,0x3080,0x091,
        0x491,0x1080,0x1080,
        0x4A2,0x2080,0x2080,
        0x691,0x3C4,0x1A2,
        0x1A2,0x3C4,0x691,
        0x2080,0x2080,0x4A2,
        0x1080,0x1080,0x491,
        0x091,0x3080,0x4A2,
        0x0A2,0x3080,0x491,
        0x6080,0x1C4,0x5A2,
        0x1A2,0x3C4,0x691
    },
    /* 8 */
    {
        0x6080,0x0C4,0x5080, 0x5080,0x2C4, 0x291,
        0x5A2,0x3C4,0x291,  0x291,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
        0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2, 0x0A2,0x3080,0x491,
        0x5A2,0x1C4,0x6080, 0x6080,0x1C4,0x5A2,

        0x691,0x3C4,0x1A2, 0x5A2,0x1C4,0x6080,
        0x1A2,0x3C4,0x291, 0x691,0x3C4,0x1A2
    },
    /* 9 */
    {
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
        0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
        0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
        0x6A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
        0x291,0x3C4,0x1A2, 0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
    },
    /* 10 */
    {
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x491,0x1080,0x80, 0x80,0x1080,0x091,
        0x4A2,0x2080,0x2080, 0x4080,0x0C4,0x5080,
        0x691,0x3C4,0x1A2, 0x5A2,0x3C4,0x291,
        0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080,
        0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x5080,0x0C4,0x4080, 0x2080,0x2080,0x4A2,
        0x091,0x1080,0x80, 0x80,0x1080,0x491,
        0x4A2,0x2080,0x2080, 0x4080,0x1C4,0x5A2
    },
    /* 11 */
    {
        0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x4080,0x0C4,0x4080, 0x2080,0x2080,0x4A2,
        0x4080,0x0C4,0x4080, 0x1080,0x1080,0x491,
        0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2,
        0x491,0x1080,0x1080,0x4080,0x0C4,0x4080,
        0x4A2,0x2080,0x2080,0x4080,0x0C4,0x4080,
        0x691,0x3C4,0x1A2, 0x5A2,0x3C4,0x291,
        0x1A2,0x3C4,0x691, 0x491,0x1080,0x1080,
        0x2080,0x2080,0x4A2, 0x4A2,0x2080,0x2080,
        0x1080,0x1080,0x491, 0x691,0x3C4,0x1A2
    },
    /* 12 */
    {
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
        0x291,0x3C4,0x5A2,  0x0A2,0x3080,0x491,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
        0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
        0x491,0x3080,0x0A2, 0x5A2,0x3C4,0x291,
        0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2
    },
    /* 13 */
    {
        0x691,0x3C4,0x5A2,  0x5A2,0x3C4,0x691,
        0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x291,
        0x5080,0x2C4,0x291, 0x291,0x2C4,0x5080,
        0x291,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
        0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x691,0x3C4,0x5A2
    },
    /* 14 */
    {
        0x291,0x3C4,0x5A2, 0x5A2,0x3C4,0x691,
        0x5A2,0x3C4,0x691, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2,0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491,0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2,0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x091,0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080,0x5080,0x2C4,0x291,
        0x5080,0x2C4,0x291,0x291,0x2C4,0x5080,
        0x291,0x2C4,0x5080,0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691,0x291,0x3C4,0x5A2
    },
    /* 15 */
    {
        0x691,0x2C4,0x5080, 0x5080,0x0C4,0x4080,
        0x1A2,0x3C4,0x691,  0x291,0x2C4,0x4080,
        0x2080,0x2080,0x4A2,0x5A2,0x3C4,0x691,
        0x1080,0x1080,0x491,0x491,0x3080,0x4A2,
        0x091,0x3080,0x4A2, 0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x491, 0x491,0x1080,0x1080,
        0x691,0x3C4,0x5A2,  0x4A2,0x2080,0x2080,
        0x4080,0x2C4,0x291, 0x691,0x3C4,0x1A2,
        0x4080,0x0C4,0x5080,0x5080,0x2C4,0x691,
        0x5A2,0x3C4,0x291,  0x291,0x3C4,0x5A2
    },
    /* 16 */
    {
        0x4080,0x1C4,0x1A2, 0x1A2,0x1C4,0x4080,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x4080,
        0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x691,
        0x1A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x2080,0x2080,0x4A2,0x4A2,0x3080,0x091,
        0x1080,0x1080,0x491,0x491,0x1080,0x1080,
        0x091,0x3080,0x4A2, 0x4A2,0x2080,0x2080,
        0x4A2,0x3080,0x491, 0x691,0x3C4,0x1A2,
        0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x4080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
    },
    /* 17 */
    {
        0x291,0x3C4,0x1A2,  0x1A2,0x3C4,0x291,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
        0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x691,
        0x5080,0x2C4,0x691, 0x491,0x3080,0x4A2,
        0x091,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x491, 0x691,0x2C4,0x5080,
        0x691,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2
    },
    /* 18 */
    {
        0x091,0x3080,0x4A2, 0x5A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x091,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080,
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
        0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
        0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080
    },
    /* 19 */
    {
        0x691,0x3C4,0x1A2,  0x1A2,0x3C4,0x691,
        0x5A2,0x1C4,0x4080, 0x2080,0x2080,0x4A2,
        0x691,0x2C4,0x4080, 0x1080,0x1080,0x091,
        0x5080,0x2C4,0x691, 0x291,0x2C4,0x5080,
        0x091,0x3080,0x4A2, 0x5A2,0x3C4,0x691,

        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x4A2,0x3080,0x091,
        0x5080,0x2C4,0x291, 0x691,0x2C4,0x5080,
        0x091,0x1080,0x1080,0x4080,0x2C4,0x691,
        0x4A2,0x2080,0x2080,0x4080,0x1C4,0x5A2
    },
    /* 20 */
    {
        0x291,0x3C4,0x5A2,  0x5A2,0x3C4,0x691,
        0x5A2,0x3C4,0x691,  0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691, 0x091,0x3080,0x4A2,
        0x091,0x3080,0x4A2, 0x4A2,0x3080,0x091,
        0x4A2,0x3080,0x091, 0x691,0x2C4,0x5080,
        0x691,0x2C4,0x5080, 0x5080,0x2C4,0x691,
        0x5080,0x2C4,0x691, 0x291,0x3C4,0x5A2
    },
    /* 21 */
    {
        0x691,0x3C4,0x1A2,  0x5A2,0x3C4,0x291,
        0x5080,0x2C4,0x691, 0x691,0x2C4,0x5080,
        0x291,0x3C4,0x5A2,  0x1A2,0x3C4,0x691,
        0x5080,0x0C4,0x6080,0x6080,0x1C4,0x5A2,
        0x091,0x3080,0x0A2, 0x5A2,0x3C4,0x691,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x491,0x3080,0x4A2, 0x4A2,0x3080,0x491,
        0x4A2,0x3080,0x491, 0x491,0x3080,0x4A2,
        0x691,0x3C4,0x5A2,  0x0A2,0x3080,0x091,
        0x5A2,0x1C4,0x6080, 0x6080,0x0C4,0x5080
    }
};

DBL PavementPattern::hexagonal (const Vector3d& EPoint) const
{
    unsigned short how;
    long xv,zv;
    DBL return_value=0;
    DBL value;
    DBL value1;
    DBL value2;
    DBL dist1;
    DBL dist2;
    DBL dist3;
    DBL x,z;
    int lng;

    x=EPoint[X];
    z=EPoint[Z];

    xv = floor(x);
    zv = floor(z/SQRT3);
    x -= xv;
    z -= SQRT3*zv;
    /* x,z is in { [0.0, 1.0 [, [0.0, SQRT3 [ }
     ** There is some mirror to reduce the problem
     */
    zv *= 2;
    xv *= 2;
    if ( z > SQRT3_2 )
    {
        z -= SQRT3_2;
        if (x>0.5)
        {
            x -= 0.5;
            xv++;
        }
        else
        {
            x += 0.5;
            xv--;
        }
        zv++;
    }
    if ((x == 0.0)||(z/x>SQRT3))
    {
        z = SQRT3_2 - z;
        x = 0.5 -x;
        xv--;
    }
    if ((x == 1.0)||(z/(1.0-x)>SQRT3))
    {
        z = SQRT3_2 - z;
        x = 1.5 -x;
        xv++;
    }
    switch(Tile)
    {
        case 5:
            switch(Number-1)
            {
                case 0:
                case 2:
                case 3:
                case 5:
                case 6:
                case 19:
                    setWrapInt (zv, 10);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 60 */
                    zv = 0;
                    break;
                case 1:
                case 4:
                case 9:
                    zv -= 2*(wrapInt (xv, 30)/6);
                    setWrapInt (zv, 10);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 60 */
                    zv = 0;
                    break;
                case 7:
                    zv -= 7*(wrapInt (xv, 60)/3);
                    setWrapInt (zv, 20);
                    setWrapInt (xv, 3);
                    xv += 3*zv; /* 60 */
                    zv = 0;
                    break;
                case 8:
                case 10:
                case 13:
                case 14:
                case 15:
                case 17:
                case 20:
                case 21:
                    zv += 2*(wrapInt (xv, 30)/6);
                    setWrapInt (zv, 10);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 60 */
                    zv = 0;
                    break;
                case 11:
                case 16:
                    zv -= 6*(wrapInt (xv, 30)/6);
                    setWrapInt (zv, 10);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 60 */
                    zv = 0;
                    break;
                case 12:
                case 18:
                    zv += 6*(wrapInt (xv, 30)/6);
                    setWrapInt (zv, 10);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 60 */
                    zv = 0;
                    break;
            }
            how = hexpentagon[Number-1][xv];
            break;
        case 4:
            switch(Number-1)
            {
                case 0:
                    setWrapInt (zv, 8);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 48 */
                    zv = 0;
                    break;
                case 3:
                    zv -= 2*(wrapInt (xv, 24)/6);
                    setWrapInt (zv, 8);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 48 */
                    zv = 0;
                    break;
                case 2:
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 12);
                    xv += 12*zv; /* 24 */
                    zv = 0;
                    break;
                case 5:
                case 4:
                    zv -= 2*(wrapInt (xv, 24)/6);
                    setWrapInt (zv, 8);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 48 */
                    zv = 0;
                    break;
                case 1:
                    zv += 2*(wrapInt (xv, 12)/6);
                    setWrapInt (zv, 8);
                    setWrapInt (xv, 6);
                    xv += 6*zv; /* 48 */
                    zv = 0;
                    break;
                case 6:
                    setWrapInt (zv, 4);
                    setWrapInt (xv, 12);
                    xv += 12*zv; /* 48 */
                    zv = 0;
                    break;
            }
            how = hextetragon[Number-1][xv];
            break;
        case 3:
            switch(Number-1)
            {
                case 0:
                    setWrapInt (zv, 6);
                    setWrapInt (xv, 6);
                    xv += 6*zv;
                    zv = 0;
                    break;
                case 1:
                    zv += 2*(wrapInt (xv, 18)/6);
                    setWrapInt (zv, 6);
                    setWrapInt (xv, 6);
                    xv += 6*zv;
                    zv = 0;
                    break;
                case 2:
                    setWrapInt (zv, 2);
                    setWrapInt (xv, 18);
                    xv += 18*zv;
                    zv = 0;
                    break;
            }
            how = hextrigon[Number-1][xv];
            break;
        case 2:
            setWrapInt (zv, 2);
            setWrapInt (xv, 6);
            how = hexdigon[Number-1][xv+6*zv];
            break;
        case 1:
        default:
            setWrapInt (zv, 2);
            xv += 3*zv;
            setWrapInt (xv, 6);
            lng = 0;
            how = hexmonogon[Number-1][xv];
            break;
    }

    /*
     **      / \
     **     1   \
     **    /     2
     **   /       \
     **  -----3-----
     ** height = sqrt(3)/2
     */
    if (how & 0x01)
    {
        dist1 = 1.0 - (fabs(SQRT3 * x - z) * SQRT3 );
        return_value = max(return_value,dist1);
    }
    if (how & 0x02)
    {
        dist2 = 1.0 - (fabs(SQRT3 * (1.0-x) - z) * SQRT3  );
        return_value = max(return_value,dist2);
    }
    if (how & 0x04)
    {
        dist3 = 1.0 - (z * 2.0 * SQRT3 );
        return_value = max(return_value,dist3);
    }
    switch(Interior)
    {
        case 1:
            dist1 = (1.0 - (fabs(SQRT3 * z + x) ));
            dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) ));
            dist3 = (1.0 - (x * 2.0 ));
            if ( (((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000)) &&
                 (dist1<0)&&(dist2<0) )
            {
                value1 = (3.0 / 2.0 *(fabs(SQRT3 * z + x) ) - 2.0);
                value2 = (3.0 / 2.0 *(fabs(SQRT3 * z - x + 1.0) ) - 2.0);
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            if ((((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
                    &&(dist1>0)&&(dist3>0))
            {
                value1 = (1.0 - 3.0 / 2.0 * (fabs(SQRT3 * z + x) ));
                value2 = (1.0 - (x * 3.0 ));
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            if ((((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
                    &&(dist3<0)&&(dist2>0))
            {
                value1 = (1.0 - 3.0 / 2.0 *(fabs(SQRT3 * z - x + 1.0) ));
                value2 = ((x * 3.0 ) - 2.0);
                value =  min(value1,value2);
                return_value = max(return_value,value);
            }
            break;
        case 2:
            if (((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000))
            {
                dist1 = x - 0.5;
                dist2 = z - SQRT3_2;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2))*3.0 );
                return_value = max(return_value,dist3);
            }
            if (((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
            {
                dist1 = x;
                dist2 = z;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
                return_value = max(return_value,dist3);
            }
            if (((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
            {
                dist1 = x - 1.0;
                dist2 = z;
                dist3 = 1.0 - (sqrt((dist1*dist1+dist2*dist2)) *3.0);
                return_value = max(return_value,dist3);
            }
            break;
        case 0:
            if (((how & 0x83) == 0x00)||((how & 0x4000) == 0x4000))
            {
                dist3 = 1.0 - ((SQRT3_2 - z) * 2.0 * SQRT3  );
                return_value = max(return_value,dist3);
            }
            if (((how & 0x85) == 0x00)||((how & 0x2000) == 0x2000))
            {
                dist2 = 1.0 - (fabs(SQRT3 * x + z) * SQRT3 );
                return_value = max(return_value,dist2);
            }
            if (((how & 0x86) == 0x00)||((how & 0x1000) == 0x1000))
            {
                dist1 = 1.0 - (fabs(SQRT3 * (x -1.0) - z) * SQRT3 );
                return_value = max(return_value,dist1);
            }
            break;
    }
    dist1 = (1.0 - (fabs(SQRT3 * z + x) ));
    dist2 = (1.0 - (fabs(SQRT3 * z - x + 1.0) ));
    dist3 = (1.0 - (x * 2.0 ));
    switch(Form)
    {
        case 2:
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value1 = x;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
            {
                value1 = x - 0.5;
                value2 = z- SQRT3_2;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
            {
                value1 = x - 1.0;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
            {
                value1 = x - 0.5;
                value2 = z - SQRT3_2;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
            {
                value1 = x - 1.0;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x420) == 0x420)&&(dist3<0)&&(dist2>0))
            {
                value1 = x;
                value2 = z;
                value = 2.0*SQRT3*(sqrt(value1*value1+value2*value2 ) - SQRT3/3.0 );
                value = max(value,0.0);
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        case 1:
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value = -dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x140) == 0x140)&&(dist1>0)&&(dist3>0))
            {
                value = dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x210) == 0x210)&&(dist1<0)&&(dist2<0))
            {
                value = -dist2 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x240) == 0x240)&&(dist3<0)&&(dist2>0))
            {
                value = dist2 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x410) == 0x410)&&(dist1>0)&&(dist3>0))
            {
                value = dist3 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x420) == 0x420)&&(dist2>0)&&(dist3<0))
            {
                value = -dist3 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            if (((how & 0x120) == 0x120)&&(dist1<0)&&(dist2<0))
            {
                value = -dist1 * 2;
                return_value = min(value,1.0);
                return return_value;
            }
            break;
        default:
        case 0:
            break;
    }
    return return_value;
}

/*****************************************************************************
*
* FUNCTION
*
*   tiling_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   Classical tiling patterns
*
* CHANGES
*
*
******************************************************************************/
DBL TilingPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    switch(tilingType)
    {
        case 27: return tiling_penrose(EPoint, true, false);
        case 26: return tiling_penrose(EPoint, false, false);
        case 25: return tiling_penrose1(EPoint, false);
        case 24: return tiling_dodeca_hex_5(EPoint);
        case 23: return tiling_dodeca_hex(EPoint);
        case 22: return tiling_dodeca_tri(EPoint);
        case 21: return tiling_square_tri(EPoint);
        case 20: return tiling_hexa_tri_left(EPoint);
        case 19: return tiling_hexa_tri_right(EPoint);
        case 18: return tiling_rectangle_pair(EPoint);
        case 17: return tiling_hexa_square_triangle_6(EPoint);
        case 16: return tiling_hexa_square_triangle(EPoint);
        case 15: return tiling_square_double(EPoint);
        case 14: return tiling_square_internal_5(EPoint);
        case 13: return tiling_square_internal(EPoint);
        case 12: return tiling_rectangle_square(EPoint);
        case 11: return tiling_square_rectangle(EPoint);
        case 10: return tiling_square_offset(EPoint);
        case 9: return tiling_hexa_triangle(EPoint);
        case 8: return tiling_square_triangle(EPoint);
        case 7: return tiling_octa_square(EPoint);
        case 6: return tiling_rectangle(EPoint);
        case 5: return tiling_rhombus(EPoint);
        case 4: return tiling_lozenge(EPoint);
        case 3: return tiling_triangle(EPoint);
        case 2: return tiling_hexagon(EPoint);
        case 1:
        default:
            return tiling_square(EPoint);
    }
}

/*****************************************************************************
*
* FUNCTION
*
*   pavement_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   Classical tiling patterns
*
* CHANGES
*
*
******************************************************************************/
DBL PavementPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    switch(Side)
    {
        case 6:
            return hexagonal(EPoint);

        case 4:
            return tetragonal(EPoint);

        case 3:
        default:
            return trigonal(EPoint);
    }
}


// Agate grabs the z point in space value and 'mixes' that value with a
// Turbulence() value calculated with default octaves, omega and lambda values
// - unless those are modified by stand alone keywords. The result is a U shape
// distribution. One heavy toward zero and one.

DBL AgatePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);
    DBL value, turb_val;

    TurbulenceWarp Turb;
    Turb.Octaves = agateOctaves;
    Turb.Omega   = agateOmega;
    Turb.Lambda  = agateLambda;
    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    turb_val = agateFrequency * Turbulence(EPoint,&Turb,noise_generator);

    value = 0.5 * (cycloidal(1.3 * turb_val + 1.1 * EPoint.z()) + 1.0);

    value = max(0.0,min(1.0,value));

    value = pow(value, 0.77);

    return(value);
}


DBL BoxedPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value = max(0.0,1.0 - max(fabs(EPoint.x()), max(fabs(EPoint.y()), fabs(EPoint.z()))));
    return(value);
}


/*****************************************************************************
*
* FUNCTION
*
*   brick_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*   TPat   -- Texture pattern struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0 or 1.0
*
* AUTHOR
*
*   Dan Farmer
*
* DESCRIPTION
*
* CHANGES
*
*   Oct 1994    : adapted from pigment by [CY]
*
******************************************************************************/

DBL BrickPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int ibrickx, ibricky, ibrickz;
    DBL brickheight, brickwidth, brickdepth;
    DBL brickmortar, mortarheight, mortarwidth, mortardepth;
    DBL brickx, bricky, brickz;
    DBL x, y, z, fudgit;

    fudgit = gkDBL_epsilon + mortar;

    x =  EPoint[X]+fudgit;
    y =  EPoint[Y]+fudgit;
    z =  EPoint[Z]+fudgit;

    brickwidth  = brickSize[X];
    brickheight = brickSize[Y];
    brickdepth  = brickSize[Z];
    brickmortar = mortar;

    mortarwidth  = brickmortar / brickwidth;
    mortarheight = brickmortar / brickheight;
    mortardepth  = brickmortar / brickdepth;

    /* 1) Check mortar layers in the X-Z plane (ie: top view) */

    bricky = y / brickheight;
    ibricky = (int) bricky;
    bricky -= (DBL) ibricky;

    if (bricky < 0.0)
    {
        bricky += 1.0;
    }

    if (bricky <= mortarheight)
    {
        return(0.0);
    }

    bricky = (y / brickheight) * 0.5;
    ibricky = (int) bricky;
    bricky -= (DBL) ibricky;

    if (bricky < 0.0)
    {
        bricky += 1.0;
    }


    /* 2) Check ODD mortar layers in the Y-Z plane (ends) */

    brickx = (x / brickwidth);
    ibrickx = (int) brickx;
    brickx -= (DBL) ibrickx;

    if (brickx < 0.0)
    {
        brickx += 1.0;
    }

    if ((brickx <= mortarwidth) && (bricky <= 0.5))
    {
        return(0.0);
    }

    /* 3) Check EVEN mortar layers in the Y-Z plane (ends) */

    brickx = (x / brickwidth) + 0.5;
    ibrickx = (int) brickx;
    brickx -= (DBL) ibrickx;

    if (brickx < 0.0)
    {
        brickx += 1.0;
    }

    if ((brickx <= mortarwidth) && (bricky > 0.5))
    {
        return(0.0);
    }

    /* 4) Check ODD mortar layers in the Y-X plane (facing) */

    brickz = (z / brickdepth);
    ibrickz = (int) brickz;
    brickz -= (DBL) ibrickz;

    if (brickz < 0.0)
    {
        brickz += 1.0;
    }

    if ((brickz <= mortardepth) && (bricky > 0.5))
    {
        return(0.0);
    }

    /* 5) Check EVEN mortar layers in the X-Y plane (facing) */

    brickz = (z / brickdepth) + 0.5;
    ibrickz = (int) brickz;
    brickz -= (DBL) ibrickz;

    if (brickz < 0.0)
    {
        brickz += 1.0;
    }

    if ((brickz <= mortardepth) && (bricky <= 0.5))
    {
        return(0.0);
    }

    /* If we've gotten this far, colour me brick. */

    return(1.0);
}


DBL CellsPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    // Select a random value based on the cube from which this came.

    // Use floor for values, instead of just truncating. This eliminates
    // duplicated cells around the axes

    DBL value = min(gPatternRands(Hash3d(
                        (int)floor(EPoint.x()+gkMinIsectDepthReturned),
                        (int)floor(EPoint.y()+gkMinIsectDepthReturned),
                        (int)floor(EPoint.z()+gkMinIsectDepthReturned)
                    )
                ),1.0);

    return value;
}


DBL CheckerPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int value;

    value = (int)(floor(EPoint.x()+gkMinIsectDepthReturned) +
                  floor(EPoint.y()+gkMinIsectDepthReturned) +
                  floor(EPoint.z()+gkMinIsectDepthReturned));

    if (value & 1)
    {
        return (1.0);
    }
    else
    {
        return (0.0);
    }
}


/*****************************************************************************
*
* FUNCTION
*
*   crackle_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern is evaluated.
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Jim McElhiney
*
* DESCRIPTION
*
*   "crackle":
*
*   New colour function by Jim McElhiney,
*     CompuServe 71201,1326, aka mcelhiney@acm.org
*
*   Large scale, without turbulence, makes a pretty good stone wall.
*   Small scale, without turbulence, makes a pretty good crackle ceramic glaze.
*   Highly turbulent (with moderate displacement) makes a good marble, solving
*   the problem of apparent parallel layers in Perlin's method.
*   2 octaves of full-displacement turbulence make a great "drizzled paint"
*   pattern, like a 1950's counter top.
*   Rule of thumb:  put a single colour transition near 0 in your colour map.
*
*   Mathematically, the set crackle(p)=0 is a 3D Voronoi diagram of a field of
*   semirandom points, and crackle(p)>0 is distance from set along shortest path.
*   (A Voronoi diagram is the locus of points equidistant from their 2 nearest
*   neighbours from a set of disjoint points, like the membranes in suds are
*   to the centres of the bubbles).
*
*   The original "crackle" specific source code and examples are in the public domain.
*
* CHANGES
*   Oct 1994    : adapted from pigment by [CY]
*   Other changes: enhanced by Ron Parker, Integer math by Nathan Kopp
*
******************************************************************************/
static int IntPickInCube(int tvx, int tvy, int tvz, Vector3d& p1);

DBL CracklePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    Vector3d tmpPoint = EPoint;
    DBL sum, minsum, minsum2, minsum3, tf;
    int minVecIdx = 0;
    Vector3d dv;

    int flox, floy, floz;

    bool UseSquare = ( crackleMetric == 2);
    bool UseUnity  = ( crackleMetric == 1);

    if (repeat.x())
        tmpPoint.x() = wrap(tmpPoint.x(), DBL(repeat.x()));
    if (repeat.y())
        tmpPoint.y() = wrap(tmpPoint.y(), DBL(repeat.y()));
    if (repeat.z())
        tmpPoint.z() = wrap(tmpPoint.z(), DBL(repeat.z()));

    // Originally epsilon macro with 1e-10 value used below.

    flox = (int)floor(tmpPoint.x() - gkMinIsectDepthReturned);
    floy = (int)floor(tmpPoint.y() - gkMinIsectDepthReturned);
    floz = (int)floor(tmpPoint.z() - gkMinIsectDepthReturned);

    // Check to see if the input point is in the same unit cube as the last
    // call to this function, to use cache of cubelets for speed.

    CrackleCellCoord ccoord(flox, floy, floz, repeat.x(), repeat.y(), repeat.z());
    pThread->Stats()[CrackleCache_Tests]++;

    CrackleCacheEntry dummy_entry;
    CrackleCacheEntry* entry = &dummy_entry;

    if (pThread->mpCrackleCache->Lookup(entry, ccoord))
    {
        // Cache hit. `entry` now points to the cached entry.
        pThread->Stats()[CrackleCache_Tests_Succeeded]++;
    }
    else
    {
        // Cache miss. `entry` now points to a pristine entry set up in the
        // cache, or to `dummy_entry` if the cache is too crowded already.
        // In either case we need to fill in the blanks.

        // Calculate the random points for this new
        // cube and its 80 neighbours which differ in any axis by 1 or 2.
        // Why distance of 2?  If there is 1 point in each cube, located
        // randomly, it is possible for the closest random point to be in the
        // cube 2 over, or the one two over and one up.  It is NOT possible
        // for it to be two over and two up.  Picture a 3x3x3 cube with 9 more
        // cubes glued onto each face.

        // TODO - Note that we're currently re-computing each cell up to 81
        //        times - once as a main cell and 80 times as a neighbor -
        //        even in the best case scenario. Wouldn't it be more efficient
        //        to just cache the individual cells?

        int *pc = gaCrackleCubeTable;
        for (int i = 0; i < 81; i++, pc += 3)
        {
            Vector3d wrappingOffset(0.0);
            int cacheX = flox + pc[X];
            int cacheY = floy + pc[Y];
            int cacheZ = floz + pc[Z];
            if (repeat.x())
            {
                int wrapped = wrapInt(cacheX, repeat.x());
                wrappingOffset.x() += (cacheX - wrapped);
                cacheX = wrapped;
            }
            if (repeat.y())
            {
                int wrapped = wrapInt(cacheY, repeat.y());
                wrappingOffset.y() += (cacheY - wrapped);
                cacheY = wrapped;
            }
            if (repeat.z())
            {
                int wrapped = wrapInt(cacheZ, repeat.z());
                wrappingOffset.z() += (cacheZ - wrapped);
                cacheZ = wrapped;
            }
            IntPickInCube(cacheX, cacheY, cacheZ, entry->aCellNuclei[i]);
            entry->aCellNuclei[i] += wrappingOffset;
        }
    }

    // Find the 3 points with the 3 shortest distances from the input point.
    // Set up the loop so the invariant is true:  minsum <= minsum2 <= minsum3
    dv = entry->aCellNuclei[0] - tmpPoint;

    if(UseSquare)
    {
        minsum = dv.lengthSqr();

        dv = entry->aCellNuclei[1] - tmpPoint;
        minsum2 = dv.lengthSqr();

        dv = entry->aCellNuclei[2] - tmpPoint;
        minsum3  = dv.lengthSqr();
    }
    else if(UseUnity)
    {
        minsum = fabs(dv[X]) + fabs(dv[Y]) + fabs(dv[Z]);

        dv = entry->aCellNuclei[1] - tmpPoint;
        minsum2 = fabs(dv[X]) + fabs(dv[Y]) + fabs(dv[Z]);

        dv = entry->aCellNuclei[2] - tmpPoint;
        minsum3 = fabs(dv[X]) + fabs(dv[Y]) + fabs(dv[Z]);
    }
    else
    {
        minsum = pow(fabs(dv[X]), crackleMetric) +
                 pow(fabs(dv[Y]), crackleMetric) +
                 pow(fabs(dv[Z]), crackleMetric);

        dv = entry->aCellNuclei[1] - tmpPoint;
        minsum2 = pow(fabs(dv[X]), crackleMetric) +
                  pow(fabs(dv[Y]), crackleMetric) +
                  pow(fabs(dv[Z]), crackleMetric);

        dv = entry->aCellNuclei[2] - tmpPoint;
        minsum3 = pow(fabs(dv[X]), crackleMetric) +
                  pow(fabs(dv[Y]), crackleMetric) +
                  pow(fabs(dv[Z]), crackleMetric);
    }

    // sort the 3 computed sums
    if(minsum2 < minsum)
    {
        tf = minsum; minsum = minsum2; minsum2 = tf;
        minVecIdx = 1;
    }

    if(minsum3 < minsum)
    {
        tf = minsum; minsum = minsum3; minsum3 = tf;
        minVecIdx = 2;
    }

    if(minsum3 < minsum2)
    {
        tf = minsum2; minsum2 = minsum3; minsum3 = tf;
    }

    // Loop for the 81 cubelets to find closest and 2nd closest.
    for(int i = 3; i < 81; i++)
    {
        dv = entry->aCellNuclei[i] - tmpPoint;

        if(UseSquare)
            sum = dv.lengthSqr();
        else if(UseUnity)
            sum = fabs(dv[X]) + fabs(dv[Y]) + fabs(dv[Z]);
        else
            sum = pow(fabs(dv[X]), crackleMetric) +
                  pow(fabs(dv[Y]), crackleMetric) +
                  pow(fabs(dv[Z]), crackleMetric);

        if(sum < minsum)
        {
            minsum3 = minsum2;
            minsum2 = minsum;
            minsum = sum;
            minVecIdx = i;
        }
        else if(sum < minsum2)
        {
            minsum3 = minsum2;
            minsum2 = sum;
        }
        else if( sum < minsum3 )
        {
            minsum3 = sum;
        }
    }

    if (crackleOffset)
    {
        if(UseSquare)
        {
            minsum += crackleOffset*crackleOffset;
            minsum2 += crackleOffset*crackleOffset;
            minsum3 += crackleOffset*crackleOffset;
        }
        else if (UseUnity)
        {
            minsum += crackleOffset;
            minsum2 += crackleOffset;
            minsum3 += crackleOffset;
        }
        else
        {
            minsum += pow( crackleOffset, crackleMetric );
            minsum2 += pow( crackleOffset, crackleMetric );
            minsum3 += pow( crackleOffset, crackleMetric );
        }
    }

    if(crackleIsSolid)
    {
        tf = Noise( entry->aCellNuclei[minVecIdx], GetNoiseGen(pThread) );
    }
    else if(UseSquare)
    {
        tf = crackleForm[X]*sqrt(minsum) +
             crackleForm[Y]*sqrt(minsum2) +
             crackleForm[Z]*sqrt(minsum3);
    }
    else if(UseUnity)
    {
        tf = crackleForm[X]*minsum +
             crackleForm[Y]*minsum2 +
             crackleForm[Z]*minsum3;
    }
    else
    {
        tf = crackleForm[X]*pow(minsum, 1.0/crackleMetric) +
             crackleForm[Y]*pow(minsum2, 1.0/crackleMetric) +
             crackleForm[Z]*pow(minsum3, 1.0/crackleMetric);
    }

    return max(min(tf, 1.), 0.);
}


DBL CylindricalPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    // EPoint -- Point in 3d space for evaluation.
    DBL value = max(0.0,1.0-sqrt(Sqr(EPoint.x()) + Sqr(EPoint.z())));
    return(value);
}


//-------------------------------------
// density_file pattern.
//
// Author Dieter Bayer
//
// Dec 1996 : Creation.
//

inline DBL fblob(DBL v, DBL s)
{
  //-------- exp expensive in this DF3 interp usage so experimented some
  //         with the following fast exp() method from:
  // http://www.schraudolph.org/pubs/Schraudolph99.pdf
  // https://gist.github.com/andersx/8057b2a6fd3d715d35eb <- less accurate, faster still
  //         Below +-17.8% in iso usage, small inaccuracies visible.
  //         Suppose inaccuracy might be OK for media applications.
  // exp(x)=2^(x*log2(e))=2^(xi-xf) look up tables often used for xi,xf
  // IEEE-745 (+-1)(1+m)2^(x-x0) where x0 is 1023, m=11bits, x=52bits.
  // int(32bit)=A*x+B-C; A=S/ln(s); B=S*1023; C=60801; S=2^20;
  // Pad 32bit int with another 32bit int to form 64bit set.
  // Cast the 64bits to a double.
  //
  // double tmpDBL=(v*s);
  // double S=pow(2,20);
  // double A=S/log(2);
  // double B=S*1023.0;
  // long int tmpLInt = (long int) (A*tmpDBL+B-60801)*pow(2,32);
  // return 1.0/reinterpret_cast<double&>(tmpLInt);
  //
  //         Perhaps with vectorization something like 2015 IBM paper would work & be faster.
  // https://www.researchgate.net/profile/A_Cristiano_I_Malossi/publication/272178514_Fast_Exponential_Computation_on_SIMD_Architectures/links/54de344f0cf22a26721fbdc9.pdf
  //         Also, glibc is to get a vectorized exp as part of fastmath optimization.
  //
  return 1.0/exp(v*s);
}

inline void calcAllDiffsSqrd(double *aryd, double this_d, size_t *ary)
{  //---- Do all our dist calculations for axis at this point
   size_t n;
   DBL tmpVal;
   for(n = 0; n<8; n++)
   {
       tmpVal=(DBL)ary[n]-this_d;
       aryd[n]=tmpVal*tmpVal;
   }
}

inline float intp3(float t, float fa, float fb, float fc, float fd)
{
    float b,d,e,f;

    b = (fc - fa) * 0.5;
    d = (fd - fb) * 0.5;
    e = 2.0 * (fb - fc) + b + d;
    f = -3.0 * (fb - fc) - 2.0 * b - d;

    return ((e * t + f) * t + b) * t + fb;
}

inline float intp3_2(float t, float fa, float fb, float fc, float fd)
{
    float b,e,f;

    e = fd - fc - fa + fb;
    f = fa - fb - e;
    b = fc - fa;

    return ((e * t + f) * t + b) * t + fb;
}

#define zmax(i,imax) (((i)<0)?(imax-1):((i) % (imax)))

DBL DensityFilePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    size_t x, y, z;
    size_t x1, y1, z1;
    size_t x2, y2, z2;
    DBL Ex, Ey, Ez;
    DBL this_x, this_y, this_z;
    DBL xx, yy, zz;
    DBL xi, yi;
    DBL f111, f112, f121, f122, f211, f212, f221, f222;
    float intpd2[4][4];
    DBL density = 0.0;
    DENSITY_FILE_DATA *Data;
    size_t k0, k1, k2, k3, i, j, k, ii, jj, kk;
    size_t rx[8],ry[8],rz[8];
    DBL    rxd[8],ryd[8],rzd[8];
    size_t startidx, stopidx;
    DBL    runningDistSqrd,runningval,strengthBias;
    unsigned short tmpUshort;
    unsigned int   tmpUint;
    unsigned long  tmpUlong;

    Ex=EPoint[X];
    Ey=EPoint[Y];
    Ez=EPoint[Z];

    if ((densityFile != nullptr) && ((Data = densityFile->Data) != nullptr) &&
       (Data->Sx) && (Data->Sy) && (Data->Sz))
    {
        Ex=EPoint[X];
        Ey=EPoint[Y];
        Ez=EPoint[Z];

        if((Ex >= 0.0) && (Ex < 1.0) && (Ey >= 0.0) && (Ey < 1.0) && (Ez >= 0.0) && (Ez < 1.0))
        {
            switch (densityFile->Interpolation)
            {
                case kDensityFileInterpolation_None:
                    x = (size_t)(Ex * (DBL)Data->Sx);
                    y = (size_t)(Ey * (DBL)Data->Sy);
                    z = (size_t)(Ez * (DBL)Data->Sz);

                    density = 0.0;
                    if(Data->Type == 4)
                    {
                        density = (DBL)Data->Density32[z * Data->Sy * Data->Sx + y * Data->Sx + x] / (DBL)UNSIGNED32_MAX;
                    }
                    else if(Data->Type==2)
                    {
                        density = (DBL)Data->Density16[z * Data->Sy * Data->Sx + y * Data->Sx + x] / (DBL)UNSIGNED16_MAX;
                    }
                    else if(Data->Type == 1)
                    {
                        density = (DBL)Data->Density8[z * Data->Sy * Data->Sx + y * Data->Sx + x] / (DBL)UNSIGNED8_MAX;
                    }
                    break;
                case kDensityFileInterpolation_Trilinear:
                    Ex = max(0.0,Ex-(1.0/(DBL)Data->Sx/2.0)); // Adjustment to voxel center
                    Ey = max(0.0,Ey-(1.0/(DBL)Data->Sy/2.0));
                    Ez = max(0.0,Ez-(1.0/(DBL)Data->Sz/2.0));
                    // FALLTHROUGH
                case kDensityFileInterpolation_Trilinear_Shftd: // (11 vs 1) As means to provide backward shift left-forward-down compatibility
                    xx = Ex * (DBL)(Data->Sx);
                    yy = Ey * (DBL)(Data->Sy);
                    zz = Ez * (DBL)(Data->Sz);

                    x1 = (size_t)xx;
                    y1 = (size_t)yy;
                    z1 = (size_t)zz;

                    x2 = (x1 + 1) % Data->Sx;
                    y2 = (y1 + 1) % Data->Sy;
                    z2 = (z1 + 1) % Data->Sz;

                    xx -= floor(xx);
                    yy -= floor(yy);
                    zz -= floor(zz);

                    xi = 1.0 - xx;
                    yi = 1.0 - yy;

                    if(Data->Type == 4)
                    {
                        f111 = (DBL)Data->Density32[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED32_MAX;
                        f112 = (DBL)Data->Density32[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED32_MAX;
                        f121 = (DBL)Data->Density32[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED32_MAX;
                        f122 = (DBL)Data->Density32[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED32_MAX;
                        f211 = (DBL)Data->Density32[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED32_MAX;
                        f212 = (DBL)Data->Density32[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED32_MAX;
                        f221 = (DBL)Data->Density32[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED32_MAX;
                        f222 = (DBL)Data->Density32[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED32_MAX;
                    }
                    else if(Data->Type == 2)
                    {
                        f111 = (DBL)Data->Density16[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED16_MAX;
                        f112 = (DBL)Data->Density16[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED16_MAX;
                        f121 = (DBL)Data->Density16[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED16_MAX;
                        f122 = (DBL)Data->Density16[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED16_MAX;
                        f211 = (DBL)Data->Density16[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED16_MAX;
                        f212 = (DBL)Data->Density16[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED16_MAX;
                        f221 = (DBL)Data->Density16[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED16_MAX;
                        f222 = (DBL)Data->Density16[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED16_MAX;
                    }
                    else if(Data->Type == 1)
                    {
                        f111 = (DBL)Data->Density8[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED8_MAX;
                        f112 = (DBL)Data->Density8[z1 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED8_MAX;
                        f121 = (DBL)Data->Density8[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED8_MAX;
                        f122 = (DBL)Data->Density8[z1 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED8_MAX;
                        f211 = (DBL)Data->Density8[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x1] / (DBL)UNSIGNED8_MAX;
                        f212 = (DBL)Data->Density8[z2 * Data->Sy * Data->Sx + y1 * Data->Sx + x2] / (DBL)UNSIGNED8_MAX;
                        f221 = (DBL)Data->Density8[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x1] / (DBL)UNSIGNED8_MAX;
                        f222 = (DBL)Data->Density8[z2 * Data->Sy * Data->Sx + y2 * Data->Sx + x2] / (DBL)UNSIGNED8_MAX;
                    }
                    else
                        POV_PATTERN_ASSERT (false);

                    density = ((f111 * xi + f112 * xx) * yi + (f121 * xi + f122 * xx) * yy) * (1.0 - zz) +
                              ((f211 * xi + f212 * xx) * yi + (f221 * xi + f222 * xx) * yy) * zz;
                    if (density < 0.0) density = 0.0; // Note - Clamp for >1.0 in pattern wave modification code.
                    break;
                case kDensityFileInterpolation_Tricubic:
                default:
                    Ex = max(0.0,Ex-(1.0/(DBL)Data->Sx/2.0)); // Adjustment to voxel center
                    Ey = max(0.0,Ey-(1.0/(DBL)Data->Sy/2.0));
                    Ez = max(0.0,Ez-(1.0/(DBL)Data->Sz/2.0));
                    // FALLTHROUGH
                case kDensityFileInterpolation_Tricubic_Shftd: // (12 vs 2) As means to provide backward shift left-forward-down compatibility
                    xx = Ex * (DBL)(Data->Sx);
                    yy = Ey * (DBL)(Data->Sy);
                    zz = Ez * (DBL)(Data->Sz);

                    x1 = (size_t)xx;
                    y1 = (size_t)yy;
                    z1 = (size_t)zz;

                    xx -= floor(xx);
                    yy -= floor(yy);
                    zz -= floor(zz);

                    k0 = zmax(-1+z1, Data->Sz);
                    k1 = zmax(   z1, Data->Sz);
                    k2 = zmax( 1+z1, Data->Sz);
                    k3 = zmax( 2+z1, Data->Sz);

                    if(Data->Type == 4)
                    {
                        for(i = 0; i < 4; i++)
                        {
                            ii = zmax(i + x1 - 1, Data->Sx);
                            for(j = 0; j < 4; j++)
                            {
                                jj = zmax(j + y1 - 1, Data->Sy);
                                intpd2[i][j] = intp3(zz,
                                                     Data->Density32[k0 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED32_MAX,
                                                     Data->Density32[k1 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED32_MAX,
                                                     Data->Density32[k2 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED32_MAX,
                                                     Data->Density32[k3 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED32_MAX);
                            }
                        }
                    }
                    else if(Data->Type == 2)
                    {
                        for(i = 0; i < 4; i++)
                        {
                            ii = zmax(i + x1 - 1, Data->Sx);
                            for(j = 0; j < 4; j++)
                            {
                                jj = zmax(j + y1 - 1, Data->Sy);
                                intpd2[i][j] = intp3(zz,
                                                     Data->Density16[k0 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED16_MAX,
                                                     Data->Density16[k1 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED16_MAX,
                                                     Data->Density16[k2 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED16_MAX,
                                                     Data->Density16[k3 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED16_MAX);
                            }
                        }
                    }
                    else if(Data->Type == 1)
                    {
                        for(i = 0; i < 4; i++)
                        {
                            ii = zmax(i + x1 - 1, Data->Sx);
                            for(j = 0; j < 4; j++)
                            {
                                jj = zmax(j + y1 - 1, Data->Sy);
                                intpd2[i][j] = intp3(zz,
                                                     Data->Density8[k0 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED8_MAX,
                                                     Data->Density8[k1 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED8_MAX,
                                                     Data->Density8[k2 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED8_MAX,
                                                     Data->Density8[k3 * Data->Sy * Data->Sx + jj * Data->Sx + ii] / (DBL)UNSIGNED8_MAX);
                            }
                        }
                    }

                    for(i = 0; i < 4; i++)
                        intpd2[0][i] = intp3(yy, intpd2[i][0], intpd2[i][1],  intpd2[i][2], intpd2[i][3]);

                    density = intp3(xx, intpd2[0][0], intpd2[0][1], intpd2[0][2], intpd2[0][3]);
                    if (density < 0.0) density = 0.0;
                    break;
                case kDensityFileInterpolation_BlobFour:
                case kDensityFileInterpolation_BlobSix:
                case kDensityFileInterpolation_BlobEight:
                    Ex -= 1.0/(DBL)Data->Sx/2.0; // Adjustment to voxel center
                    Ey -= 1.0/(DBL)Data->Sy/2.0;
                    Ez -= 1.0/(DBL)Data->Sz/2.0;

                    this_x = Ex * (DBL)(Data->Sx);
                    this_y = Ey * (DBL)(Data->Sy);
                    this_z = Ez * (DBL)(Data->Sz);

                    x1 = (size_t)this_x;
                    y1 = (size_t)this_y;
                    z1 = (size_t)this_z;

                    rx[3]=max((int)0,(int)x1);        ry[3]=max((int)0,(int)y1);        rz[3]=max((int)0,(int)z1);
                    rx[2]=max((int)0,(int)x1-(int)1); ry[2]=max((int)0,(int)y1-(int)1); rz[2]=max((int)0,(int)z1-(int)1);
                    rx[1]=max((int)0,(int)x1-(int)2); ry[1]=max((int)0,(int)y1-(int)2); rz[1]=max((int)0,(int)z1-(int)2);
                    rx[0]=max((int)0,(int)x1-(int)3); ry[0]=max((int)0,(int)y1-(int)3); rz[0]=max((int)0,(int)z1-(int)3);
                    //---
                    rx[4]=min(Data->Sx-1,x1+1); ry[4]=min(Data->Sy-1,y1+1); rz[4]=min(Data->Sz-1,z1+1);
                    rx[5]=min(Data->Sx-1,x1+2); ry[5]=min(Data->Sy-1,y1+2); rz[5]=min(Data->Sz-1,z1+2);
                    rx[6]=min(Data->Sx-1,x1+3); ry[6]=min(Data->Sy-1,y1+3); rz[6]=min(Data->Sz-1,z1+3);
                    rx[7]=min(Data->Sx-1,x1+4); ry[7]=min(Data->Sy-1,y1+4); rz[7]=min(Data->Sz-1,z1+4);

                    calcAllDiffsSqrd(rxd,this_x,rx);
                    calcAllDiffsSqrd(ryd,this_y,ry);
                    calcAllDiffsSqrd(rzd,this_z,rz);

                    switch (densityFile->Interpolation)
                    {
                        case kDensityFileInterpolation_BlobFour:
                            startidx     = 2;
                            stopidx      = 5;
                            strengthBias = 2.4748737341529163; // Via manual tuning. Inexact due exp() & potential uses.
                            break;
                        case kDensityFileInterpolation_BlobSix:
                            startidx     = 1;
                            stopidx      = 6;
                            strengthBias = 1.2374368670764582;
                            break;
                        case kDensityFileInterpolation_BlobEight:
                        default:
                            startidx     = 0;
                            stopidx      = 7;
                            strengthBias = 0.6187184335382291;
                            break;
                    }
                    runningval=0.0;
                    if(Data->Type == 4)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUlong=Data->Density32[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if (tmpUlong)
                                    {
                                        f111 = strengthBias/((DBL)tmpUlong/(DBL)UNSIGNED32_MAX);
                                        runningval+=fblob(rzd[i]+ryd[j]+rxd[k],f111);
                                    }
                                }
                            }
                        }
                    }
                    else if(Data->Type == 2)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUint=Data->Density16[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if (tmpUint)
                                    {
                                        f111 = strengthBias/((DBL)tmpUint/(DBL)UNSIGNED16_MAX);
                                        runningval+=fblob(rzd[i]+ryd[j]+rxd[k],f111);
                                    }
                                }
                            }
                        }
                    }
                    else if(Data->Type == 1)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUshort=Data->Density8[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if (tmpUshort)
                                    {
                                        f111 = strengthBias/((DBL)tmpUshort/(DBL)UNSIGNED8_MAX);
                                        runningval+=fblob(rzd[i]+ryd[j]+rxd[k],f111);
                                    }
                                }
                            }
                        }
                    }
                    density = min(1.0,runningval);
                    break;  // Break for last case in blob switch
                case kDensityFileInterpolation_NearestNonZeroR:
                case kDensityFileInterpolation_NearestNonZero:
                    Ex -= 1.0/(DBL)Data->Sx/2.0; // Adjustment to voxel center
                    Ey -= 1.0/(DBL)Data->Sy/2.0;
                    Ez -= 1.0/(DBL)Data->Sz/2.0;

                    this_x = Ex * (DBL)(Data->Sx);
                    this_y = Ey * (DBL)(Data->Sy);
                    this_z = Ez * (DBL)(Data->Sz);

                    x1 = (size_t)this_x;
                    y1 = (size_t)this_y;
                    z1 = (size_t)this_z;

                    rx[3]=max((int)0,(int)x1);        ry[3]=max((int)0,(int)y1);        rz[3]=max((int)0,(int)z1);
                    rx[2]=max((int)0,(int)x1-(int)1); ry[2]=max((int)0,(int)y1-(int)1); rz[2]=max((int)0,(int)z1-(int)1);
                    rx[1]=max((int)0,(int)x1-(int)2); ry[1]=max((int)0,(int)y1-(int)2); rz[1]=max((int)0,(int)z1-(int)2);
                    rx[0]=max((int)0,(int)x1-(int)3); ry[0]=max((int)0,(int)y1-(int)3); rz[0]=max((int)0,(int)z1-(int)3);
                    //---
                    rx[4]=min(Data->Sx-1,x1+1); ry[4]=min(Data->Sy-1,y1+1); rz[4]=min(Data->Sz-1,z1+1);
                    rx[5]=min(Data->Sx-1,x1+2); ry[5]=min(Data->Sy-1,y1+2); rz[5]=min(Data->Sz-1,z1+2);
                    rx[6]=min(Data->Sx-1,x1+3); ry[6]=min(Data->Sy-1,y1+3); rz[6]=min(Data->Sz-1,z1+3);
                    rx[7]=min(Data->Sx-1,x1+4); ry[7]=min(Data->Sy-1,y1+4); rz[7]=min(Data->Sz-1,z1+4);

                    calcAllDiffsSqrd(rxd,this_x,rx);
                    calcAllDiffsSqrd(ryd,this_y,ry);
                    calcAllDiffsSqrd(rzd,this_z,rz);

                    startidx     = 0;
                    stopidx      = 7;

                    runningval=0.0;
                    runningDistSqrd=1e30;
                    if(Data->Type == 4)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUlong=Data->Density32[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if ((tmpUlong) && ((rzd[i]+ryd[j]+rxd[k])<runningDistSqrd))
                                    {
                                        switch (densityFile->Interpolation)
                                        {
                                            case kDensityFileInterpolation_NearestNonZeroR:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k])+0.50*(-0.5+(DBL)std::rand()/RAND_MAX);
                                                break;
                                            case kDensityFileInterpolation_NearestNonZero:
                                            default:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k]);
                                                break;
                                        }
                                        runningval=(DBL)tmpUlong/(DBL)UNSIGNED32_MAX;
                                    }
                                }
                            }
                        }
                    }
                    else if(Data->Type == 2)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUint=Data->Density16[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if ((tmpUint) && ((rzd[i]+ryd[j]+rxd[k])<runningDistSqrd))
                                    {
                                        switch (densityFile->Interpolation)
                                        {
                                            case kDensityFileInterpolation_NearestNonZeroR:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k])+0.50*(-0.5+(DBL)std::rand()/RAND_MAX);
                                                break;
                                            case kDensityFileInterpolation_NearestNonZero:
                                            default:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k]);
                                                break;
                                        }
                                        runningval=(DBL)tmpUint/(DBL)UNSIGNED16_MAX;
                                    }
                                }
                            }
                        }
                    }
                    else if(Data->Type == 1)
                    {
                        for(i = startidx; i <= stopidx; i++)
                        {
                            for(j = startidx; j <= stopidx; j++)
                            {
                                for(k = startidx; k <= stopidx; k++)
                                {
                                    tmpUshort=Data->Density8[rz[i] * Data->Sy * Data->Sx + ry[j] * Data->Sx + rx[k]];
                                    if ((tmpUshort) && ((rzd[i]+ryd[j]+rxd[k])<runningDistSqrd))
                                    {
                                        switch (densityFile->Interpolation)
                                        {
                                            case kDensityFileInterpolation_NearestNonZeroR:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k])+0.50*(-0.5+(DBL)std::rand()/RAND_MAX);
                                                break;
                                            case kDensityFileInterpolation_NearestNonZero:
                                            default:
                                                runningDistSqrd=(rzd[i]+ryd[j]+rxd[k]);
                                                break;
                                        }
                                        runningval=(DBL)tmpUshort/(DBL)UNSIGNED8_MAX;
                                    }
                                }
                            }
                        }
                    }
                    density = min(1.0-gkDBL_epsilon,runningval);
                    break;  // Break for nearest non zero search
            }
        }
        else
            density = 0.0;
    }

    return density;
}


DBL FunctionPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    // The EPoint here == x,y,z values where the function is evaluated.
    DBL value = GenericScalarFunctionInstance(pFn, pThread).Evaluate(EPoint);
    return (value);
}


DBL GradientPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    // The gradient vector is normalized by the parser. The dot product
    // of any 3d point in space with the gradient vector becomes a value
    // representing the length of the projection of the EPoint 'vector'
    // onto the unit 1 length gradient vector. The projection can be
    // and often is negative in value.
    DBL value = dot(EPoint, gradient);

    return value;
}


DBL GranitePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);

    int i;
    DBL temp, value = 0.0, freq = 1.0;
    Vector3d tv1, tv2;

    tv1 = EPoint * 4.0;

    for (i = 0; i < 6; freq *= 2.0, i++)
    {
        tv2 = tv1 * freq;

        // TODO - This distinction (with minor variations that seem to be more
        // of an inconsistency rather than intentional) appears in other places
        // as well; make it a function.

        switch (noise_generator)
        {
            case kNoiseGen_Original:
                temp = 0.5 - Noise(tv2,noise_generator);
                temp = fabs(temp);
                break;

            default:
                temp = 1.0 - 2.0 * Noise(tv2,noise_generator); // TODO similar code clips the result
                temp = fabs(temp);
                if (temp>0.5) temp=0.5;
                break;
        }

        value += temp / freq;

        // Note. See f_granite inbuilt code for similar approach mixing noise 2&3 with
        // too a set of post pattern creation filters for fun and profit or...
    }
    return(value);
}


/*****************************************************************************
*
* FUNCTION
*
*   hexagon_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0 or 2.0
*
* AUTHOR
*
*   Ernest MacDougal Campbell III
*
* DESCRIPTION
*
*   TriHex pattern -- Ernest MacDougal Campbell III (EMC3) 11/23/92
*
*   Creates a hexagon pattern in the XZ plane.
*
*   This algorithm is hard to explain.  First it scales the point to make
*   a few of the later calculations easier, then maps some points to be
*   closer to the Origin.  A small area in the first quadrant is subdivided
*   into a 6 x 6 grid.  The position of the point mapped into that grid
*   determines its colour.  For some points, just the grid location is enough,
*   but for others, we have to calculate which half of the block it's in
*   (this is where the atan2() function comes in handy).
*
* CHANGES
*
*   Nov 1992 : Creation.
*   Oct 1994 : adapted from pigment by [CY]
*
******************************************************************************/

const DBL xfactor = 0.5;         /* each triangle is split in half for the grid */
const DBL zfactor = 0.866025404; /* sqrt(3)/2 -- Height of an equilateral triangle */

DBL HexagonPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int xm, zm;
    int brkindx;
    DBL xs, zs, xl, zl, value = 0.0;
    DBL x=EPoint[X];
    DBL z=EPoint[Z];

    // Keep all numbers positive.  Also, if z is negative, map it in such a way
    // as to avoid mirroring across the x-axis.  The value 5.196152424 is
    // (sqrt(3)/2) * 6 (because the grid is 6 blocks high)

    x = fabs(x);

    /* Avoid mirroring across x-axis. */

    z = z < 0.0 ? 5.196152424 - fabs(z) : z;    /// TODO FIXME - magic number! Use an aptly named const instead.

    /* Scale point to make calcs easier. */

    xs = x/xfactor;
    zs = z/zfactor;

    /* Map points into the 6 x 6 grid where the basic formula works. */

    xs -= floor(xs/6.0) * 6.0;
    zs -= floor(zs/6.0) * 6.0;

    /* Get a block in the 6 x 6 grid. */

    // FLOOR() faster and safe here because xs, zs all positive at this point.
    xm = (int) FLOOR(xs) % 6;
    zm = (int) FLOOR(zs) % 6;

    switch (xm)
    {
        /* These are easy cases: Colour depends only on xm and zm. */

        case 0:
        case 5:

            switch (zm)
            {
                case 0:
                case 5: value = 0; break;

                case 1:
                case 2: value = 1; break;

                case 3:
                case 4: value = 2; break;
            }

            break;

        case 2:
        case 3:

            switch (zm)
            {
                case 0:
                case 1: value = 2; break;

                case 2:
                case 3: value = 0; break;

                case 4:
                case 5: value = 1; break;
            }

            break;

        /* These cases are harder.  These blocks are divided diagonally
         * by the angled edges of the hexagons.  Some slope positive, and
         * others negative.  We flip the x value of the negatively sloped
         * pieces.  Then we check to see if the point in question falls
         * in the upper or lower half of the block.  That info, plus the
         * z status of the block determines the colour.
         */

        case 1:
        case 4:

            /* Map the point into the block at the origin. */

            xl = xs-xm;
            zl = zs-zm;

            /* These blocks have negative slopes so we flip it horizontally. */

            if (((xm + zm) % 2) == 1)
            {
                xl = 1.0 - xl;
            }

            /* Avoid a divide-by-zero error. */

            if (xl == 0.0)
            {
                xl = 0.0001; // TODO FIXME - magic number! Should be epsilon I guess, or use nextafter()
            }

            /* Is the angle less-than or greater-than 45 degrees? */

            brkindx = (zl / xl) < 1.0;

            /* was...
             * brkindx = (atan2(zl,xl) < (45 * M_PI_180));
             * ...but because of the mapping, it's easier and cheaper,
             * CPU-wise, to just use a good ol' slope.
             */

            switch (brkindx)
            {
                case true:

                    switch (zm)
                    {
                        case 0:
                        case 3: value = 0; break;

                        case 2:
                        case 5: value = 1; break;

                        case 1:
                        case 4: value = 2; break;
                    }

                    break;

                case false:

                    switch (zm)
                    {
                        case 0:
                        case 3: value = 2; break;

                        case 2:
                        case 5: value = 0; break;

                        case 1:
                        case 4: value = 1; break;
                    }

                    break;
            }
    }

    value = fmod(value, 3.0);

    return(value);
}


/*****************************************************************************
*
* FUNCTION
*
*   cubic_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0, 2.0, 3.0, 4.0 or 5.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   Creates a cubic pattern. The six texture elements are mapped to the six
*   pyramids centered on the six axes.
*
* CHANGES
*
*   Nov 2007 : Creation.
*
******************************************************************************/

DBL CubicPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    const DBL x = EPoint[X], y = EPoint[Y], z = EPoint[Z];
    const DBL ax = fabs(x), ay = fabs(y), az = fabs(z);

    if(x >= 0 && x >= ay && x >= az) return 0.0;
    if(y >= 0 && y >= ax && y >= az) return 1.0;
    if(z >= 0 && z >= ax && z >= ay) return 2.0;
    if(x < 0 && x <= -ay && x <= -az) return 3.0;
    if(y < 0 && y <= -ax && y <= -az) return 4.0;
    return 5.0;
}

/*****************************************************************************
*
* FUNCTION
*
*   square_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0, 2.0 or 3.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*   Paving the XZ plan with 4 'colours', in square
*
* CHANGES
*
******************************************************************************/

DBL SquarePattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int valueX,valueZ;

    valueX = (int)(floor(EPoint[X]));
    valueZ = (int)(floor(EPoint[Z]));

    if (valueX & 1)
    {
        if (valueZ & 1)
        {
            return (2.0);
        }
        else
        {
            return (3.0);
        }
    }
    else
    {
        if (valueZ & 1)
        {
            return (1.0);
        }
        else
        {
            return (0.0);
        }
    }
}

/*****************************************************************************
*
* FUNCTION
*
*   triangular_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value exactly 0.0, 1.0, 2.0, 3.0, 4.0 or 5.0
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*   Paving the XZ plan with 6 'colours', in triangle around the origin
*
* CHANGES
*
******************************************************************************/

DBL TriangularPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL answer;
    DBL x,z;
    DBL xs,zs;
    int a,b;
    DBL k,slop1,slop2;
    int mask;

    x=EPoint[X];
    z=EPoint[Z];
  /* Fold the space to a basic rectangle */
    xs = x-3.0*floor(x/3.0);
    zs = z-SQRT3*floor(z/SQRT3);

    /* xs,zs is in { [0.0, 3.0 [, [0.0, SQRT3 [ }
     ** but there is some symmetry to simplify the testing
     */

    a = (int)floor(xs);
    xs -= a;
    b = (zs <SQRT3_2 ? 0: 1);
    if (b)
    {
        zs = SQRT3 - zs; /* mirror */
    }

    k = 1.0 - xs;
    if ((xs != 0.0)&&( k != 0.0 )) /* second condition should never occurs */
    {
        slop1 = zs/xs;
        slop2 = zs/k; /* just in case */
        switch( (slop1<SQRT3?1:0)+(slop2<SQRT3?2:0))
        {
            case 3:
                answer = 0.0;
                break;
            case 2:
                answer = 1.0;
                break;
            case 1:
                answer = 3.0;
                break;
            default:
                POV_PATTERN_ASSERT (false);
                break;
        }
    }
    else
    {
        answer = 1.0;
    }
    mask = (int) answer;
    answer = (mask & 1) ? fmod(answer+2.0*a,6.0): fmod(6.0+answer-2.0*a,6.0);
    if (b)
    {
        answer = 5.0 - answer;
    }

    return answer;
}




/*****************************************************************************
*
* FUNCTION
*
*   julia_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL JuliaPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, dist2, mindist2,
        cr = juliaCoord[U], ci = juliaCoord[V];

    a = EPoint[X]; a2 = Sqr(a);
    b = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b  = 2.0 * a * b + ci;
        a  = a2 - b2 + cr;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   julia3_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Julia3Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, dist2, mindist2,
        cr = juliaCoord[U], ci = juliaCoord[V];

    a = EPoint[X]; a2 = Sqr(a);
    b = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b = 3.0*a2*b - b2*b + ci;
        a = a2*a - 3.0*a*b2 + cr;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   julia4_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Julia4Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, dist2, mindist2,
        cr = juliaCoord[U], ci = juliaCoord[V];

    a = EPoint[X]; a2 = Sqr(a);
    b = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b = 4.0 * (a2*a*b - a*b2*b) + ci;
        a = a2*a2 - 6.0*a2*b2 + b2*b2 + cr;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   juliax_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL JuliaXPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf=0, x, y, dist2, mindist2,
        cr = juliaCoord[U], ci = juliaCoord[V];
    int* binomial_coeff;

    a = x = EPoint[X];
    b = y = EPoint[Y];
    mindist2 = a*a+b*b;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    binomial_coeff = &(gaBinomialCoefficients[(fractalExponent+1)*fractalExponent/2]);

    for (col = 0; col < it_max; col++)
    {
        // Calculate (a+bi)^fractalExponent
        DBL new_a = pow(a, fractalExponent);
        for(int k=2; k<=fractalExponent; k+=2)
        {
            new_a += binomial_coeff[k]*pow(a, fractalExponent-k)*pow(b, k);
        }
        DBL new_b = 0;
        for(int l=1; l<=fractalExponent; l+=2)
        {
            new_b += binomial_coeff[l]*pow(a, fractalExponent-l)*pow(b, l);
        }

        a = new_a + cr;
        b = new_b + ci;

        dist2 = a*a+b*b;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


// Author: Scott Taylor
//
// Jul 1991 : Creation.
// Oct 1994 : adapted from pigment by [CY]
//
double LeopardPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    // This form didn't work with Zortech 386 compiler
    // value = Sqr((sin(x)+sin(y)+sin(z))/3);
    // So we break it down. TODO. After time, review return to original form restored Nov 09, 2020

  //double temp1 = sin(EPoint[X]);
  //double temp2 = sin(EPoint[Y]);
  //double temp3 = sin(EPoint[Z]);

  //double value = Sqr((temp1 + temp2 + temp3) / 3.0);

    double value = ( sin(EPoint.x()) + sin(EPoint.y()) + sin(EPoint.z()) ) / 3.0;

    return(value*value);
}


/*****************************************************************************
*
* FUNCTION
*
*   magnet1m_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Magnet1MPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, x, y, tmp, tmp1r, tmp1i, tmp2r, tmp2i, dist2, mindist2;

    x = EPoint[X];
    y = EPoint[Y];
    a = a2 = 0;
    b = b2 = 0;
    mindist2 = 10000;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        tmp1r = a2-b2 + x-1;
        tmp1i = 2*a*b + y;
        tmp2r = 2*a + x-2;
        tmp2i = 2*b + y;
        tmp = tmp2r*tmp2r + tmp2i*tmp2i;
        a = (tmp1r*tmp2r + tmp1i*tmp2i) / tmp;
        b = (tmp1i*tmp2r - tmp1r*tmp2i) / tmp;
        b2 = b*b;
        b = 2*a*b;
        a = a*a-b2;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        tmp1r = a-1;
        if(dist2 > 10000.0 || tmp1r*tmp1r+b2 < 1/10000.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   magnet1j_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Magnet1JPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, tmp, tmp1r, tmp1i, tmp2r, tmp2i, dist2, mindist2,
        cr = juliaCoord[U], ci = juliaCoord[V];

    a = EPoint[X]; a2 = Sqr(a);
    b = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        tmp1r = a2-b2 + cr-1;
        tmp1i = 2*a*b + ci;
        tmp2r = 2*a + cr-2;
        tmp2i = 2*b + ci;
        tmp = tmp2r*tmp2r + tmp2i*tmp2i;
        a = (tmp1r*tmp2r + tmp1i*tmp2i) / tmp;
        b = (tmp1i*tmp2r - tmp1r*tmp2i) / tmp;
        b2 = b*b;
        b = 2*a*b;
        a = a*a-b2;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        tmp1r = a-1;
        if(dist2 > 10000.0 || tmp1r*tmp1r+b2 < 1/10000.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   magnet2m_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Magnet2MPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, x, y, tmp, tmp1r, tmp1i, tmp2r, tmp2i,
        c1r, c2r, c1c2r, c1c2i, dist2, mindist2;

    x = EPoint[X];
    y = EPoint[Y];
    a = a2 = 0;
    b = b2 = 0;
    mindist2 = 10000;

    c1r = x-1; c2r = x-2;
    c1c2r = c1r*c2r-y*y;
    c1c2i = (c1r+c2r)*y;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        tmp1r = a2*a-3*a*b2 + 3*(a*c1r-b*y) + c1c2r;
        tmp1i = 3*a2*b-b2*b + 3*(a*y+b*c1r) + c1c2i;
        tmp2r = 3*(a2-b2) + 3*(a*c2r-b*y) + c1c2r + 1;
        tmp2i = 6*a*b + 3*(a*y+b*c2r) + c1c2i;
        tmp = tmp2r*tmp2r + tmp2i*tmp2i;
        a = (tmp1r*tmp2r + tmp1i*tmp2i) / tmp;
        b = (tmp1i*tmp2r - tmp1r*tmp2i) / tmp;
        b2 = b*b;
        b = 2*a*b;
        a = a*a-b2;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        tmp1r = a-1;
        if(dist2 > 10000.0 || tmp1r*tmp1r+b2 < 1/10000.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   magnet2j_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Magnet2JPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, tmp, tmp1r, tmp1i, tmp2r, tmp2i, c1r,c2r,c1c2r,c1c2i,
        cr = juliaCoord[U], ci = juliaCoord[V],
        dist2, mindist2;

    a = EPoint[X]; a2 = Sqr(a);
    b = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    c1r = cr-1, c2r = cr-2;
    c1c2r = c1r*c2r-ci*ci;
    c1c2i = (c1r+c2r)*ci;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        tmp1r = a2*a-3*a*b2 + 3*(a*c1r-b*ci) + c1c2r;
        tmp1i = 3*a2*b-b2*b + 3*(a*ci+b*c1r) + c1c2i;
        tmp2r = 3*(a2-b2) + 3*(a*c2r-b*ci) + c1c2r + 1;
        tmp2i = 6*a*b + 3*(a*ci+b*c2r) + c1c2i;
        tmp = tmp2r*tmp2r + tmp2i*tmp2i;
        a = (tmp1r*tmp2r + tmp1i*tmp2i) / tmp;
        b = (tmp1i*tmp2r - tmp1r*tmp2i) / tmp;
        b2 = b*b;
        b = 2*a*b;
        a = a*a-b2;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        tmp1r = a-1;
        if(dist2 > 10000.0 || tmp1r*tmp1r+b2 < 1/10000.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   mandel_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   submitted by user, name lost (sorry)
*
* DESCRIPTION
*
*   The mandel pattern computes the standard Mandelbrot fractal pattern and
*   projects it onto the X-Y plane.  It uses the X and Y coordinates to compute
*   the Mandelbrot set.
*
* CHANGES
*
*   Oct 1994 : adapted from pigment by [CY]
*   May 2001 : updated with code from Warp [trf]
*
******************************************************************************/

DBL Mandel2Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, x, y, dist2, mindist2;

    a = x = EPoint[X]; a2 = Sqr(a);
    b = y = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b  = 2.0 * a * b + y;
        a  = a2 - b2 + x;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   mandel3_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Mandel3Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, x, y, dist2, mindist2;

    a = x = EPoint[X]; a2 = Sqr(a);
    b = y = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b = 3.0*a2*b - b2*b + y;
        a = a2*a - 3.0*a*b2 + x;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   mandel4_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL Mandel4Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf, a2, b2, x, y, dist2, mindist2;

    a = x = EPoint[X]; a2 = Sqr(a);
    b = y = EPoint[Y]; b2 = Sqr(b);
    mindist2 = a2+b2;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    for (col = 0; col < it_max; col++)
    {
        b = 4.0 * (a2*a*b - a*b2*b) + y;
        a = a2*a2 - 6.0*a2*b2 + b2*b2 + x;

        a2 = Sqr(a);
        b2 = Sqr(b);
        dist2 = a2+b2;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


/*****************************************************************************
*
* FUNCTION
*
*   mandelx_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*   is evaluated.
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0
*
* AUTHOR
*
*   Nieminen Juha
*
* DESCRIPTION
*
*   -
*
* CHANGES
*
*   -
*
******************************************************************************/

DBL MandelXPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection, const Ray *pRay, TraceThreadData *pThread) const
{
    int it_max, col;
    DBL a, b, cf=0, x, y, dist2, mindist2;
    int* binomial_coeff;

    a = x = EPoint[X];
    b = y = EPoint[Y];
    mindist2 = a*a+b*b;

    it_max = maxIterations;
    POV_PATTERN_ASSERT (it_max > 0);

    binomial_coeff = &(gaBinomialCoefficients[(fractalExponent+1)*fractalExponent/2]);

    for (col = 0; col < it_max; col++)
    {
        // Calculate (a+bi)^fractalExponent
        DBL new_a = pow(a, fractalExponent);
        for(int k=2; k<=fractalExponent; k+=2)
        {
            new_a += binomial_coeff[k]*pow(a, fractalExponent-k)*pow(b, k);
        }
        DBL new_b = 0;
        for(int l=1; l<=fractalExponent; l+=2)
        {
            new_b += binomial_coeff[l]*pow(a, fractalExponent-l)*pow(b, l);
        }

        a = new_a + x;
        b = new_b + y;

        dist2 = a*a+b*b;

        if(dist2 < mindist2) mindist2 = dist2;
        if(dist2 > 4.0)
        {
            cf = ExteriorColour(col, a, b);
            break;
        }
    }

    if(col == it_max)
        cf = InteriorColour(a, b, mindist2);

    return(cf);
}


// Oct 1994 : adapted from pigment by [CY]
//
DBL MarblePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);
    DBL turb_val;

    TurbulenceWarp Turb;
    Turb.Octaves = marbleOctaves;
    Turb.Omega   = marbleOmega;
    Turb.Lambda  = marbleLambda;
    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    turb_val = marbleFrequency * Turbulence(EPoint,&Turb,noise_generator);

    return (EPoint.x() + turb_val);
}


DBL NoisePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    return Noise(EPoint, GetNoiseGen(pThread));
}


double NoiseCubedPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    // Traditional noise_cubed (dents scalar value pattern) : type 0 ip_scale 1.0
    // Single call to Noise value. Value shaped then cubed  : type 1 ip_scale 1.0
    // Three calls to Noise. value1*value2*value3           : type 2 ip_scale 1.0
    // Three calls to Noise. Each value shaped then cubed   : type 3 ip_scale 1.0
    //
    // The ip_scale value can be used to scale the distribution wider (or narrower).
    // Only old noise type 1 bangs against he 0 and 1 limits. 2,3 more centered
    // around 0.5. Always values nearer 1.0 are being "selected" with the cubing.

    double scale = noiseCubedScale;
    auto lambdaScaleNoiseDistVal = [&scale] (double val) -> double
    {
        val = val - 0.5;
        val *= scale;
        val += 0.5;
        return (val);
    };

    auto lambdaModePowShapeVal = [] (double val) -> double
    {
        val = max(0.0+gkMinIsectDepthReturned,min(1.0-gkMinIsectDepthReturned,val));
        val = pow(val,0.95/val);
        val = pow(val,0.50/val);
        return (val);
    };

    int    noise_generator = GetNoiseGen(pThread);
    double value = Noise(EPoint,noise_generator);
    if (noiseCubedScale != 1.0)
        value = lambdaScaleNoiseDistVal(value);

    if (noiseCubedType <= 1)
    {
        if (noiseCubedType == 1)
            value = lambdaModePowShapeVal(value);

        value = value * value * value;
    }
    else
    {
        double x = EPoint.x(), y = EPoint.y(), z = EPoint.z();
        double value2 = Noise(Vector3d(x+y, y+z, z+x),noise_generator);
        double value3 = Noise(Vector3d(x-z, y-x, z-y),noise_generator);

        if (noiseCubedScale != 1.0)
        {
            value2 = lambdaScaleNoiseDistVal(value2);
            value3 = lambdaScaleNoiseDistVal(value3);
        }

        if (noiseCubedType == 3)
        {
            value  = lambdaModePowShapeVal(value);
            value2 = lambdaModePowShapeVal(value2);
            value3 = lambdaModePowShapeVal(value3);
        }
        value = value * value2 * value3;

    }

    if (noiseCubedType & 0x01)  // If mode is 1 or 3
       value = pow(value,0.125/value);

    return (value);
}


DBL ObjectPattern::Evaluate(const Vector3d& EPoint, const Intersection *pIsection,
                            const Ray *pRay, TraceThreadData *pThread) const
{
    if (pObject != nullptr)
    {
        if(Inside_BBox(EPoint, pObject->BBox) && Inside_Object(EPoint, pObject, pThread))
            return 1.0;
        else
            return 0.0;
    }

    return 0.0;
}


//-------- HardObjectPattern related code ----------

inline bool HardObjectPattern::Inside(const Vector3d& EPoint, const Vector3d& MoveBy, TraceThreadData *pThread) const
{
    return Inside_Object(EPoint+MoveBy, pObject, pThread);

    // Note: On the fly generated samples like below helpful to result , but
    // what speed, smoothness tradeoff makes sense?  Also leaning toward a
    // slightly perturbed sample over the opposite side of 'sphere' example
    // below if done.

    //return Inside_Object(EPoint+MoveBy, pObject, pThread) &&
    //       Inside_Object(EPoint-MoveBy, pObject, pThread);
}

//---- Notes on samples as header to fit paper ref in <120 chars.
// Note: Tried various methods and settled on E.B. Saff and A.B.J. Kuijlaars method.
// See:
// https://perswww.kuleuven.be/~u0017946/publications/Papers97/art97a-Saff-Kuijlaars-MI/Saff-Kuijlaars-MathIntel97.pdf
// Once the initial spherical points created, some tcl code was used to pick one
// then always the next chosen is the furthest from all already picked.
//
// Note: The 0,0,0 point is not always needed, but is useful to have when experimenting
// with alternative overall methods. Once sure it not really needed it should be pulled.

bool HardObjectPattern::SphericalSamplesInside(const Vector3d& EPoint, const DBL& TestRadius,
                                               const size_t& samples, TraceThreadData *pThread) const
{
    size_t n;

    static const Vector3d SampleVect[kHard_ObjectSamples_Max+1] = {
        Vector3d(0,0,0),
        Vector3d(1.22465e-16,0,-1), Vector3d(0,0,1),
        Vector3d(0.37235,-0.928092,0.001001), Vector3d(-0.936124,-0.350631,-0.027027),
        Vector3d(0.919791,0.391333,0.029029), Vector3d(-0.327382,0.943287,-0.0550551),
        Vector3d(-0.709217,0.398464,0.581582), Vector3d(-0.357071,-0.780259,0.513514),
        Vector3d(0.744444,-0.285423,-0.603604), Vector3d(-0.752215,0.347944,-0.55956),
        Vector3d(0.778961,-0.322915,0.537538), Vector3d(-0.307444,-0.786562,-0.535536),
        Vector3d(0.267919,0.782859,0.561562), Vector3d(0.407745,0.702263,-0.583584),
        Vector3d(-0.612571,-0.231495,0.755756), Vector3d(0.237273,-0.587346,0.773774),
        Vector3d(-0.641525,-0.270789,-0.717718), Vector3d(-0.20767,0.642352,-0.737738),
        Vector3d(0.547103,0.271558,0.791792), Vector3d(0.205977,-0.599039,-0.773774),
        Vector3d(-0.970789,0.228539,0.0730731), Vector3d(0.818664,-0.565657,-0.0990991),
        Vector3d(0.342344,0.937178,-0.0670671), Vector3d(-0.186568,-0.982401,-0.00900901),
        Vector3d(-0.290308,0.767481,0.571572), Vector3d(0.758351,0.269352,-0.593594),
        Vector3d(0.977044,-0.0755282,-0.199199), Vector3d(-0.701619,0.691189,0.173173),
        Vector3d(0.0159741,0.503472,0.863864), Vector3d(0.467541,0.0124303,-0.883884),
        Vector3d(0.631481,0.732172,0.255255), Vector3d(-0.610074,-0.786857,0.0930931),
        Vector3d(0.519879,-0.741915,-0.423423), Vector3d(-0.910074,-0.110662,0.399399),
        Vector3d(0.158136,-0.89107,0.425425), Vector3d(-0.714374,-0.60861,-0.345345),
        Vector3d(0.864565,0.220722,0.451451), Vector3d(-0.430824,0.267553,-0.861862),
        Vector3d(0.406914,-0.156848,0.8999), Vector3d(0.566754,-0.700661,0.433433),
        Vector3d(-0.17241,-0.400568,-0.8999), Vector3d(-0.59228,0.715044,-0.371371),
        Vector3d(0.16915,0.455777,-0.873874), Vector3d(-0.380014,0.308915,0.871872),
        Vector3d(0.0420132,0.922633,-0.383383), Vector3d(0.00371852,0.963615,0.267267),
        Vector3d(0.0970411,-0.896911,-0.431431), Vector3d(-0.18395,-0.523658,0.831832),
        Vector3d(-0.743853,-0.529813,0.407407), Vector3d(0.740768,0.604353,-0.293293),
        Vector3d(-0.933706,0.0220926,-0.357357), Vector3d(0.952475,-0.276822,0.127127),
        Vector3d(-0.836462,0.526189,-0.153153), Vector3d(0.641209,0.531442,0.553554),
        Vector3d(0.230795,-0.266003,-0.935936), Vector3d(-0.456119,-0.858812,-0.233233),
        Vector3d(-0.373788,-0.0542569,0.925926), Vector3d(-0.653438,0.104375,0.74975),
        Vector3d(0.700182,-0.0189257,0.713714), Vector3d(0.989366,0.0585693,0.133133),
        Vector3d(-0.687139,0.0637546,-0.723724), Vector3d(0.508066,0.420403,-0.751752),
        Vector3d(-0.394268,-0.154554,-0.905906), Vector3d(-0.882203,-0.313479,-0.351351),
        Vector3d(0.488165,-0.415003,-0.767768), Vector3d(-0.294485,0.877989,-0.377377),
        Vector3d(-0.434866,0.857388,0.275275), Vector3d(0.903641,0.306354,-0.299299),
        Vector3d(0.650638,-0.750302,0.117117), Vector3d(0.00801046,0.998102,-0.0610611),
        Vector3d(0.3383,0.904044,0.261261), Vector3d(-0.443776,-0.544545,-0.711712),
        Vector3d(0.258251,0.186282,0.947948), Vector3d(-0.158613,-0.934961,0.317317),
        Vector3d(-0.890521,0.228502,0.393393), Vector3d(0.0505397,-0.348525,0.935936),
        Vector3d(0.553427,-0.442378,0.705706), Vector3d(0.313122,0.515241,0.797798),
        Vector3d(0.879786,-0.359251,-0.311311), Vector3d(-0.463718,-0.561071,0.685686),
        Vector3d(-0.0763678,-0.789015,0.60961), Vector3d(-0.104246,0.274374,-0.955956),
        Vector3d(0.636816,0.767546,-0.0730731), Vector3d(-0.480948,0.584317,-0.653654),
        Vector3d(-0.590932,0.645789,0.483483), Vector3d(0.0500487,0.746355,-0.663664),
        Vector3d(0.0040655,0.756783,0.653654), Vector3d(0.830851,-0.506181,0.231231),
        Vector3d(0.201209,0.184781,-0.961962), Vector3d(0.729397,-0.0235848,-0.683684),
        Vector3d(0.483962,0.826587,-0.287287), Vector3d(-0.999315,-0.0167009,-0.033033),
        Vector3d(-0.0285202,-0.709921,-0.703704), Vector3d(-0.356286,0.570829,0.73974),
        Vector3d(-0.837296,0.466421,0.285285), Vector3d(-0.342314,-0.318701,0.883884),
        Vector3d(0.140205,-0.984312,0.107107), Vector3d(0.396888,-0.857518,0.327327),
        Vector3d(0.793677,0.49376,0.355355), Vector3d(-0.915373,0.309695,-0.257257),
        Vector3d(0.887774,-0.115912,0.445445), Vector3d(-0.862378,-0.467121,0.195195),
        Vector3d(-0.624948,0.779124,-0.049049), Vector3d(-0.130914,-0.933676,-0.333333),
        Vector3d(-0.828729,-0.0948664,-0.551552), Vector3d(-0.15532,0.249046,0.955956),
        Vector3d(0.370209,-0.697441,-0.613614), Vector3d(-0.824314,-0.550256,-0.133133),
        Vector3d(0.317832,-0.890588,-0.325325), Vector3d(0.912566,0.0352607,-0.407407),
        Vector3d(-0.726417,-0.417969,-0.545546), Vector3d(0.580149,-0.807701,-0.105105),
        Vector3d(0.0498978,-0.992098,-0.115115), Vector3d(-0.478716,-0.820924,0.311311),
        Vector3d(-0.210826,0.96381,0.163163), Vector3d(0.977861,0.191178,-0.0850851),
        Vector3d(-0.714442,-0.367204,0.595596), Vector3d(0.185562,-0.144351,0.971972),
        Vector3d(-0.0812985,0.877048,0.473473), Vector3d(0.797407,0.586219,0.143143),
        Vector3d(0.717815,0.279572,0.637638), Vector3d(0.363731,-0.697315,0.617618),
        Vector3d(0.726188,0.474496,-0.497497), Vector3d(-0.806678,-0.0406761,0.58959),
        Vector3d(-0.947464,-0.256434,0.191191), Vector3d(-0.853196,-0.33064,0.403403),
        Vector3d(-0.415601,-0.904349,0.0970971), Vector3d(0.722568,-0.535298,0.437437),
        Vector3d(0.695612,-0.644541,-0.317317), Vector3d(0.471933,0.682941,0.557558),
        Vector3d(-0.677786,-0.723831,-0.129129), Vector3d(-0.749262,0.551043,-0.367367),
        Vector3d(0.265511,0.882856,-0.387387), Vector3d(0.0160771,-0.638119,0.76977),
        Vector3d(-0.58981,-0.599047,-0.541542), Vector3d(0.554947,-0.566047,-0.60961),
        Vector3d(-0.204836,0.452601,0.867868), Vector3d(0.280464,-0.344544,0.895896),
        Vector3d(-0.820965,0.131811,-0.555556), Vector3d(-0.26768,0.422639,-0.865866),
        Vector3d(-0.0531084,0.490414,-0.86987), Vector3d(0.0637693,0.299535,0.951952),
        Vector3d(-0.894141,0.442426,0.0690691), Vector3d(0.415278,0.230988,-0.87988),
        Vector3d(0.0480704,-0.441656,-0.895896), Vector3d(0.661392,-0.242636,0.70971),
        Vector3d(-0.0828666,-0.210969,-0.973974), Vector3d(-0.266338,-0.673344,0.68969),
        Vector3d(-0.510712,0.057035,-0.857858), Vector3d(-0.789032,0.185792,0.585586),
        Vector3d(-0.981809,-0.0318244,0.187187), Vector3d(-0.568547,0.379801,-0.72973),
        Vector3d(0.339122,0.57085,-0.747748), Vector3d(0.925968,-0.365432,-0.0950951),
        Vector3d(0.150349,-0.77516,0.613614), Vector3d(0.934072,0.259545,0.245245),
        Vector3d(0.606098,-0.22206,-0.763764), Vector3d(0.793771,0.603313,-0.0770771),
        Vector3d(0.616179,0.221712,-0.755756), Vector3d(-0.2503,0.781445,-0.571572),
        Vector3d(-0.249585,-0.660952,-0.707708), Vector3d(-0.584069,0.32051,0.745746),
        Vector3d(-0.550591,-0.661248,0.50951), Vector3d(-0.274897,0.134979,-0.951952),
        Vector3d(0.257396,-0.0256328,-0.965966), Vector3d(-0.0876054,-0.842496,-0.531532),
        Vector3d(-0.38166,-0.363443,-0.84985), Vector3d(-0.123112,-0.227493,0.965966),
        Vector3d(-0.772835,-0.628323,0.0890891), Vector3d(0.155751,-0.770902,-0.617618),
        Vector3d(0.422322,0.0678392,0.903904), Vector3d(-0.660698,-0.684865,0.307307),
        Vector3d(0.58006,0.64807,-0.493493), Vector3d(0.935443,0.0655394,0.347347),
        Vector3d(0.19059,0.656635,0.72973), Vector3d(0.89175,-0.188473,-0.411411),
        Vector3d(0.4123,-0.204117,-0.887888), Vector3d(-0.147917,0.660919,0.735736),
        Vector3d(-0.958367,-0.13931,-0.249249), Vector3d(0.704191,-0.48822,-0.515516),
        Vector3d(0.467844,0.805661,0.363363), Vector3d(-0.471686,0.101754,0.875876),
        Vector3d(0.485544,0.490378,0.723724), Vector3d(0.900303,-0.270074,0.341341),
        Vector3d(0.861516,0.471167,-0.189189), Vector3d(-0.205637,-0.0356713,-0.977978),
        Vector3d(0.147044,0.917575,0.369369), Vector3d(-0.492439,0.827646,-0.269269),
        Vector3d(-0.162606,0.0730391,0.983984), Vector3d(-0.627434,0.537329,-0.563564),
        Vector3d(0.198772,0.790302,-0.57958), Vector3d(0.141756,0.974637,-0.173173),
        Vector3d(-0.510987,0.546299,0.663664), Vector3d(-0.0268653,-0.976636,0.213213),
        Vector3d(0.967967,-0.0761872,0.239239), Vector3d(0.142518,0.988674,0.047047),
        Vector3d(-0.514002,0.855753,0.0590591), Vector3d(-0.197892,0.965864,-0.167167),
        Vector3d(0.465849,0.883912,0.041041), Vector3d(-0.548312,-0.76344,-0.341341),
        Vector3d(0.849324,0.158534,-0.503504), Vector3d(-0.301993,0.875435,0.377377),
        Vector3d(0.839333,0.00901721,0.543544), Vector3d(0.206171,-0.953181,-0.221221),
        Vector3d(0.339607,-0.43133,-0.835836), Vector3d(0.233073,0.343134,0.90991),
        Vector3d(0.742552,-0.669728,0.00900901), Vector3d(-0.984361,0.0998265,-0.145145),
        Vector3d(-0.0905524,-0.528893,-0.843844), Vector3d(-0.391766,-0.672758,-0.627628),
        Vector3d(0.986305,-0.163831,0.019019), Vector3d(0.514395,-0.827991,0.223223),
        Vector3d(-0.492723,-0.424251,0.75976), Vector3d(-0.475077,0.741704,-0.473473),
        Vector3d(-0.572525,-0.010539,0.81982), Vector3d(0.425973,-0.326304,0.843844),
        Vector3d(0.309118,-0.925413,0.219219), Vector3d(-0.244823,-0.94208,-0.229229),
        Vector3d(-0.158371,0.863144,-0.479479), Vector3d(0.413591,0.310569,0.855856),
        Vector3d(-0.74909,0.537397,0.387387), Vector3d(-0.0609468,0.958855,-0.277277),
        Vector3d(-0.505343,0.774065,0.381381), Vector3d(0.859118,-0.0659711,-0.507508),
        Vector3d(0.673333,-0.659933,0.333333), Vector3d(0.551434,0.820777,0.149149),
        Vector3d(-0.606605,-0.0438359,-0.793794), Vector3d(-0.575453,0.179925,-0.797798),
        Vector3d(-0.784983,0.617878,-0.045045), Vector3d(-0.733088,0.213831,-0.645646),
        Vector3d(0.421345,-0.880503,-0.217217), Vector3d(-0.349958,-0.873918,-0.337337),
        Vector3d(0.270386,0.693622,-0.667668), Vector3d(-0.693726,0.702884,-0.157157),
        Vector3d(0.240552,0.958155,0.155155), Vector3d(0.87453,-0.484798,0.013013),
        Vector3d(-0.504805,-0.863103,-0.015015), Vector3d(0.634933,-0.348122,-0.68969),
        Vector3d(-0.347995,0.170176,0.921922), Vector3d(-0.960961,-0.239152,-0.139139),
        Vector3d(0.758957,-0.165985,0.62963), Vector3d(0.0684079,-0.944496,0.321321),
        Vector3d(0.426346,-0.461835,0.777778), Vector3d(0.335275,0.68421,0.647648),
        Vector3d(0.464634,0.577038,-0.671672), Vector3d(-0.695046,0.261636,0.66967),
        Vector3d(-0.413887,0.895586,-0.163163), Vector3d(-0.182603,-0.760099,-0.623624),
        Vector3d(0.735333,0.6768,0.035035), Vector3d(0.545949,-0.561718,0.621622),
        Vector3d(-0.94692,0.319089,-0.039039), Vector3d(-0.580639,-0.550762,0.5996),
        Vector3d(-0.0845936,0.995107,0.0510511), Vector3d(0.129438,0.655813,-0.743744),
        Vector3d(-0.0358627,-0.454764,0.88989), Vector3d(0.308289,-0.792957,0.525526),
        Vector3d(-0.852784,-0.462159,-0.243243), Vector3d(-0.945682,0.144341,0.291291),
        Vector3d(-0.251856,-0.944881,0.209209), Vector3d(0.68158,-0.379528,0.625626),
        Vector3d(-0.240897,-0.866382,-0.437437), Vector3d(-0.0040782,0.17821,-0.983984),
        Vector3d(0.0325502,-0.137333,0.98999), Vector3d(-0.449369,-0.776658,-0.441441),
        Vector3d(-0.291673,-0.860633,0.417417), Vector3d(-0.0686903,-0.90426,0.421421),
        Vector3d(-0.846045,-0.186303,0.499499), Vector3d(-0.313903,-0.56403,0.763764),
        Vector3d(0.210654,0.319358,-0.923924), Vector3d(-0.264618,-0.187519,-0.945946),
        Vector3d(-0.406115,0.777927,0.479479), Vector3d(0.329021,-0.0267444,0.943944),
        Vector3d(0.527001,-0.00503356,0.84985), Vector3d(0.250125,-0.813179,-0.525526),
        Vector3d(-0.488234,0.656003,0.575576), Vector3d(0.126371,-0.06285,-0.98999),
        Vector3d(-0.867696,0.0398549,0.495495), Vector3d(0.363817,0.806833,0.465465),
        Vector3d(-0.623299,0.730412,0.279279), Vector3d(-0.935019,0.304815,0.181181),
        Vector3d(-0.394053,-0.910804,-0.123123), Vector3d(0.803786,-0.582458,0.121121),
        Vector3d(-0.0263803,0.817323,-0.575576), Vector3d(-0.766936,-0.0102491,-0.641642),
        Vector3d(0.352701,0.454726,-0.817818), Vector3d(-0.537438,0.213554,0.815816),
        Vector3d(0.403447,-0.594338,-0.695696), Vector3d(-0.30552,0.267244,-0.913914),
        Vector3d(-0.400641,-0.689311,0.603604), Vector3d(0.839311,0.366652,-0.401401),
        Vector3d(0.128658,0.164335,0.977978), Vector3d(-0.38269,0.455464,-0.803804),
        Vector3d(0.0986995,0.589383,0.801802), Vector3d(-0.990696,-0.110762,0.0790791),
        Vector3d(-0.0972912,0.3894,0.915916), Vector3d(-0.711222,-0.186857,0.677678),
        Vector3d(0.0420013,0.582407,-0.811812), Vector3d(-0.616875,-0.393415,0.681682),
        Vector3d(0.387794,-0.597688,0.701702), Vector3d(0.560996,0.807748,-0.181181),
        Vector3d(-0.498084,-0.362357,-0.787788), Vector3d(0.800443,0.0465604,-0.597598),
        Vector3d(-0.0199788,-0.974102,-0.225225), Vector3d(0.463342,-0.878931,0.113113),
        Vector3d(0.614537,0.407193,-0.675676), Vector3d(-0.738018,0.0386454,0.673674),
        Vector3d(-0.325547,0.677398,0.65966), Vector3d(-0.337468,-0.521343,-0.783784),
        Vector3d(0.55848,-0.27734,0.781782), Vector3d(0.815797,-0.402374,-0.415415),
        Vector3d(-0.83018,0.309843,-0.463463), Vector3d(0.257403,-0.477925,0.83984),
        Vector3d(0.558335,0.689431,0.461461), Vector3d(0.274054,-0.955274,-0.111111),
        Vector3d(0.584206,0.559861,-0.587588), Vector3d(-0.733794,-0.234445,-0.637638),
        Vector3d(0.932602,-0.298263,-0.203203), Vector3d(-0.636814,-0.439291,-0.633634),
        Vector3d(0.565806,0.724641,-0.393393), Vector3d(0.0880466,-0.848683,0.521522),
        Vector3d(-0.818168,-0.356079,-0.451451), Vector3d(-0.0869113,-0.990866,0.103103),
        Vector3d(0.549592,-0.654254,-0.51952), Vector3d(0.838024,-0.504758,-0.207207),
        Vector3d(-0.789481,0.370297,0.489489), Vector3d(0.941161,-0.140636,-0.307307),
        Vector3d(-0.139095,-0.844291,0.517518), Vector3d(-0.284636,0.697473,-0.657658),
        Vector3d(0.34495,0.141277,-0.927928), Vector3d(0.771391,0.0588583,0.633634),
        Vector3d(0.945994,0.260332,-0.193193), Vector3d(-0.934597,-0.195302,0.297297),
        Vector3d(0.558932,-0.047923,-0.827828), Vector3d(0.613892,-0.76005,-0.213213),
        Vector3d(-0.899254,0.246495,-0.361361), Vector3d(0.649895,0.669699,0.359359),
        Vector3d(-0.805224,-0.509531,0.303303), Vector3d(-0.883697,0.0893076,-0.459459),
        Vector3d(0.71664,0.526459,0.457457), Vector3d(-0.000434588,0.392106,-0.91992),
        Vector3d(0.613523,0.0545902,0.787788), Vector3d(0.66586,0.203745,0.717718),
        Vector3d(-0.56059,-0.802774,0.203203), Vector3d(0.181689,-0.69297,0.697698),
        Vector3d(0.134866,-0.525814,-0.83984), Vector3d(-0.458663,-0.328086,0.825826),
        Vector3d(0.393992,0.777927,-0.489489), Vector3d(-0.418805,0.406896,0.811812),
        Vector3d(0.847052,0.469443,0.249249), Vector3d(0.132888,-0.203713,-0.96997),
        Vector3d(0.730842,0.656945,-0.185185), Vector3d(0.949193,0.0838985,-0.303303),
        Vector3d(-0.73108,-0.652566,0.199199), Vector3d(0.995484,-0.0327951,-0.0890891),
        Vector3d(0.705759,0.19985,-0.67968), Vector3d(-0.45863,-0.124378,0.87988),
        Vector3d(-0.895419,-0.437019,0.0850851), Vector3d(-0.496854,-0.763037,0.413413),
        Vector3d(0.890744,0.288318,0.351351), Vector3d(-0.750231,0.606503,-0.263263),
        Vector3d(0.0444188,0.823506,0.565566), Vector3d(-0.0896873,-0.323576,-0.941942),
        Vector3d(-0.703623,-0.552001,-0.447447), Vector3d(-0.840777,0.511568,0.177177),
        Vector3d(0.114447,-0.0534211,0.991992), Vector3d(-0.879561,-0.137596,-0.455455),
        Vector3d(0.236745,0.452319,0.85986), Vector3d(-0.161212,-0.0429239,0.985986),
        Vector3d(0.763228,0.339821,0.54955), Vector3d(0.353595,-0.0804568,-0.931932),
        Vector3d(0.180709,0.855364,-0.485485), Vector3d(0.689237,-0.590796,-0.419419),
        Vector3d(0.505919,-0.680914,0.52953), Vector3d(0.203826,0.0769788,0.975976),
        Vector3d(0.196737,-0.686815,-0.6997), Vector3d(0.0415651,-0.547404,0.835836),
        Vector3d(-0.273983,0.058396,0.95996), Vector3d(0.260593,-0.369621,-0.891892),
        Vector3d(-0.0106965,-0.250846,0.967968), Vector3d(-0.409047,0.0688781,-0.90991),
        Vector3d(0.246233,-0.236394,0.93994), Vector3d(-0.21423,-0.15773,0.963964),
        Vector3d(-0.234112,0.165856,0.957958), Vector3d(0.165728,0.945206,-0.281281),
        Vector3d(0.0183769,0.198246,0.97998), Vector3d(-0.410284,0.541573,-0.733734),
        Vector3d(0.157763,-0.308842,0.937938), Vector3d(-0.0637302,-0.351718,0.933934),
        Vector3d(-0.335723,-0.161987,0.927928), Vector3d(0.128805,0.389648,0.911912),
        Vector3d(-0.0438893,-0.718932,0.693694), Vector3d(0.385263,0.175804,0.905906),
        Vector3d(-0.252561,-0.389126,0.885886), Vector3d(-0.300792,0.390961,0.86987),
        Vector3d(0.333087,0.391322,0.857858), Vector3d(0.516808,-0.118589,0.847848),
        Vector3d(0.153132,-0.524002,0.837838), Vector3d(-0.0724272,-0.547243,0.833834),
        Vector3d(-0.381128,-0.411634,0.827828), Vector3d(-0.5178,-0.230647,0.823824),
        Vector3d(-0.487652,0.316074,0.813814), Vector3d(-0.33362,0.482603,0.80981),
        Vector3d(0.209449,0.56254,0.7998), Vector3d(0.597256,-0.170201,0.783784),
        Vector3d(0.128959,-0.622686,0.771772), Vector3d(-0.645642,-0.122479,0.753754),
        Vector3d(-0.522809,0.416552,0.743744), Vector3d(-0.255917,0.6247,0.737738),
        Vector3d(-0.0353734,0.678516,0.733734), Vector3d(0.0785267,0.677054,0.731732),
        Vector3d(0.396923,0.561938,0.725726), Vector3d(0.561179,0.40521,0.721722),
        Vector3d(-0.650187,0.597374,-0.469469), Vector3d(0.0699324,-0.714924,0.695696),
        Vector3d(-0.156918,-0.704939,0.691692), Vector3d(-0.672094,-0.293811,0.67968),
        Vector3d(-0.733359,-0.0751439,0.675676), Vector3d(-0.220115,0.720441,0.657658),
        Vector3d(0.117699,0.749331,0.651652), Vector3d(-0.109399,0.747093,0.655656),
        Vector3d(0.66726,0.381608,0.63964), Vector3d(0.752768,0.171197,0.635636),
        Vector3d(-0.642326,-0.728791,-0.237237), Vector3d(0.0374904,-0.790269,0.611612),
        Vector3d(-0.654076,-0.463747,0.597598), Vector3d(-0.760545,-0.263094,0.593594),
        Vector3d(-0.805845,0.0731805,0.587588), Vector3d(-0.647838,0.494361,0.57958),
        Vector3d(-0.181688,0.801611,0.56957), Vector3d(0.157574,0.810905,0.563564),
        Vector3d(0.561771,0.613002,0.555556), Vector3d(0.803576,0.23336,0.547548),
        Vector3d(0.829094,0.122406,0.545546), Vector3d(0.814142,-0.214638,0.53954),
        Vector3d(0.200034,-0.828197,0.523524), Vector3d(-0.0256712,-0.854073,0.51952),
        Vector3d(-0.250215,-0.819534,0.515516), Vector3d(-0.829939,0.263888,0.491491),
        Vector3d(-0.856167,0.15311,0.493493), Vector3d(-0.735501,0.470525,0.487487),
        Vector3d(-0.668936,0.562875,0.485485), Vector3d(0.778889,0.431152,0.455455),
        Vector3d(0.642722,0.61303,0.459459), Vector3d(0.828462,0.328679,0.453453),
        Vector3d(0.78358,-0.439198,0.439439), Vector3d(0.474467,-0.767299,0.431431),
        Vector3d(0.268515,-0.863253,0.427427), Vector3d(-0.181552,-0.889453,0.419419),
        Vector3d(-0.804739,-0.433638,0.405405), Vector3d(-0.671473,-0.617663,0.409409),
        Vector3d(-0.917642,0.00291265,0.397397), Vector3d(-0.67781,0.62614,0.385385),
        Vector3d(-0.856276,0.337053,0.391391), Vector3d(-0.192535,0.906655,0.375375),
        Vector3d(0.0336883,0.927873,0.371371), Vector3d(0.258289,0.893492,0.367367),
        Vector3d(0.848473,0.393998,0.353353), Vector3d(0.727158,0.58612,0.357357),
        Vector3d(0.919874,0.178289,0.349349), Vector3d(0.925635,-0.15911,0.343343),
        Vector3d(0.861741,-0.377161,0.339339), Vector3d(-0.376844,-0.871679,0.313313),
        Vector3d(-0.573787,-0.758351,0.309309), Vector3d(-0.738211,-0.601526,0.305305),
        Vector3d(-0.860786,-0.410201,0.301301), Vector3d(-0.904113,-0.304957,0.299299),
        Vector3d(-0.951812,-0.0827982,0.295295), Vector3d(-0.955521,0.0309544,0.293293),
        Vector3d(-0.109886,0.956775,0.269269), Vector3d(0.117336,0.957009,0.265265),
        Vector3d(0.713506,0.653278,0.253253), Vector3d(0.442571,0.858438,0.259259),
        Vector3d(0.958546,0.1484,0.243243), Vector3d(0.896739,0.367054,0.247247),
        Vector3d(0.766447,-0.60001,0.229229), Vector3d(0.607126,-0.762018,0.225225),
        Vector3d(0.199375,-0.955545,0.217217), Vector3d(-0.463438,-0.862042,0.205205),
        Vector3d(-0.360061,-0.909627,0.207207), Vector3d(-0.650207,-0.732632,0.201201),
        Vector3d(-0.802121,-0.563662,0.197197), Vector3d(-0.911043,-0.36425,0.193193),
        Vector3d(-0.971157,-0.145126,0.189189), Vector3d(-0.979281,0.0819484,0.185185),
        Vector3d(-0.963613,0.194665,0.183183), Vector3d(-0.776399,0.605408,0.175175),
        Vector3d(-0.893889,0.410923,0.179179), Vector3d(-0.617441,0.767768,0.171171),
        Vector3d(-0.320318,0.932801,0.165165), Vector3d(0.914741,-0.384178,0.125125),
        Vector3d(0.251715,-0.961631,0.109109), Vector3d(0.560681,-0.819991,0.115115),
        Vector3d(0.0268348,-0.994099,0.105105), Vector3d(-0.309586,-0.945693,0.0990991),
        Vector3d(-0.8396,-0.536178,0.0870871), Vector3d(-0.695992,-0.712248,0.0910911),
        Vector3d(-0.97146,-0.222914,0.0810811), Vector3d(-0.990356,0.116445,0.0750751),
        Vector3d(-0.838063,0.541436,0.0670671), Vector3d(-0.413332,0.908791,0.0570571),
        Vector3d(0.869304,0.493303,0.031031), Vector3d(0.958359,0.284285,0.027027),
        Vector3d(0.545341,0.835196,-0.0710711), Vector3d(0.233465,0.970186,-0.0650651),
        Vector3d(0.961797,-0.257451,-0.0930931), Vector3d(0.483987,-0.868496,-0.107107),
        Vector3d(0.668735,-0.736317,-0.103103), Vector3d(0.163032,-0.980115,-0.113113),
        Vector3d(-0.063864,-0.991063,-0.117117), Vector3d(-0.287308,-0.950149,-0.121121),
        Vector3d(-0.590587,-0.796898,-0.127127), Vector3d(-0.75604,-0.641256,-0.131131),
        Vector3d(-0.966103,0.212114,-0.147147), Vector3d(-0.891664,0.426718,-0.151151),
        Vector3d(0.464157,0.867441,-0.179179), Vector3d(0.801568,0.567846,-0.187187),
        Vector3d(0.65029,0.73727,-0.183183), Vector3d(0.909875,0.368202,-0.191191),
        Vector3d(0.969379,0.149006,-0.195195), Vector3d(0.773414,-0.598383,-0.209209),
        Vector3d(0.961303,-0.188189,-0.201201), Vector3d(0.891325,-0.404265,-0.205205),
        Vector3d(0.521133,-0.825895,-0.215215), Vector3d(0.0937074,-0.970253,-0.223223),
        Vector3d(0.361073,0.915551,-0.177177), Vector3d(-0.900567,-0.358934,-0.245245),
        Vector3d(-0.967574,-0.0259356,-0.251251), Vector3d(-0.586896,0.764278,-0.267267),
        Vector3d(0.382841,0.878661,-0.285285), Vector3d(0.664203,0.688464,-0.291291),
        Vector3d(0.806797,0.511742,-0.295295), Vector3d(0.861349,0.411938,-0.297297),
        Vector3d(0.933065,0.196487,-0.301301), Vector3d(0.917024,-0.251782,-0.309309),
        Vector3d(0.829972,-0.461498,-0.313313), Vector3d(0.768288,-0.557054,-0.315315),
        Vector3d(-0.782155,-0.517285,-0.347347), Vector3d(0.563028,0.743248,0.361361),
        Vector3d(-0.912978,-0.203993,-0.353353), Vector3d(-0.838377,-0.418425,-0.349349),
        Vector3d(-0.930234,-0.0915822,-0.355355), Vector3d(-0.923332,0.135346,-0.359359),
        Vector3d(-0.811576,0.455909,-0.365365), Vector3d(-0.86182,0.353883,-0.363363),
        Vector3d(-0.400185,0.83603,-0.375375), Vector3d(0.788188,0.468231,-0.399399),
        Vector3d(0.877448,0.259521,-0.403403), Vector3d(0.860468,-0.2978,-0.413413),
        Vector3d(0.609306,-0.671677,-0.421421), Vector3d(0.422356,-0.800393,-0.425425),
        Vector3d(0.209255,-0.878523,-0.429429), Vector3d(-0.0165941,-0.901033,-0.433433),
        Vector3d(-0.543431,-0.71277,-0.443443), Vector3d(-0.85585,-0.2488,-0.453453),
        Vector3d(-0.8889,-0.0242762,-0.457457), Vector3d(-0.864022,0.201295,-0.461461),
        Vector3d(0.0683409,0.872682,-0.483483), Vector3d(0.289878,0.823606,-0.487487),
        Vector3d(0.491275,0.719086,-0.491491), Vector3d(0.658818,0.566077,-0.495495),
        Vector3d(0.780998,0.37489,-0.499499), Vector3d(0.63242,-0.576385,-0.517518),
        Vector3d(0.457157,-0.720432,-0.521522), Vector3d(0.356742,-0.77373,-0.523524),
        Vector3d(0.0259458,-0.847895,-0.52953), Vector3d(-0.809439,-0.206889,-0.54955),
        Vector3d(-0.696347,0.446934,-0.561562), Vector3d(-0.546762,0.617403,-0.565566),
        Vector3d(-0.356393,0.740658,-0.56957), Vector3d(0.0871396,0.811672,-0.577578),
        Vector3d(0.655758,0.471556,-0.58959), Vector3d(0.714201,0.37408,-0.591592),
        Vector3d(0.77891,-0.177127,-0.601602), Vector3d(0.631334,-0.481903,-0.607608),
        Vector3d(0.46731,-0.638398,-0.611612), Vector3d(0.265636,-0.741927,-0.615616),
        Vector3d(-0.566882,-0.528854,-0.631632), Vector3d(-0.692866,-0.340447,-0.635636),
        Vector3d(-0.758677,-0.123575,-0.63964), Vector3d(-0.387062,0.648304,-0.655656),
        Vector3d(-0.17598,0.730671,-0.65966), Vector3d(-0.832603,0.0187385,-0.553554),
        Vector3d(-0.0635609,0.747104,-0.661662), Vector3d(0.668184,0.307056,-0.677678),
        Vector3d(-0.864309,-0.0739358,0.497497), Vector3d(0.30377,-0.6488,-0.697698),
        Vector3d(0.0850316,-0.707379,-0.701702), Vector3d(-0.590817,-0.372406,-0.715716),
        Vector3d(-0.675033,-0.162279,-0.71972), Vector3d(0.139199,-0.838056,-0.527528),
        Vector3d(-0.625427,0.281519,-0.727728), Vector3d(-0.496179,0.467306,-0.731732),
        Vector3d(-0.313217,0.60049,-0.735736), Vector3d(-0.0966002,0.665923,-0.73974),
        Vector3d(0.0168496,0.670474,-0.741742), Vector3d(-0.502825,0.717874,0.481481),
        Vector3d(0.952891,-0.188991,0.237237), Vector3d(-0.423261,0.618913,0.661662),
        Vector3d(0.570741,0.325745,-0.753754), Vector3d(0.642958,0.111394,-0.757758),
        Vector3d(0.637909,-0.113096,-0.761762), Vector3d(0.310546,-0.554914,-0.771772),
        Vector3d(-0.425104,-0.449252,-0.785786), Vector3d(-0.553873,-0.263547,-0.78979),
        Vector3d(-0.529111,0.283482,-0.7998), Vector3d(-0.464104,0.376459,-0.801802),
        Vector3d(0.708775,0.439806,0.551552), Vector3d(0.153298,0.560542,-0.813814),
        Vector3d(0.258218,0.517463,-0.815816), Vector3d(0.253125,0.951442,-0.175175),
        Vector3d(0.25986,-0.743969,0.615616), Vector3d(0.53496,-0.158745,-0.82983),
        Vector3d(-0.694044,0.717165,0.0630631), Vector3d(0.242169,-0.489267,-0.837838),
        Vector3d(0.3128,0.0856874,0.945946), Vector3d(0.0223162,-0.539263,-0.841842),
        Vector3d(-0.722765,-0.648364,-0.239239), Vector3d(-0.447915,-0.271515,-0.851852),
        Vector3d(-0.35803,0.354336,-0.863864), Vector3d(-0.164281,0.468847,-0.867868),
        Vector3d(0.724878,0.562696,-0.397397), Vector3d(0.060043,0.486039,-0.871872),
        Vector3d(0.351914,0.324788,-0.877878), Vector3d(0.698366,-0.683867,-0.211211),
        Vector3d(0.454635,0.124866,-0.881882), Vector3d(-0.0852669,0.981892,-0.169169),
        Vector3d(-0.911087,0.11655,0.395395), Vector3d(0.346793,-0.296362,-0.88989),
        Vector3d(-0.26963,-0.909877,0.315315), Vector3d(0.15891,-0.419167,-0.893894),
        Vector3d(-0.141023,-0.694328,-0.705706), Vector3d(-0.238167,-0.576276,-0.781782),
        Vector3d(0.430075,0.502902,-0.74975), Vector3d(-0.26738,-0.339236,-0.901902),
        Vector3d(-0.416871,-0.0438351,-0.907908), Vector3d(-0.0973496,0.490713,0.865866),
        Vector3d(-0.775116,-0.315256,-0.547548), Vector3d(-0.112708,0.380426,-0.917918),
        Vector3d(0.110484,0.371286,-0.921922), Vector3d(-0.793242,-0.559079,-0.241241),
        Vector3d(0.539017,0.175429,-0.823824), Vector3d(-0.888485,-0.222422,0.401401),
        Vector3d(0.886625,0.109046,0.449449), Vector3d(-0.636041,-0.691063,-0.343343),
        Vector3d(0.0284545,0.98483,-0.171171), Vector3d(-0.628063,0.215431,0.747748),
        Vector3d(0.758425,-0.500554,-0.417417), Vector3d(-0.524992,0.834126,0.169169),
        Vector3d(0.924722,-0.299256,0.235235), Vector3d(0.546136,0.497895,-0.673674),
        Vector3d(-0.524301,-0.464458,-0.713714), Vector3d(0.307013,-0.183058,-0.933934),
        Vector3d(-0.397324,-0.818269,0.415415), Vector3d(-0.782713,0.413162,-0.465465),
        Vector3d(-0.287246,0.266903,0.91992), Vector3d(-0.173904,0.945506,-0.275275),
        Vector3d(0.0215337,-0.34066,-0.93994), Vector3d(0.555682,-0.32376,-0.765766),
        Vector3d(-0.482527,0.16676,-0.85986), Vector3d(0.464856,0.754394,0.463463),
        Vector3d(0.129258,0.490394,0.861862), Vector3d(0.612982,-0.722695,-0.319319),
        Vector3d(-0.725131,0.151798,0.671672), Vector3d(-0.425508,0.889381,0.167167),
        Vector3d(-0.140287,-0.967321,0.211211), Vector3d(0.689889,-0.13236,0.711712),
        Vector3d(0.615794,-0.0577767,0.785786), Vector3d(-0.307269,-0.0835489,-0.947948),
        Vector3d(-0.556233,-0.123344,0.821822), Vector3d(-0.881707,-0.45203,-0.135135),
        Vector3d(-0.188952,-0.270678,-0.943944), Vector3d(0.650366,0.648603,-0.395395),
        Vector3d(-0.564192,0.506989,-0.651652), Vector3d(-0.284493,0.918905,-0.273273),
        Vector3d(-0.31108,0.0287086,-0.94995), Vector3d(-0.922443,0.255756,0.289289),
        Vector3d(-0.217916,0.974298,-0.0570571), Vector3d(-0.49283,-0.167487,-0.853854),
        Vector3d(0.692182,0.0929266,0.715716), Vector3d(0.0952121,-0.623784,-0.775776),
        Vector3d(0.666709,-0.520423,0.533534), Vector3d(-0.202836,0.220973,-0.953954),
        Vector3d(-0.307279,0.950025,0.0550551), Vector3d(-0.105632,0.99265,-0.0590591),
        Vector3d(-0.0453251,0.875284,-0.481481), Vector3d(0.0291457,0.998371,0.049049),
        Vector3d(-0.532228,0.84506,-0.0510511), Vector3d(0.49329,-0.524837,-0.693694),
        Vector3d(0.822287,0.268962,-0.501502), Vector3d(0.121535,0.990582,-0.0630631),
        Vector3d(-0.60798,0.7916,0.0610611), Vector3d(-0.553035,-0.79926,-0.235235),
        Vector3d(0.362295,0.931069,0.043043), Vector3d(-0.771073,0.633414,0.0650651),
        Vector3d(0.00710517,0.286821,-0.957958), Vector3d(0.185643,-0.408032,0.893894),
        Vector3d(0.446756,0.891986,-0.0690691), Vector3d(-0.221889,0.936577,0.271271),
        Vector3d(0.228934,0.724945,0.64965), Vector3d(0.883853,-0.405471,0.233233),
        Vector3d(-0.30792,0.936966,-0.165165), Vector3d(0.560277,0.0640442,-0.825826),
        Vector3d(0.0451379,-0.904807,0.423423), Vector3d(0.653585,0.755946,0.037037),
        Vector3d(-0.938577,0.337673,0.0710711), Vector3d(-0.197223,0.978922,0.0530531),
        Vector3d(-0.393155,0.718639,0.573574), Vector3d(-0.34315,-0.255354,-0.903904),
        Vector3d(-0.561097,-0.333126,0.757758), Vector3d(-0.791487,-0.153518,0.591592),
        Vector3d(0.719987,0.689915,-0.0750751), Vector3d(0.406036,0.44927,0.795796),
        Vector3d(-0.573688,0.580763,0.577578), Vector3d(0.773452,-0.0530408,0.631632),
        Vector3d(-0.34798,-0.828132,-0.439439), Vector3d(0.975861,0.17156,0.135135),
        Vector3d(0.650204,-0.000248327,-0.75976), Vector3d(0.521582,-0.790383,-0.321321),
        Vector3d(0.0428385,-0.783732,-0.61962), Vector3d(0.114627,0.255612,-0.95996),
        Vector3d(-0.18887,-0.771454,0.607608), Vector3d(0.857205,0.508867,-0.0790791),
        Vector3d(-0.997021,0.0028512,0.0770771), Vector3d(0.291274,0.240459,-0.925926),
        Vector3d(0.405602,-0.492891,-0.76977), Vector3d(-0.758356,0.103052,-0.643644),
        Vector3d(-0.0453895,-0.94656,0.319319), Vector3d(0.366461,0.0306028,-0.92993),
        Vector3d(0.288553,-0.653573,0.6997), Vector3d(0.909461,0.407807,-0.0810811),
        Vector3d(-0.991005,-0.130174,-0.031031), Vector3d(-0.665304,0.175194,-0.725726),
        Vector3d(0.320683,0.269937,0.907908), Vector3d(0.901997,0.148485,-0.405405),
        Vector3d(-0.432608,0.90002,-0.0530531), Vector3d(0.726322,0.0881278,-0.681682),
        Vector3d(0.949856,0.301448,-0.0830831), Vector3d(-0.297266,-0.439075,-0.847848),
        Vector3d(0.410868,-0.743574,0.527528), Vector3d(-0.939563,-0.332141,0.0830831),
        Vector3d(-0.371075,0.17527,-0.911912), Vector3d(-0.767145,-0.457693,-0.449449),
        Vector3d(-0.722387,0.50954,-0.467467), Vector3d(-0.391167,0.879432,-0.271271),
        Vector3d(0.763654,-0.391326,-0.513514), Vector3d(0.602181,0.475052,0.641642),
        Vector3d(-0.184439,0.906672,-0.379379), Vector3d(0.578194,0.762891,-0.289289),
        Vector3d(0.998493,-0.0507026,0.021021), Vector3d(-0.0984911,0.982001,0.161161),
        Vector3d(-0.935926,-0.250821,-0.247247), Vector3d(0.162232,0.728402,-0.665666),
        Vector3d(0.365758,0.855996,0.365365), Vector3d(0.985069,-0.146087,-0.0910911),
        Vector3d(-0.330725,0.903296,0.273273), Vector3d(-0.452482,-0.824687,-0.339339),
        Vector3d(-0.457783,-0.72718,0.511512), Vector3d(0.489071,-0.262422,-0.831832),
        Vector3d(-0.88614,0.363623,0.287287), Vector3d(-0.287757,0.517565,-0.805806),
        Vector3d(-0.412325,-0.228631,0.881882), Vector3d(0.937353,-0.0458955,0.345345),
        Vector3d(0.810524,-0.478805,0.337337), Vector3d(-0.0506787,0.295641,0.953954),
        Vector3d(0.422724,-0.84662,-0.323323), Vector3d(0.512787,0.106761,0.851852),
        Vector3d(0.21543,-0.128949,-0.967968), Vector3d(0.969828,0.0351534,0.241241),
        Vector3d(-0.0977486,-0.633228,0.767768), Vector3d(0.95186,-0.0274027,-0.305305),
        Vector3d(0.878049,-0.468617,-0.0970971), Vector3d(0.318268,-0.846175,-0.427427),
        Vector3d(-0.516202,-0.851171,0.0950951), Vector3d(0.0868704,-0.972695,0.215215),
        Vector3d(-0.0714271,0.582328,-0.80981), Vector3d(0.453292,-0.0986532,-0.885886),
        Vector3d(0.979713,0.0357204,-0.197197), Vector3d(0.894307,-0.00226645,0.447447),
        Vector3d(-0.514385,0.84228,-0.161161), Vector3d(0.993108,0.0784345,-0.0870871),
        Vector3d(-0.590494,-0.156151,-0.791792), Vector3d(-0.409745,-0.736995,-0.537538),
        Vector3d(-0.709566,0.703067,-0.047047), Vector3d(-0.446125,0.500791,0.741742),
        Vector3d(0.015191,0.987136,0.159159), Vector3d(0.25405,0.966142,0.045045),
        Vector3d(0.748588,-0.655282,-0.101101), Vector3d(-0.455843,0.68562,-0.567568),
        Vector3d(0.48256,-0.227333,0.845846), Vector3d(0.132296,-0.320578,-0.937938),
        Vector3d(-0.495592,-0.626469,0.601602), Vector3d(0.500982,0.637265,-0.585586),
        Vector3d(0.355968,-0.258972,0.897898), Vector3d(0.433032,0.374672,-0.81982),
        Vector3d(-0.199543,-0.974659,0.101101), Vector3d(0.729509,-0.425462,0.535536),
        Vector3d(-0.0151338,0.594702,0.803804), Vector3d(0.649901,-0.622917,0.435435),
        Vector3d(-0.139588,0.807173,-0.573574), Vector3d(-0.0717045,0.921633,-0.381381),
        Vector3d(-0.35291,-0.906635,-0.231231), Vector3d(-0.484603,-0.607229,-0.62963),
        Vector3d(0.472503,0.789654,-0.391391), Vector3d(0.7873,0.159449,-0.595596),
        Vector3d(0.0257989,-0.233677,-0.971972), Vector3d(-0.130548,-0.612291,-0.77978),
        Vector3d(-0.514151,-0.056208,-0.855856), Vector3d(0.30577,-0.138745,0.941942),
        Vector3d(0.252009,0.0852334,-0.963964), Vector3d(-0.0648403,-0.435402,-0.897898),
        Vector3d(0.747411,-0.573521,0.335335), Vector3d(-0.406793,0.831018,0.379379),
        Vector3d(0.31588,-0.923126,-0.219219), Vector3d(-0.608079,0.777759,-0.159159),
        Vector3d(-0.85022,0.524664,-0.043043), Vector3d(-0.675799,0.637857,-0.369369),
        Vector3d(0.373367,0.739926,0.55956), Vector3d(-0.808886,0.440544,0.389389),
        Vector3d(-0.290375,-0.724068,-0.625626), Vector3d(0.381508,-0.917904,-0.109109),
        Vector3d(-0.7566,0.294934,0.583584), Vector3d(0.268279,0.401083,-0.875876),
        Vector3d(0.128708,0.979151,0.157157), Vector3d(0.181136,-0.928791,0.323323),
        Vector3d(-0.302443,0.824951,0.477477), Vector3d(0.306343,0.753603,-0.581582),
        Vector3d(0.809782,-0.287419,-0.511512), Vector3d(-0.217154,0.337553,-0.915916),
        Vector3d(0.264512,-0.964382,-0.001001), Vector3d(0.590439,0.166173,0.78979),
        Vector3d(-0.664261,-0.513145,-0.543544), Vector3d(-0.0707424,-0.780117,-0.621622),
        Vector3d(-0.904434,0.424635,-0.041041), Vector3d(-0.813199,-0.295303,0.501502),
        Vector3d(0.256881,0.845864,0.467467), Vector3d(0.683296,-0.24534,-0.687688),
        Vector3d(-0.634947,0.418088,-0.64965), Vector3d(0.563367,0.825284,0.039039),
        Vector3d(0.571084,-0.44207,-0.691692), Vector3d(0.153249,-0.988183,-0.003003),
        Vector3d(0.3599,-0.926351,0.111111), Vector3d(0.495996,0.280352,-0.821822),
        Vector3d(-0.288295,-0.477775,0.82983), Vector3d(-0.794009,0.242238,-0.557558),
        Vector3d(0.728101,-0.275595,0.627628), Vector3d(0.040002,-0.999187,-0.00500501),
        Vector3d(-0.963413,0.0877352,-0.253253), Vector3d(0.715029,-0.136268,-0.685686),
        Vector3d(-0.588713,-0.695814,0.411411), Vector3d(0.620419,-0.475577,0.623624),
        Vector3d(0.475366,-0.879783,0.003003), Vector3d(-0.673227,0.690218,-0.265265),
        Vector3d(-0.242243,-0.910422,-0.335335), Vector3d(-0.129843,-0.890807,-0.435435),
        Vector3d(0.785694,0.565294,0.251251), Vector3d(0.834389,-0.102604,0.541542),
        Vector3d(-0.198714,-0.495033,-0.845846), Vector3d(-0.601562,0.0695129,-0.795796),
        Vector3d(-0.0737614,-0.997251,-0.00700701), Vector3d(0.797653,-0.0650342,-0.5996),
        Vector3d(-0.0181909,-0.628276,-0.777778), Vector3d(0.0527957,0.958757,-0.279279),
        Vector3d(0.208419,-0.921639,-0.327327), Vector3d(0.572225,-0.820081,0.00500501),
        Vector3d(0.229383,0.937057,0.263263), Vector3d(-0.566138,0.103255,0.817818),
        Vector3d(0.371891,0.842658,-0.389389), Vector3d(-0.977129,0.209397,-0.037037),
        Vector3d(0.691529,-0.685678,0.227227), Vector3d(0.423062,-0.354598,-0.833834),
        Vector3d(-0.235401,0.540401,0.807808), Vector3d(0.459772,-0.636145,0.61962),
        Vector3d(-0.690392,-0.0497598,-0.721722), Vector3d(-0.17676,-0.977019,-0.119119),
        Vector3d(-0.504447,-0.674115,-0.53954), Vector3d(-0.182722,0.56041,-0.807808),
        Vector3d(-0.0800756,0.924219,0.373373), Vector3d(0.661674,-0.749759,0.00700701),
        Vector3d(0.476857,-0.526696,0.703704), Vector3d(0.291161,-0.89966,0.325325),
        Vector3d(-0.532851,0.799492,0.277277), Vector3d(-0.375057,0.795773,-0.475475),
        Vector3d(-0.296956,-0.954828,-0.011011), Vector3d(-0.994669,0.0969808,-0.035035),
        Vector3d(-0.945935,0.200131,-0.255255), Vector3d(0.237916,0.622302,-0.745746),
        Vector3d(0.694924,-0.387714,-0.605606), Vector3d(0.732031,-0.670777,0.119119),
        Vector3d(-0.776606,0.562702,0.283283), Vector3d(0.807551,0.588872,0.033033),
        Vector3d(-0.770191,0.618654,-0.155155), Vector3d(0.523977,0.557827,0.643644),
        Vector3d(-0.567295,0.675197,-0.471471), Vector3d(-0.297703,-0.73798,0.605606),
        Vector3d(0.909171,-0.0760996,-0.409409), Vector3d(-0.35139,-0.610604,-0.70971),
        Vector3d(-0.403495,-0.914889,-0.013013), Vector3d(-0.268899,0.836486,-0.477477),
        Vector3d(-0.691657,0.319631,-0.647648), Vector3d(0.724619,0.673692,0.145145),
        Vector3d(0.371984,0.642783,-0.66967), Vector3d(-0.989606,-0.0138154,-0.143143),
        Vector3d(0.813812,-0.581025,0.011011), Vector3d(-0.133338,-0.96467,-0.227227),
        Vector3d(0.861622,0.0455129,-0.505506), Vector3d(0.297705,0.617887,0.727728),
        Vector3d(0.841783,-0.178329,-0.50951), Vector3d(0.349239,0.924433,0.153153),
        Vector3d(-0.167445,-0.139402,-0.975976), Vector3d(0.859669,0.490967,0.141141),
        Vector3d(-0.199409,-0.821935,-0.533534), Vector3d(-0.495591,-0.859496,-0.125125),
        Vector3d(0.0960679,-0.939315,-0.329329), Vector3d(-0.628644,-0.637483,-0.445445),
        Vector3d(0.864959,-0.486505,0.123123), Vector3d(-0.98177,-0.127308,-0.141141),
        Vector3d(0.154991,0.909646,-0.385385), Vector3d(-0.499948,0.781437,-0.373373),
        Vector3d(0.831945,-0.336151,0.441441), Vector3d(-0.935068,0.321562,-0.149149),
        Vector3d(-0.599574,-0.800138,-0.017017), Vector3d(0.276262,0.918384,-0.283283),
        Vector3d(0.429591,-0.0449904,0.901902), Vector3d(-0.409486,-0.502036,0.761762),
        Vector3d(0.910585,0.389197,0.139139), Vector3d(-0.816844,0.514304,-0.261261),
        Vector3d(0.923919,-0.382294,0.015015), Vector3d(-0.96985,-0.241967,-0.029029),
        Vector3d(-0.686573,-0.726812,-0.019019), Vector3d(0.589379,-0.736785,0.331331),
        Vector3d(-0.0175935,-0.94335,-0.331331), Vector3d(-0.872145,0.414906,-0.259259),
        Vector3d(0.145815,0.870825,0.469469), Vector3d(0.949486,0.282257,0.137137),
        Vector3d(-0.927458,-0.347871,-0.137137), Vector3d(0.961339,-0.274841,0.017017),
        Vector3d(-0.764674,-0.644074,-0.021021), Vector3d(0.414604,-0.882703,0.221221),
        Vector3d(0.642265,0.752225,0.147147), Vector3d(-0.0693935,0.820397,0.567568),
        Vector3d(-0.890264,-0.454757,-0.025025), Vector3d(-0.832866,-0.552995,-0.023023),
        Vector3d(-0.706204,-0.495721,0.505506), Vector3d(0.97766,-0.165849,0.129129),
        Vector3d(0.984508,0.173547,0.025025), Vector3d(0.453325,0.878436,0.151151),
        Vector3d(0.374521,-0.821781,0.429429), Vector3d(0.989962,-0.0527215,0.131131),
        Vector3d(0.997899,0.0605544,0.023023), Vector3d(-0.596144,0.705428,0.383383),
        Vector3d(-0.546861,-0.483239,0.683684), Vector3d(0.540752,0.800879,0.257257),
        Vector3d(-0.704936,0.651112,0.281281), Vector3d(0.49678,-0.802964,0.329329),
        Vector3d(-0.633879,-0.583638,0.507508), Vector3d(-0.19356,0.858171,0.475475),
        Vector3d(0.0324598,0.881284,0.471471), Vector3d(0.866869,-0.227808,0.443443),
        Vector3d(-0.201182,0.341983,0.917918), Vector3d(-0.76632,-0.399045,0.503504),
        Vector3d(-0.148319,-0.435496,0.887888), Vector3d(-0.369447,-0.624976,0.687688),
        Vector3d(0.434353,0.628076,0.645646), Vector3d(0.591721,-0.606086,0.531532),
        Vector3d(0.337501,-0.533165,0.775776), Vector3d(-0.659379,-0.00938952,0.751752),
        Vector3d(-0.648515,0.365578,0.667668), Vector3d(0.61549,-0.346875,0.707708),
        Vector3d(-0.478732,-0.0120935,0.877878), Vector3d(-0.586669,0.461203,0.665666),
        Vector3d(0.621815,0.308787,0.71972), Vector3d(-0.185355,0.0726829,-0.97998),
        Vector3d(-0.127893,0.578205,0.805806), Vector3d(0.349744,-0.411073,0.841842),
        Vector3d(0.500846,-0.375628,0.77978), Vector3d(-0.208905,-0.608245,0.765766),
        Vector3d(0.484936,0.367054,0.793794), Vector3d(-0.26677,-0.253109,0.92993),
        Vector3d(0.474427,0.214133,0.853854), Vector3d(-0.377855,0.0599151,0.923924),
        Vector3d(0.0778614,-0.445496,0.891892), Vector3d(0.015744,0.405603,0.913914),
        Vector3d(-0.438094,0.210755,0.873874), Vector3d(0.223806,-0.0358572,0.973974),
        Vector3d(-0.267367,-0.0560734,0.961962), Vector3d(-0.0616725,-0.0904016,-0.993994),
        Vector3d(0.0902568,0.0618842,0.993994), Vector3d(0.171594,0.261056,0.94995),
        Vector3d(-0.110614,0.153218,-0.981982), Vector3d(-0.173136,-0.318633,0.931932),
        Vector3d(0.100383,-0.221543,0.96997), Vector3d(0.0963564,0.136188,-0.985986),
        Vector3d(-0.0922508,0.164928,0.981982), Vector3d(-0.0838939,-0.129775,0.987988),
        Vector3d(0.148823,0.0416121,-0.987988), Vector3d(0.0376468,-0.12056,-0.991992),
        Vector3d(-0.0891903,0.00608871,-0.995996), Vector3d(-0.0267314,0.0853077,0.995996),
        Vector3d(-0.0143696,0.0615915,-0.997998), Vector3d(-0.0544768,-0.032129,0.997998),
    };

    for(n = 0; n<=samples; n++)
        if(!Inside(EPoint, SampleVect[n]*TestRadius, pThread))
            return false;

    return true;
}

DBL HardObjectPattern::RadiusFitsInside(const Vector3d& EPoint, DBL CurrentRadius, DBL CurrentRadiusDelta,
                                        size_t& Depth, const size_t& MaxDepth, TraceThreadData *pThread) const
{
    if ((Depth > MaxDepth) || (CurrentRadiusDelta < gkDBL_epsilon))
        return CurrentRadius;
    else
    {
       Depth++;

       if(SphericalSamplesInside(EPoint, CurrentRadius, samples, pThread))
           return RadiusFitsInside(EPoint,CurrentRadius+CurrentRadiusDelta,CurrentRadiusDelta*0.5,
                                   Depth,MaxDepth,pThread);
       else
           return RadiusFitsInside(EPoint,CurrentRadius-CurrentRadiusDelta,CurrentRadiusDelta*0.5,
                                   Depth,MaxDepth,pThread);
    }

    return 0.0;
}

DBL HardObjectPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
                                   const Ray *pRay, TraceThreadData *pThread) const
{
    size_t Depth=0;

    if(pObject != NULL)
    {
        if(Inside_BBox(EPoint, pObject->BBox) && Inside_Object(EPoint, pObject, pThread))
        {
            if(SphericalSamplesInside(EPoint, radius, samples, pThread))
            {
                return 1.0; // Where initial spherical samples inside object at full radius.
            }
            else
            {
                return min(1.0,RadiusFitsInside(EPoint,radius,radius*0.5,Depth,recursion_limit,pThread)/radius);
            }
        }
        else
        {
            return 0.0;
        }
    }

    return 0.0;
}


//-------- SoftObjectPattern related code ----------

inline DBL SoftObjectPattern::Blob(const DBL v, const DBL s) const
{
    return 1.0/exp(v*s);
}

inline void SoftObjectPattern::CalcAllDiffsSqrd(DBL *aryd, const DBL axisVal, const DBL spacing) const
{  //---- Do all normalized dist calculations for axis at this point for use later.
    size_t n;
    DBL nrmVal = 3.5+(fmod(axisVal,spacing)*(1/spacing)-0.5);
    DBL tmpVal;
    for(n = 0; n<8; n++)
    {
        tmpVal=(DBL)(n)-nrmVal;
        aryd[n]=tmpVal*tmpVal;
    }
}

DBL SoftObjectPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
                                   const Ray *pRay, TraceThreadData *pThread) const
{
    //--- strength and spacing defaulted but expect usually set in SDL.

    DBL runningval   = 0.0;
    size_t startidx  = 0;
    size_t stopidx   = 7;
    DBL strengthBias = 0.6187184335382291; // Experimentally determined.
    DBL val          = strengthBias/strength;
    DBL rxd[8],ryd[8],rzd[8];

    if(pObject != NULL)
    {
        Vector3d VirtBBoxLL;
        Vector3d VirtBBoxSize;

        VirtBBoxLL.x() = floor((EPoint.x()/spacing-3.0))*spacing; // LL - Lower Left
        VirtBBoxLL.y() = floor((EPoint.y()/spacing-3.0))*spacing;
        VirtBBoxLL.z() = floor((EPoint.z()/spacing-3.0))*spacing;

        //---------
        // The following BBOX pre-filter mostly helps - 5% faster in testing
        // where shape <2/3 volume/pattern surface.

        VirtBBoxSize.x() =
        VirtBBoxSize.y() =
        VirtBBoxSize.z() = 8.0*spacing;

        if((VirtBBoxLL.x()-gkMinIsectDepthReturned > (DBL)pObject->BBox.lowerLeft.x()+(DBL)pObject->BBox.size.x()) ||
           (VirtBBoxLL.y()-gkMinIsectDepthReturned > (DBL)pObject->BBox.lowerLeft.y()+(DBL)pObject->BBox.size.y()) ||
           (VirtBBoxLL.z()-gkMinIsectDepthReturned > (DBL)pObject->BBox.lowerLeft.z()+(DBL)pObject->BBox.size.z())
          )
            return(0.0);
        if((VirtBBoxLL.x()+VirtBBoxSize.x()+gkMinIsectDepthReturned < (DBL)pObject->BBox.lowerLeft.x()) ||
           (VirtBBoxLL.y()+VirtBBoxSize.y()+gkMinIsectDepthReturned < (DBL)pObject->BBox.lowerLeft.y()) ||
           (VirtBBoxLL.z()+VirtBBoxSize.z()+gkMinIsectDepthReturned < (DBL)pObject->BBox.lowerLeft.z())
          )
            return(0.0);
        //--------- End BBOX pre-filter

        size_t i, j, k;
        Vector3d thisVoxel;

        CalcAllDiffsSqrd(rxd,EPoint.x()-VirtBBoxLL.x(),spacing);
        CalcAllDiffsSqrd(ryd,EPoint.y()-VirtBBoxLL.y(),spacing);
        CalcAllDiffsSqrd(rzd,EPoint.z()-VirtBBoxLL.z(),spacing);

        for(i = startidx; i <= stopidx; i++)
        {
            thisVoxel.z() = VirtBBoxLL.z() + i * spacing;
            for(j = startidx; j <= stopidx; j++)
            {
                thisVoxel.y() = VirtBBoxLL.y() + j * spacing;
                for(k = startidx; k <= stopidx; k++)
                {
                    thisVoxel.x() = VirtBBoxLL.x() + k * spacing;
                    if(Inside_BBox(thisVoxel, pObject->BBox) && Inside_Object(thisVoxel, pObject, pThread))
                    {
                        runningval+=Blob(rzd[i]+ryd[j]+rxd[k],val);
                        if(runningval > 1.0)
                            break;
                    }
                }
            }
        }
    }

    return min(1.0,runningval);
}


// Original Author Scott Taylor
DBL OnionPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value = (fmod(EPoint.length(), 1.0));
    return value;
}


DBL PigmentPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value;
    TransColour Col;
    int colour_found=false;

    if (pPigment)
        // TODO ALPHA - we're discarding transparency information, so maybe we want to pre-multiply if there's alpha in there?
        colour_found = Compute_Pigment(Col, pPigment, EPoint, pIsection, pRay, pThread);

    if(!colour_found)
        value = 0.0;
    else
        value = Col.Greyscale();

    return value;
}


DBL PlanarPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    // EPoint -- Point in 3d space for evaluation.
    DBL value = max(0.0,1.0 - fabs(EPoint.y()));
    return value;
}


DBL PotentialPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    if (pObject != nullptr)
        return pObject->GetPotential (EPoint, subtractThreshold, pThread);

    return 0.0;
}


// Pattern intended as a seed pattern for -1 to 1 function interval --> seedfi.
// function_interval must be used to work with negative values as returned.
// It can be used as a normal 0-1 scalar indexed pattern too.
double SeedfiPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    double x = EPoint.x();
    double y = EPoint.y();
    double z = EPoint.z();

    double xm = floor(x)+(ceil(floor(x)+gkMinIsectDepthReturned)-floor(x))/2.0;
    double ym = floor(y)+(ceil(floor(y)+gkMinIsectDepthReturned)-floor(y))/2.0;
    double zm = floor(z)+(ceil(floor(z)+gkMinIsectDepthReturned)-floor(z))/2.0;

    x = (x-xm);
    y = (y-ym);
    z = (z-zm);

    double value = Vector3d(x,y,z).length() + seedfiOffset;
    value = max(-1.0,min(1.0,value*2.0-1.0));

    return (value);
}


DBL RadialPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    DBL angle = atan2(EPoint.z(),EPoint.x());
    if (angle > 0.0)
    {
        angle = TWO_M_PI - angle;
    }

    DBL value = fabs(angle)/TWO_M_PI;

    return(value);
}


// Note this pattern is only used for pigments and textures.  Normals have a
// specialized pattern for this.
//
// Nov 1994 : adapted from normal by [CY]
//
DBL RipplesPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL length, index;
    DBL scalar = 0.0;
    Vector3d point;

    for (unsigned int i = 0; i < pThread->numberOfWaves; i++)
    {
        point = EPoint - pThread->waveSources[i];
        length = point.length();

        // Code v1.0->v3.8 set to 1.0 if length zero.
        if (fabs(length) <= gkDBL_epsilon)
            length = gkDBL_epsilon;

        index = length * frequency + phase;

        scalar += cycloidal(index);
    }

    scalar = 0.5*(1.0+(scalar / (DBL)pThread->numberOfWaves));

    return(scalar);
}


/*****************************************************************************
*
* FUNCTION
*
*   slope_pattern
*
* INPUT
*
*   EPoint -- The point in 3d space at which the pattern
*             is evaluated.
*   TPat   -- Texture pattern struct
*   Intersection - intersection struct
*
* OUTPUT
*
* RETURNS
*
*   DBL value in the range 0.0 to 1.0, 0.0 if intersection struct is `nullptr`
*
* AUTHOR
*
*   -hdf-
*
* DESCRIPTION   :
*
*   calculates the surface slope from surface normal vector
*
* CHANGES
*
*   Apr 1998 : written by H.-D. Fink
*   May 1998 : modified by M.C. Andrews - now combines slope and 'gradient'.
*
******************************************************************************/

DBL SlopePattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value, value1, value2;

    if (pIsection == nullptr) return 0.0; /* just in case ... */

    if (pointAt)
    {
        Vector3d vect;
        vect = slopeDirection - pIsection->IPoint.normalized();
        value1 = dot(pIsection->PNormal, vect);
    }
    else
    {
        if (slopeAxis > 0)
            /* short case 1: slope vector in x, y or z direction */
            value1 = pIsection->PNormal[slopeAxis - 1];
        else if (slopeAxis < 0)
            /* short case 2: slope vector in negative x, y or z direction */
            value1 = -pIsection->PNormal[-slopeAxis - 1];
        else
            /* projection slope onto normal vector */
            value1 = dot(pIsection->PNormal, slopeDirection);
    }

    /* Clamp to 1.0. */
    /* should never be necessary since both vectors are normalized */
    if      (value1 >  1.0) value1 =  1.0;
    else if (value1 < -1.0) value1 = -1.0;

    value1 = asin(value1) / M_PI * 2;
    value1 = (value1 + 1.0) * 0.5;        /* normalize to [0..1] interval */

    /* If set, use offset and scalings for slope and altitude. */
    if (0.0 != slopeModWidth)
    {
        value1 = (value1 - slopeModLow) / slopeModWidth;
    }

    if (!altitudeLen)
    {
        /* Clamp to 1.0. */
        if ( value1 == 1.0 )
        {
            value1= value1- gkDBL_epsilon;
        }
        else
        {
            value1 = (value1 < 0.0) ? 1.0 + fmod(value1, 1.0) : fmod(value1, 1.0);
        }
        return value1; /* no altitude defined */
    }

    /* Calculate projection of EPoint along altitude vector */
    if (altitudeAxis > 0)
        /* short case 1: altitude vector in x, y or z direction */
        value2 = EPoint[altitudeAxis - 1];
    else if (altitudeAxis < 0)
        /* short case 2: altitude vector in negative x, y or z direction */
        value2 = -EPoint[-altitudeAxis - 1];
    else
        /* projection of EPoint along altitude vector */
        value2 = dot(EPoint, altitudeDirection);

    if (0.0 != altitudeModWidth)
    {
        value2 = (value2 - altitudeModLow) / altitudeModWidth;
    }

    value = slopeLen * value1 + altitudeLen * value2;

    /* Clamp to 1.0. */
    if ( value - 1.0 < gkDBL_epsilon && value >= 1.0 )
    {
        /* 1.0 is a very common value to get *exactly*.  We don't want to wrap
           it to the bottom end of the map. */
        value = value - gkDBL_epsilon;
    }
    else
    {
        value = (value < 0.0) ? 1.0 + fmod(value, 1.0) : fmod(value, 1.0);
    }
    return value;

}


// Author: J. Grimbert
//
// Return a value related to ray angle of incidence with the surface after the
// ray has seen any normal block perturbations.
//
// Returns the angle between the, possibly perturbed normal, at the intersection
// and the ray as a value between -1 and 1, where 1 a normal pointing exactly
// back at the incoming ray and -1 completely away. NOTE! This is different
// than v3.7 and v3.8 which handle well only perturbed normals which do not
// invert with respect to the ray direction.
//
// Note! Unless povr's function_interval keyword is used the wave modifier code
// will do what it always does with -1 to 0 and flip it to 0 to 1.
//
// Mar 2010 : modified by [CLi]
//
DBL AOIPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL       cosAngle, angle;

    if ((pIsection == nullptr) || (pRay == nullptr))
        return 0.0;

    // Perturbed normal normalization below necessary for when no_bump_scale
    // used in normal block. Otherwise, lengthSqr values are normalized but
    // differ by around 4.4e-15 in testing.

    Vector3d a = pIsection->PNormal.normalized();

    if (0)
    {   // 0.36% slower. In testing no differences seen, but if somebody reports aoi issue, perhap try at (1)
        Vector3d b = pRay->Direction.normalized(); // Already normalized, but there is a TINY bit of noise
      //if ((fabs(b.lengthSqr()-pRay->Direction.lengthSqr())) > 4.4e-15)  // In limited testing: 4.4e-15
      //    throw POV_EXCEPTION_STRING("aoi Ray Direction not perfectly normalized");

        cosAngle = dot(b,a);

    }
    else
    {
        cosAngle = dot(pRay->Direction,a);

        // clip to [-1.0; 1.0], just to be sure given using already normalized,
        // but not perfectly normalized, ray direction vector just above.

        cosAngle = clip(cosAngle, -1.0, 1.0);
    }

//  v3.7 and v3.8 both return the "angle" as a value 0 to 0.5 where 0 is the normal pointing
//  directly back at the ray. Perpendicular values are at 0.5. See comments above why the
//  change.

//  angle = acos(cosAngle) / M_PI;
//  return angle;

    return (cosAngle * -1.0);
}


// Author Dieter Bayer
//
// Aug 1994 : Creation.
// Oct 1994 : adapted from pigment by [CY]
//
DBL Spiral1Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);
    DBL x = EPoint.x();
    DBL y = EPoint.y();
    DBL z = EPoint.z();

    DBL turb_val = 0.0;
    if (spiralFrequency != 0.0)
    {
        TurbulenceWarp Turb;
        Turb.Octaves = spiralOctaves;
        Turb.Omega   = spiralOmega;
        Turb.Lambda  = spiralLambda;
        // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
        // Turbulence() uses only octaves, omega, lambda values.
        turb_val = spiralFrequency * Turbulence(EPoint,&Turb,noise_generator);
    }

    DBL rad = sqrt(x*x + y*y);

    // Get 0-1 rotation in x,y-plane via (0...2*PI/2*pi)

    // Natural to use atan2, but because it has an effective internal radius
    // calculation (sqrt), it is better to use acos or asin. Set up with acos
    // to match atan2(y,x) results which IS a change from traditional POV-Ray
    // spiral1 and spiral2 orientation.

    //  DBL angle = atan2(y,x);
    ////DBL angle = atan2(-x,-y); // Use to match old spiral1 results. (or rotate z*90 scale <1,-1,1>)
    //  if (angle > 0.0)
    //  {
    //      angle = TWO_M_PI - angle;
    //  }
    //  DBL rot = fabs(angle)/TWO_M_PI;

    DBL angle = 0.0;
    if (rad != 0.0)
    {
        if(y>=0)
            angle = TWO_M_PI - acos(x/rad);
        else
            angle = acos(x/rad);
    }

    return(z + rad + ((DBL)arms * angle/TWO_M_PI) + turb_val);
}


// Author Dieter Bayer
//
// Aug 1994 : Creation.
// Oct 1994 : adapted from pigment by [CY]
//
DBL Spiral2Pattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);
    DBL x = EPoint.x();
    DBL y = EPoint.y();
    DBL z = EPoint.z();

    DBL turb_val = 0.0;
    if (spiralFrequency != 0.0)
    {
        TurbulenceWarp Turb;
        Turb.Octaves = spiralOctaves;
        Turb.Omega   = spiralOmega;
        Turb.Lambda  = spiralLambda;
        // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
        // Turbulence() uses only octaves, omega, lambda values.
        turb_val = spiralFrequency * Turbulence(EPoint,&Turb,noise_generator);
    }

    DBL rad = sqrt(x*x + y*y);

    // See spiral1 pattern code for additional documentation.

    DBL angle = 0.0;
    if (rad != 0.0)
    {
        if(y>=0)
            angle = TWO_M_PI - acos(x/rad);
        else
            angle = acos(x/rad);
    }

    // Need Triangle_Wave call because normal shape wave modifier 'triangle_wave' not yet run.
    DBL tmpVal = Triangle_Wave(z + rad + ((DBL)arms * angle/TWO_M_PI) + turb_val);

    return(Triangle_Wave(rad) + tmpVal);
}


DBL SphericalPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
    const Ray *pRay, TraceThreadData *pThread) const
{
    DBL value = max(0.0,1.0 - EPoint.length());
    return(value);
}


// Note this pattern is only used for pigments and textures.  Normals have a
// specialized pattern for this.
//
// Nov 1994 : adapted from normal by [CY]
//
DBL WavesPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    DBL length, index;
    DBL scalar = 0.0;
    Vector3d point;

    for (unsigned int i = 0; i < pThread->numberOfWaves; i++)
    {
        point = EPoint - pThread->waveSources[i];
        length = point.length();

        // Code v1.0->v3.8 set to 1.0 if length zero.
        if (fabs(length) <= gkDBL_epsilon)
            length = gkDBL_epsilon;

        index = length * frequency * pThread->waveFrequencies[i] + phase;

        scalar += cycloidal(index)/pThread->waveFrequencies[i];
    }

    scalar = 0.2*(2.5+(scalar / (DBL)pThread->numberOfWaves));

    return(scalar);
}


// point[Z] += z; Deleted per David Buck -- BP 7/91
// Oct 1994 : adapted from pigment by [CY]
//
DBL WoodPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);
    DBL x = EPoint.x();
    DBL y = EPoint.y();
    DBL z = EPoint.z();

    TurbulenceWarp Turb;
    Turb.Octaves = woodOctaves;
    Turb.Omega   = woodOmega;
    Turb.Lambda  = woodLambda;
    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    DBL xoffset = woodFrequency * (x + Turbulence(EPoint,&Turb,noise_generator));
    DBL yoffset = woodFrequency * (y + Turbulence(Vector3d(x,y+M_PI,z),&Turb,noise_generator));

    xoffset = sin(xoffset*TWO_M_PI);  // This was long a call to cycloidal. In testing
    yoffset = sin(yoffset*TWO_M_PI);  // see no difference in result.

    // For whatever reason std::hypot consistently ~1% slower?
    // return std::hypot(x+xoffset,y+yoffset);
    x += xoffset;
    y += yoffset;
    return sqrt(x*x+y*y);
}


// Author: POV-Ray Team
// Nov 1994 : adapted from normal by [CY]
// May 2021 : The povr branch dropped the normal block version of this pattern
//            due it having bias issues. It can still be run as type 0 of the
//            bevy normal collection pattern.
//
DBL WrinklesPattern::EvaluateRaw(const Vector3d& EPoint, const Intersection *pIsection,
        const Ray *pRay, TraceThreadData *pThread) const
{
    int noise_generator = GetNoiseGen(pThread);

    int i;
    DBL lambda = 2.0;
    DBL omega = 0.5;
    DBL value;
    Vector3d temp;
    DBL noise;

    // TODO - This distinction (with minor variations that seem to be more of an inconsistency rather than intentional)
    // appears in other places as well; make it a function.
    switch (noise_generator)
    {
        case kNoiseGen_Original:
            value = Noise(EPoint, noise_generator);
            break;

        default:
            noise = Noise(EPoint, noise_generator)*2.0-0.5;
            value = min(max(noise,0.0),1.0);
            break;
    }

    for (i = 1; i < 10; i++)
    {
        temp = EPoint * lambda;

        // TODO - This distinction (with minor variations that seem to be more of an inconsistency rather than intentional)
        // appears in other places as well; make it a function.
        switch (noise_generator)
        {
            case kNoiseGen_Original:
                value += omega * Noise(temp, noise_generator);
                break;

            default:
                noise = Noise(temp, noise_generator)*2.0-0.5;
                value += omega * min(max(noise,0.0),1.0);
                break;
        }

        lambda *= 2.0;

        omega *= 0.5;
    }

    return(value/2.0);
}


//--- IntPickInCube(tvx,tvy,tvz, p1)
//
// A version of PickInCube that takes integers for input which is now the version
// crackle uses.
//
// long integer hash function used, to speed up cacheing.
//
// Authors:
//  Original PickInCube by Jim McElhiney
//  This integer one modified by Nathan Kopp
//
// Original description:
//
// A subroutine to go with crackle.
//
// Pick a random point in the same unit-sized cube as tv, in a
// predictable way, so that when called again with another point in
// the same unit cube, p1 is picked to be the same.
//
static int IntPickInCube(int tvx, int tvy, int tvz, Vector3d& p1)
{
    size_t seed;

    seed = size_t(Hash3d(tvx&0xFFF,tvy&0xFFF,tvz&0xFFF));

    p1[X] = tvx + gPatternRands(seed);
    p1[Y] = tvy + gPatternRands(seed + 1);
    p1[Z] = tvz + gPatternRands(seed + 2);

    return (int)seed;
}


/*****************************************************************************
*
* FUNCTION
*
*   NewHash(tvx, tvy, tvz)
*
* INPUT
*
*   3D integer coordinates of the cell
*
* OUTPUT
*
* RETURNS
*
*   long integer hash value
*
* AUTHOR
*
*   Christoph Hormann based on MechSim Hash function by Daniel Jungmann
*
* DESCRIPTION
*
*   New Hash function for the crackle pattern.
*
* CHANGES
*
*   -- Aug 2005 : Creation
*
******************************************************************************/

#ifndef HAVE_BOOST_HASH
// NOTE: Keep this around - we may need it when we get rid of boost::hash.
static unsigned long int NewHash(long int tvx, long int tvy, long int tvz)
{
    unsigned long int seed;
    long int r;

    tvx *= 73856093L;
    tvy *= 19349663L;
    tvz *= 83492791L;

    r = tvx ^ tvy ^ tvz;
    seed = abs(r);
    if (tvx<0) seed += std::numeric_limits<unsigned long>::max()/2;
    if (tvy<0) seed += std::numeric_limits<unsigned long>::max()/4;
    if (tvz<0) seed += std::numeric_limits<unsigned long>::max()/8;

    return (seed);
}
#endif


DBL FractalPattern::ExteriorColour(int iters, DBL a, DBL b) const
{
    switch(exteriorType)
    {
        case 0:
            return exteriorFactor;
        case 1:
            POV_PATTERN_ASSERT (maxIterations > 0);
            return (DBL)iters / (DBL)maxIterations;
        case 2:
            return a * exteriorFactor;
        case 3:
            return b * exteriorFactor;
        case 4:
            return a*a * exteriorFactor;
        case 5:
            return b*b * exteriorFactor;
        case 6:
            return sqrt(a*a+b*b) * exteriorFactor;
        case 7: // range 0.. (n-1)/n
            return (DBL)( iters % (unsigned int)exteriorFactor )/exteriorFactor;
        case 8: // range 0.. 1
            return (DBL)( iters % (unsigned int)(1+exteriorFactor) )/exteriorFactor;
    }
    return 0;
}


DBL FractalPattern::InteriorColour(DBL a, DBL b, DBL mindist2) const
{
    switch(interiorType)
    {
        case 0:
            return interiorFactor;
        case 1:
            return sqrt(mindist2) * interiorFactor;
        case 2:
            return a * interiorFactor;
        case 3:
            return b * interiorFactor;
        case 4:
            return a*a * interiorFactor;
        case 5:
            return b*b * interiorFactor;
        case 6:
            return a*a+b*b * interiorFactor;
    }
    return 0;
}


// Author: Dieter Bayer
//
// Create a density file structure.
//
//  Dec 1996 : Creation.
//
DENSITY_FILE *Create_Density_File()
{
    DENSITY_FILE *New;

    New = new DENSITY_FILE;

    New->Interpolation = kDensityFileInterpolation_None;

    New->Data = new DENSITY_FILE_DATA;

    New->Data->References = 1;

    New->Data->Name = nullptr;

    New->Data->Sx =
    New->Data->Sy =
    New->Data->Sz = 0;

    New->Data->Type = 0;

    New->Data->Density32 = nullptr;
    New->Data->Density16 = nullptr;
    New->Data->Density8 = nullptr;

    return (New);
}


// Author: Dieter Bayer
//
// Copy a density file structure.
//
//  Dec 1996 : Creation.
//
DENSITY_FILE *Copy_Density_File(DENSITY_FILE *Old)
{
    DENSITY_FILE *New;

    if (Old != nullptr)
    {
        New = new DENSITY_FILE;

        *New = *Old;

        New->Data->References++;
    }
    else
    {
        New = nullptr;
    }

    return(New);
}


// Author: Dieter Bayer
//
//  Destroy a density file structure.
//
//  Dec 1996 : Creation.
//
void Destroy_Density_File(DENSITY_FILE *Density_File)
{
    if (Density_File != nullptr)
    {
        if (Density_File->Data)
        {
            if((--(Density_File->Data->References)) == 0)
            {
                if (Density_File->Data->Name)
                    POV_FREE(Density_File->Data->Name);

                if(Density_File->Data->Type == 4)
                {
                    if (Density_File->Data->Density32)
                        delete[] Density_File->Data->Density32;
                }
                else if(Density_File->Data->Type == 2)
                {
                    if (Density_File->Data->Density16)
                        delete[] Density_File->Data->Density16;
                }
                else if(Density_File->Data->Type == 1)
                {
                    if (Density_File->Data->Density8)
                        delete[] Density_File->Data->Density8;
                }

                delete Density_File->Data;
            }
        }

        delete Density_File;
    }
}


void Read_Density_File(IStream *file, DENSITY_FILE *df)
{
    POV_PATTERN_ASSERT (file);

    size_t x, y, z, sx, sy, sz, len;

    if (df == nullptr)
        return;

    /* Allocate and read density file. */

    if ((df != nullptr) && (df->Data->Name != nullptr))
    {
        sx = df->Data->Sx = ReadUShort(file);
        sy = df->Data->Sy = ReadUShort(file);
        sz = df->Data->Sz = ReadUShort(file);

        // All could be zero and render clean. Check formally per point evaluation in density_pattern.
        if (sx==0 || sy==0 || sz==0)
            throw POV_EXCEPTION_STRING("Invalid density file. One or more dimensions is 0.");

        file->seekg(0, IOBase::seek_end);
        len = file->tellg() - 6;
        file->seekg(6);

        // figure out the data size
        if((sx * sy * sz * 4) == len)
        {
            df->Data->Type = 4;

            POV_UINT32 *map = new POV_UINT32[sx * sy * sz];

            for (z = 0; z < sz; z++)
            {
                for (y = 0; y < sy; y++)
                {
                    for (x = 0; x < sx; x++)
                    {
                        map[z * sy * sx + y * sx + x] = ReadUInt(file);
                    }
                }
            }

            df->Data->Density32 = map;
        }
        else if((sx * sy * sz * 2) == len)
        {
            df->Data->Type = 2;

            POV_UINT16 *map = new POV_UINT16[sx * sy * sz];

            for (z = 0; z < sz; z++)
            {
                for (y = 0; y < sy; y++)
                {
                    for (x = 0; x < sx; x++)
                        map[z * sy * sx + y * sx + x] = ReadUShort(file);
                }
            }

            df->Data->Density16 = map;
        }
        else if((sx * sy * sz) == len)
        {
            df->Data->Type = 1;

            POV_UINT8 *map = new POV_UINT8[sx * sy * sz];

            for (z = 0; z < sz; z++)
            {
                for (y = 0; y < sy; y++)
                    file->read(&(map[z * sy * sx + y * sx]), sizeof(unsigned char) * sx);
            }

            df->Data->Density8 = map;
        }
        else
            throw POV_EXCEPTION_STRING("Invalid density file size");
    }
    else
        throw POV_EXCEPTION_STRING("Null density file data structure or file name");
}


static unsigned short ReadUShort(IStream *pInFile)
{
    int i0 = 0, i1 = 0;

    if ((i0 = pInFile->Read_Byte ()) == EOF || (i1 = pInFile->Read_Byte ()) == EOF)
    {
        throw POV_EXCEPTION_STRING("Error reading density_file");
    }

    return (((unsigned short)i0 << 8) | (unsigned short)i1);
}


static unsigned int ReadUInt(IStream *pInFile)
{
    int i0 = 0, i1 = 0, i2 = 0, i3 = 0;

    if ((i0 = pInFile->Read_Byte ()) == EOF || (i1 = pInFile->Read_Byte ()) == EOF ||
        (i2 = pInFile->Read_Byte ()) == EOF || (i3 = pInFile->Read_Byte ()) == EOF)
    {
        throw POV_EXCEPTION_STRING("Error reading density_file");
    }

    return (((unsigned int)i0 << 24) | ((unsigned int)i1 << 16) | ((unsigned int)i2 << 8) | (unsigned int)i3);
}


static void InitializeBinomialCoefficients(void)
{
    int* ptr = gaBinomialCoefficients;
    *ptr = 1; ++ptr;

    for(unsigned n=1; n<=kFractalMaxExponent; ++n)
    {
        *ptr = 1; ++ptr;
        for(unsigned k=1; k<n; ++k)
        {
            *ptr = *(ptr-(n+1)) + *(ptr-n); ++ptr;
        }
        *ptr = 1; ++ptr;
    }
    ptr = gaBinomialCoefficients+1;
    for(unsigned m=1; m<=kFractalMaxExponent; ++m)
    {
        ++ptr;
        for(unsigned k=1; k<m; ++k)
        {
            if((k&2)!=0) *ptr = -(*ptr);
            ++ptr;
        }
        if((m&2)!=0) *ptr = -(*ptr);
        ++ptr;
    }
}


static void InitialiseCrackleCubeTable(void)
{
    int     *p = gaCrackleCubeTable;

    // the crackle cube table is a list of offsets in the range -2 ... 2 which
    // are applied to the EPoint while evaluating the Crackle pattern, in order
    // to look up points in close-by cubes. consider the EPoint to be in a cube
    // at the center of a 3x3 grid of cubes; candidate cubes are that are within
    // a "3d knight move" away (i.e. not more than 2 units). we use a lookup table
    // to speed up iteration of the cube list by avoiding branch tests.
    for(int addx = -2; addx <= 2; addx++)
    {
        for(int addy = -2; addy <= 2; addy++)
        {
            for(int addz = -2; addz <= 2; addz++)
            {
                if((abs(addx) == 2) + (abs(addy) == 2) + (abs(addz) == 2) <= 1)
                {
                    *p++ = addx;
                    *p++ = addy;
                    *p++ = addz;
                }
            }
        }
    }
}


// This should be called once, at povray start
void InitializePatternGenerators(void)
{
    InitializeBinomialCoefficients();
    InitialiseCrackleCubeTable();
}

}
// end of namespace pov
